@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1999-2015 Douglas Gilbert.
+ * Copyright (c) 1999-2018 Douglas Gilbert.
  * All rights reserved.
  * Use of this source code is governed by a BSD-style
  * license that can be found in the BSD_LICENSE file.
@@ -17,23 +17,26 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <stdarg.h>
+#include <stdbool.h>
 #include <string.h>
 #include <unistd.h>
-#include "sg_lib.h"
-#include "sg_cmds_basic.h"
-#include "sg_pt.h"
 
 #ifdef HAVE_CONFIG_H
 #include "config.h"
 #endif
 
+#include "sg_lib.h"
+#include "sg_cmds_basic.h"
+#include "sg_pt.h"
+#include "sg_unaligned.h"
+
 /* Needs to be after config.h */
 #ifdef SG_LIB_LINUX
 #include <errno.h>
 #endif
 
 
-static const char * version_str = "1.70 20150511";
+static const char * const version_str = "1.85 20180302";
 
 
 #define SENSE_BUFF_LEN 64       /* Arbitrary, could be larger */
@@ -61,7 +64,7 @@ sg_cmds_version()
     return version_str;
 }
 
-#ifdef __GNUC__
+#if defined(__GNUC__) || defined(__clang__)
 static int pr2ws(const char * fmt, ...)
         __attribute__ ((format (printf, 1, 2)));
 #else
@@ -84,7 +87,7 @@ pr2ws(const char * fmt, ...)
 /* Returns file descriptor >= 0 if successful. If error in Unix returns
    negated errno. */
 int
-sg_cmds_open_device(const char * device_name, int read_only, int verbose)
+sg_cmds_open_device(const char * device_name, bool read_only, int verbose)
 {
     /* The following 2 lines are temporary. It is to avoid a NULL pointer
      * crash when an old utility is used with a newer library built after
@@ -110,14 +113,16 @@ sg_cmds_close_device(int device_fd)
     return scsi_pt_close_device(device_fd);
 }
 
+static const char * const pass_through_s = "pass-through";
+
 static int
 sg_cmds_process_helper(const char * leadin, int mx_di_len, int resid,
-                       const unsigned char * sbp, int slen, int noisy,
+                       const uint8_t * sbp, int slen, bool noisy,
                        int verbose, int * o_sense_cat)
 {
     int scat, got;
-    int n = 0;
-    int check_data_in = 0;
+    bool n = false;
+    bool check_data_in = false;
     char b[512];
 
     scat = sg_err_category_sense(sbp, slen);
@@ -131,12 +136,17 @@ sg_cmds_process_helper(const char * leadin, int mx_di_len, int resid,
     case SG_LIB_CAT_PROTECTION:
     case SG_LIB_CAT_NO_SENSE:
     case SG_LIB_CAT_MISCOMPARE:
-        n = 0;
+        n = false;
         break;
     case SG_LIB_CAT_RECOVERED:
     case SG_LIB_CAT_MEDIUM_HARD:
-        ++check_data_in;
-        /* drop through */
+        check_data_in = true;
+#if defined(__GNUC__)
+#if (__GNUC__ >= 7)
+        __attribute__((fallthrough));
+        /* FALL THROUGH */
+#endif
+#endif
     case SG_LIB_CAT_UNIT_ATTENTION:
     case SG_LIB_CAT_SENSE:
     default:
@@ -144,14 +154,16 @@ sg_cmds_process_helper(const char * leadin, int mx_di_len, int resid,
         break;
     }
     if (verbose || n) {
-        sg_get_sense_str(leadin, sbp, slen, (verbose > 1),
+        if (leadin && (strlen(leadin) > 0))
+            pr2ws("%s:\n", leadin);
+        sg_get_sense_str(NULL, sbp, slen, (verbose > 1),
                          sizeof(b), b);
         pr2ws("%s", b);
         if ((mx_di_len > 0) && (resid > 0)) {
             got = mx_di_len - resid;
             if ((verbose > 2) || check_data_in || (got > 0))
-                pr2ws("    pass-through requested %d bytes (data-in) but "
-                      "got %d bytes\n", mx_di_len, got);
+                pr2ws("    %s requested %d bytes (data-in) but got %d "
+                      "bytes\n", pass_through_s, mx_di_len, got);
         }
     }
     if (o_sense_cat)
@@ -172,10 +184,11 @@ sg_cmds_process_helper(const char * leadin, int mx_di_len, int resid,
  * categories also have data in bytes received; -2 is still returned. */
 int
 sg_cmds_process_resp(struct sg_pt_base * ptvp, const char * leadin,
-                     int pt_res, int mx_di_len, const unsigned char * sbp,
-                     int noisy, int verbose, int * o_sense_cat)
+                     int pt_res, int mx_di_len, const uint8_t * sbp,
+                     bool noisy, int verbose, int * o_sense_cat)
 {
     int got, cat, duration, slen, resid, resp_code, sstat;
+    bool transport_sense;
     char b[1024];
 
     if (NULL == leadin)
@@ -183,7 +196,7 @@ sg_cmds_process_resp(struct sg_pt_base * ptvp, const char * leadin,
     if (pt_res < 0) {
 #ifdef SG_LIB_LINUX
         if (verbose)
-            pr2ws("%s: pass through os error: %s\n", leadin,
+            pr2ws("%s: %s os error: %s\n", leadin, pass_through_s,
                   safe_strerror(-pt_res));
         if ((-ENXIO == pt_res) && o_sense_cat) {
             if (verbose > 2)
@@ -191,19 +204,19 @@ sg_cmds_process_resp(struct sg_pt_base * ptvp, const char * leadin,
             *o_sense_cat = SG_LIB_CAT_NOT_READY;
             return -2;
         } else if (noisy && (0 == verbose))
-            pr2ws("%s: pass through os error: %s\n", leadin,
+            pr2ws("%s: %s os error: %s\n", leadin, pass_through_s,
                   safe_strerror(-pt_res));
 #else
         if (noisy || verbose)
-            pr2ws("%s: pass through os error: %s\n", leadin,
+            pr2ws("%s: %s os error: %s\n", leadin, pass_through_s,
                   safe_strerror(-pt_res));
 #endif
         return -1;
     } else if (SCSI_PT_DO_BAD_PARAMS == pt_res) {
-        pr2ws("%s: bad pass through setup\n", leadin);
+        pr2ws("%s: bad %s setup\n", leadin, pass_through_s);
         return -1;
     } else if (SCSI_PT_DO_TIMEOUT == pt_res) {
-        pr2ws("%s: pass through timeout\n", leadin);
+        pr2ws("%s: %s timeout\n", leadin, pass_through_s);
         return -1;
     }
     if ((verbose > 2) && ((duration = get_scsi_pt_duration_ms(ptvp)) >= 0))
@@ -212,7 +225,7 @@ sg_cmds_process_resp(struct sg_pt_base * ptvp, const char * leadin,
     slen = get_scsi_pt_sense_len(ptvp);
     switch ((cat = get_scsi_pt_result_category(ptvp))) {
     case SCSI_PT_RESULT_GOOD:
-        if (slen > 7) {
+        if (sbp && (slen > 7)) {
             resp_code = sbp[0] & 0x7f;
             /* SBC referrals can have status=GOOD and sense_key=COMPLETED */
             if (resp_code >= 0x70) {
@@ -227,18 +240,44 @@ sg_cmds_process_resp(struct sg_pt_base * ptvp, const char * leadin,
         }
         if (mx_di_len > 0) {
             got = mx_di_len - resid;
-            if ((verbose > 1) && (resid > 0))
-                pr2ws("    %s: pass-through requested %d bytes (data-in) "
-                      "but got %d bytes\n", leadin, mx_di_len, got);
-            return got;
+            if ((verbose > 1) && (resid != 0))
+                pr2ws("    %s: %s requested %d bytes (data-in) but got %d "
+                      "bytes\n", leadin, pass_through_s, mx_di_len, got);
+            if (got >= 0)
+                return got;
+            else {
+                if (verbose)
+                    pr2ws("    %s: %s can't get negative bytes, say it got "
+                          "none\n", leadin, pass_through_s);
+                return 0;
+            }
         } else
             return 0;
     case SCSI_PT_RESULT_STATUS: /* other than GOOD and CHECK CONDITION */
         sstat = get_scsi_pt_status_response(ptvp);
-        if ((SAM_STAT_RESERVATION_CONFLICT == sstat) && o_sense_cat) {
-            /* treat this SCSI status as "sense" category */
-            *o_sense_cat = SG_LIB_CAT_RES_CONFLICT;
-            return -2;
+        if (o_sense_cat) {
+            switch (sstat) {
+            case SAM_STAT_RESERVATION_CONFLICT:
+                *o_sense_cat = SG_LIB_CAT_RES_CONFLICT;
+                return -2;
+            case SAM_STAT_CONDITION_MET:
+                *o_sense_cat = SG_LIB_CAT_CONDITION_MET;
+                return -2;
+            case SAM_STAT_BUSY:
+                *o_sense_cat = SG_LIB_CAT_BUSY;
+                return -2;
+            case SAM_STAT_TASK_SET_FULL:
+                *o_sense_cat = SG_LIB_CAT_TS_FULL;
+                return -2;
+            case SAM_STAT_ACA_ACTIVE:
+                *o_sense_cat = SG_LIB_CAT_ACA_ACTIVE;
+                return -2;
+            case SAM_STAT_TASK_ABORTED:
+                *o_sense_cat = SG_LIB_CAT_TASK_ABORTED;
+                return -2;
+            default:
+                break;
+            }
         }
         if (verbose || noisy) {
             sg_get_scsi_status_str(sstat, sizeof(b), b);
@@ -253,8 +292,13 @@ sg_cmds_process_resp(struct sg_pt_base * ptvp, const char * leadin,
             get_scsi_pt_transport_err_str(ptvp, sizeof(b), b);
             pr2ws("%s: transport: %s\n", leadin, b);
         }
-        if ((SAM_STAT_CHECK_CONDITION == get_scsi_pt_status_response(ptvp))
-            && (slen > 0))
+#ifdef SG_LIB_LINUX
+        transport_sense = (slen > 0);
+#else
+        transport_sense = ((SAM_STAT_CHECK_CONDITION ==
+                            get_scsi_pt_status_response(ptvp)) && (slen > 0));
+#endif
+        if (transport_sense)
             return sg_cmds_process_helper(leadin, mx_di_len, resid, sbp,
                                           slen, noisy, verbose, o_sense_cat);
         else
@@ -266,58 +310,81 @@ sg_cmds_process_resp(struct sg_pt_base * ptvp, const char * leadin,
         }
         return -1;
     default:
-        pr2ws("%s: unknown pass through result category (%d)\n", leadin, cat);
+        pr2ws("%s: unknown %s result category (%d)\n", leadin, pass_through_s,
+               cat);
         return -1;
     }
 }
 
-/* Invokes a SCSI INQUIRY command and yields the response. Returns 0 when
- * successful, various SG_LIB_CAT_* positive values or -1 -> other errors */
-int
-sg_ll_inquiry(int sg_fd, int cmddt, int evpd, int pg_op, void * resp,
-              int mx_resp_len, int noisy, int verbose)
+bool
+sg_cmds_is_nvme(const struct sg_pt_base * ptvp)
+{
+    return pt_device_is_nvme(ptvp);
+}
+
+static struct sg_pt_base *
+create_pt_obj(const char * cname)
+{
+    struct sg_pt_base * ptvp = construct_scsi_pt_obj();
+    if (NULL == ptvp)
+        pr2ws("%s: out of memory\n", cname);
+    return ptvp;
+}
+
+static const char * const inquiry_s = "inquiry";
+
+/* Returns 0 on success, while positive values are SG_LIB_CAT_* errors
+ * (e.g. SG_LIB_CAT_MALFORMED). If OS error, returns negated errno or -1. */
+static int
+sg_ll_inquiry_com(int sg_fd, bool cmddt, bool evpd, int pg_op, void * resp,
+                  int mx_resp_len, int timeout_secs, int * residp,
+                  bool noisy, int verbose)
 {
     int res, ret, k, sense_cat, resid;
-    unsigned char inqCmdBlk[INQUIRY_CMDLEN] = {INQUIRY_CMD, 0, 0, 0, 0, 0};
-    unsigned char sense_b[SENSE_BUFF_LEN];
-    unsigned char * up;
+    uint8_t inq_cdb[INQUIRY_CMDLEN] = {INQUIRY_CMD, 0, 0, 0, 0, 0};
+    uint8_t sense_b[SENSE_BUFF_LEN];
+    uint8_t * up;
     struct sg_pt_base * ptvp;
 
     if (cmddt)
-        inqCmdBlk[1] |= 2;
+        inq_cdb[1] |= 0x2;
     if (evpd)
-        inqCmdBlk[1] |= 1;
-    inqCmdBlk[2] = (unsigned char)pg_op;
-    /* 16 bit allocation length (was 8) is a recent SPC-3 addition */
-    inqCmdBlk[3] = (unsigned char)((mx_resp_len >> 8) & 0xff);
-    inqCmdBlk[4] = (unsigned char)(mx_resp_len & 0xff);
+        inq_cdb[1] |= 0x1;
+    inq_cdb[2] = (uint8_t)pg_op;
+    /* 16 bit allocation length (was 8, increased in spc3r09, 200209) */
+    sg_put_unaligned_be16((uint16_t)mx_resp_len, inq_cdb + 3);
     if (verbose) {
-        pr2ws("    inquiry cdb: ");
+        pr2ws("    %s cdb: ", inquiry_s);
         for (k = 0; k < INQUIRY_CMDLEN; ++k)
-            pr2ws("%02x ", inqCmdBlk[k]);
+            pr2ws("%02x ", inq_cdb[k]);
         pr2ws("\n");
     }
     if (resp && (mx_resp_len > 0)) {
-        up = (unsigned char *)resp;
+        up = (uint8_t *)resp;
         up[0] = 0x7f;   /* defensive prefill */
         if (mx_resp_len > 4)
             up[4] = 0;
     }
+    if (timeout_secs <= 0)
+        timeout_secs = DEF_PT_TIMEOUT;
     ptvp = construct_scsi_pt_obj();
     if (NULL == ptvp) {
-        pr2ws("inquiry: out of memory\n");
+        pr2ws("%s: out of memory\n", __func__);
+        if (residp)
+            *residp = 0;
         return -1;
     }
-    set_scsi_pt_cdb(ptvp, inqCmdBlk, sizeof(inqCmdBlk));
+    set_scsi_pt_cdb(ptvp, inq_cdb, sizeof(inq_cdb));
     set_scsi_pt_sense(ptvp, sense_b, sizeof(sense_b));
-    set_scsi_pt_data_in(ptvp, (unsigned char *)resp, mx_resp_len);
-    res = do_scsi_pt(ptvp, sg_fd, DEF_PT_TIMEOUT, verbose);
-    ret = sg_cmds_process_resp(ptvp, "inquiry", res, mx_resp_len, sense_b,
+    set_scsi_pt_data_in(ptvp, (uint8_t *)resp, mx_resp_len);
+    res = do_scsi_pt(ptvp, sg_fd, timeout_secs, verbose);
+    ret = sg_cmds_process_resp(ptvp, inquiry_s, res, mx_resp_len, sense_b,
                                noisy, verbose, &sense_cat);
     resid = get_scsi_pt_resid(ptvp);
-    destruct_scsi_pt_obj(ptvp);
+    if (residp)
+        *residp = resid;
     if (-1 == ret)
-        ;
+        ret = sg_convert_errno(get_scsi_pt_os_err(ptvp));
     else if (-2 == ret) {
         switch (sense_cat) {
         case SG_LIB_CAT_RECOVERED:
@@ -330,79 +397,52 @@ sg_ll_inquiry(int sg_fd, int cmddt, int evpd, int pg_op, void * resp,
         }
     } else if (ret < 4) {
         if (verbose)
-            pr2ws("inquiry: got too few bytes (%d)\n", ret);
+            pr2ws("%s: got too few bytes (%d)\n", __func__, ret);
         ret = SG_LIB_CAT_MALFORMED;
     } else
         ret = 0;
+    destruct_scsi_pt_obj(ptvp);
 
     if (resid > 0) {
         if (resid > mx_resp_len) {
-            pr2ws("inquiry: resid (%d) should never exceed requested "
-                  "len=%d\n", resid, mx_resp_len);
+            pr2ws("%s resid (%d) should never exceed requested "
+                    "len=%d\n", inquiry_s, resid, mx_resp_len);
             return ret ? ret : SG_LIB_CAT_MALFORMED;
         }
-        /* zero unfilled section of response buffer */
-        memset((unsigned char *)resp + (mx_resp_len - resid), 0, resid);
+        /* zero unfilled section of response buffer, based on resid */
+        memset((uint8_t *)resp + (mx_resp_len - resid), 0, resid);
     }
     return ret;
 }
 
+/* Invokes a SCSI INQUIRY command and yields the response. Returns 0 when
+ * successful, various SG_LIB_CAT_* positive values, negated errno or
+ * -1 -> other errors. The CMDDT field is obsolete in the INQUIRY cdb. */
+int
+sg_ll_inquiry(int sg_fd, bool cmddt, bool evpd, int pg_op, void * resp,
+              int mx_resp_len, bool noisy, int verbose)
+{
+    return sg_ll_inquiry_com(sg_fd, cmddt, evpd, pg_op, resp, mx_resp_len,
+                             0 /* timeout_sec */, NULL, noisy, verbose);
+}
+
 /* Yields most of first 36 bytes of a standard INQUIRY (evpd==0) response.
- * Returns 0 when successful, various SG_LIB_CAT_* positive values or
- * -1 -> other errors */
+ * Returns 0 when successful, various SG_LIB_CAT_* positive values, negated
+ * errno or -1 -> other errors */
 int
 sg_simple_inquiry(int sg_fd, struct sg_simple_inquiry_resp * inq_data,
-                  int noisy, int verbose)
+                  bool noisy, int verbose)
 {
-    int res, ret, k, sense_cat;
-    unsigned char inqCmdBlk[INQUIRY_CMDLEN] = {INQUIRY_CMD, 0, 0, 0, 0, 0};
-    unsigned char sense_b[SENSE_BUFF_LEN];
-    unsigned char inq_resp[SAFE_STD_INQ_RESP_LEN];
-    struct sg_pt_base * ptvp;
+    int ret;
+    uint8_t inq_resp[SAFE_STD_INQ_RESP_LEN];
 
     if (inq_data) {
         memset(inq_data, 0, sizeof(* inq_data));
         inq_data->peripheral_qualifier = 0x3;
         inq_data->peripheral_type = 0x1f;
     }
-    inqCmdBlk[4] = (unsigned char)sizeof(inq_resp);
-    if (verbose) {
-        pr2ws("    inquiry cdb: ");
-        for (k = 0; k < INQUIRY_CMDLEN; ++k)
-            pr2ws("%02x ", inqCmdBlk[k]);
-        pr2ws("\n");
-    }
-    memset(inq_resp, 0, sizeof(inq_resp));
-    inq_resp[0] = 0x7f; /* defensive prefill */
-    ptvp = construct_scsi_pt_obj();
-    if (NULL == ptvp) {
-        pr2ws("inquiry: out of memory\n");
-        return -1;
-    }
-    set_scsi_pt_cdb(ptvp, inqCmdBlk, sizeof(inqCmdBlk));
-    set_scsi_pt_sense(ptvp, sense_b, sizeof(sense_b));
-    set_scsi_pt_data_in(ptvp, inq_resp, sizeof(inq_resp));
-    res = do_scsi_pt(ptvp, sg_fd, DEF_PT_TIMEOUT, verbose);
-    ret = sg_cmds_process_resp(ptvp, "inquiry", res, sizeof(inq_resp),
-                               sense_b, noisy, verbose, &sense_cat);
-    if (-1 == ret)
-        ;
-    else if (-2 == ret) {
-        switch (sense_cat) {
-        case SG_LIB_CAT_RECOVERED:
-        case SG_LIB_CAT_NO_SENSE:
-            ret = 0;
-            break;
-        default:
-            ret = sense_cat;
-            break;
-        }
-    } else if (ret < 4) {
-        if (verbose)
-            pr2ws("inquiry: got too few bytes (%d)\n", ret);
-        ret = SG_LIB_CAT_MALFORMED;
-    } else
-        ret = 0;
+    ret = sg_ll_inquiry_com(sg_fd, false, false, 0, inq_resp,
+                            sizeof(inq_resp), 0, NULL, noisy, verbose);
 
     if (inq_data && (0 == ret)) {
         inq_data->peripheral_qualifier = (inq_resp[0] >> 5) & 0x7;
@@ -417,10 +457,29 @@ sg_simple_inquiry(int sg_fd, struct sg_simple_inquiry_resp * inq_data,
         memcpy(inq_data->product, inq_resp + 16, 16);
         memcpy(inq_data->revision, inq_resp + 32, 4);
     }
-    destruct_scsi_pt_obj(ptvp);
     return ret;
 }
 
+/* Invokes a SCSI INQUIRY command and yields the response. Returns 0 when
+ * successful, various SG_LIB_CAT_* positive values or -1 -> other errors.
+ * The CMDDT field is obsolete in the INQUIRY cdb (since spc3r16 in 2003) so
+ * an argument to set it has been removed (use the REPORT SUPPORTED OPERATION
+ * CODES command instead). Adds the ability to set the command abort timeout
+ * and the ability to report the residual count. If timeout_secs is zero
+ * or less the default command abort timeout (60 seconds) is used.
+ * If residp is non-NULL then the residual value is written where residp
+ * points. A residual value of 0 implies mx_resp_len bytes have be written
+ * where resp points. If the residual value equals mx_resp_len then no
+ * bytes have been written. */
+int
+sg_ll_inquiry_v2(int sg_fd, bool evpd, int pg_op, void * resp,
+                 int mx_resp_len, int timeout_secs, int * residp,
+                 bool noisy, int verbose)
+{
+    return sg_ll_inquiry_com(sg_fd, false, evpd, pg_op, resp, mx_resp_len,
+                             timeout_secs, residp, noisy, verbose);
+}
+
 /* Invokes a SCSI TEST UNIT READY command.
  * 'pack_id' is just for diagnostics, safe to set to 0.
  * Looks for progress indicator if 'progress' non-NULL;
@@ -429,33 +488,31 @@ sg_simple_inquiry(int sg_fd, struct sg_simple_inquiry_resp * inq_data,
  * -1 -> other errors */
 int
 sg_ll_test_unit_ready_progress(int sg_fd, int pack_id, int * progress,
-                               int noisy, int verbose)
+                               bool noisy, int verbose)
 {
+    static const char * const tur_s = "test unit ready";
     int res, ret, k, sense_cat;
-    unsigned char turCmdBlk[TUR_CMDLEN] = {TUR_CMD, 0, 0, 0, 0, 0};
-    unsigned char sense_b[SENSE_BUFF_LEN];
+    uint8_t tur_cdb[TUR_CMDLEN] = {TUR_CMD, 0, 0, 0, 0, 0};
+    uint8_t sense_b[SENSE_BUFF_LEN];
     struct sg_pt_base * ptvp;
 
     if (verbose) {
-        pr2ws("    test unit ready cdb: ");
+        pr2ws("    %s cdb: ", tur_s);
         for (k = 0; k < TUR_CMDLEN; ++k)
-            pr2ws("%02x ", turCmdBlk[k]);
+            pr2ws("%02x ", tur_cdb[k]);
         pr2ws("\n");
     }
 
-    ptvp = construct_scsi_pt_obj();
-    if (NULL == ptvp) {
-        pr2ws("test unit ready: out of memory\n");
+    if (NULL == ((ptvp = create_pt_obj(tur_s))))
         return -1;
-    }
-    set_scsi_pt_cdb(ptvp, turCmdBlk, sizeof(turCmdBlk));
+    set_scsi_pt_cdb(ptvp, tur_cdb, sizeof(tur_cdb));
     set_scsi_pt_sense(ptvp, sense_b, sizeof(sense_b));
     set_scsi_pt_packet_id(ptvp, pack_id);
     res = do_scsi_pt(ptvp, sg_fd, DEF_PT_TIMEOUT, verbose);
-    ret = sg_cmds_process_resp(ptvp, "test unit ready", res, 0, sense_b,
+    ret = sg_cmds_process_resp(ptvp, tur_s, res, SG_NO_DATA_IN, sense_b,
                                noisy, verbose, &sense_cat);
     if (-1 == ret)
-        ;
+        ret = sg_convert_errno(get_scsi_pt_os_err(ptvp));
     else if (-2 == ret) {
         if (progress) {
             int slen = get_scsi_pt_sense_len(ptvp);
@@ -484,7 +541,7 @@ sg_ll_test_unit_ready_progress(int sg_fd, int pack_id, int * progress,
  * Returns 0 when successful, various SG_LIB_CAT_* positive values or
  * -1 -> other errors */
 int
-sg_ll_test_unit_ready(int sg_fd, int pack_id, int noisy, int verbose)
+sg_ll_test_unit_ready(int sg_fd, int pack_id, bool noisy, int verbose)
 {
     return sg_ll_test_unit_ready_progress(sg_fd, pack_id, NULL, noisy,
                                           verbose);
@@ -493,42 +550,40 @@ sg_ll_test_unit_ready(int sg_fd, int pack_id, int noisy, int verbose)
 /* Invokes a SCSI REQUEST SENSE command. Returns 0 when successful, various
  * SG_LIB_CAT_* positive values or -1 -> other errors */
 int
-sg_ll_request_sense(int sg_fd, int desc, void * resp, int mx_resp_len,
-                    int noisy, int verbose)
+sg_ll_request_sense(int sg_fd, bool desc, void * resp, int mx_resp_len,
+                    bool noisy, int verbose)
 {
+    static const char * const rq_s = "request sense";
     int k, ret, res, sense_cat;
-    unsigned char rsCmdBlk[REQUEST_SENSE_CMDLEN] =
+    uint8_t rs_cdb[REQUEST_SENSE_CMDLEN] =
         {REQUEST_SENSE_CMD, 0, 0, 0, 0, 0};
-    unsigned char sense_b[SENSE_BUFF_LEN];
+    uint8_t sense_b[SENSE_BUFF_LEN];
     struct sg_pt_base * ptvp;
 
     if (desc)
-        rsCmdBlk[1] |= 0x1;
+        rs_cdb[1] |= 0x1;
     if (mx_resp_len > 0xff) {
         pr2ws("mx_resp_len cannot exceed 255\n");
         return -1;
     }
-    rsCmdBlk[4] = mx_resp_len & 0xff;
+    rs_cdb[4] = mx_resp_len & 0xff;
     if (verbose) {
-        pr2ws("    Request Sense cmd: ");
+        pr2ws("    %s cmd: ", rq_s);
         for (k = 0; k < REQUEST_SENSE_CMDLEN; ++k)
-            pr2ws("%02x ", rsCmdBlk[k]);
+            pr2ws("%02x ", rs_cdb[k]);
         pr2ws("\n");
     }
 
-    ptvp = construct_scsi_pt_obj();
-    if (NULL == ptvp) {
-        pr2ws("request sense: out of memory\n");
+    if (NULL == ((ptvp = create_pt_obj(rq_s))))
         return -1;
-    }
-    set_scsi_pt_cdb(ptvp, rsCmdBlk, sizeof(rsCmdBlk));
+    set_scsi_pt_cdb(ptvp, rs_cdb, sizeof(rs_cdb));
     set_scsi_pt_sense(ptvp, sense_b, sizeof(sense_b));
-    set_scsi_pt_data_in(ptvp, (unsigned char *)resp, mx_resp_len);
+    set_scsi_pt_data_in(ptvp, (uint8_t *)resp, mx_resp_len);
     res = do_scsi_pt(ptvp, sg_fd, DEF_PT_TIMEOUT, verbose);
-    ret = sg_cmds_process_resp(ptvp, "request sense", res, mx_resp_len,
-                               sense_b, noisy, verbose, &sense_cat);
+    ret = sg_cmds_process_resp(ptvp, rq_s, res, mx_resp_len, sense_b, noisy,
+                               verbose, &sense_cat);
     if (-1 == ret)
-        ;
+        ret = sg_convert_errno(get_scsi_pt_os_err(ptvp));
     else if (-2 == ret) {
         switch (sense_cat) {
         case SG_LIB_CAT_RECOVERED:
@@ -542,8 +597,8 @@ sg_ll_request_sense(int sg_fd, int desc, void * resp, int mx_resp_len,
     } else {
         if ((mx_resp_len >= 8) && (ret < 8)) {
             if (verbose)
-                pr2ws("    request sense: got %d bytes in response, too "
-                      "short\n", ret);
+                pr2ws("    %s: got %d bytes in response, too short\n", rq_s,
+                      ret);
             ret = -1;
         } else
             ret = 0;
@@ -556,39 +611,34 @@ sg_ll_request_sense(int sg_fd, int desc, void * resp, int mx_resp_len,
  * various SG_LIB_CAT_* positive values or -1 -> other errors */
 int
 sg_ll_report_luns(int sg_fd, int select_report, void * resp, int mx_resp_len,
-                  int noisy, int verbose)
+                  bool noisy, int verbose)
 {
+    static const char * const report_luns_s = "report luns";
     int k, ret, res, sense_cat;
-    unsigned char rlCmdBlk[REPORT_LUNS_CMDLEN] =
+    uint8_t rl_cdb[REPORT_LUNS_CMDLEN] =
                          {REPORT_LUNS_CMD, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
-    unsigned char sense_b[SENSE_BUFF_LEN];
+    uint8_t sense_b[SENSE_BUFF_LEN];
     struct sg_pt_base * ptvp;
 
-    rlCmdBlk[2] = select_report & 0xff;
-    rlCmdBlk[6] = (mx_resp_len >> 24) & 0xff;
-    rlCmdBlk[7] = (mx_resp_len >> 16) & 0xff;
-    rlCmdBlk[8] = (mx_resp_len >> 8) & 0xff;
-    rlCmdBlk[9] = mx_resp_len & 0xff;
+    rl_cdb[2] = select_report & 0xff;
+    sg_put_unaligned_be32((uint32_t)mx_resp_len, rl_cdb + 6);
     if (verbose) {
-        pr2ws("    report luns cdb: ");
+        pr2ws("    %s cdb: ", report_luns_s);
         for (k = 0; k < REPORT_LUNS_CMDLEN; ++k)
-            pr2ws("%02x ", rlCmdBlk[k]);
+            pr2ws("%02x ", rl_cdb[k]);
         pr2ws("\n");
     }
 
-    ptvp = construct_scsi_pt_obj();
-    if (NULL == ptvp) {
-        pr2ws("report luns: out of memory\n");
+    if (NULL == ((ptvp = create_pt_obj(report_luns_s))))
         return -1;
-    }
-    set_scsi_pt_cdb(ptvp, rlCmdBlk, sizeof(rlCmdBlk));
+    set_scsi_pt_cdb(ptvp, rl_cdb, sizeof(rl_cdb));
     set_scsi_pt_sense(ptvp, sense_b, sizeof(sense_b));
-    set_scsi_pt_data_in(ptvp, (unsigned char *)resp, mx_resp_len);
+    set_scsi_pt_data_in(ptvp, (uint8_t *)resp, mx_resp_len);
     res = do_scsi_pt(ptvp, sg_fd, DEF_PT_TIMEOUT, verbose);
-    ret = sg_cmds_process_resp(ptvp, "report luns", res, mx_resp_len,
+    ret = sg_cmds_process_resp(ptvp, report_luns_s, res, mx_resp_len,
                                sense_b, noisy, verbose, &sense_cat);
     if (-1 == ret)
-        ;
+        ret = sg_convert_errno(get_scsi_pt_os_err(ptvp));
     else if (-2 == ret) {
         switch (sense_cat) {
         case SG_LIB_CAT_RECOVERED:
