@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2008-2015 Douglas Gilbert.
+ * Copyright (c) 2008-2018 Douglas Gilbert.
  * All rights reserved.
  * Use of this source code is governed by a BSD-style
  * license that can be found in the BSD_LICENSE file.
@@ -13,14 +13,15 @@
 #define __STDC_FORMAT_MACROS 1
 #include <inttypes.h>
 
+#ifdef HAVE_CONFIG_H
+#include "config.h"
+#endif
+
 #include "sg_lib.h"
 #include "sg_cmds_basic.h"
 #include "sg_cmds_mmc.h"
 #include "sg_pt.h"
-
-#ifdef HAVE_CONFIG_H
-#include "config.h"
-#endif
+#include "sg_unaligned.h"
 
 
 #define SENSE_BUFF_LEN 64       /* Arbitrary, could be larger */
@@ -36,7 +37,7 @@
 #define SET_STREAMING_CMD 0xb6
 #define SET_STREAMING_CMDLEN 12
 
-#ifdef __GNUC__
+#if defined(__GNUC__) || defined(__clang__)
 static int pr2ws(const char * fmt, ...)
         __attribute__ ((format (printf, 1, 2)));
 #else
@@ -56,6 +57,15 @@ pr2ws(const char * fmt, ...)
     return n;
 }
 
+static struct sg_pt_base *
+create_pt_obj(const char * cname)
+{
+    struct sg_pt_base * ptvp = construct_scsi_pt_obj();
+    if (NULL == ptvp)
+        pr2ws("%s: out of memory\n", cname);
+    return ptvp;
+}
+
 /* Invokes a SCSI SET CD SPEED command (MMC).
  * Return of 0 -> success, SG_LIB_CAT_INVALID_OP -> command not supported,
  * SG_LIB_CAT_ILLEGAL_REQ -> bad field in cdb, SG_LIB_CAT_UNIT_ATTENTION,
@@ -63,38 +73,34 @@ pr2ws(const char * fmt, ...)
  * -1 -> other failure */
 int
 sg_ll_set_cd_speed(int sg_fd, int rot_control, int drv_read_speed,
-                   int drv_write_speed, int noisy, int verbose)
+                   int drv_write_speed, bool noisy, int verbose)
 {
+    static const char * const cdb_name_s = "set cd speed";
     int res, ret, k, sense_cat;
-    unsigned char scsCmdBlk[SET_CD_SPEED_CMDLEN] = {SET_CD_SPEED_CMD, 0,
+    uint8_t scsCmdBlk[SET_CD_SPEED_CMDLEN] = {SET_CD_SPEED_CMD, 0,
                                          0, 0, 0, 0, 0, 0, 0, 0, 0 ,0};
-    unsigned char sense_b[SENSE_BUFF_LEN];
+    uint8_t sense_b[SENSE_BUFF_LEN];
     struct sg_pt_base * ptvp;
 
     scsCmdBlk[1] |= (rot_control & 0x3);
-    scsCmdBlk[2] = (drv_read_speed >> 8) & 0xff;
-    scsCmdBlk[3] = drv_read_speed & 0xff;
-    scsCmdBlk[4] = (drv_write_speed >> 8) & 0xff;
-    scsCmdBlk[5] = drv_write_speed & 0xff;
+    sg_put_unaligned_be16((uint16_t)drv_read_speed, scsCmdBlk + 2);
+    sg_put_unaligned_be16((uint16_t)drv_write_speed, scsCmdBlk + 4);
 
     if (verbose) {
-        pr2ws("    set cd speed cdb: ");
+        pr2ws("    %s cdb: ", cdb_name_s);
         for (k = 0; k < SET_CD_SPEED_CMDLEN; ++k)
             pr2ws("%02x ", scsCmdBlk[k]);
         pr2ws("\n");
     }
-    ptvp = construct_scsi_pt_obj();
-    if (NULL == ptvp) {
-        pr2ws("set cd speed: out of memory\n");
+    if (NULL == ((ptvp = create_pt_obj(cdb_name_s))))
         return -1;
-    }
     set_scsi_pt_cdb(ptvp, scsCmdBlk, sizeof(scsCmdBlk));
     set_scsi_pt_sense(ptvp, sense_b, sizeof(sense_b));
     res = do_scsi_pt(ptvp, sg_fd, DEF_PT_TIMEOUT, verbose);
-    ret = sg_cmds_process_resp(ptvp, "set cd speed", res, 0,
-                               sense_b, noisy, verbose, &sense_cat);
+    ret = sg_cmds_process_resp(ptvp, cdb_name_s, res, SG_NO_DATA_IN, sense_b,
+                               noisy, verbose, &sense_cat);
     if (-1 == ret)
-        ;
+        ret = sg_convert_errno(get_scsi_pt_os_err(ptvp));
     else if (-2 == ret) {
         switch (sense_cat) {
         case SG_LIB_CAT_NOT_READY:
@@ -125,12 +131,13 @@ sg_ll_set_cd_speed(int sg_fd, int rot_control, int drv_read_speed,
  * SG_LIB_CAT_UNIT_ATTENTION, SG_LIB_CAT_ABORTED_COMMAND, else -1 */
 int
 sg_ll_get_config(int sg_fd, int rt, int starting, void * resp,
-                 int mx_resp_len, int noisy, int verbose)
+                 int mx_resp_len, bool noisy, int verbose)
 {
+    static const char * const cdb_name_s = "get configuration";
     int res, k, ret, sense_cat;
-    unsigned char gcCmdBlk[GET_CONFIG_CMD_LEN] = {GET_CONFIG_CMD, 0, 0, 0,
+    uint8_t gcCmdBlk[GET_CONFIG_CMD_LEN] = {GET_CONFIG_CMD, 0, 0, 0,
                                                   0, 0, 0, 0, 0, 0};
-    unsigned char sense_b[SENSE_BUFF_LEN];
+    uint8_t sense_b[SENSE_BUFF_LEN];
     struct sg_pt_base * ptvp;
 
     if ((rt < 0) || (rt > 3)) {
@@ -142,35 +149,30 @@ sg_ll_get_config(int sg_fd, int rt, int starting, void * resp,
         pr2ws("Bad starting field number: 0x%x\n", starting);
         return -1;
     }
-    gcCmdBlk[2] = (unsigned char)((starting >> 8) & 0xff);
-    gcCmdBlk[3] = (unsigned char)(starting & 0xff);
+    sg_put_unaligned_be16((uint16_t)starting, gcCmdBlk + 2);
     if ((mx_resp_len < 0) || (mx_resp_len > 0xffff)) {
         pr2ws("Bad mx_resp_len: 0x%x\n", starting);
         return -1;
     }
-    gcCmdBlk[7] = (unsigned char)((mx_resp_len >> 8) & 0xff);
-    gcCmdBlk[8] = (unsigned char)(mx_resp_len & 0xff);
+    sg_put_unaligned_be16((uint16_t)mx_resp_len, gcCmdBlk + 7);
 
     if (verbose) {
-        pr2ws("    Get Configuration cdb: ");
+        pr2ws("    %s cdb: ", cdb_name_s);
         for (k = 0; k < GET_CONFIG_CMD_LEN; ++k)
             pr2ws("%02x ", gcCmdBlk[k]);
         pr2ws("\n");
     }
 
-    ptvp = construct_scsi_pt_obj();
-    if (NULL == ptvp) {
-        pr2ws("get configuration: out of memory\n");
+    if (NULL == ((ptvp = create_pt_obj(cdb_name_s))))
         return -1;
-    }
     set_scsi_pt_cdb(ptvp, gcCmdBlk, sizeof(gcCmdBlk));
     set_scsi_pt_sense(ptvp, sense_b, sizeof(sense_b));
-    set_scsi_pt_data_in(ptvp, (unsigned char *)resp, mx_resp_len);
+    set_scsi_pt_data_in(ptvp, (uint8_t *)resp, mx_resp_len);
     res = do_scsi_pt(ptvp, sg_fd, DEF_PT_TIMEOUT, verbose);
-    ret = sg_cmds_process_resp(ptvp, "get configuration", res, mx_resp_len,
+    ret = sg_cmds_process_resp(ptvp, cdb_name_s, res, mx_resp_len,
                                sense_b, noisy, verbose, &sense_cat);
     if (-1 == ret)
-        ;
+        ret = sg_convert_errno(get_scsi_pt_os_err(ptvp));
     else if (-2 == ret) {
         switch (sense_cat) {
         case SG_LIB_CAT_INVALID_OP:
@@ -189,18 +191,23 @@ sg_ll_get_config(int sg_fd, int rt, int starting, void * resp,
         }
     } else {
         if ((verbose > 2) && (ret > 3)) {
-            unsigned char * ucp;
+            uint8_t * bp;
             int len;
 
-            ucp = (unsigned char *)resp;
-            len = (ucp[0] << 24) + (ucp[1] << 16) + (ucp[2] << 8) + ucp[3] +
-                  4;
+            bp = (uint8_t *)resp;
+            len = sg_get_unaligned_be32(bp + 0);
             if (len < 0)
                 len = 0;
             len = (ret < len) ? ret : len;
-            pr2ws("    get configuration: response%s\n",
-                  (len > 256 ? ", first 256 bytes" : ""));
-            dStrHexErr((const char *)resp, (len > 256 ? 256 : len), -1);
+            pr2ws("    %s: response:\n", cdb_name_s);
+            if (3 == verbose) {
+                pr2ws("%s:\n", (len > 256 ? ", first 256 bytes" : ""));
+                hex2stderr((const uint8_t *)resp, (len > 256 ? 256 : len),
+                           -1);
+            } else {
+                pr2ws(":\n");
+                hex2stderr((const uint8_t *)resp, len, 0);
+            }
         }
         ret = 0;
     }
@@ -215,12 +222,13 @@ sg_ll_get_config(int sg_fd, int rt, int starting, void * resp,
 int
 sg_ll_get_performance(int sg_fd, int data_type, unsigned int starting_lba,
                       int max_num_desc, int ttype, void * resp,
-                      int mx_resp_len, int noisy, int verbose)
+                      int mx_resp_len, bool noisy, int verbose)
 {
+    static const char * const cdb_name_s = "get performance";
     int res, k, ret, sense_cat;
-    unsigned char gpCmdBlk[GET_PERFORMANCE_CMD_LEN] = {GET_PERFORMANCE_CMD, 0,
+    uint8_t gpCmdBlk[GET_PERFORMANCE_CMD_LEN] = {GET_PERFORMANCE_CMD, 0,
                                         0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
-    unsigned char sense_b[SENSE_BUFF_LEN];
+    uint8_t sense_b[SENSE_BUFF_LEN];
     struct sg_pt_base * ptvp;
 
     if ((data_type < 0) || (data_type > 0x1f)) {
@@ -228,42 +236,35 @@ sg_ll_get_performance(int sg_fd, int data_type, unsigned int starting_lba,
         return -1;
     }
     gpCmdBlk[1] = (data_type & 0x1f);
-    gpCmdBlk[2] = (unsigned char)((starting_lba >> 24) & 0xff);
-    gpCmdBlk[3] = (unsigned char)((starting_lba >> 16) & 0xff);
-    gpCmdBlk[4] = (unsigned char)((starting_lba >> 8) & 0xff);
-    gpCmdBlk[3] = (unsigned char)(starting_lba & 0xff);
+    sg_put_unaligned_be32((uint32_t)starting_lba, gpCmdBlk + 2);
     if ((max_num_desc < 0) || (max_num_desc > 0xffff)) {
         pr2ws("Bad max_num_desc: 0x%x\n", max_num_desc);
         return -1;
     }
-    gpCmdBlk[8] = (unsigned char)((max_num_desc >> 8) & 0xff);
-    gpCmdBlk[9] = (unsigned char)(max_num_desc & 0xff);
+    sg_put_unaligned_be16((uint16_t)max_num_desc, gpCmdBlk + 8);
     if ((ttype < 0) || (ttype > 0xff)) {
         pr2ws("Bad type: 0x%x\n", ttype);
         return -1;
     }
-    gpCmdBlk[10] = (unsigned char)ttype;
+    gpCmdBlk[10] = (uint8_t)ttype;
 
     if (verbose) {
-        pr2ws("    Get Performance cdb: ");
+        pr2ws("    %s cdb: ", cdb_name_s);
         for (k = 0; k < GET_PERFORMANCE_CMD_LEN; ++k)
             pr2ws("%02x ", gpCmdBlk[k]);
         pr2ws("\n");
     }
 
-    ptvp = construct_scsi_pt_obj();
-    if (NULL == ptvp) {
-        pr2ws("get performance: out of memory\n");
+    if (NULL == ((ptvp = create_pt_obj(cdb_name_s))))
         return -1;
-    }
     set_scsi_pt_cdb(ptvp, gpCmdBlk, sizeof(gpCmdBlk));
     set_scsi_pt_sense(ptvp, sense_b, sizeof(sense_b));
-    set_scsi_pt_data_in(ptvp, (unsigned char *)resp, mx_resp_len);
+    set_scsi_pt_data_in(ptvp, (uint8_t *)resp, mx_resp_len);
     res = do_scsi_pt(ptvp, sg_fd, DEF_PT_TIMEOUT, verbose);
-    ret = sg_cmds_process_resp(ptvp, "get performance", res, mx_resp_len,
-                               sense_b, noisy, verbose, &sense_cat);
+    ret = sg_cmds_process_resp(ptvp, cdb_name_s, res, mx_resp_len, sense_b,
+                               noisy, verbose, &sense_cat);
     if (-1 == ret)
-        ;
+        ret = sg_convert_errno(get_scsi_pt_os_err(ptvp));
     else if (-2 == ret) {
         switch (sense_cat) {
         case SG_LIB_CAT_INVALID_OP:
@@ -282,18 +283,23 @@ sg_ll_get_performance(int sg_fd, int data_type, unsigned int starting_lba,
         }
     } else {
         if ((verbose > 2) && (ret > 3)) {
-            unsigned char * ucp;
+            uint8_t * bp;
             int len;
 
-            ucp = (unsigned char *)resp;
-            len = (ucp[0] << 24) + (ucp[1] << 16) + (ucp[2] << 8) + ucp[3] +
-                  4;
+            bp = (uint8_t *)resp;
+            len = sg_get_unaligned_be32(bp + 0);
             if (len < 0)
                 len = 0;
             len = (ret < len) ? ret : len;
-            pr2ws("    get performance:: response%s\n",
-                  (len > 256 ? ", first 256 bytes" : ""));
-            dStrHexErr((const char *)resp, (len > 256 ? 256 : len), -1);
+            pr2ws("    %s: response", cdb_name_s);
+            if (3 == verbose) {
+                pr2ws("%s:\n", (len > 256 ? ", first 256 bytes" : ""));
+                hex2stderr((const uint8_t *)resp, (len > 256 ? 256 : len),
+                           -1);
+            } else {
+                pr2ws(":\n");
+                hex2stderr((const uint8_t *)resp, len, 0);
+            }
         }
         ret = 0;
     }
@@ -308,41 +314,38 @@ sg_ll_get_performance(int sg_fd, int data_type, unsigned int starting_lba,
  * -1 -> other failure */
 int
 sg_ll_set_streaming(int sg_fd, int type, void * paramp, int param_len,
-                    int noisy, int verbose)
+                    bool noisy, int verbose)
 {
+    static const char * const cdb_name_s = "set streaming";
     int k, res, ret, sense_cat;
-    unsigned char ssCmdBlk[SET_STREAMING_CMDLEN] =
+    uint8_t ssCmdBlk[SET_STREAMING_CMDLEN] =
                  {SET_STREAMING_CMD, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
-    unsigned char sense_b[SENSE_BUFF_LEN];
+    uint8_t sense_b[SENSE_BUFF_LEN];
     struct sg_pt_base * ptvp;
 
     ssCmdBlk[8] = type;
-    ssCmdBlk[9] = (param_len >> 8) & 0xff;
-    ssCmdBlk[10] = param_len & 0xff;
+    sg_put_unaligned_be16((uint16_t)param_len, ssCmdBlk + 9);
     if (verbose) {
-        pr2ws("    set streaming cdb: ");
+        pr2ws("    %s cdb: ", cdb_name_s);
         for (k = 0; k < SET_STREAMING_CMDLEN; ++k)
             pr2ws("%02x ", ssCmdBlk[k]);
         pr2ws("\n");
         if ((verbose > 1) && paramp && param_len) {
-            pr2ws("    set streaming parameter list:\n");
-            dStrHexErr((const char *)paramp, param_len, -1);
+            pr2ws("    %s parameter list:\n", cdb_name_s);
+            hex2stderr((const uint8_t *)paramp, param_len, -1);
         }
     }
 
-    ptvp = construct_scsi_pt_obj();
-    if (NULL == ptvp) {
-        pr2ws("set streaming: out of memory\n");
+    if (NULL == ((ptvp = create_pt_obj(cdb_name_s))))
         return -1;
-    }
     set_scsi_pt_cdb(ptvp, ssCmdBlk, sizeof(ssCmdBlk));
     set_scsi_pt_sense(ptvp, sense_b, sizeof(sense_b));
-    set_scsi_pt_data_out(ptvp, (unsigned char *)paramp, param_len);
+    set_scsi_pt_data_out(ptvp, (uint8_t *)paramp, param_len);
     res = do_scsi_pt(ptvp, sg_fd, DEF_PT_TIMEOUT, verbose);
-    ret = sg_cmds_process_resp(ptvp, "set streaming", res, 0,
-                               sense_b, noisy, verbose, &sense_cat);
+    ret = sg_cmds_process_resp(ptvp, cdb_name_s, res, SG_NO_DATA_IN, sense_b,
+                               noisy, verbose, &sense_cat);
     if (-1 == ret)
-        ;
+        ret = sg_convert_errno(get_scsi_pt_os_err(ptvp));
     else if (-2 == ret) {
         switch (sense_cat) {
         case SG_LIB_CAT_NOT_READY:
