@@ -1,11 +1,10 @@
 #!/bin/bash
-# Skript to rescan SCSI bus, using the 
-# scsi add-single-device mechanism
+# Script to rescan SCSI bus, using the scsi add-single-device mechanism.
 # (c) 1998--2010 Kurt Garloff <kurt@garloff.de>, GNU GPL v2 or v3
 # (c) 2006--2015 Hannes Reinecke, GNU GPL v2 or later
 # $Id: rescan-scsi-bus.sh,v 1.57 2012/03/31 14:08:48 garloff Exp $
 
-VERSION="20160201"
+VERSION="20160511"
 SCAN_WILD_CARD=4294967295
 
 setcolor ()
@@ -17,7 +16,7 @@ setcolor ()
   norm="\e[0;0m"
 }
 
-unsetcolor () 
+unsetcolor ()
 {
   red=""; green=""
   yellow=""; norm=""
@@ -44,11 +43,10 @@ print_and_scroll_back ()
   return $LN
 }
 
-# Overwrite a text of length $1 (fallback to $LN) with whitespace
+# Overwrite a text of length $LN with whitespace
 white_out ()
 {
   BK=""; WH=""
-  if test -n "$1"; then LN=$1; fi
   declare -i cntr=0
   while test $cntr -lt $LN; do BK="$BK\e[D"; WH="$WH "; let cntr+=1; done
   echo -en "$WH$BK"
@@ -57,9 +55,9 @@ white_out ()
 # Return hosts. sysfs must be mounted
 findhosts_26 ()
 {
-  hosts=`find /sys/class/scsi_host/host* -maxdepth 4 -type d -o -type l 2> /dev/null | awk -F'/' '{print $5}' | sed -e 's~host~~' | sort -nu` 
-  scsi_host_data=`echo "$hosts" | sed -e 's~^~/sys/class/scsi_host/host~'` 
-  for hostdir in $scsi_host_data; do 
+  hosts=`find /sys/class/scsi_host/host* -maxdepth 4 -type d -o -type l 2> /dev/null | awk -F'/' '{print $5}' | sed -e 's~host~~' | sort -nu`
+  scsi_host_data=`echo "$hosts" | sed -e 's~^~/sys/class/scsi_host/host~'`
+  for hostdir in $scsi_host_data; do
     hostno=${hostdir#/sys/class/scsi_host/host}
     if [ -f $hostdir/isp_name ] ; then
       hostname="qla2xxx"
@@ -70,7 +68,7 @@ findhosts_26 ()
     fi
     #hosts="$hosts $hostno"
     echo_debug "Host adapter $hostno ($hostname) found."
-  done  
+  done
   if [ -z "$hosts" ] ; then
     echo "No SCSI host adapters found in sysfs"
     exit 1;
@@ -91,8 +89,8 @@ findhosts ()
       if test $name = add_map -o $name = map -o $name = mod_parm; then continue; fi
       num=$name
       driverinfo=$driver
-      if test -r $hostdir/status; then
-        num=$(printf '%d\n' `sed -n 's/SCSI host number://p' $hostdir/status`)
+      if test -r "$hostdir/status"; then
+        num=$(printf '%d\n' "$(sed -n 's/SCSI host number://p' "$hostdir/status")")
         driverinfo="$driver:$name"
       fi
       hosts="$hosts $num"
@@ -127,7 +125,7 @@ printtype ()
 
 print02i()
 {
-    if [ "$1" = "*" ] ; then 
+    if [ "$1" = "*" ] ; then
         echo "00"
     else
         printf "%02i" "$1"
@@ -135,10 +133,10 @@ print02i()
 }
 
 # Get /proc/scsi/scsi info for device $host:$channel:$id:$lun
-# Optional parameter: Number of lines after first (default = 2), 
+# Optional parameter: Number of lines after first (default = 2),
 # result in SCSISTR, return code 1 means empty.
 procscsiscsi ()
-{  
+{
   if test -z "$1"; then LN=2; else LN=$1; fi
   CHANNEL=`print02i "$channel"`
   ID=`print02i "$id"`
@@ -168,7 +166,7 @@ $SCSITMP"
     fi
   else
     grepstr="scsi$host Channel: $CHANNEL Id: $ID Lun: $LUN"
-    SCSISTR=`cat /proc/scsi/scsi | grep -A$LN -e"$grepstr"`
+    SCSISTR=$(grep -A "$LN" -e "$grepstr" /proc/scsi/scsi)
   fi
   if test -z "$SCSISTR"; then return 1; else return 0; fi
 }
@@ -176,9 +174,11 @@ $SCSITMP"
 # Find sg device with 2.6 sysfs support
 sgdevice26 ()
 {
-  if test -e /sys/class/scsi_device/$host\:$channel\:$id\:$lun/device/generic; then
-    SGDEV=`readlink /sys/class/scsi_device/$host\:$channel\:$id\:$lun/device/generic`
-    SGDEV=`basename $SGDEV`
+  local gendev
+
+  gendev=/sys/class/scsi_device/${host}:${channel}:${id}:${lun}/device/generic
+  if test -e "$gendev"; then
+    SGDEV=$(basename "$(readlink "$gendev")")
   else
     for SGDEV in /sys/class/scsi_generic/sg*; do
       DEV=`readlink $SGDEV/device`
@@ -187,7 +187,7 @@ sgdevice26 ()
       fi
     done
     SGDEV=""
-  fi  
+  fi
 }
 
 # Find sg device with 2.4 report-devs extensions
@@ -205,7 +205,7 @@ sgdevice ()
   SGDEV=
   if test -d /sys/class/scsi_device; then
     sgdevice26
-  else  
+  else
     DRV=`grep 'Attached drivers:' /proc/scsi/scsi 2>/dev/null`
     repdevstat=$((1-$?))
     if [ $repdevstat = 0 ]; then
@@ -213,7 +213,7 @@ sgdevice ()
       DRV=`grep 'Attached drivers:' /proc/scsi/scsi 2>/dev/null`
       if [ $? = 1 ]; then return; fi
     fi
-    if ! `echo $DRV | grep 'drivers: sg' >/dev/null`; then
+    if ! echo "$DRV" | grep -q 'drivers: sg'; then
       modprobe sg
     fi
     sgdevice24
@@ -223,6 +223,26 @@ sgdevice ()
   fi
 }
 
+# Whether or not the RMB (removable) bit has been set in the INQUIRY response.
+# Uses ${host}, ${channel}, ${id} and ${lun}. Assumes that sg_device() has
+# already been called. How to test this function: copy/paste this function
+# in a shell and run
+# (cd /sys/class/scsi_device && for d in *; do set ${d//:/ }; echo -n "$d $(</sys/class/scsi_device/${d}/device/block/*/removable) <> "; SGDEV=bsg/$d host=$1 channel=$2 id=$3 lun=$4 is_removable; done)
+is_removable ()
+{
+  local b p
+
+  p=/sys/class/scsi_device/${host}:${channel}:${id}:${lun}/device/inquiry
+  # Extract the second byte of the INQUIRY response and check bit 7 (mask 0x80).
+  b=$(od -tx1 -j1 -N1 "$p" 2>/dev/null |
+           { read -r offset byte rest; echo -n "$byte"; })
+  if [ -n "$b" ]; then
+    echo $(((0x$b & 0x80) != 0))
+  else
+    sg_inq /dev/$SGDEV 2>/dev/null | sed -n 's/^.*RMB=\([0-9]*\).*$/\1/p'
+  fi
+}
+
 # Test if SCSI device is still responding to commands
 # Return values:
 #   0 device is present
@@ -230,7 +250,10 @@ sgdevice ()
 #   2 device has been removed
 testonline ()
 {
+  local ctr RC RMB
+
   : testonline
+  ctr=0
   RC=0
   # Set default values
   IPTYPE=31
@@ -242,7 +265,7 @@ testonline ()
   RC=$?
 
   # Handle in progress of becoming ready and unit attention
-  while test $RC = 2 -o $RC = 6 && test $ctr -le 30; do
+  while [ $RC = 2 -o $RC = 6 ] && [ $ctr -lt $timeout ] ; do
     if test $RC = 2 -a "$RMB" != "1"; then echo -n "."; let LN+=1; sleep 1
     else sleep 0.02; fi
     let ctr+=1
@@ -250,7 +273,7 @@ testonline ()
     RC=$?
     # Check for removable device; TEST UNIT READY obviously will
     # fail for a removable device with no medium
-    RMB=`sg_inq /dev/$SGDEV 2>/dev/null | grep 'RMB=' | sed 's/^.*RMB=\(.\).*$/\1/'`
+    RMB=$(is_removable)
     print_and_scroll_back "$host:$channel:$id:$lun $SGDEV ($RMB) "
     test $RC = 2 -a "$RMB" = "1" && break
   done
@@ -284,9 +307,19 @@ testonline ()
     return 2
   fi
   TMPSTR=`echo "$SCSISTR" | grep 'Vendor:'`
-  if [ "$TMPSTR" != "$STR" ]; then
-    echo -e "\e[A\e[A\e[A\e[A${red}$SGDEV changed: ${bold}\nfrom:${SCSISTR#* } \nto: $STR ${norm} \n\n\n"
-    return 1
+  if test $ignore_rev -eq 0 ; then
+      if [ "$TMPSTR" != "$STR" ]; then
+        echo -e "\e[A\e[A\e[A\e[A${red}$SGDEV changed: ${bold}\nfrom:${SCSISTR#* } \nto: $STR ${norm} \n\n\n"
+        return 1
+      fi
+  else
+      # Ignore disk revision change
+      local old_str_no_rev=`echo "$TMPSTR" | sed -e 's/.\{4\}$//'`
+      local new_str_no_rev=`echo "$STR" | sed -e 's/.\{4\}$//'`
+      if [ "$old_str_no_rev" != "$new_str_no_rev" ]; then
+        echo -e "\e[A\e[A\e[A\e[A${red}$SGDEV changed: ${bold}\nfrom:${SCSISTR#* } \nto: $STR ${norm} \n\n\n"
+        return 1
+      fi
   fi
   TMPSTR=`echo "$SCSISTR" | sed -n 's/.*Type: *\(.*\) *ANSI.*/\1/p' | sed 's/ *$//g'`
   if [ "$TMPSTR" != "$TYPE" ] ; then
@@ -296,7 +329,7 @@ testonline ()
   return $RC
 }
 
-# Test if SCSI device $host $channen $id $lun exists
+# Test if SCSI device $host $channel $id $lun exists
 # Outputs description from /proc/scsi/scsi (unless arg passed)
 # Returns SCSISTR (empty if no dev)
 testexist ()
@@ -337,37 +370,37 @@ chanlist ()
 # Returns the list of existing targets per host
 idlist ()
 {
-  local hcil
-  local target
   local tmpid
   local newid
+  local oldid
 
-  idsearch=$(ls /sys/class/scsi_device/ | sed -n "s/${host}:${channel}:\([0-9]*\):[0-9]*/\1/p" | uniq)
+  oldlist=$(ls /sys/class/scsi_device/ | sed -n "s/${host}:${channel}:\([0-9]*:[0-9]*\)/\1/p" | uniq)
+  # Rescan LUN 0 to check if we found new targets
   echo "${channel} - -" > /sys/class/scsi_host/host${host}/scan
-  # Rescan to check if we found new targets
-  newsearch=$(ls /sys/class/scsi_device/ | sed -n "s/${host}:${channel}:\([0-9]*\):[0-9]*/\1/p" | uniq)
-  for id in $newsearch ; do
-    newid=$id
-    for tmpid in $idsearch ; do
-      if test $id -eq $tmpid ; then
-        newid=
+  newlist=$(ls /sys/class/scsi_device/ | sed -n "s/${host}:${channel}:\([0-9]*:[0-9]*\)/\1/p" | uniq)
+  for newid in $newlist ; do
+    oldid=$newid
+    for tmpid in $oldlist ; do
+      if test $newid = $tmpid ; then
+        oldid=
         break
       fi
     done
-    if test -n "$newid" ; then
-      id=$newid
-      for dev in /sys/class/scsi_device/${host}:${channel}:${newid}:* ; do
-        [ -d $dev ] || continue;
-        hcil=${dev##*/}
-        lun=${hcil##*:}
+    if test -n "$oldid" ; then
+      id=${oldid%%:*}
+      lun=${oldid##*:}
+      dev=/sys/class/scsi_device/${host}:${channel}:${id}:${lun}
+      if [ -d $dev ] ; then
+	hcil=${dev##*/}
         printf "\r${green}NEW: $norm"
         testexist
         if test "$SCSISTR" ; then
           incrfound "$hcil"
         fi
-      done
+      fi
     fi
   done
+  idsearch=$(ls /sys/class/scsi_device/ | sed -n "s/${host}:${channel}:\([0-9]*\):[0-9]*/\1/p" | uniq)
 }
 
 # Returns the list of existing LUNs from device $host $channel $id $lun
@@ -382,9 +415,9 @@ getluns()
   if test $? != 0 -o -z "$LLUN"; then echo 0; return 1; fi
   for lun in $LLUN ; do
       # Swap LUN number
-      l0=$(printf '%u' 0x$lun)
+      l0=0x$lun
       l1=$(( ($l0 >> 48) & 0xffff ))
-      l2=$(( ($l0 >> 32) & 0xffff )) 
+      l2=$(( ($l0 >> 32) & 0xffff ))
       l3=$(( ($l0 >> 16) & 0xffff ))
       l4=$(( $l0 & 0xffff ))
       l0=$(( ( ( ($l4 * 0xffff) + $l3 ) * 0xffff + $l2 ) * 0xffff + $l1 ))
@@ -397,7 +430,7 @@ getluns()
 udevadm_settle()
 {
   local tmo=60
-  if test -x /sbin/udevadm; then 
+  if test -x /sbin/udevadm; then
     print_and_scroll_back " Calling udevadm settle (can take a while) "
     # Loop for up to 60 seconds if sd devices still are settling..
     # This allows us to continue if udev events are stuck on multipaths in recovery mode
@@ -494,7 +527,7 @@ dolunscan()
     if test -z "$SCSISTR" -a $RC != 1 -a "$remappedlun0" != "1"; then
       printf "\r${red}DEL: $norm\r\n\n"
       # In the event we're replacing with a well known node, we need to let it continue, to create the replacement node
-      test "$remappedlun0" != "2" && return 1
+      test "$remappedlun0" != "2" && return 2
     fi
   fi
   if test -z "$SCSISTR" -o -n "$remappedlun0"; then
@@ -512,15 +545,16 @@ dolunscan()
       # Device not present
       printf "\r\e[A";
       # Optimization: if lun==0, stop here (only if in non-remove mode)
-      if test $lun = 0 -a -z "$remove" -a $optscan = 1; then 
-        break;
+      if test $lun = 0 -a -z "$remove" -a $optscan = 1; then
+        return 1;
       fi
-    else 
+    else
       if test "$remappedlun0" != "2" ; then
         incrfound "$host:$channel:$id:$lun"
       fi
     fi
   fi
+  return 0;
 }
 
 # Perform report lun scan on $host $channel $id using REPORT_LUNS
@@ -562,13 +596,12 @@ doreportlun()
   REPLUNSTAT=$?
   lunremove=
   #echo "getluns reports " $targetluns
-  olddev=`find /sys/class/scsi_device/ -name $host:$channel:$id:* 2>/dev/null | sort -t: -k4 -n`
-  oldluns=`echo "$olddev" | awk -F'/' '{print $5}' | awk -F':' '{print $4}'`
+  olddev=`find /sys/class/scsi_device/ -name "$host:$channel:$id:*" 2>/dev/null | sort -t: -k4 -n`
   oldtargets="$targetluns"
   # OK -- if we don't have a LUN to send a REPORT_LUNS to, we could
   # fall back to wildcard scanning. Same thing if the device does not
   # support REPORT_LUNS
-  # TODO: We might be better off to ALWAYS use wildcard scanning if 
+  # TODO: We might be better off to ALWAYS use wildcard scanning if
   # it works
   if test "$REPLUNSTAT" = "1"; then
     if test -e /sys/class/scsi_host/host${host}/scan; then
@@ -577,7 +610,7 @@ doreportlun()
     else
       echo "scsi add-single-device $host $channel $id $SCAN_WILD_CARD" > /proc/scsi/scsi
     fi
-    targetluns=`find /sys/class/scsi_device/ -name $host:$channel:$id:* 2>/dev/null | awk -F'/' '{print $5}' | awk -F':' '{print $4}' | sort -n`
+    targetluns=`find /sys/class/scsi_device/ -name "$host:$channel:$id:*" 2>/dev/null | awk -F'/' '{print $5}' | awk -F':' '{print $4}' | sort -n`
     let found+=`echo "$targetluns" | wc -l`
     let found-=`echo "$olddev" | wc -l`
   fi
@@ -593,11 +626,12 @@ doreportlun()
       if test $tmplun -eq $lun ; then
         inlist=1
         dolunscan $lun0added
+        [ $? -eq 1 ] && break
       else
         newsearch="$newsearch $tmplun"
       fi
     done
-    # OK, we have now done a lunscan on $lun and 
+    # OK, we have now done a lunscan on $lun and
     # $newsearch is the old $targetluns without $lun
     if [ -z "$inlist" ]; then
       # Stale lun
@@ -609,6 +643,7 @@ doreportlun()
   # Add new ones and check stale ones
   for lun in $targetluns $lunremove; do
     dolunscan $lun0added
+    [ $? -eq 1 ] && break
   done
 }
 
@@ -620,7 +655,11 @@ dosearch ()
   fi
   for channel in $channelsearch; do
     if test -z "$idsearch" ; then
-      idlist
+      if test -z "$lunsearch" ; then
+	idlist
+      else
+	idsearch=$(ls /sys/class/scsi_device/ | sed -n "s/${host}:${channel}:\([0-9]*\):[0-9]*/\1/p" | uniq)
+      fi
     fi
     for id in $idsearch; do
       if test -z "$lunsearch" ; then
@@ -628,19 +667,20 @@ dosearch ()
       else
         for lun in $lunsearch; do
           dolunscan
+          [ $? -eq 1 ] && break
         done
       fi
     done
   done
 }
- 
+
 expandlist ()
 {
   list=$1
   result=""
   first=${list%%,*}
   rest=${list#*,}
-  while test ! -z "$first"; do 
+  while test ! -z "$first"; do
     beg=${first%%-*};
     if test "$beg" = "$first"; then
       result="$result $beg";
@@ -689,7 +729,17 @@ searchexisting()
     else
       match=1
     fi
-    test $match -eq 1 && doreportlun
+
+    test $match -eq 0 && continue
+
+    if [ -z "$lunsearch" ] ; then
+      doreportlun
+    else
+      for lun in $lunsearch ; do
+        dolunscan
+        [ $? -eq 1 ] && break
+      done
+    fi
   done
 }
 
@@ -701,8 +751,6 @@ findremapped()
   local sddev=
   local id_serial=
   local id_serial_old=
-  local sysfs_devpath=
-  local mpath_uuid=
   local remapped=
   mpaths=""
   local tmpfile=$(mktemp /tmp/rescan-scsi-bus.XXXXXXXX 2> /dev/null)
@@ -820,11 +868,11 @@ incrrmvd()
 findsddev()
 {
   local hctl="$1"
-  local sddev=  
+  local sddev=
 
   if test ! -e /sys/class/scsi_device/$hctl/device/block ; then
     return 1
-  fi 
+  fi
 
   sddev=`ls /sys/class/scsi_device/$hctl/device/block`
   echo $sddev
@@ -862,13 +910,13 @@ findmultipath()
   local maj_min=`cat /sys/block/$dev/dev`
   for mp in $($DMSETUP ls --target=multipath | cut -f 1) ; do
     [ "$mp" = "No" ] && break;
-    if $($DMSETUP status $mp | grep -q " $maj_min ") ; then
+    if $DMSETUP status $mp | grep -q " $maj_min "; then
       # With two arguments, look up current uuid from sysfs
       # if it doesn't match what was passed, this multipath
       # device is not updated, so this is a remapped LUN
       if [ -n "$find_mismatch" ] ; then
-        mp2=`$MULTIPATH -l $mp | egrep -o dm-[0-9]+`
-        mp2=`cat /sys/block/$mp2/dm/uuid | cut -f2 -d-`
+        mp2=$($MULTIPATH -l "$mp" | egrep -o "dm-[0-9]+")
+        mp2=$(cut -f2 -d- "/sys/block/$mp2/dm/uuid")
         if [ "$find_mismatch" != "$mp2" ] ; then
           addmpathtolist $mp
           found_dup=1
@@ -969,7 +1017,7 @@ flushmpaths()
 # Find resized luns
 findresized()
 {
-  local devs=`ls /sys/class/scsi_device/`
+  local devs=
   local size=
   local new_size=
   local sysfs_path=
@@ -979,6 +1027,14 @@ findresized()
   local mpathsize=
   declare -a mpathsizes
 
+  if [ -z "$lunsearch" ] ; then
+    devs=`ls /sys/class/scsi_device/`
+  else
+    for lun in $lunsearch ; do
+      devs="$devs `(cd /sys/class/scsi_device/ && ls -d *:${lun})`"
+    done
+  fi
+
   for hctl in $devs ; do
     sysfs_path="/sys/class/scsi_device/$hctl/device"
     if [ -d "$sysfs_path/block" ] ; then
@@ -1038,6 +1094,7 @@ if test @$1 = @--help -o @$1 = @-h -o @$1 = @-?; then
     echo " -m      update multipath devices           [default: disabled]"
     echo " -r      enables removing of devices        [default: disabled]"
     echo " -s      look for resized disks and reload associated multipath devices, if applicable"
+    echo " -t SECS timeout for testing if device is online. Test is skipped if 0 [default: 30]"
     echo " -u      look for existing disks that have been remapped"
     echo " -V      print version date then exit"
     echo " -w      scan for target device IDs 0--15   [default: 0--7]"
@@ -1046,15 +1103,15 @@ if test @$1 = @--help -o @$1 = @-h -o @$1 = @-?; then
     echo "--channels=LIST: Scan only channel(s) in LIST"
     echo "--color:         use coloured prefixes OLD/NEW/DEL"
     echo "--flush:         same as -f"
-    echo "--forceremove:   Remove and readd every device (DANGEROUS)"
-    echo "--forcerescan:   Rescan existing devices"
+    echo "--forceremove:   Remove stale devices (DANGEROUS)"
+    echo "--forcerescan:   Remove and readd existing devices (DANGEROUS)"
     echo "--help:          print this usage message then exit"
     echo "--hosts=LIST:    Scan only host(s) in LIST"
     echo "--ids=LIST:      Scan only target ID(s) in LIST"
     echo "--issue-lip:     same as -i"
     echo "--issue-lip-wait=SECS:     same as -I"
     echo "--largelun:      Tell kernel to support LUNs > 7 even on SCSI2 devs"
-    echo "--luns=LIST:     Scan only lun(s) in LIST"  
+    echo "--luns=LIST:     Scan only lun(s) in LIST"
     echo "--multipath:     same as -m"
     echo "--nooptscan:     don't stop looking for LUNs is 0 is not found"
     echo "--remove:        same as -r"
@@ -1062,9 +1119,11 @@ if test @$1 = @--help -o @$1 = @-h -o @$1 = @-?; then
     echo "--resize:        same as -s"
     echo "--sparselun:     Tell kernel to support sparse LUN numbering"
     echo "--sync/nosync:   Issue a sync / no sync [default: sync if remove]"
+    echo "--timeout=SECS:  same as -t"
     echo "--update:        same as -u"
     echo "--version:       same as -V"
     echo "--wide:          same as -w"
+    echo "--ignore-rev:    Ignore the revision change"
     echo ""
     echo "Host numbers may thus be specified either directly on cmd line (deprecated)"
     echo "or with the --hosts=LIST parameter (recommended)."
@@ -1102,8 +1161,8 @@ if test -x /usr/bin/sg_inq; then
   fi
 else
   echo "WARN: /usr/bin/sg_inq not present -- please install sg3_utils"
-  echo " or rescan-scsi-bus.sh might not fully work."     
-fi    
+  echo " or rescan-scsi-bus.sh might not fully work."
+fi
 
 # defaults
 unsetcolor
@@ -1122,7 +1181,9 @@ sync=1
 existing_targets=1
 mp_enable=
 lipreset=-1
+timeout=30
 declare -i scan_flags=0
+ignore_rev=0
 
 # Scan options
 opt="$1"
@@ -1139,6 +1200,7 @@ while test ! -z "$opt" -a -z "${opt##-*}"; do
     c) opt_channelsearch="0 1" ;;
     r) remove=1 ;;
     s) resize=1; mp_enable=1 ;;
+    t) timeout=$2; shift ;;
     i) lipreset=0 ;;
     I) shift; lipreset=$opt ;;
     u) update=1 ;;
@@ -1148,41 +1210,43 @@ while test ! -z "$opt" -a -z "${opt##-*}"; do
     -forcerescan) remove=1; forcerescan=1 ;;
     -forceremove) remove=1; forceremove=1 ;;
     -hosts=*)     arg=${opt#-hosts=};   hosts=`expandlist $arg` ;;
-    -channels=*)  arg=${opt#-channels=};opt_channelsearch=`expandlist $arg` ;; 
+    -channels=*)  arg=${opt#-channels=};opt_channelsearch=`expandlist $arg` ;;
     -ids=*)   arg=${opt#-ids=};         opt_idsearch=`expandlist $arg` ; filter_ids=1;;
-    -luns=*)  arg=${opt#-luns=};        lunsearch=`expandlist $arg` ;; 
+    -luns=*)  arg=${opt#-luns=};        lunsearch=`expandlist $arg` ;;
     -color) setcolor ;;
     -nooptscan) optscan=0 ;;
     -issue-lip) lipreset=0 ;;
-    -issue-lip-wait) lipreset=${opt#-issue-lip-wait=};;
+    -issue-lip-wait=*) lipreset=${opt#-issue-lip-wait=};;
     -sync) sync=2 ;;
     -nosync) sync=0 ;;
     -multipath) mp_enable=1 ;;
     -attachpq3) scan_flags=$(($scan_flags|0x1000000)) ;;
     -reportlun2) scan_flags=$(($scan_flags|0x20000)) ;;
     -resize) resize=1;;
+    -timeout=*) timeout=${opt#-timeout=};;
     -largelun) scan_flags=$(($scan_flags|0x200)) ;;
     -sparselun) scan_flags=$((scan_flags|0x40)) ;;
     -update) update=1;;
     -wide) opt_idsearch=`seq 0 15` ;;
+    -ignore-rev) ignore_rev=1;;
     *) echo "Unknown option -$opt !" ;;
   esac
   shift
   opt="$1"
-done    
+done
 
 if [ -z "$hosts" ] ; then
-  if test -d /sys/class/scsi_host; then 
+  if test -d /sys/class/scsi_host; then
     findhosts_26
-  else  
+  else
     findhosts
-  fi  
+  fi
 fi
 
 if [ -d /sys/class/scsi_host -a ! -w /sys/class/scsi_host ]; then
   echo "You need to run scsi-rescan-bus.sh as root"
   exit 2
-fi  
+fi
 if test "$sync" = 1 -a "$remove" = 1; then sync=2; fi
 if test "$sync" = 2; then echo "Syncing file systems"; sync; fi
 if test -w /sys/module/scsi_mod/parameters/default_dev_flags -a $scan_flags != 0; then
@@ -1195,7 +1259,7 @@ if test -w /sys/module/scsi_mod/parameters/default_dev_flags -a $scan_flags != 0
   else
     unset OLD_SCANFLAGS
   fi
-fi  
+fi
 DMSETUP=$(which dmsetup)
 [ -z "$DMSETUP" ] && flush= && mp_enable=
 MULTIPATH=$(which multipath)
@@ -1283,7 +1347,7 @@ if test -n "$mp_enable" -a $rmvd_found -gt 0 ; then
       echo "Trying to discover new multipath mappings for newly discovered devices... "
       $MULTIPATH | grep "create:" 2> /dev/null
     fi
-  fi 
+  fi
 fi
 
 echo "$found new or changed device(s) found.          "
