@@ -1,5 +1,5 @@
 /*
-*  Copyright (c) 2012-2015, Kaminario Technologies LTD
+*  Copyright (c) 2012-2018, Kaminario Technologies LTD
 *  All rights reserved.
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions are met:
@@ -29,15 +29,16 @@
  */
 
 #ifndef __sun
-#define _XOPEN_SOURCE 500
 #ifndef _GNU_SOURCE
-#define _GNU_SOURCE
+#define _GNU_SOURCE 1
 #endif
 #endif
 
 #include <unistd.h>
 #include <fcntl.h>
 #include <stdlib.h>
+#include <stdio.h>
+#include <stdbool.h>
 #include <string.h>
 #include <errno.h>
 #define __STDC_FORMAT_MACROS 1
@@ -50,8 +51,10 @@
 #include "sg_lib.h"
 #include "sg_cmds_basic.h"
 #include "sg_pt.h"
+#include "sg_unaligned.h"
+#include "sg_pr2serr.h"
 
-static const char * version_str = "1.10 20150511";
+static const char * version_str = "1.22 20180219";
 
 #define DEF_BLOCK_SIZE 512
 #define DEF_NUM_BLOCKS (1)
@@ -70,6 +73,7 @@ static struct option long_options[] = {
         {"fua", no_argument, 0, 'f'},
         {"fua_nv", no_argument, 0, 'F'},
         {"group", required_argument, 0, 'g'},
+        {"grpnum", required_argument, 0, 'g'},
         {"help", no_argument, 0, 'h'},
         {"in", required_argument, 0, 'i'},
         {"inc", required_argument, 0, 'C'},
@@ -86,23 +90,23 @@ static struct option long_options[] = {
 };
 
 struct caw_flags {
-        int dpo;
-        int fua;
-        int fua_nv;
+        bool dpo;
+        bool fua;
+        bool fua_nv;
         int group;
         int wrprotect;
 };
 
 struct opts_t {
-        const char * ifn;
-        const char * wfn;
-        int wfn_given;
-        uint64_t lba;
+        bool quiet;
+        bool wfn_given;
         int numblocks;
-        int quiet;
         int verbose;
         int timeout;
         int xfer_len;
+        uint64_t lba;
+        const char * ifn;
+        const char * wfn;
         const char * device_name;
         struct caw_flags flags;
 };
@@ -111,14 +115,14 @@ struct opts_t {
 static void
 usage()
 {
-        fprintf(stderr, "Usage: "
-                "sg_compare_and_write [--dpo] [--fua] [--fua_nv] "
-                "[--group=GN] [--help]\n"
-                "                            --in=IF [--inw=WF] --lba=LBA "
+        pr2serr("Usage: sg_compare_and_write [--dpo] [--fua] [--fua_nv] "
+                "[--grpnum=GN] [--help]\n"
+                "                            --in=IF|--inc=IF [--inw=WF] "
+                "--lba=LBA "
                 "[--num=NUM]\n"
                 "                            [--quiet] [--timeout=TO] "
                 "[--verbose] [--version]\n"
-                "                            [--wrpotect=WP] [--xferlen=LEN] "
+                "                            [--wrprotect=WP] [--xferlen=LEN] "
                 "DEVICE\n"
                 "  where:\n"
                 "    --dpo|-d            set the dpo bit in cdb (def: "
@@ -127,7 +131,7 @@ usage()
                 "clear)\n"
                 "    --fua_nv|-F         set the fua_nv bit in cdb (def: "
                 "clear)\n"
-                "    --group=GN|-g GN    GN is GROUP NUMBER to set in "
+                "    --grpnum=GN|-g GN    GN is GROUP NUMBER to set in "
                 "cdb (def: 0)\n"
                 "    --help|-h           print out usage message\n"
                 "    --in=IF|-i IF       IF is a file containing a compare "
@@ -135,6 +139,7 @@ usage()
                 "                        optionally a write buffer (when "
                 "--inw=WF is\n"
                 "                        not given)\n"
+                "    --inc=IF|-C IF      The same as the --in option\n"
                 "    --inw=WF|-D WF      WF is a file containing a write "
                 "buffer\n"
                 "    --lba=LBA|-l LBA    LBA of the first block to compare "
@@ -162,9 +167,9 @@ usage()
 static int
 parse_args(int argc, char* argv[], struct opts_t * op)
 {
+        bool lba_given = false;
+        bool if_given = false;
         int c;
-        int lba_given = 0;
-        int if_given = 0;
         int64_t ll;
 
         op->numblocks = DEF_NUM_BLOCKS;
@@ -184,27 +189,27 @@ parse_args(int argc, char* argv[], struct opts_t * op)
                 case 'C':
                 case 'i':
                         op->ifn = optarg;
-                        if_given = 1;
+                        if_given = true;
                         break;
                 case 'd':
-                        op->flags.dpo = 1;
+                        op->flags.dpo = true;
                         break;
                 case 'D':
                         op->wfn = optarg;
-                        op->wfn_given = 1;
+                        op->wfn_given = true;
                         break;
                 case 'F':
-                        op->flags.fua_nv = 1;
+                        op->flags.fua_nv = true;
                         break;
                 case 'f':
-                        op->flags.fua = 1;
+                        op->flags.fua = true;
                         break;
                 case 'g':
                         op->flags.group = sg_get_num(optarg);
                         if ((op->flags.group < 0) ||
-                            (op->flags.group > 31))  {
-                                fprintf(stderr, "argument to '--group' "
-                                        "expected to be 0 to 31\n");
+                            (op->flags.group > 63))  {
+                                pr2serr("argument to '--grpnum=' expected to "
+                                        "be 0 to 63\n");
                                 goto out_err_no_usage;
                         }
                         break;
@@ -215,28 +220,27 @@ parse_args(int argc, char* argv[], struct opts_t * op)
                 case 'l':
                         ll = sg_get_llnum(optarg);
                         if (-1 == ll) {
-                                fprintf(stderr, "bad argument to '--lba'\n");
+                                pr2serr("bad argument to '--lba'\n");
                                 goto out_err_no_usage;
                         }
                         op->lba = (uint64_t)ll;
-                        lba_given = 1;
+                        lba_given = true;
                         break;
                 case 'n':
                         op->numblocks = sg_get_num(optarg);
                         if ((op->numblocks < 0) || (op->numblocks > 255))  {
-                                fprintf(stderr, "bad argument to '--num', "
-                                        "expect 0 to 255\n");
+                                pr2serr("bad argument to '--num', expect 0 "
+                                        "to 255\n");
                                 goto out_err_no_usage;
                         }
                         break;
                 case 'q':
-                        ++op->quiet;
+                        op->quiet = true;
                         break;
                 case 't':
                         op->timeout = sg_get_num(optarg);
                         if (op->timeout < 0)  {
-                                fprintf(stderr, "bad argument to "
-                                        "'--timeout'\n");
+                                pr2serr("bad argument to '--timeout'\n");
                                 goto out_err_no_usage;
                         }
                         break;
@@ -244,27 +248,25 @@ parse_args(int argc, char* argv[], struct opts_t * op)
                         ++op->verbose;
                         break;
                 case 'V':
-                        fprintf(stderr, ME "version: %s\n", version_str);
+                        pr2serr(ME "version: %s\n", version_str);
                         exit(0);
                 case 'w':
                         op->flags.wrprotect = sg_get_num(optarg);
                         if (op->flags.wrprotect >> 3) {
-                                fprintf(stderr, "bad argument to "
-                                        "'--wrprotect' not in range 0-7\n");
+                                pr2serr("bad argument to '--wrprotect' not "
+                                        "in range 0-7\n");
                                 goto out_err_no_usage;
                         }
                         break;
                 case 'x':
                         op->xfer_len = sg_get_num(optarg);
                         if (op->xfer_len < 0) {
-                                fprintf(stderr, "bad argument to "
-                                        "'--xferlen'\n");
+                                pr2serr("bad argument to '--xferlen'\n");
                                 goto out_err_no_usage;
                         }
                         break;
                 default:
-                        fprintf(stderr, "unrecognised option code 0x%x ??\n",
-                                c);
+                        pr2serr("unrecognised option code 0x%x ??\n", c);
                         goto out_err;
                 }
         }
@@ -275,21 +277,21 @@ parse_args(int argc, char* argv[], struct opts_t * op)
                 }
                 if (optind < argc) {
                         for (; optind < argc; ++optind)
-                                fprintf(stderr, "Unexpected extra argument: "
-                                        "%s\n", argv[optind]);
+                                pr2serr("Unexpected extra argument: %s\n",
+                                        argv[optind]);
                         goto out_err;
                 }
         }
         if (NULL == op->device_name) {
-                fprintf(stderr, "missing device name!\n");
+                pr2serr("missing device name!\n");
                 goto out_err;
         }
-        if (!if_given) {
-                fprintf(stderr, "missing input file\n");
+        if (! if_given) {
+                pr2serr("missing input file\n");
                 goto out_err;
         }
-        if (!lba_given) {
-                fprintf(stderr, "missing lba\n");
+        if (! lba_given) {
+                pr2serr("missing lba\n");
                 goto out_err;
         }
         if (0 == op->xfer_len)
@@ -310,7 +312,7 @@ out_err_no_usage:
 #define WRPROTECT_SHIFT (5)
 
 static int
-sg_build_scsi_cdb(unsigned char * cdbp, unsigned int blocks,
+sg_build_scsi_cdb(uint8_t * cdbp, unsigned int blocks,
                   int64_t start_block, struct caw_flags flags)
 {
         memset(cdbp, 0, COMPARE_AND_WRITE_CDB_SIZE);
@@ -322,42 +324,35 @@ sg_build_scsi_cdb(unsigned char * cdbp, unsigned int blocks,
                 cdbp[1] |= FLAG_FUA;
         if (flags.fua_nv)
                 cdbp[1] |= FLAG_FUA_NV;
-        cdbp[2] = (unsigned char)((start_block >> 56) & 0xff);
-        cdbp[3] = (unsigned char)((start_block >> 48) & 0xff);
-        cdbp[4] = (unsigned char)((start_block >> 40) & 0xff);
-        cdbp[5] = (unsigned char)((start_block >> 32) & 0xff);
-        cdbp[6] = (unsigned char)((start_block >> 24) & 0xff);
-        cdbp[7] = (unsigned char)((start_block >> 16) & 0xff);
-        cdbp[8] = (unsigned char)((start_block >> 8) & 0xff);
-        cdbp[9] = (unsigned char)(start_block & 0xff);
+        sg_put_unaligned_be64((uint64_t)start_block, cdbp + 2);
         /* cdbp[10-12] are reserved */
-        cdbp[13] = (unsigned char)(blocks & 0xff);
-        cdbp[14] = (unsigned char)(flags.group & 0x1f);
+        cdbp[13] = (uint8_t)(blocks & 0xff);
+        cdbp[14] = (uint8_t)(flags.group & 0x1f);
         return 0;
 }
 
 /* Returns 0 for success, SG_LIB_CAT_MISCOMPARE if compare fails,
  * various other SG_LIB_CAT_*, otherwise -1 . */
 static int
-sg_compare_and_write(int sg_fd, unsigned char * buff, int blocks,
-                     int64_t lba, int xfer_len, struct caw_flags flags,
-                     int noisy, int verbose)
+sg_ll_compare_and_write(int sg_fd, uint8_t * buff, int blocks,
+                        int64_t lba, int xfer_len, struct caw_flags flags,
+                        bool noisy, int verbose)
 {
-        int k, sense_cat, valid, slen, res, ret;
-        unsigned char cawCmd[COMPARE_AND_WRITE_CDB_SIZE];
-        unsigned char sense_b[SENSE_BUFF_LEN];
-        struct sg_pt_base * ptvp;
+        bool valid;
+        int k, sense_cat, slen, res, ret;
         uint64_t ull = 0;
+        struct sg_pt_base * ptvp;
+        uint8_t cawCmd[COMPARE_AND_WRITE_CDB_SIZE];
+        uint8_t sense_b[SENSE_BUFF_LEN];
 
         if (sg_build_scsi_cdb(cawCmd, blocks, lba, flags)) {
-                fprintf(stderr, ME "bad cdb build, lba=0x%" PRIx64 ", "
-                        "blocks=%d\n", lba, blocks);
+                pr2serr(ME "bad cdb build, lba=0x%" PRIx64 ", blocks=%d\n",
+                        lba, blocks);
                 return -1;
         }
         ptvp = construct_scsi_pt_obj();
         if (NULL == ptvp) {
-                fprintf(stderr, "Could not construct scsit_pt_obj, out of "
-                        "memory\n");
+                pr2serr("Could not construct scsit_pt_obj, out of memory\n");
                 return -1;
         }
 
@@ -365,21 +360,21 @@ sg_compare_and_write(int sg_fd, unsigned char * buff, int blocks,
         set_scsi_pt_sense(ptvp, sense_b, sizeof(sense_b));
         set_scsi_pt_data_out(ptvp, buff, xfer_len);
         if (verbose > 1) {
-                fprintf(stderr, "    Compare and write cdb: ");
+                pr2serr("    Compare and write cdb: ");
                 for (k = 0; k < COMPARE_AND_WRITE_CDB_SIZE; ++k)
-                        fprintf(stderr, "%02x ", cawCmd[k]);
-                fprintf(stderr, "\n");
+                        pr2serr("%02x ", cawCmd[k]);
+                pr2serr("\n");
         }
         if ((verbose > 2) && (xfer_len > 0)) {
-                fprintf(stderr, "    Data-out buffer contents:\n");
-                dStrHexErr((const char *)buff, xfer_len, 1);
+                pr2serr("    Data-out buffer contents:\n");
+                hex2stderr(buff, xfer_len, 1);
         }
         res = do_scsi_pt(ptvp, sg_fd, DEF_TIMEOUT_SECS, verbose);
-        ret = sg_cmds_process_resp(ptvp, "COMPARE AND WRITE", res, 0,
-                                   sense_b, noisy, verbose,
+        ret = sg_cmds_process_resp(ptvp, "COMPARE AND WRITE", res,
+                                   SG_NO_DATA_IN, sense_b, noisy, verbose,
                                    &sense_cat);
         if (-1 == ret)
-                ;
+                ret = sg_convert_errno(get_scsi_pt_os_err(ptvp));
         else if (-2 == ret) {
                 switch (sense_cat) {
                 case SG_LIB_CAT_RECOVERED:
@@ -391,13 +386,11 @@ sg_compare_and_write(int sg_fd, unsigned char * buff, int blocks,
                         valid = sg_get_sense_info_fld(sense_b, slen,
                                                       &ull);
                         if (valid)
-                                fprintf(stderr, "Medium or hardware "
-                                        "error starting at lba=%"
-                                        PRIu64 " [0x%" PRIx64 "]\n",
-                                        ull, ull);
+                                pr2serr("Medium or hardware error starting "
+                                        "at lba=%" PRIu64 " [0x%" PRIx64
+                                        "]\n", ull, ull);
                         else
-                                fprintf(stderr, "Medium or hardware "
-                                        "error\n");
+                                pr2serr("Medium or hardware error\n");
                         ret = sense_cat;
                         break;
                 case SG_LIB_CAT_MISCOMPARE:
@@ -407,11 +400,10 @@ sg_compare_and_write(int sg_fd, unsigned char * buff, int blocks,
                         slen = get_scsi_pt_sense_len(ptvp);
                         valid = sg_get_sense_info_fld(sense_b, slen, &ull);
                         if (valid)
-                                fprintf(stderr, "Miscompare at byte offset: %"
-                                        PRIu64 " [0x%" PRIx64 "]\n", ull,
-                                        ull);
+                                pr2serr("Miscompare at byte offset: %" PRIu64
+                                        " [0x%" PRIx64 "]\n", ull, ull);
                         else
-                                fprintf(stderr, "Miscompare reported\n");
+                                pr2serr("Miscompare reported\n");
                         break;
                 default:
                         ret = sense_cat;
@@ -425,7 +417,7 @@ sg_compare_and_write(int sg_fd, unsigned char * buff, int blocks,
 }
 
 static int
-open_if(const char * fn, int got_stdin)
+open_if(const char * fn, bool got_stdin)
 {
         int fd;
 
@@ -434,7 +426,7 @@ open_if(const char * fn, int got_stdin)
         else {
                 fd = open(fn, O_RDONLY);
                 if (fd < 0) {
-                        fprintf(stderr, ME "open error: %s: %s\n", fn,
+                        pr2serr(ME "open error: %s: %s\n", fn,
                                 safe_strerror(errno));
                         return -SG_LIB_FILE_ERROR;
                 }
@@ -449,9 +441,10 @@ open_if(const char * fn, int got_stdin)
 static int
 open_dev(const char * outf, int verbose)
 {
-        int sg_fd = sg_cmds_open_device(outf, 0 /* rw */, verbose);
+        int sg_fd = sg_cmds_open_device(outf, false /* rw */, verbose);
+
         if (sg_fd < 0) {
-                fprintf(stderr, ME "open error: %s: %s\n", outf,
+                pr2serr(ME "open error: %s: %s\n", outf,
                         safe_strerror(-sg_fd));
                 return -SG_LIB_FILE_ERROR;
         }
@@ -463,31 +456,31 @@ open_dev(const char * outf, int verbose)
 int
 main(int argc, char * argv[])
 {
-        int res, half_xlen, ifn_stdin;
+        bool ifn_stdin;
+        int res, half_xlen;
         int infd = -1;
         int wfd = -1;
         int devfd = -1;
-        unsigned char * wrkBuff = NULL;
-        struct opts_t opts;
+        uint8_t * wrkBuff = NULL;
         struct opts_t * op;
+        struct opts_t opts;
 
         op = &opts;
         memset(op, 0, sizeof(opts));
         res = parse_args(argc, argv, op);
         if (res != 0) {
-                fprintf(stderr, "Failed parsing args\n");
+                pr2serr("Failed parsing args\n");
                 goto out;
         }
 
         if (op->verbose) {
-                fprintf(stderr, "Running COMPARE AND WRITE command with the "
+                pr2serr("Running COMPARE AND WRITE command with the "
                         "following options:\n  in=%s ", op->ifn);
                 if (op->wfn_given)
-                        fprintf(stderr, "inw=%s ", op->wfn);
-                fprintf(stderr, "device=%s\n  lba=0x%" PRIx64
-                        " num_blocks=%d xfer_len=%d timeout=%d\n",
-                        op->device_name, op->lba, op->numblocks,
-                        op->xfer_len, op->timeout);
+                        pr2serr("inw=%s ", op->wfn);
+                pr2serr("device=%s\n  lba=0x%" PRIx64 " num_blocks=%d "
+                        "xfer_len=%d timeout=%d\n", op->device_name,
+                        op->lba, op->numblocks, op->xfer_len, op->timeout);
         }
         ifn_stdin = ((1 == strlen(op->ifn)) && ('-' == op->ifn[0]));
         infd = open_if(op->ifn, ifn_stdin);
@@ -497,12 +490,11 @@ main(int argc, char * argv[])
         }
         if (op->wfn_given) {
                 if ((1 == strlen(op->wfn)) && ('-' == op->wfn[0])) {
-                        fprintf(stderr, ME "don't allow stdin for write "
-                                "file\n");
+                        pr2serr(ME "don't allow stdin for write file\n");
                         res = SG_LIB_FILE_ERROR;
                         goto out;
                 }
-                wfd = open_if(op->wfn, 0);
+                wfd = open_if(op->wfn, false);
                 if (wfd < 0) {
                         res = -wfd;
                         goto out;
@@ -515,9 +507,9 @@ main(int argc, char * argv[])
                 goto out;
         }
 
-        wrkBuff = (unsigned char *)malloc(op->xfer_len);
+        wrkBuff = (uint8_t *)malloc(op->xfer_len);
         if (0 == wrkBuff) {
-                fprintf(stderr, "Not enough user memory\n");
+                pr2serr("Not enough user memory\n");
                 res = SG_LIB_CAT_OTHER;
                 goto out;
         }
@@ -526,35 +518,36 @@ main(int argc, char * argv[])
                 half_xlen = op->xfer_len / 2;
                 res = read(infd, wrkBuff, half_xlen);
                 if (res < 0) {
-                        fprintf(stderr, "Could not read from %s", op->ifn);
+                        pr2serr("Could not read from %s", op->ifn);
                         goto out;
                 } else if (res < half_xlen) {
-                        fprintf(stderr, "Read only %d bytes (expected %d) "
-                                "from %s\n", res, half_xlen, op->ifn);
+                        pr2serr("Read only %d bytes (expected %d) from %s\n",
+                                res, half_xlen, op->ifn);
                         goto out;
                 }
                 res = read(wfd, wrkBuff + half_xlen, half_xlen);
                 if (res < 0) {
-                        fprintf(stderr, "Could not read from %s", op->wfn);
+                        pr2serr("Could not read from %s", op->wfn);
                         goto out;
                 } else if (res < half_xlen) {
-                        fprintf(stderr, "Read only %d bytes (expected %d) "
-                                "from %s\n", res, half_xlen, op->wfn);
+                        pr2serr("Read only %d bytes (expected %d) from %s\n",
+                                res, half_xlen, op->wfn);
                         goto out;
                 }
         } else {
                 res = read(infd, wrkBuff, op->xfer_len);
                 if (res < 0) {
-                        fprintf(stderr, "Could not read from %s", op->ifn);
+                        pr2serr("Could not read from %s", op->ifn);
                         goto out;
                 } else if (res < op->xfer_len) {
-                        fprintf(stderr, "Read only %d bytes (expected %d) "
-                                "from %s\n", res, op->xfer_len, op->ifn);
+                        pr2serr("Read only %d bytes (expected %d) from %s\n",
+                                res, op->xfer_len, op->ifn);
                         goto out;
                 }
         }
-        res = sg_compare_and_write(devfd, wrkBuff, op->numblocks, op->lba,
-                op->xfer_len, op->flags, !op->quiet, op->verbose);
+        res = sg_ll_compare_and_write(devfd, wrkBuff, op->numblocks, op->lba,
+                                      op->xfer_len, op->flags, ! op->quiet,
+                                      op->verbose);
 
 out:
         if (0 != res) {
@@ -568,7 +561,7 @@ out:
                 default:
                         sg_get_category_sense_str(res, sizeof(b), b,
                                                   op->verbose);
-                        fprintf(stderr, ME "SCSI COMPARE AND WRITE: %s\n", b);
+                        pr2serr(ME "SCSI COMPARE AND WRITE: %s\n", b);
                         break;
                 }
         }
