@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2011-2015 Hannes Reinecke, SUSE Labs
+ * Copyright (c) 2011-2018 Hannes Reinecke, SUSE Labs
  * All rights reserved.
  * Use of this source code is governed by a BSD-style
  * license that can be found in the BSD_LICENSE file.
@@ -9,6 +9,7 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <stdarg.h>
+#include <stdbool.h>
 #include <string.h>
 #include <errno.h>
 #include <getopt.h>
@@ -22,6 +23,7 @@
 #include "sg_cmds_basic.h"
 #include "sg_cmds_extra.h"
 #include "sg_unaligned.h"
+#include "sg_pr2serr.h"
 
 /* A utility program for the Linux OS SCSI subsystem.
    *  Copyright (C) 2004-2010 D. Gilbert
@@ -36,7 +38,7 @@
    and the optional list identifier passed as the list_id argument.
 */
 
-static const char * version_str = "1.12 20150227";
+static const char * version_str = "1.18 20180219";
 
 
 #define MAX_XFER_LEN 10000
@@ -100,33 +102,13 @@ struct descriptor_type segment_descriptor_codes [] = {
     { -1, "" }
 };
 
-#ifdef __GNUC__
-static int pr2serr(const char * fmt, ...)
-        __attribute__ ((format (printf, 1, 2)));
-#else
-static int pr2serr(const char * fmt, ...);
-#endif
-
-
-static int
-pr2serr(const char * fmt, ...)
-{
-    va_list args;
-    int n;
-
-    va_start(args, fmt);
-    n = vfprintf(stderr, fmt, args);
-    va_end(args);
-    return n;
-}
-
 
 static void
-scsi_failed_segment_details(unsigned char *rcBuff, unsigned int rcBuffLen)
+scsi_failed_segment_details(uint8_t *rcBuff, unsigned int rcBuffLen)
 {
+    int senseLen;
     unsigned int len;
     char senseBuff[1024];
-    int senseLen;
 
     if (rcBuffLen < 4) {
         pr2serr("  <<not enough data to procedd report>>\n");
@@ -149,7 +131,7 @@ scsi_failed_segment_details(unsigned char *rcBuff, unsigned int rcBuffLen)
 }
 
 static void
-scsi_copy_status(unsigned char *rcBuff, unsigned int rcBuffLen)
+scsi_copy_status(uint8_t *rcBuff, unsigned int rcBuffLen)
 {
     unsigned int len;
 
@@ -185,7 +167,7 @@ scsi_copy_status(unsigned char *rcBuff, unsigned int rcBuffLen)
 }
 
 static void
-scsi_operating_parameters(unsigned char *rcBuff, unsigned int rcBuffLen)
+scsi_operating_parameters(uint8_t *rcBuff, unsigned int rcBuffLen)
 {
     unsigned int len, n;
 
@@ -267,17 +249,18 @@ scsi_operating_parameters(unsigned char *rcBuff, unsigned int rcBuffLen)
 }
 
 static struct option long_options[] = {
-        {"failed", 0, 0, 'f'},
-        {"help", 0, 0, 'h'},
-        {"hex", 0, 0, 'H'},
-        {"list_id", 1, 0, 'l'},
-        {"params", 0, 0, 'p'},
-        {"readonly", 0, 0, 'R'},
-        {"receive", 0, 0, 'r'},
-        {"status", 0, 0, 's'},
-        {"verbose", 0, 0, 'v'},
-        {"version", 0, 0, 'V'},
-        {"xfer_len", 1, 0, 'x'},
+        {"failed", no_argument, 0, 'f'},
+        {"help", no_argument, 0, 'h'},
+        {"hex", no_argument, 0, 'H'},
+        {"list_id", required_argument, 0, 'l'},
+        {"list-id", required_argument, 0, 'l'},
+        {"params", no_argument, 0, 'p'},
+        {"readonly", no_argument, 0, 'R'},
+        {"receive", no_argument, 0, 'r'},
+        {"status", no_argument, 0, 's'},
+        {"verbose", no_argument, 0, 'v'},
+        {"version", no_argument, 0, 'V'},
+        {"xfer_len", required_argument, 0, 'x'},
         {0, 0, 0, 0},
 };
 
@@ -321,18 +304,18 @@ static const char * rec_copy_name_arr[] = {
 int
 main(int argc, char * argv[])
 {
+    bool do_hex = false;
+    bool o_readonly = false;
     int sg_fd, res, c, k;
-    unsigned char * cpResultBuff = NULL;
-    int xfer_len = 520;
+    int ret = 1;
     int sa = 3;
-    uint32_t list_id = 0;
-    int do_hex = 0;
-    int o_readonly = 0;
     int verbose = 0;
+    int xfer_len = 520;
+    uint32_t list_id = 0;
     const char * cp;
+    uint8_t * cpResultBuff = NULL;
     const char * device_name = NULL;
     char file_name[256];
-    int ret = 1;
 
     memset(file_name, 0, sizeof file_name);
     while (1) {
@@ -348,7 +331,7 @@ main(int argc, char * argv[])
             sa = 4;
             break;
         case 'H':
-            do_hex = 1;
+            do_hex = true;
             break;
         case 'h':
         case '?':
@@ -369,7 +352,7 @@ main(int argc, char * argv[])
             sa = 1;
             break;
         case 'R':
-            ++o_readonly;
+            o_readonly = true;
             break;
         case 's':
             sa = 0;
@@ -418,7 +401,7 @@ main(int argc, char * argv[])
         return SG_LIB_SYNTAX_ERROR;
     }
 
-    if (NULL == (cpResultBuff = (unsigned char *)malloc(xfer_len))) {
+    if (NULL == (cpResultBuff = (uint8_t *)malloc(xfer_len))) {
             pr2serr(ME "out of memory\n");
             return SG_LIB_FILE_ERROR;
     }
@@ -428,7 +411,8 @@ main(int argc, char * argv[])
     if (sg_fd < 0) {
         pr2serr(ME "open error: %s: %s\n", device_name,
                 safe_strerror(-sg_fd));
-        return SG_LIB_FILE_ERROR;
+        ret = SG_LIB_FILE_ERROR;
+        goto finish;
     }
 
     if ((sa < 0) || (sa >= (int)(sizeof(rec_copy_name_arr) / sizeof(char *))))
@@ -440,7 +424,7 @@ main(int argc, char * argv[])
                 PRIu32 "\n", cp, device_name, xfer_len, xfer_len, list_id);
 
     res = sg_ll_receive_copy_results(sg_fd, sa, list_id, cpResultBuff,
-                                     xfer_len, 1, verbose);
+                                     xfer_len, true, verbose);
     ret = res;
     if (res) {
         char b[80];
@@ -449,27 +433,22 @@ main(int argc, char * argv[])
         pr2serr("  SCSI %s failed: %s\n", cp, b);
         goto finish;
     }
-    if (1 == do_hex) {
-        dStrHex((const char *)cpResultBuff, xfer_len, 1);
-        res = 0;
+    if (do_hex) {
+        hex2stdout(cpResultBuff, xfer_len, 1);
         goto finish;
     }
     switch (sa) {
     case 4: /* Failed segment details */
         scsi_failed_segment_details(cpResultBuff, xfer_len);
-        res = 0;
         break;
     case 3: /* Operating parameters */
         scsi_operating_parameters(cpResultBuff, xfer_len);
-        res = 0;
         break;
     case 0: /* Copy status */
         scsi_copy_status(cpResultBuff, xfer_len);
-        res = 0;
         break;
     default:
-        dStrHex((const char *)cpResultBuff, xfer_len, 1);
-        res = 0;
+        hex2stdout(cpResultBuff, xfer_len, 1);
         break;
     }
 
