@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2010-2015 Douglas Gilbert.
+ * Copyright (c) 2010-2018 Douglas Gilbert.
  * All rights reserved.
  * Use of this source code is governed by a BSD-style
  * license that can be found in the BSD_LICENSE file.
@@ -9,6 +9,7 @@
 #include <fcntl.h>
 #include <stdio.h>
 #include <stdlib.h>
+#include <stdbool.h>
 #include <string.h>
 #include <errno.h>
 #include <limits.h>
@@ -23,14 +24,17 @@
 #include "config.h"
 #endif
 #include "sg_lib.h"
+#include "sg_pr2serr.h"
+#include "sg_unaligned.h"
 
 
-static const char * version_str = "1.07 20150123";
+static const char * version_str = "1.15 20180302";
 
 #define MAX_SENSE_LEN 1024 /* max descriptor format actually: 256+8 */
 
 static struct option long_options[] = {
     {"binary", required_argument, 0, 'b'},
+    {"cdb", no_argument, 0, 'c'},
     {"file", required_argument, 0, 'f'},
     {"help", no_argument, 0, 'h'},
     {"hex", no_argument, 0, 'H'},
@@ -43,20 +47,21 @@ static struct option long_options[] = {
 };
 
 struct opts_t {
-    int do_binary;
+    bool do_binary;
+    bool do_cdb;
+    bool do_help;
+    bool do_hex;
+    bool no_space;
+    bool do_status;
+    bool do_version;
+    bool file_given;
     const char * fname;
-    int do_file;
-    int do_help;
-    int do_hex;
-    int no_space;
-    int do_status;
+    int sense_len;
     int sstatus;
     int do_verbose;
-    int do_version;
     const char * wfname;
-    unsigned char sense[MAX_SENSE_LEN + 4];
     const char * no_space_str;
-    int sense_len;
+    uint8_t sense[MAX_SENSE_LEN + 4];
 };
 
 static char concat_buff[1024];
@@ -65,17 +70,18 @@ static char concat_buff[1024];
 static void
 usage()
 {
-  fprintf(stderr, "Usage: "
-          "sg_decode_sense [--binary=FN] [--file=FN] [--help] [--hex] "
-          "[--nospace]\n"
-          "                       [--status=SS] [--verbose] [--version] "
-          "[--write=WFN]\n"
-          "                       [H1 H2 H3 ...]\n"
+  pr2serr("Usage: sg_decode_sense [--binary=FN] [--cdb] [--file=FN] "
+          "[--help] [--hex]\n"
+          "                       [--nospace] [--status=SS] [--verbose] "
+          "[--version]\n"
+          "                       [--write=WFN] H1 H2 H3 ...\n"
           "  where:\n"
           "    --binary=FN|-b FN     FN is a file name to read sense "
           "data in\n"
           "                          binary from. If FN is '-' then read "
           "from stdin\n"
+          "    --cdb|-c              decode given hex as cdb rather than "
+          "sense data\n"
           "    --file=FN|-f FN       FN is a file name from which to read "
           "sense data\n"
           "                          in ASCII hexadecimal. Interpret '-' "
@@ -98,73 +104,77 @@ usage()
           "Decodes SCSI sense data given on the command line as a sequence "
           "of\nhexadecimal bytes (H1 H2 H3 ...) . Alternatively the sense "
           "data can\nbe in a binary file or in a file containing ASCII "
-          "hexadecimal.\n"
+          "hexadecimal. If\n'--cdb' is given then interpret hex as SCSI CDB "
+          "rather than sense data.\n"
           );
 }
 
 static int
-process_cl(struct opts_t *optsp, int argc, char *argv[])
+parse_cmd_line(struct opts_t *op, int argc, char *argv[])
 {
     int c;
     unsigned int ui;
-    char * opt;
-    char *endptr;
     long val;
+    char * avp;
+    char *endptr;
 
     while (1) {
-        c = getopt_long(argc, argv, "b:f:hHns:vVw:", long_options, NULL);
+        c = getopt_long(argc, argv, "b:cf:hHns:vVw:", long_options, NULL);
         if (c == -1)
             break;
 
         switch (c) {
         case 'b':
-            if (optsp->fname) {
-                fprintf(stderr, "expect only one '--binary=FN' or "
-                        "'--file=FN' option\n");
+            if (op->fname) {
+                pr2serr("expect only one '--binary=FN' or '--file=FN' "
+                        "option\n");
                 return SG_LIB_SYNTAX_ERROR;
             }
-            ++optsp->do_binary;
-            optsp->fname = optarg;
+            op->do_binary = true;
+            op->fname = optarg;
+            break;
+        case 'c':
+            op->do_cdb = true;
             break;
         case 'f':
-            if (optsp->fname) {
-                fprintf(stderr, "expect only one '--binary=FN' or "
-                        "'--file=FN' option\n");
+            if (op->fname) {
+                pr2serr("expect only one '--binary=FN' or '--file=FN' "
+                        "option\n");
                 return SG_LIB_SYNTAX_ERROR;
             }
-            ++optsp->do_file;
-            optsp->fname = optarg;
+            op->file_given = true;
+            op->fname = optarg;
             break;
         case 'h':
         case '?':
-            optsp->do_help = 1;
+            op->do_help = true;
             return 0;
         case 'H':
-            ++optsp->do_hex;
+            op->do_hex = true;
             break;
         case 'n':
-            ++optsp->no_space;
+            op->no_space = true;
             break;
         case 's':
             if (1 != sscanf(optarg, "%x", &ui)) {
-                fprintf(stderr, "'--status=SS' expects a byte value\n");
+                pr2serr("'--status=SS' expects a byte value\n");
                 return SG_LIB_SYNTAX_ERROR;
             }
             if (ui > 0xff) {
-                fprintf(stderr, "'--status=SS' byte value exceeds FF\n");
+                pr2serr("'--status=SS' byte value exceeds FF\n");
                 return SG_LIB_SYNTAX_ERROR;
             }
-            ++optsp->do_status;
-            optsp->sstatus = ui;
+            op->do_status = true;
+            op->sstatus = ui;
             break;
         case 'v':
-            ++optsp->do_verbose;
+            ++op->do_verbose;
             break;
         case 'V':
-            optsp->do_version = 1;
-            return 0;
+            op->do_version = true;
+            break;
         case 'w':
-            optsp->wfname = optarg;
+            op->wfname = optarg;
             break;
         default:
             return SG_LIB_SYNTAX_ERROR;
@@ -172,42 +182,41 @@ process_cl(struct opts_t *optsp, int argc, char *argv[])
     }
 
     while (optind < argc) {
-        opt = argv[optind++];
-        if (optsp->no_space) {
-            if (optsp->no_space_str) {
+        avp = argv[optind++];
+        if (op->no_space) {
+            if (op->no_space_str) {
                 if ('\0' == concat_buff[0]) {
-                    if (strlen(optsp->no_space_str) > sizeof(concat_buff)) {
-                        fprintf(stderr, "'--nospace' concat_buff overflow\n");
+                    if (strlen(op->no_space_str) > sizeof(concat_buff)) {
+                        pr2serr("'--nospace' concat_buff overflow\n");
                         return SG_LIB_SYNTAX_ERROR;
                     }
-                    strcpy(concat_buff, optsp->no_space_str);
+                    strcpy(concat_buff, op->no_space_str);
                 }
-                if ((strlen(concat_buff) + strlen(opt)) >=
+                if ((strlen(concat_buff) + strlen(avp)) >=
                     sizeof(concat_buff)) {
-                    fprintf(stderr, "'--nospace' concat_buff overflow\n");
+                    pr2serr("'--nospace' concat_buff overflow\n");
                     return SG_LIB_SYNTAX_ERROR;
                 }
-                if (optsp->do_version)
-                    fprintf(stderr, "'--nospace' and found whitespace so "
+                if (op->do_version)
+                    pr2serr("'--nospace' and found whitespace so "
                             "concatenate\n");
-                strcat(concat_buff, opt);
-                optsp->no_space_str = concat_buff;
+                strcat(concat_buff, avp);
+                op->no_space_str = concat_buff;
             } else
-                optsp->no_space_str = opt;
+                op->no_space_str = avp;
             continue;
         }
-        val = strtol(opt, &endptr, 16);
-        if (*opt == '\0' || *endptr != '\0' || val < 0x00 || val > 0xff) {
-            fprintf(stderr, "Invalid byte '%s'\n", opt);
+        val = strtol(avp, &endptr, 16);
+        if (*avp == '\0' || *endptr != '\0' || val < 0x00 || val > 0xff) {
+            pr2serr("Invalid byte '%s'\n", avp);
             return SG_LIB_SYNTAX_ERROR;
         }
 
-        if (optsp->sense_len > MAX_SENSE_LEN) {
-            fprintf(stderr, "sense data too long (max. %d bytes)\n",
-                    MAX_SENSE_LEN);
+        if (op->sense_len > MAX_SENSE_LEN) {
+            pr2serr("sense data too long (max. %d bytes)\n", MAX_SENSE_LEN);
             return SG_LIB_SYNTAX_ERROR;
         }
-        optsp->sense[optsp->sense_len++] = (unsigned char)val;
+        op->sense[op->sense_len++] = (uint8_t)val;
     }
     return 0;
 }
@@ -218,16 +227,17 @@ process_cl(struct opts_t *optsp, int argc, char *argv[])
  * digits is expected, 2 per byte. Everything from and including a '#'
  * on a line is ignored.  Returns 0 if ok, or 1 if error. */
 static int
-f2hex_arr(const char * fname, int no_space, unsigned char * mp_arr,
+f2hex_arr(const char * fname, bool no_space, uint8_t * mp_arr,
           int * mp_arr_len, int max_arr_len)
 {
-    int fn_len, in_len, k, j, m, split_line;
+    bool split_line;
+    int fn_len, in_len, k, j, m;
+    int off = 0;
     unsigned int h;
     const char * lcp;
     FILE * fp;
     char line[512];
     char carry_over[4];
-    int off = 0;
 
     if ((NULL == fname) || (NULL == mp_arr) || (NULL == mp_arr_len))
         return 1;
@@ -239,7 +249,7 @@ f2hex_arr(const char * fname, int no_space, unsigned char * mp_arr,
     else {
         fp = fopen(fname, "r");
         if (NULL == fp) {
-            fprintf(stderr, "Unable to open %s for reading\n", fname);
+            pr2serr("Unable to open %s for reading\n", fname);
             return 1;
         }
     }
@@ -253,9 +263,9 @@ f2hex_arr(const char * fname, int no_space, unsigned char * mp_arr,
             if ('\n' == line[in_len - 1]) {
                 --in_len;
                 line[in_len] = '\0';
-                split_line = 0;
+                split_line = false;
             } else
-                split_line = 1;
+                split_line = true;
         }
         if (in_len < 1) {
             carry_over[0] = 0;
@@ -268,8 +278,8 @@ f2hex_arr(const char * fname, int no_space, unsigned char * mp_arr,
                 if (1 == sscanf(carry_over, "%x", &h))
                     mp_arr[off - 1] = h;       /* back up and overwrite */
                 else {
-                    fprintf(stderr, "f2hex_arr: carry_over error ['%s'] "
-                            "around line %d\n", carry_over, j + 1);
+                    pr2serr("f2hex_arr: carry_over error ['%s'] around line "
+                            "%d\n", carry_over, j + 1);
                     goto bad;
                 }
                 lcp = line + 1;
@@ -289,20 +299,20 @@ f2hex_arr(const char * fname, int no_space, unsigned char * mp_arr,
             continue;
         k = strspn(lcp, "0123456789aAbBcCdDeEfF ,\t");
         if ((k < in_len) && ('#' != lcp[k]) && ('\r' != lcp[k])) {
-            fprintf(stderr, "f2hex_arr: syntax error at line %d, pos %d\n",
-                    j + 1, m + k + 1);
+            pr2serr("f2hex_arr: syntax error at line %d, pos %d\n", j + 1,
+                    m + k + 1);
             goto bad;
         }
         if (no_space) {
             for (k = 0; isxdigit(*lcp) && isxdigit(*(lcp + 1));
                  ++k, lcp += 2) {
                 if (1 != sscanf(lcp, "%2x", &h)) {
-                    fprintf(stderr, "f2hex_arr: bad hex number in line %d, "
-                            "pos %d\n", j + 1, (int)(lcp - line + 1));
+                    pr2serr("f2hex_arr: bad hex number in line %d, pos %d\n",
+                            j + 1, (int)(lcp - line + 1));
                     goto bad;
                 }
                 if ((off + k) >= max_arr_len) {
-                    fprintf(stderr, "f2hex_arr: array length exceeded\n");
+                    pr2serr("f2hex_arr: array length exceeded\n");
                     goto bad;
                 }
                 mp_arr[off + k] = h;
@@ -314,8 +324,8 @@ f2hex_arr(const char * fname, int no_space, unsigned char * mp_arr,
             for (k = 0; k < 1024; ++k) {
                 if (1 == sscanf(lcp, "%x", &h)) {
                     if (h > 0xff) {
-                        fprintf(stderr, "f2hex_arr: hex number larger than "
-                                "0xff in line %d, pos %d\n", j + 1,
+                        pr2serr("f2hex_arr: hex number larger than 0xff in "
+                                "line %d, pos %d\n", j + 1,
                                 (int)(lcp - line + 1));
                         goto bad;
                     }
@@ -324,7 +334,7 @@ f2hex_arr(const char * fname, int no_space, unsigned char * mp_arr,
                         carry_over[0] = *lcp;
                     }
                     if ((off + k) >= max_arr_len) {
-                        fprintf(stderr, "f2hex_arr: array length exceeded\n");
+                        pr2serr("f2hex_arr: array length exceeded\n");
                         goto bad;
                     }
                     mp_arr[off + k] = h;
@@ -339,8 +349,8 @@ f2hex_arr(const char * fname, int no_space, unsigned char * mp_arr,
                         --k;
                         break;
                     }
-                    fprintf(stderr, "f2hex_arr: error in line %d, at pos "
-                            "%d\n", j + 1, (int)(lcp - line + 1));
+                    pr2serr("f2hex_arr: error in line %d, at pos %d\n", j + 1,
+                            (int)(lcp - line + 1));
                     goto bad;
                 }
             }
@@ -358,30 +368,36 @@ bad:
 }
 
 static void
-write2wfn(FILE * fp, struct opts_t * optsp)
+write2wfn(FILE * fp, struct opts_t * op)
 {
     int k, n;
     size_t s;
     char b[128];
 
-    if (optsp->do_hex) {
-        for (k = 0, n = 0; k < optsp->sense_len; ++k) {
-            n += sprintf(b + n, "0x%02x,", optsp->sense[k]);
+    if (op->do_hex) {
+        for (k = 0, n = 0; k < op->sense_len; ++k) {
+            n += sprintf(b + n, "0x%02x,", op->sense[k]);
             if (15 == (k % 16)) {
                 b[n] = '\n';
                 s = fwrite(b, 1, n + 1, fp);
+                if ((int)s != (n + 1))
+                    pr2serr("only able to write %d of %d bytes to %s\n",
+                            (int)s, n + 1, op->wfname);
                 n = 0;
             }
         }
         if (n > 0) {
             b[n] = '\n';
             s = fwrite(b, 1, n + 1, fp);
+            if ((int)s != (n + 1))
+                pr2serr("only able to write %d of %d bytes to %s\n", (int)s,
+                        n + 1, op->wfname);
         }
     } else {
-        s = fwrite(optsp->sense, 1, optsp->sense_len, fp);
-        if ((int)s != optsp->sense_len)
-            fprintf(stderr, "only able to write %d of %d bytes to %s\n",
-                    (int)s, optsp->sense_len, optsp->wfname);
+        s = fwrite(op->sense, 1, op->sense_len, fp);
+        if ((int)s != op->sense_len)
+            pr2serr("only able to write %d of %d bytes to %s\n", (int)s,
+                    op->sense_len, op->wfname);
     }
 }
 
@@ -393,98 +409,110 @@ main(int argc, char *argv[])
     int ret = 0;
     unsigned int ui;
     size_t s;
-    struct opts_t opts;
-    char b[2048];
+    struct opts_t * op;
     FILE * fp = NULL;
     const char * cp;
+    char b[2048];
+    struct opts_t opts;
 
-    memset(&opts, 0, sizeof(opts));
+    op = &opts;
+    memset(op, 0, sizeof(opts));
     memset(b, 0, sizeof(b));
-    ret = process_cl(&opts, argc, argv);
+    ret = parse_cmd_line(op, argc, argv);
     if (ret != 0) {
         usage();
         return ret;
-    } else if (opts.do_help) {
+    } else if (op->do_help) {
         usage();
         return 0;
-    } else if (opts.do_version) {
-        fprintf(stderr, "version: %s\n", version_str);
+    } else if (op->do_version) {
+        pr2serr("version: %s\n", version_str);
         return 0;
     }
 
 
-    if (opts.do_status) {
-        sg_get_scsi_status_str(opts.sstatus, sizeof(b) - 1, b);
+    if (op->do_status) {
+        sg_get_scsi_status_str(op->sstatus, sizeof(b) - 1, b);
         printf("SCSI status: %s\n", b);
     }
 
-    if ((0 == opts.sense_len) && opts.no_space_str) {
-        if (opts.do_verbose > 2)
-            fprintf(stderr, "no_space str: %s\n", opts.no_space_str);
-        cp = opts.no_space_str;
+    if ((0 == op->sense_len) && op->no_space_str) {
+        if (op->do_verbose > 2)
+            pr2serr("no_space str: %s\n", op->no_space_str);
+        cp = op->no_space_str;
         for (k = 0; isxdigit(cp[k]) && isxdigit(cp[k + 1]); k += 2) {
             if (1 != sscanf(cp + k, "%2x", &ui)) {
-                fprintf(stderr, "bad no_space hex string: %s\n", cp);
+                pr2serr("bad no_space hex string: %s\n", cp);
                 return SG_LIB_SYNTAX_ERROR;
             }
-            opts.sense[opts.sense_len++] = (unsigned char)ui;
+            op->sense[op->sense_len++] = (uint8_t)ui;
         }
     }
 
-    if ((0 == opts.sense_len) && (! opts.do_binary) && (! opts.do_file)) {
-        if (opts.do_status)
+    if ((0 == op->sense_len) && (! op->do_binary) && (! op->file_given)) {
+        if (op->do_status)
             return 0;
-        fprintf(stderr, ">> Need sense data on the command line or in a "
-                "file\n\n");
+        pr2serr(">> Need sense data on the command line or in a file\n\n");
         usage();
         return SG_LIB_SYNTAX_ERROR;
     }
-    if (opts.sense_len && (opts.do_binary || opts.do_file)) {
-        fprintf(stderr, ">> Need sense data on command line or in a file, "
-                "not both\n\n");
+    if (op->sense_len && (op->do_binary || op->file_given)) {
+        pr2serr(">> Need sense data on command line or in a file, not "
+                "both\n\n");
         return SG_LIB_SYNTAX_ERROR;
     }
-    if (opts.do_binary && opts.do_file) {
-        fprintf(stderr, ">> Either a binary file or a ASCII hexadecimal, "
-                "file not both\n\n");
+    if (op->do_binary && op->file_given) {
+        pr2serr(">> Either a binary file or a ASCII hexadecimal, file not "
+                "both\n\n");
         return SG_LIB_SYNTAX_ERROR;
     }
 
-    if (opts.do_binary) {
-        fp = fopen(opts.fname, "r");
+    if (op->do_binary) {
+        fp = fopen(op->fname, "r");
         if (NULL == fp) {
-            fprintf(stderr, "unable to open file: %s\n", opts.fname);
+            pr2serr("unable to open file: %s\n", op->fname);
             return SG_LIB_SYNTAX_ERROR;
         }
-        s = fread(opts.sense, 1, MAX_SENSE_LEN, fp);
+        s = fread(op->sense, 1, MAX_SENSE_LEN, fp);
         fclose(fp);
         if (0 == s) {
-            fprintf(stderr, "read nothing from file: %s\n", opts.fname);
+            pr2serr("read nothing from file: %s\n", op->fname);
             return SG_LIB_SYNTAX_ERROR;
         }
-        opts.sense_len = s;
-    } else if (opts.do_file) {
-        ret = f2hex_arr(opts.fname, opts.no_space, opts.sense,
-                        &opts.sense_len, MAX_SENSE_LEN);
+        op->sense_len = s;
+    } else if (op->file_given) {
+        ret = f2hex_arr(op->fname, op->no_space, op->sense, &op->sense_len,
+                        MAX_SENSE_LEN);
         if (ret) {
-            fprintf(stderr, "unable to decode ASCII hex from file: %s\n",
-                    opts.fname);
+            pr2serr("unable to decode ASCII hex from file: %s\n", op->fname);
             return SG_LIB_SYNTAX_ERROR;
         }
     }
 
-    if (opts.sense_len) {
-        if (opts.wfname) {
-            if ((fp = fopen(opts.wfname, "w"))) {
-                write2wfn(fp, &opts);
+    if (op->sense_len) {
+        if (op->wfname) {
+            if ((fp = fopen(op->wfname, "w"))) {
+                write2wfn(fp, op);
                 fclose(fp);
             } else {
                 perror("open");
-                fprintf(stderr, "trying to write to %s\n", opts.wfname);
+                pr2serr("trying to write to %s\n", op->wfname);
             }
         }
-        sg_get_sense_str(NULL, opts.sense, opts.sense_len, opts.do_verbose,
-                         sizeof(b) - 1, b);
+        if (op->do_cdb) {
+            int sa, opcode;
+
+            opcode = op->sense[0];
+            if ((0x75 == opcode) || (0x7e == opcode) || (op->sense_len > 16))
+                sa = sg_get_unaligned_be16(op->sense + 8);
+            else if (op->sense_len > 1)
+                sa = op->sense[1] & 0x1f;
+            else
+                sa = 0;
+            sg_get_opcode_sa_name(opcode, sa, 0, sizeof(b), b);
+        } else
+            sg_get_sense_str(NULL, op->sense, op->sense_len,
+                             op->do_verbose, sizeof(b) - 1, b);
         printf("%s\n", b);
     }
 
