@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2004-2013 Douglas Gilbert.
+ * Copyright (c) 2004-2017 Douglas Gilbert.
  * All rights reserved.
  * Use of this source code is governed by a BSD-style
  * license that can be found in the BSD_LICENSE file.
@@ -9,6 +9,8 @@
 #include <fcntl.h>
 #include <stdio.h>
 #include <stdlib.h>
+#include <stdarg.h>
+#include <stdbool.h>
 #include <string.h>
 #include <getopt.h>
 
@@ -17,6 +19,7 @@
 #endif
 #include "sg_lib.h"
 #include "sg_cmds_basic.h"
+#include "sg_pr2serr.h"
 
 /* A utility program originally written for the Linux OS SCSI subsystem.
  *
@@ -24,45 +27,44 @@
  * given SCSI device.
  */
 
-static const char * version_str = "1.06 20130507";
+static const char * version_str = "1.10 20171006";
 
 #define ME "sg_prevent: "
 
 
 static struct option long_options[] = {
-        {"allow", 0, 0, 'a'},
-        {"help", 0, 0, 'h'},
-        {"prevent", 1, 0, 'p'},
-        {"verbose", 0, 0, 'v'},
-        {"version", 0, 0, 'V'},
+        {"allow", no_argument, 0, 'a'},
+        {"help", no_argument, 0, 'h'},
+        {"prevent", required_argument, 0, 'p'},
+        {"verbose", no_argument, 0, 'v'},
+        {"version", no_argument, 0, 'V'},
         {0, 0, 0, 0},
 };
 
 static void usage()
 {
-    fprintf(stderr, "Usage: "
-          "sg_prevent [--allow] [--help] [--prevent=PC] [--verbose] "
-          "[--version]\n"
-          "                  DEVICE\n"
-          "  where:\n"
-          "    --allow|-a            allow media removal\n"
-          "    --help|-h             print usage message then exit\n"
-          "    --prevent=PC|-p PC    prevent code value (def: 1 -> "
-          "prevent)\n"
-          "                            0 -> allow, 1 -> prevent\n"
-          "                            2 -> persistent allow, 3 -> "
-          "persistent prevent\n"
-          "    --verbose|-v          increase verbosity\n"
-          "    --version|-V          print version string and exit\n\n"
-          "Performs a SCSI PREVENT ALLOW MEDIUM REMOVAL command\n"
-          );
+    pr2serr("Usage: "
+            "sg_prevent [--allow] [--help] [--prevent=PC] [--verbose] "
+            "[--version]\n"
+            "                  DEVICE\n"
+            "  where:\n"
+            "    --allow|-a            allow media removal\n"
+            "    --help|-h             print usage message then exit\n"
+            "    --prevent=PC|-p PC    prevent code value (def: 1 -> "
+            "prevent)\n"
+            "                            0 -> allow, 1 -> prevent\n"
+            "                            2 -> persistent allow, 3 -> "
+            "persistent prevent\n"
+            "    --verbose|-v          increase verbosity\n"
+            "    --version|-V          print version string and exit\n\n"
+            "Performs a SCSI PREVENT ALLOW MEDIUM REMOVAL command\n");
 
 }
 
 int main(int argc, char * argv[])
 {
     int sg_fd, res, c;
-    int allow = 0;
+    bool allow = false;
     int prevent = -1;
     int verbose = 0;
     const char * device_name = NULL;
@@ -78,7 +80,7 @@ int main(int argc, char * argv[])
 
         switch (c) {
         case 'a':
-            allow = 1;
+            allow = true;
             break;
         case 'h':
         case '?':
@@ -87,7 +89,7 @@ int main(int argc, char * argv[])
         case 'p':
            prevent = sg_get_num(optarg);
            if ((prevent < 0) || (prevent > 3)) {
-                fprintf(stderr, "bad argument to '--prevent'\n");
+                pr2serr("bad argument to '--prevent'\n");
                 return SG_LIB_SYNTAX_ERROR;
             }
             break;
@@ -95,10 +97,10 @@ int main(int argc, char * argv[])
             ++verbose;
             break;
         case 'V':
-            fprintf(stderr, ME "version: %s\n", version_str);
+            pr2serr(ME "version: %s\n", version_str);
             return 0;
         default:
-            fprintf(stderr, "unrecognised option code 0x%x ??\n", c);
+            pr2serr("unrecognised option code 0x%x ??\n", c);
             usage();
             return SG_LIB_SYNTAX_ERROR;
         }
@@ -110,19 +112,18 @@ int main(int argc, char * argv[])
         }
         if (optind < argc) {
             for (; optind < argc; ++optind)
-                fprintf(stderr, "Unexpected extra argument: %s\n",
-                        argv[optind]);
+                pr2serr("Unexpected extra argument: %s\n", argv[optind]);
             usage();
             return SG_LIB_SYNTAX_ERROR;
         }
     }
     if (NULL == device_name) {
-        fprintf(stderr, "missing device name!\n");
+        pr2serr("missing device name!\n");
         usage();
         return SG_LIB_SYNTAX_ERROR;
     }
     if (allow && (prevent >= 0)) {
-        fprintf(stderr, "can't give both '--allow' and '--prevent='\n");
+        pr2serr("can't give both '--allow' and '--prevent='\n");
         usage();
         return SG_LIB_SYNTAX_ERROR;
     }
@@ -131,34 +132,23 @@ int main(int argc, char * argv[])
     else if (prevent < 0)
         prevent = 1;    /* default is to prevent, as utility name suggests */
 
-    sg_fd = sg_cmds_open_device(device_name, 0 /* rw */, verbose);
+    sg_fd = sg_cmds_open_device(device_name, false /* rw */, verbose);
     if (sg_fd < 0) {
-        fprintf(stderr, ME "open error: %s: %s\n", device_name,
+        pr2serr(ME "open error: %s: %s\n", device_name,
                 safe_strerror(-sg_fd));
         return SG_LIB_FILE_ERROR;
     }
-    res = sg_ll_prevent_allow(sg_fd, prevent, 1, verbose);
+    res = sg_ll_prevent_allow(sg_fd, prevent, true, verbose);
     ret = res;
-    if (0 == res)
-        ;
-    else if (SG_LIB_CAT_NOT_READY == res)
-        fprintf(stderr, "Device not ready\n");
-    else if (SG_LIB_CAT_UNIT_ATTENTION == res)
-        fprintf(stderr, "Unit attention\n");
-    else if (SG_LIB_CAT_ABORTED_COMMAND == res)
-        fprintf(stderr, "Aborted command\n");
-    else if (SG_LIB_CAT_INVALID_OP == res)
-        fprintf(stderr, "Prevent allow medium removal command not "
-                "supported\n");
-    else if (SG_LIB_CAT_ILLEGAL_REQ == res)
-        fprintf(stderr, "Prevent allow medium removal, bad field in "
-                "command\n");
-    else
-        fprintf(stderr, "Prevent allow medium removal command failed\n");
+    if (res) {
+        char b[80];
 
+        sg_get_category_sense_str(res, sizeof(b), b, verbose);
+        pr2serr("Prevent allow medium removal: %s\n", b);
+    }
     res = sg_cmds_close_device(sg_fd);
     if (res < 0) {
-        fprintf(stderr, "close error: %s\n", safe_strerror(-res));
+        pr2serr("close error: %s\n", safe_strerror(-res));
         if (0 == ret)
             return SG_LIB_FILE_ERROR;
     }
