@@ -1,5 +1,5 @@
 /* A utility program originally written for the Linux OS SCSI subsystem.
- *  Copyright (C) 1999-2014 D. Gilbert
+ *  Copyright (C) 1999-2018 D. Gilbert
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
  *  the Free Software Foundation; either version 2, or (at your option)
@@ -11,15 +11,17 @@
  */
 
 
-#define _XOPEN_SOURCE 500
+#define _XOPEN_SOURCE 600
 #ifndef _GNU_SOURCE
-#define _GNU_SOURCE
+#define _GNU_SOURCE 1
 #endif
 
 #include <unistd.h>
 #include <fcntl.h>
 #include <stdio.h>
 #include <stdlib.h>
+#include <stdarg.h>
+#include <stdbool.h>
 #include <string.h>
 #include <errno.h>
 #include <getopt.h>
@@ -36,6 +38,8 @@
 #endif
 #include "sg_lib.h"
 #include "sg_io_linux.h"
+#include "sg_unaligned.h"
+#include "sg_pr2serr.h"
 
 #define RB_MODE_DESC 3
 #define RB_MODE_DATA 2
@@ -53,7 +57,7 @@
 #endif
 
 
-static const char * version_str = "4.92 20141227";
+static const char * version_str = "5.03 20180302";
 
 static struct option long_options[] = {
         {"buffer", required_argument, 0, 'b'},
@@ -72,51 +76,52 @@ static struct option long_options[] = {
 };
 
 struct opts_t {
+    bool do_dio;
+    bool do_echo;
+    bool do_mmap;
+    bool do_quick;
+    bool do_time;
+    bool do_version;
+    bool opt_new;
     int do_buffer;
-    int do_dio;
-    int do_echo;
     int do_help;
-    int do_mmap;
-    int do_quick;
-    int64_t do_size;
-    int do_time;
     int do_verbose;
-    int do_version;
+    int64_t do_size;
     const char * device_name;
-    int opt_new;
 };
 
 
 static void
 usage()
 {
-    fprintf(stderr, "Usage: sg_rbuf [--buffer=EACH] [--dio] [--echo] "
+    pr2serr("Usage: sg_rbuf [--buffer=EACH] [--dio] [--echo] "
             "[--help] [--mmap]\n"
             "               [--quick] [--size=OVERALL] [--time] [--verbose] "
             "[--version]\n"
-            "               DEVICE\n");
-    fprintf(stderr, "  where:\n"
+            "               SG_DEVICE\n");
+    pr2serr("  where:\n"
             "    --buffer=EACH|-b EACH    buffer size to use (in bytes)\n"
             "    --dio|-d        requests dio ('-q' overrides it)\n"
             "    --echo|-e       use echo buffer (def: use data mode)\n"
             "    --help|-h       print usage message then exit\n"
             "    --mmap|-m       requests mmap-ed IO (overrides -q, -d)\n"
             "    --quick|-q      quick, don't xfer to user space\n");
-    fprintf(stderr,
-            "    --size=OVERALL|-s OVERALL    total size to read (in bytes)\n"
+    pr2serr("    --size=OVERALL|-s OVERALL    total size to read (in bytes)\n"
             "                    default: 200 MiB\n"
             "    --time|-t       time the data transfer\n"
             "    --verbose|-v    increase verbosity (more debug)\n"
+            "    --old|-O        use old interface (use as first option)\n"
             "    --version|-V    print version string then exit\n\n"
             "Use SCSI READ BUFFER command (data or echo buffer mode, buffer "
-            "id 0)\nrepeatedly\n");
+            "id 0)\nrepeatedly. This utility only works with Linux sg "
+            "devices.\n");
 }
 
 static void
 usage_old()
 {
     printf("Usage: sg_rbuf [-b=EACH_KIB] [-d] [-m] [-q] [-s=OVERALL_MIB] "
-           "[-t] [-v] [-V]\n               DEVICE\n");
+           "[-t] [-v] [-V]\n               SG_DEVICE\n");
     printf("  where:\n");
     printf("    -b=EACH_KIB    num is buffer size to use (in KiB)\n");
     printf("    -d       requests dio ('-q' overrides it)\n");
@@ -128,9 +133,11 @@ usage_old()
     printf("             maximum total size is 4000 MiB\n");
     printf("    -t       time the data transfer\n");
     printf("    -v       increase verbosity (more debug)\n");
+    printf("    -N|--new use new interface\n");
     printf("    -V       print version string then exit\n\n");
     printf("Use SCSI READ BUFFER command (data or echo buffer mode, buffer "
-           "id 0)\nrepeatedly\n");
+           "id 0)\nrepeatedly. This utility only works with Linux sg "
+            "devices.\n");
 }
 
 static void
@@ -143,7 +150,7 @@ usage_for(const struct opts_t * optsp)
 }
 
 static int
-process_cl_new(struct opts_t * optsp, int argc, char * argv[])
+new_parse_cmd_line(struct opts_t * optsp, int argc, char * argv[])
 {
     int c, n;
     int64_t nn;
@@ -160,53 +167,53 @@ process_cl_new(struct opts_t * optsp, int argc, char * argv[])
         case 'b':
             n = sg_get_num(optarg);
             if (n < 0) {
-                fprintf(stderr, "bad argument to '--buffer'\n");
+                pr2serr("bad argument to '--buffer'\n");
                 usage_for(optsp);
                 return SG_LIB_SYNTAX_ERROR;
             }
             optsp->do_buffer = n;
             break;
         case 'd':
-            ++optsp->do_dio;
+            optsp->do_dio = true;
             break;
         case 'e':
-            ++optsp->do_echo;
+            optsp->do_echo = true;
             break;
         case 'h':
         case '?':
             ++optsp->do_help;
             break;
         case 'm':
-            ++optsp->do_mmap;
+            optsp->do_mmap = true;
             break;
         case 'N':
             break;      /* ignore */
         case 'O':
-            optsp->opt_new = 0;
+            optsp->opt_new = false;
             return 0;
         case 'q':
-            ++optsp->do_quick;
+            optsp->do_quick = true;
             break;
         case 's':
            nn = sg_get_llnum(optarg);
            if (nn < 0) {
-                fprintf(stderr, "bad argument to '--size'\n");
+                pr2serr("bad argument to '--size'\n");
                 usage_for(optsp);
                 return SG_LIB_SYNTAX_ERROR;
             }
             optsp->do_size = nn;
             break;
         case 't':
-            ++optsp->do_time;
+            optsp->do_time = true;
             break;
         case 'v':
             ++optsp->do_verbose;
             break;
         case 'V':
-            ++optsp->do_version;
+            optsp->do_version = true;
             break;
         default:
-            fprintf(stderr, "unrecognised option code %c [0x%x]\n", c, c);
+            pr2serr("unrecognised option code %c [0x%x]\n", c, c);
             if (optsp->do_help)
                 break;
             usage_for(optsp);
@@ -220,8 +227,7 @@ process_cl_new(struct opts_t * optsp, int argc, char * argv[])
         }
         if (optind < argc) {
             for (; optind < argc; ++optind)
-                fprintf(stderr, "Unexpected extra argument: %s\n",
-                        argv[optind]);
+                pr2serr("Unexpected extra argument: %s\n", argv[optind]);
             usage_for(optsp);
             return SG_LIB_SYNTAX_ERROR;
         }
@@ -230,9 +236,10 @@ process_cl_new(struct opts_t * optsp, int argc, char * argv[])
 }
 
 static int
-process_cl_old(struct opts_t * optsp, int argc, char * argv[])
+old_parse_cmd_line(struct opts_t * optsp, int argc, char * argv[])
 {
-    int k, jmp_out, plen, num;
+    bool jmp_out;
+    int k, plen, num;
     int64_t nn;
     const char * cp;
 
@@ -242,40 +249,40 @@ process_cl_old(struct opts_t * optsp, int argc, char * argv[])
         if (plen <= 0)
             continue;
         if ('-' == *cp) {
-            for (--plen, ++cp, jmp_out = 0; plen > 0; --plen, ++cp) {
+            for (--plen, ++cp, jmp_out = false; plen > 0; --plen, ++cp) {
                 switch (*cp) {
                 case 'd':
-                    ++optsp->do_dio;
+                    optsp->do_dio = true;
                     break;
                 case 'e':
-                    ++optsp->do_echo;
+                    optsp->do_echo = true;
                     break;
                 case 'h':
                 case '?':
                     ++optsp->do_help;
                     break;
                 case 'm':
-                    ++optsp->do_mmap;
+                    optsp->do_mmap = true;
                     break;
                 case 'N':
-                    optsp->opt_new = 1;
+                    optsp->opt_new = true;
                     return 0;
                 case 'O':
                     break;
                 case 'q':
-                    ++optsp->do_quick;
+                    optsp->do_quick = true;
                     break;
                 case 't':
-                    ++optsp->do_time;
+                    optsp->do_time = true;
                     break;
                 case 'v':
                     ++optsp->do_verbose;
                     break;
                 case 'V':
-                    ++optsp->do_version;
+                    optsp->do_version = true;
                     break;
                 default:
-                    jmp_out = 1;
+                    jmp_out = true;
                     break;
                 }
                 if (jmp_out)
@@ -304,15 +311,15 @@ process_cl_old(struct opts_t * optsp, int argc, char * argv[])
             } else if (0 == strncmp("-old", cp, 4))
                 ;
             else if (jmp_out) {
-                fprintf(stderr, "Unrecognized option: %s\n", cp);
+                pr2serr("Unrecognized option: %s\n", cp);
                 usage_for(optsp);
                 return SG_LIB_SYNTAX_ERROR;
             }
         } else if (0 == optsp->device_name)
             optsp->device_name = cp;
         else {
-            fprintf(stderr, "too many arguments, got: %s, not expecting: "
-                    "%s\n", optsp->device_name, cp);
+            pr2serr("too many arguments, got: %s, not expecting: %s\n",
+                    optsp->device_name, cp);
             usage_for(optsp);
             return SG_LIB_SYNTAX_ERROR;
         }
@@ -321,22 +328,22 @@ process_cl_old(struct opts_t * optsp, int argc, char * argv[])
 }
 
 static int
-process_cl(struct opts_t * optsp, int argc, char * argv[])
+parse_cmd_line(struct opts_t * optsp, int argc, char * argv[])
 {
     int res;
     char * cp;
 
     cp = getenv("SG3_UTILS_OLD_OPTS");
     if (cp) {
-        optsp->opt_new = 0;
-        res = process_cl_old(optsp, argc, argv);
+        optsp->opt_new = false;
+        res = old_parse_cmd_line(optsp, argc, argv);
         if ((0 == res) && optsp->opt_new)
-            res = process_cl_new(optsp, argc, argv);
+            res = new_parse_cmd_line(optsp, argc, argv);
     } else {
-        optsp->opt_new = 1;
-        res = process_cl_new(optsp, argc, argv);
-        if ((0 == res) && (0 == optsp->opt_new))
-            res = process_cl_old(optsp, argc, argv);
+        optsp->opt_new = true;
+        res = new_parse_cmd_line(optsp, argc, argv);
+        if ((0 == res) && (! optsp->opt_new))
+            res = old_parse_cmd_line(optsp, argc, argv);
     }
     return res;
 }
@@ -345,24 +352,24 @@ process_cl(struct opts_t * optsp, int argc, char * argv[])
 int
 main(int argc, char * argv[])
 {
+#ifdef SG_DEBUG
+    bool clear = true;
+#endif
+    bool dio_incomplete = false;
     int sg_fd, res, j;
-    unsigned int k, num;
-    unsigned char rbCmdBlk [RB_CMD_LEN];
-    unsigned char * rbBuff = NULL;
-    void * rawp = NULL;
-    unsigned char sense_buffer[32];
     int buf_capacity = 0;
     int buf_size = 0;
-    int64_t total_size = RB_DEF_SIZE;
     size_t psz;
-    int dio_incomplete = 0;
+    unsigned int k, num;
+    int64_t total_size = RB_DEF_SIZE;
+    struct opts_t * op;
+    uint8_t * rbBuff = NULL;
+    void * rawp = NULL;
+    uint8_t sense_buffer[32];
+    uint8_t rb_cdb [RB_CMD_LEN];
     struct sg_io_hdr io_hdr;
     struct timeval start_tm, end_tm;
-#ifdef SG_DEBUG
-    int clear = 1;
-#endif
     struct opts_t opts;
-    struct opts_t * op;
 
 #if defined(HAVE_SYSCONF) && defined(_SC_PAGESIZE)
     psz = sysconf(_SC_PAGESIZE); /* POSIX.1 (was getpagesize()) */
@@ -371,7 +378,7 @@ main(int argc, char * argv[])
 #endif
     op = &opts;
     memset(op, 0, sizeof(opts));
-    res = process_cl(op, argc, argv);
+    res = parse_cmd_line(op, argc, argv);
     if (res)
         return SG_LIB_SYNTAX_ERROR;
     if (op->do_help) {
@@ -379,12 +386,12 @@ main(int argc, char * argv[])
         return 0;
     }
     if (op->do_version) {
-        fprintf(stderr, "Version string: %s\n", version_str);
+        pr2serr("Version string: %s\n", version_str);
         return 0;
     }
 
     if (NULL == op->device_name) {
-        fprintf(stderr, "No DEVICE argument given\n");
+        pr2serr("No DEVICE argument given\n");
         usage_for(op);
         return SG_LIB_SYNTAX_ERROR;
     }
@@ -400,53 +407,59 @@ main(int argc, char * argv[])
         return SG_LIB_FILE_ERROR;
     }
     if (op->do_mmap) {
-        op->do_dio = 0;
-        op->do_quick = 0;
+        op->do_dio = false;
+        op->do_quick = false;
     }
     if (NULL == (rawp = malloc(512))) {
         printf("out of memory (query)\n");
         return SG_LIB_CAT_OTHER;
     }
-    rbBuff = (unsigned char *)rawp;
+    rbBuff = (uint8_t *)rawp;
 
-    memset(rbCmdBlk, 0, RB_CMD_LEN);
-    rbCmdBlk[0] = RB_OPCODE;
-    rbCmdBlk[1] = op->do_echo ? RB_MODE_ECHO_DESC : RB_MODE_DESC;
-    rbCmdBlk[8] = RB_DESC_LEN;
+    memset(rb_cdb, 0, RB_CMD_LEN);
+    rb_cdb[0] = RB_OPCODE;
+    rb_cdb[1] = op->do_echo ? RB_MODE_ECHO_DESC : RB_MODE_DESC;
+    rb_cdb[8] = RB_DESC_LEN;
     memset(&io_hdr, 0, sizeof(struct sg_io_hdr));
     io_hdr.interface_id = 'S';
-    io_hdr.cmd_len = sizeof(rbCmdBlk);
+    io_hdr.cmd_len = sizeof(rb_cdb);
     io_hdr.mx_sb_len = sizeof(sense_buffer);
     io_hdr.dxfer_direction = SG_DXFER_FROM_DEV;
     io_hdr.dxfer_len = RB_DESC_LEN;
     io_hdr.dxferp = rbBuff;
-    io_hdr.cmdp = rbCmdBlk;
+    io_hdr.cmdp = rb_cdb;
     io_hdr.sbp = sense_buffer;
     io_hdr.timeout = 60000;     /* 60000 millisecs == 60 seconds */
     if (op->do_verbose) {
-        fprintf(stderr, "    Read buffer (%sdescriptor) cdb: ",
+        pr2serr("    Read buffer (%sdescriptor) cdb: ",
                 (op->do_echo ? "echo " : ""));
         for (k = 0; k < RB_CMD_LEN; ++k)
-            fprintf(stderr, "%02x ", rbCmdBlk[k]);
-        fprintf(stderr, "\n");
+            pr2serr("%02x ", rb_cdb[k]);
+        pr2serr("\n");
     }
 
     /* do normal IO to find RB size (not dio or mmap-ed at this stage) */
     if (ioctl(sg_fd, SG_IO, &io_hdr) < 0) {
         perror("SG_IO READ BUFFER descriptor error");
-        if (rawp) free(rawp);
+        if (rawp)
+            free(rawp);
         return SG_LIB_CAT_OTHER;
     }
 
     if (op->do_verbose > 2)
-        fprintf(stderr, "      duration=%u ms\n", io_hdr.duration);
+        pr2serr("      duration=%u ms\n", io_hdr.duration);
     /* now for the error processing */
     res = sg_err_category3(&io_hdr);
     switch (res) {
     case SG_LIB_CAT_RECOVERED:
         sg_chk_n_print3("READ BUFFER descriptor, continuing", &io_hdr,
                         op->do_verbose > 1);
-        /* fall through */
+#if defined(__GNUC__)
+#if (__GNUC__ >= 7)
+        __attribute__((fallthrough));
+        /* FALL THROUGH */
+#endif
+#endif
     case SG_LIB_CAT_CLEAN:
         break;
     default: /* won't bother decoding other categories */
@@ -457,11 +470,11 @@ main(int argc, char * argv[])
     }
 
     if (op->do_echo) {
-        buf_capacity = (((0x1f & rbBuff[2]) << 8) | rbBuff[3]);
+        buf_capacity = 0x1fff & sg_get_unaligned_be16(rbBuff + 2);
         printf("READ BUFFER reports: echo buffer capacity=%d\n",
                buf_capacity);
     } else {
-        buf_capacity = ((rbBuff[1] << 16) | (rbBuff[2] << 8) | rbBuff[3]);
+        buf_capacity = sg_get_unaligned_be24(rbBuff + 1);
         printf("READ BUFFER reports: buffer capacity=%d, offset "
                "boundary=%d\n", buf_capacity, (int)rbBuff[0]);
     }
@@ -489,35 +502,34 @@ main(int argc, char * argv[])
     }
 
     if (op->do_mmap) {
-        rbBuff = (unsigned char *)mmap(NULL, buf_size, PROT_READ, MAP_SHARED,
+        rbBuff = (uint8_t *)mmap(NULL, buf_size, PROT_READ, MAP_SHARED,
                                        sg_fd, 0);
         if (MAP_FAILED == rbBuff) {
             if (ENOMEM == errno) {
-                fprintf(stderr, "mmap() out of memory, try a smaller "
-                       "buffer size than %d bytes\n", buf_size);
+                pr2serr("mmap() out of memory, try a smaller buffer size "
+                        "than %d bytes\n", buf_size);
                 if (op->opt_new)
-                    fprintf(stderr, "    [with '--buffer=EACH' where EACH "
-                            "is in bytes]\n");
+                    pr2serr("    [with '--buffer=EACH' where EACH is in "
+                            "bytes]\n");
                 else
-                    fprintf(stderr, "    [with '-b=EACH' where EACH is in "
-                            "KiB]\n");
+                    pr2serr("    [with '-b=EACH' where EACH is in KiB]\n");
             } else
                 perror("error using mmap()");
             return SG_LIB_CAT_OTHER;
         }
     }
     else { /* non mmap-ed IO */
-        rawp = (unsigned char *)malloc(buf_size + (op->do_dio ? psz : 0));
+        rawp = (uint8_t *)malloc(buf_size + (op->do_dio ? psz : 0));
         if (NULL == rawp) {
             printf("out of memory (data)\n");
             return SG_LIB_CAT_OTHER;
         }
         /* perhaps use posix_memalign() instead */
         if (op->do_dio)    /* align to page boundary */
-            rbBuff= (unsigned char *)(((uintptr_t)rawp + psz - 1) &
+            rbBuff= (uint8_t *)(((sg_uintptr_t)rawp + psz - 1) &
                                       (~(psz - 1)));
         else
-            rbBuff = (unsigned char *)rawp;
+            rbBuff = (uint8_t *)rawp;
     }
 
     num = total_size / buf_size;
@@ -528,25 +540,23 @@ main(int argc, char * argv[])
     }
     /* main data reading loop */
     for (k = 0; k < num; ++k) {
-        memset(rbCmdBlk, 0, RB_CMD_LEN);
-        rbCmdBlk[0] = RB_OPCODE;
-        rbCmdBlk[1] = op->do_echo ? RB_MODE_ECHO_DATA : RB_MODE_DATA;
-        rbCmdBlk[6] = 0xff & (buf_size >> 16);
-        rbCmdBlk[7] = 0xff & (buf_size >> 8);
-        rbCmdBlk[8] = 0xff & buf_size;
+        memset(rb_cdb, 0, RB_CMD_LEN);
+        rb_cdb[0] = RB_OPCODE;
+        rb_cdb[1] = op->do_echo ? RB_MODE_ECHO_DATA : RB_MODE_DATA;
+        sg_put_unaligned_be24((uint32_t)buf_size, rb_cdb + 6);
 #ifdef SG_DEBUG
         memset(rbBuff, 0, buf_size);
 #endif
 
         memset(&io_hdr, 0, sizeof(struct sg_io_hdr));
         io_hdr.interface_id = 'S';
-        io_hdr.cmd_len = sizeof(rbCmdBlk);
+        io_hdr.cmd_len = sizeof(rb_cdb);
         io_hdr.mx_sb_len = sizeof(sense_buffer);
         io_hdr.dxfer_direction = SG_DXFER_FROM_DEV;
         io_hdr.dxfer_len = buf_size;
         if (! op->do_mmap)
             io_hdr.dxferp = rbBuff;
-        io_hdr.cmdp = rbCmdBlk;
+        io_hdr.cmdp = rb_cdb;
         io_hdr.sbp = sense_buffer;
         io_hdr.timeout = 20000;     /* 20000 millisecs == 20 seconds */
         io_hdr.pack_id = k;
@@ -557,23 +567,22 @@ main(int argc, char * argv[])
         else if (op->do_quick)
             io_hdr.flags |= SG_FLAG_NO_DXFER;
         if (op->do_verbose > 1) {
-            fprintf(stderr, "    Read buffer (%sdata) cdb: ",
+            pr2serr("    Read buffer (%sdata) cdb: ",
                     (op->do_echo ? "echo " : ""));
             for (j = 0; j < RB_CMD_LEN; ++j)
-                fprintf(stderr, "%02x ", rbCmdBlk[j]);
-            fprintf(stderr, "\n");
+                pr2serr("%02x ", rb_cdb[j]);
+            pr2serr("\n");
         }
 
         if (ioctl(sg_fd, SG_IO, &io_hdr) < 0) {
             if (ENOMEM == errno) {
-                fprintf(stderr, "SG_IO data: out of memory, try a smaller "
-                       "buffer size than %d bytes\n", buf_size);
+                pr2serr("SG_IO data: out of memory, try a smaller buffer "
+                        "size than %d bytes\n", buf_size);
                 if (op->opt_new)
-                    fprintf(stderr, "    [with '--buffer=EACH' where EACH "
-                            "is in bytes]\n");
+                    pr2serr("    [with '--buffer=EACH' where EACH is in "
+                            "bytes]\n");
                 else
-                    fprintf(stderr, "    [with '-b=EACH' where EACH is in "
-                            "KiB]\n");
+                    pr2serr("    [with '-b=EACH' where EACH is in KiB]\n");
             } else
                 perror("SG_IO READ BUFFER data error");
             if (rawp) free(rawp);
@@ -581,8 +590,7 @@ main(int argc, char * argv[])
         }
 
         if (op->do_verbose > 2)
-            fprintf(stderr, "      duration=%u ms\n",
-                    io_hdr.duration);
+            pr2serr("      duration=%u ms\n", io_hdr.duration);
         /* now for the error processing */
         res = sg_err_category3(&io_hdr);
         switch (res) {
@@ -600,20 +608,20 @@ main(int argc, char * argv[])
         }
         if (op->do_dio &&
             ((io_hdr.info & SG_INFO_DIRECT_IO_MASK) != SG_INFO_DIRECT_IO))
-            dio_incomplete = 1;    /* flag that dio not done (completely) */
+            dio_incomplete = true;  /* flag that dio not done (completely) */
 
 #ifdef SG_DEBUG
         if (clear) {
             for (j = 0; j < buf_size; ++j) {
                 if (rbBuff[j] != 0) {
-                    clear = 0;
+                    clear = false;
                     break;
                 }
             }
         }
 #endif
     }
-    if ((op->do_time) && (start_tm.tv_sec || start_tm.tv_usec)) {
+    if (op->do_time && (start_tm.tv_sec || start_tm.tv_usec)) {
         struct timeval res_tm;
         double a, b;
 
