@@ -3,7 +3,7 @@
  *
  * Retrieve / set RDAC options.
  *
- * Copyright (C) 2006-2015 Hannes Reinecke <hare@suse.de>
+ * Copyright (C) 2006-2018 Hannes Reinecke <hare@suse.de>
  *
  * Based on sg_modes.c and sg_emc_trespass.c; credits from there apply.
  *
@@ -15,26 +15,31 @@
 
 #include <stdio.h>
 #include <stdlib.h>
+#include <stdarg.h>
+#include <stdbool.h>
 #include <unistd.h>
 #include <string.h>
 
 #ifdef HAVE_CONFIG_H
 #include "config.h"
 #endif
+
 #include "sg_lib.h"
 #include "sg_cmds_basic.h"
+#include "sg_unaligned.h"
+#include "sg_pr2serr.h"
 
 
-static const char * version_str = "1.08 20150407";
+static const char * version_str = "1.16 20180219";
 
-unsigned char mode6_hdr[] = {
+uint8_t mode6_hdr[] = {
     0x75, /* Length */
     0, /* medium */
     0, /* params */
     8, /* Block descriptor length */
 };
 
-unsigned char mode10_hdr[] = {
+uint8_t mode10_hdr[] = {
     0x01, 0x18, /* Length */
     0, /* medium */
     0, /* params */
@@ -42,7 +47,7 @@ unsigned char mode10_hdr[] = {
     0, 0, /* block descriptor length */
 };
 
-unsigned char block_descriptor[] = {
+uint8_t block_descriptor[] = {
     0, /* Density code */
     0, 0, 0, /* Number of blocks */
     0, /* Reserved */
@@ -50,37 +55,37 @@ unsigned char block_descriptor[] = {
 };
 
 struct rdac_page_common {
-    unsigned char  current_serial[16];
-    unsigned char  alternate_serial[16];
-    unsigned char  current_mode_msb;
-    unsigned char  current_mode_lsb;
-    unsigned char  alternate_mode_msb;
-    unsigned char  alternate_mode_lsb;
-    unsigned char  quiescence;
-    unsigned char  options;
+    uint8_t  current_serial[16];
+    uint8_t  alternate_serial[16];
+    uint8_t  current_mode_msb;
+    uint8_t  current_mode_lsb;
+    uint8_t  alternate_mode_msb;
+    uint8_t  alternate_mode_lsb;
+    uint8_t  quiescence;
+    uint8_t  options;
 };
 
 struct rdac_legacy_page {
-    unsigned char  page_code;
-    unsigned char  page_length;
+    uint8_t  page_code;
+    uint8_t  page_length;
     struct rdac_page_common attr;
-    unsigned char  lun_table[32];
-    unsigned char  lun_table_exp[32];
+    uint8_t  lun_table[32];
+    uint8_t  lun_table_exp[32];
     unsigned short reserved;
 };
 
 struct rdac_expanded_page {
-    unsigned char  page_code;
-    unsigned char  subpage_code;
-    unsigned char  page_length[2];
+    uint8_t  page_code;
+    uint8_t  subpage_code;
+    uint8_t  page_length[2];
     struct rdac_page_common attr;
-    unsigned char  lun_table[256];
-    unsigned char  reserved[2];
+    uint8_t  lun_table[256];
+    uint8_t  reserved[2];
 };
 
 static int do_verbose = 0;
 
-static void dump_mode_page( unsigned char *page, int len )
+static void dump_mode_page( uint8_t *page, int len )
 {
         int i, k;
 
@@ -110,12 +115,12 @@ static void dump_mode_page( unsigned char *page, int len )
 #define RDAC_FORCE_QUIESCENCE 0x2
 #define RDAC_QUIESCENCE_TIME 10
 
-static int fail_all_paths(int fd, int use_6_byte)
+static int fail_all_paths(int fd, bool use_6_byte)
 {
-        unsigned char fail_paths_pg[308];
         struct rdac_legacy_page *rdac_page;
         struct rdac_expanded_page *rdac_page_exp;
         struct rdac_page_common *rdac_common = NULL;
+        uint8_t fail_paths_pg[308];
 
         int res;
         char b[80];
@@ -134,10 +139,8 @@ static int fail_all_paths(int fd, int use_6_byte)
                                 (fail_paths_pg + 8);
                 rdac_page_exp->page_code = RDAC_CONTROLLER_PAGE | 0x40;
                 rdac_page_exp->subpage_code = 0x1;
-                rdac_page_exp->page_length[0] =
-                                EXPANDED_LUN_SPACE_PAGE_LEN >> 8;
-                rdac_page_exp->page_length[1] =
-                                EXPANDED_LUN_SPACE_PAGE_LEN & 0xFF;
+                sg_put_unaligned_be16(EXPANDED_LUN_SPACE_PAGE_LEN,
+                                      rdac_page_exp->page_length + 0);
                 rdac_common = &rdac_page_exp->attr;
         }
 
@@ -148,39 +151,46 @@ static int fail_all_paths(int fd, int use_6_byte)
         if (use_6_byte) {
                 res = sg_ll_mode_select6(fd, 1 /* pf */, 0 /* sp */,
                                         fail_paths_pg, 118,
-                                        1, (do_verbose ? 2 : 0));
+                                        true, (do_verbose ? 2 : 0));
         } else {
                 res = sg_ll_mode_select10(fd, 1 /* pf */, 0 /* sp */,
                                         fail_paths_pg, 308,
-                                        1, (do_verbose ? 2: 0));
+                                        true, (do_verbose ? 2: 0));
         }
 
         switch (res) {
         case 0:
                 if (do_verbose)
-                        fprintf(stderr, "fail paths successful\n");
+                        pr2serr("fail paths successful\n");
                 break;
         default:
                 sg_get_category_sense_str(res, sizeof(b), b, do_verbose);
-                fprintf(stderr, "fail paths failed: %s\n", b);
+                pr2serr("fail paths failed: %s\n", b);
                 break;
         }
 
         return res;
 }
 
-static int fail_this_path(int fd, int lun, int use_6_byte)
+static int fail_this_path(int fd, int lun, bool use_6_byte)
 {
-        unsigned char fail_paths_pg[308];
+        int res;
         struct rdac_legacy_page *rdac_page;
         struct rdac_expanded_page *rdac_page_exp;
         struct rdac_page_common *rdac_common = NULL;
-        int res;
+        uint8_t fail_paths_pg[308];
         char b[80];
 
-        if (use_6_byte && lun > 32) {
-                fprintf(stderr, "must use 10 byte cdb to fail luns over 32\n");
-                return -1;
+        if (use_6_byte) {
+                if (lun > 31) {
+                        pr2serr("must use 10 byte cdb to fail luns over 31\n");
+                        return -1;
+                }
+        } else {        /* 10 byte cdb case */
+                if (lun > 255) {
+                        pr2serr("lun cannot exceed 255\n");
+                        return -1;
+                }
         }
 
         memset(fail_paths_pg, 0, 308);
@@ -199,10 +209,8 @@ static int fail_this_path(int fd, int lun, int use_6_byte)
                                 (fail_paths_pg + 8);
                 rdac_page_exp->page_code = RDAC_CONTROLLER_PAGE | 0x40;
                 rdac_page_exp->subpage_code = 0x1;
-                rdac_page_exp->page_length[0] =
-                                EXPANDED_LUN_SPACE_PAGE_LEN >> 8;
-                rdac_page_exp->page_length[1] =
-                                EXPANDED_LUN_SPACE_PAGE_LEN & 0xFF;
+                sg_put_unaligned_be16(EXPANDED_LUN_SPACE_PAGE_LEN,
+                                      rdac_page_exp->page_length + 0);
                 rdac_common = &rdac_page_exp->attr;
                 memset(rdac_page_exp->lun_table, 0x0, 256);
                 rdac_page_exp->lun_table[lun] = 0x81;
@@ -215,37 +223,36 @@ static int fail_this_path(int fd, int lun, int use_6_byte)
         if (use_6_byte) {
                 res = sg_ll_mode_select6(fd, 1 /* pf */, 0 /* sp */,
                                         fail_paths_pg, 118,
-                                        1, (do_verbose ? 2 : 0));
+                                        true, (do_verbose ? 2 : 0));
         } else {
                 res = sg_ll_mode_select10(fd, 1 /* pf */, 0 /* sp */,
                                         fail_paths_pg, 308,
-                                        1, (do_verbose ? 2: 0));
+                                        true, (do_verbose ? 2: 0));
         }
 
         switch (res) {
         case 0:
                 if (do_verbose)
-                        fprintf(stderr, "fail paths successful\n");
+                        pr2serr("fail paths successful\n");
                 break;
         default:
                 sg_get_category_sense_str(res, sizeof(b), b, do_verbose);
-                fprintf(stderr, "fail paths page (lun=%d) failed: %s\n", lun,
-                        b);
+                pr2serr("fail paths page (lun=%d) failed: %s\n", lun, b);
                 break;
         }
 
         return res;
 }
 
-static void print_rdac_mode( unsigned char *ptr, int subpg)
+static void print_rdac_mode(uint8_t *ptr, bool exp_subpg)
 {
+        int i, k, bd_len, lun_table_len;
+        uint8_t * lun_table = NULL;
         struct rdac_legacy_page *legacy;
         struct rdac_expanded_page *expanded;
         struct rdac_page_common *rdac_ptr = NULL;
-        unsigned char * lun_table = NULL;
-        int i, k, bd_len, lun_table_len;
 
-        if (subpg == 1) {
+        if (exp_subpg) {
                 bd_len = ptr[7];
                 expanded = (struct rdac_expanded_page *)(ptr + 8 + bd_len);
                 rdac_ptr = &expanded->attr;
@@ -259,7 +266,7 @@ static void print_rdac_mode( unsigned char *ptr, int subpg)
                 lun_table_len = 32;
         }
 
-        printf("RDAC %s page\n", (subpg == 1) ? "Expanded" : "Legacy");
+        printf("RDAC %s page\n", exp_subpg ? "Expanded" : "Legacy");
         printf("  Controller serial: %s\n",
                rdac_ptr->current_serial);
         printf("  Alternate controller serial: %s\n",
@@ -379,14 +386,14 @@ static void usage()
 
 int main(int argc, char * argv[])
 {
-        unsigned char rsp_buff[MX_ALLOC_LEN];
+        bool fail_all = false;
+        bool fail_path = false;
+        bool use_6_byte = false;
+        int res, fd, k, resid, len, lun = -1;
+        int ret = 0;
         char **argptr;
         char * file_name = 0;
-        int res, fd, k, lun = -1;
-        int fail_all = 0;
-        int fail_path = 0;
-        int ret = 0;
-        int use_6_byte = 0;
+        uint8_t rsp_buff[MX_ALLOC_LEN];
 
         if (argc < 2) {
                 usage ();
@@ -398,28 +405,28 @@ int main(int argc, char * argv[])
                 if (!strcmp (*argptr, "-v"))
                         ++do_verbose;
                 else if (!strncmp(*argptr, "-f=",3)) {
-                        ++fail_path;
+                        fail_path = true;
                         lun = strtoul(*argptr + 3, NULL, 0);
                 }
                 else if (!strcmp(*argptr, "-a")) {
-                        ++fail_all;
+                        fail_all = true;
                 }
                 else if (!strcmp(*argptr, "-6")) {
-                        use_6_byte = 1;
+                        use_6_byte = true;
                 }
                 else if (!strcmp(*argptr, "-V")) {
-                        fprintf(stderr, "sg_rdac version: %s\n", version_str);
+                        pr2serr("sg_rdac version: %s\n", version_str);
                         return 0;
                 }
                 else if (*argv[k] == '-') {
-                        fprintf(stderr, "Unrecognized switch: %s\n", argv[k]);
+                        pr2serr("Unrecognized switch: %s\n", argv[k]);
                         file_name = 0;
                         break;
                 }
                 else if (0 == file_name)
                         file_name = argv[k];
                 else {
-                        fprintf(stderr, "too many arguments\n");
+                        pr2serr("too many arguments\n");
                         file_name = 0;
                         break;
                 }
@@ -429,10 +436,9 @@ int main(int argc, char * argv[])
                 return SG_LIB_SYNTAX_ERROR;
         }
 
-        fd = sg_cmds_open_device(file_name, 0 /* rw */, do_verbose);
+        fd = sg_cmds_open_device(file_name, false /* rw */, do_verbose);
         if (fd < 0) {
-                fprintf(stderr, "open error: %s: %s\n", file_name,
-                        safe_strerror(-fd));
+                pr2serr("open error: %s: %s\n", file_name, safe_strerror(-fd));
                 usage();
                 return SG_LIB_FILE_ERROR;
         }
@@ -442,36 +448,50 @@ int main(int argc, char * argv[])
         } else if (fail_path) {
                 res = fail_this_path(fd, lun, use_6_byte);
         } else {
+                resid = 0;
                 if (use_6_byte)
-                        res = sg_ll_mode_sense6(fd, /* DBD */ 0, /* PC */0,
-                                                0x2c, 0, rsp_buff, 252,
-                                                1, do_verbose);
+                        res = sg_ll_mode_sense6(fd, /* DBD */ false,
+                                                /* PC */ 0,
+                                                0x2c /* page */,
+                                                0 /*subpage */,
+                                                rsp_buff, 252,
+                                                true, do_verbose);
                 else
-                        res = sg_ll_mode_sense10(fd, /* llbaa */ 0,
-                                                 /* DBD */ 0,
-                                                 /* page control */0,
-                                                 0x2c, 0x1, rsp_buff, 308,
-                                                 1, do_verbose);
+                        res = sg_ll_mode_sense10_v2(fd, /* llbaa */ false,
+                                                    /* DBD */ false,
+                                                    /* page control */0,
+                                                    0x2c, 0x1 /* subpage */,
+                                                    rsp_buff, 308, 0, &resid,
+                                                    true, do_verbose);
 
-                if (!res) {
-                        if (do_verbose)
-                                dump_mode_page(rsp_buff, rsp_buff[0]);
-                        print_rdac_mode(rsp_buff, !use_6_byte);
+                if (! res) {
+                        len = sg_msense_calc_length(rsp_buff, 308, use_6_byte,
+                                                    NULL);
+                        if (resid > 0) {
+                                len = ((308 - resid) < len) ? (308 - resid) :
+                                                              len;
+                                if (len < 2)
+                                        pr2serr("MS(10) residual value (%d) "
+                                                "a worry\n", resid);
+                        }
+                        if (do_verbose && (len > 1))
+                                dump_mode_page(rsp_buff, len);
+                        print_rdac_mode(rsp_buff, ! use_6_byte);
                 } else {
                         if (SG_LIB_CAT_INVALID_OP == res)
-                                fprintf(stderr, ">>>>>> try again without "
-                                        "the '-6' switch for a 10 byte MODE "
-                                        "SENSE command\n");
+                                pr2serr(">>>>>> try again without the '-6' "
+                                        "switch for a 10 byte MODE SENSE "
+                                        "command\n");
                         else if (SG_LIB_CAT_ILLEGAL_REQ == res)
-                                fprintf(stderr, "mode sense: invalid field "
-                                        "in cdb (perhaps subpages or page "
-                                        "control (PC) not supported)\n");
+                                pr2serr("mode sense: invalid field in cdb "
+                                        "(perhaps subpages or page control "
+                                        "(PC) not supported)\n");
                         else {
                                 char b[80];
 
                                 sg_get_category_sense_str(res, sizeof(b), b,
                                                           do_verbose);
-                                fprintf(stderr, "mode sense failed: %s\n", b);
+                                pr2serr("mode sense failed: %s\n", b);
                         }
                 }
         }
@@ -479,7 +499,7 @@ int main(int argc, char * argv[])
 
         res = sg_cmds_close_device(fd);
         if (res < 0) {
-                fprintf(stderr, "close error: %s\n", safe_strerror(-res));
+                pr2serr("close error: %s\n", safe_strerror(-res));
                 if (0 == ret)
                         return SG_LIB_FILE_ERROR;
         }
