@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2006-2015 Luben Tuikov and Douglas Gilbert.
+ * Copyright (c) 2006-2018 Luben Tuikov and Douglas Gilbert.
  * All rights reserved.
  * Use of this source code is governed by a BSD-style
  * license that can be found in the BSD_LICENSE file.
@@ -9,6 +9,8 @@
 #include <fcntl.h>
 #include <stdio.h>
 #include <stdlib.h>
+#include <stdarg.h>
+#include <stdbool.h>
 #include <ctype.h>
 #include <string.h>
 #include <getopt.h>
@@ -18,18 +20,20 @@
 #ifdef HAVE_CONFIG_H
 #include "config.h"
 #endif
+
 #include "sg_lib.h"
 #include "sg_cmds_basic.h"
 #include "sg_cmds_extra.h"
-#include "sg_pt.h"      /* needed for scsi_pt_win32_direct() */
+#include "sg_pt.h"
 #include "sg_unaligned.h"
+#include "sg_pr2serr.h"
 
 /*
  * This utility issues the SCSI READ BUFFER(10 or 16) command to the given
  * device.
  */
 
-static const char * version_str = "1.13 20150105";
+static const char * version_str = "1.23 20180219";
 
 
 #ifndef SG_READ_BUFFER_10_CMD
@@ -66,8 +70,8 @@ static struct option long_options[] = {
 static void
 usage()
 {
-    fprintf(stderr, "Usage: "
-            "sg_read_buffer [--16] [--help] [--hex] [--id=ID] [--length=LEN]\n"
+    pr2serr("Usage: sg_read_buffer [--16] [--help] [--hex] [--id=ID] "
+            "[--length=LEN]\n"
             "                      [--long] [--mode=MO] [--offset=OFF] "
             "[--raw]\n"
             "                      [--readonly] [--specific=MS] [--verbose] "
@@ -84,9 +88,9 @@ usage()
             "acronym (def: 0)\n"
             "    --offset=OFF|-o OFF    buffer offset (unit: bytes, def: 0)\n"
             "    --raw|-r            output response to stdout\n"
+            "    --readonly|-R       open DEVICE read-only (def: read-write)\n"
             "    --specific=MS|-S MS    mode specific value; 3 bit field (0 "
             "to 7)\n"
-            "    --readonly|-R       open DEVICE read-only (def: read-write)\n"
             "    --verbose|-v        increase verbosity\n"
             "    --version|-V        print version string and exit\n\n"
             "Performs a SCSI READ BUFFER (10 or 16) command. Use '-m xxx' to "
@@ -129,10 +133,10 @@ print_modes(void)
 {
     const struct mode_s *mp;
 
-    fprintf(stderr, "The modes parameter argument can be numeric "
-                "(hex or decimal)\nor symbolic:\n");
+    pr2serr("The modes parameter argument can be numeric (hex or decimal)\n"
+            "or symbolic:\n");
     for (mp = modes; mp->mode_string; ++mp) {
-        fprintf(stderr, " %2d (0x%02x)  %-16s%s\n", mp->mode, mp->mode,
+        pr2serr(" %2d (0x%02x)  %-16s%s\n", mp->mode, mp->mode,
                 mp->mode_string, mp->comment);
     }
 }
@@ -140,9 +144,9 @@ print_modes(void)
 /* Invokes a SCSI READ BUFFER(10) command (spc5r02).  Return of 0 -> success,
  * various SG_LIB_CAT_* positive values or -1 -> other errors */
 static int
-ll_read_buffer_10(int sg_fd, int rb_mode, int rb_mode_sp, int rb_id,
-                  uint32_t rb_offset, void * resp, int mx_resp_len,
-                  int * residp, int noisy, int verbose)
+sg_ll_read_buffer_10(int sg_fd, int rb_mode, int rb_mode_sp, int rb_id,
+                     uint32_t rb_offset, void * resp, int mx_resp_len,
+                     int * residp, bool noisy, int verbose)
 {
     int k, ret, res, sense_cat;
     uint8_t rb10_cb[SG_READ_BUFFER_10_CMDLEN] =
@@ -157,25 +161,25 @@ ll_read_buffer_10(int sg_fd, int rb_mode, int rb_mode_sp, int rb_id,
     sg_put_unaligned_be24(rb_offset, rb10_cb + 3);
     sg_put_unaligned_be24(mx_resp_len, rb10_cb + 6);
     if (verbose) {
-        fprintf(stderr, "    Read buffer(10) cdb: ");
+        pr2serr("    Read buffer(10) cdb: ");
         for (k = 0; k < SG_READ_BUFFER_10_CMDLEN; ++k)
-            fprintf(stderr, "%02x ", rb10_cb[k]);
-        fprintf(stderr, "\n");
+            pr2serr("%02x ", rb10_cb[k]);
+        pr2serr("\n");
     }
 
     ptvp = construct_scsi_pt_obj();
     if (NULL == ptvp) {
-        fprintf(stderr, "Read buffer(10): out of memory\n");
+        pr2serr("Read buffer(10): out of memory\n");
         return -1;
     }
     set_scsi_pt_cdb(ptvp, rb10_cb, sizeof(rb10_cb));
     set_scsi_pt_sense(ptvp, sense_b, sizeof(sense_b));
-    set_scsi_pt_data_in(ptvp, (unsigned char *)resp, mx_resp_len);
+    set_scsi_pt_data_in(ptvp, (uint8_t *)resp, mx_resp_len);
     res = do_scsi_pt(ptvp, sg_fd, DEF_PT_TIMEOUT, verbose);
     ret = sg_cmds_process_resp(ptvp, "Read buffer(10)", res, mx_resp_len,
                                sense_b, noisy, verbose, &sense_cat);
     if (-1 == ret)
-        ;
+        ret = sg_convert_errno(get_scsi_pt_os_err(ptvp));
     else if (-2 == ret) {
         switch (sense_cat) {
         case SG_LIB_CAT_RECOVERED:
@@ -188,9 +192,9 @@ ll_read_buffer_10(int sg_fd, int rb_mode, int rb_mode_sp, int rb_id,
         }
     } else {
         if ((verbose > 2) && (ret > 0)) {
-            fprintf(stderr, "    Read buffer(10): response%s\n",
+            pr2serr("    Read buffer(10): response%s\n",
                     (ret > 256 ? ", first 256 bytes" : ""));
-            dStrHexErr((const char *)resp, (ret > 256 ? 256 : ret), -1);
+            hex2stderr((const uint8_t *)resp, (ret > 256 ? 256 : ret), -1);
         }
         ret = 0;
     }
@@ -203,13 +207,13 @@ ll_read_buffer_10(int sg_fd, int rb_mode, int rb_mode_sp, int rb_id,
 /* Invokes a SCSI READ BUFFER(16) command (spc5r02).  Return of 0 -> success,
  * various SG_LIB_CAT_* positive values or -1 -> other errors */
 static int
-ll_read_buffer_16(int sg_fd, int rb_mode, int rb_mode_sp, int rb_id,
-                  uint64_t rb_offset, void * resp, int mx_resp_len,
-                  int * residp, int noisy, int verbose)
+sg_ll_read_buffer_16(int sg_fd, int rb_mode, int rb_mode_sp, int rb_id,
+                     uint64_t rb_offset, void * resp, int mx_resp_len,
+                     int * residp, bool noisy, int verbose)
 {
     int k, ret, res, sense_cat;
     uint8_t rb16_cb[SG_READ_BUFFER_16_CMDLEN] =
-          {SG_READ_BUFFER_16_CMD, 0, 0, 0,  0, 0, 0, 0, 0, 0, 0, 0,
+          {SG_READ_BUFFER_16_CMD, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,
            0, 0, 0, 0};
     uint8_t sense_b[SENSE_BUFF_LEN];
     struct sg_pt_base * ptvp;
@@ -221,25 +225,25 @@ ll_read_buffer_16(int sg_fd, int rb_mode, int rb_mode_sp, int rb_id,
     sg_put_unaligned_be24(mx_resp_len, rb16_cb + 11);
     rb16_cb[14] = (uint8_t)rb_id;
     if (verbose) {
-        fprintf(stderr, "    Read buffer(16) cdb: ");
+        pr2serr("    Read buffer(16) cdb: ");
         for (k = 0; k < SG_READ_BUFFER_16_CMDLEN; ++k)
-            fprintf(stderr, "%02x ", rb16_cb[k]);
-        fprintf(stderr, "\n");
+            pr2serr("%02x ", rb16_cb[k]);
+        pr2serr("\n");
     }
 
     ptvp = construct_scsi_pt_obj();
     if (NULL == ptvp) {
-        fprintf(stderr, "Read buffer(16): out of memory\n");
+        pr2serr("Read buffer(16): out of memory\n");
         return -1;
     }
     set_scsi_pt_cdb(ptvp, rb16_cb, sizeof(rb16_cb));
     set_scsi_pt_sense(ptvp, sense_b, sizeof(sense_b));
-    set_scsi_pt_data_in(ptvp, (unsigned char *)resp, mx_resp_len);
+    set_scsi_pt_data_in(ptvp, (uint8_t *)resp, mx_resp_len);
     res = do_scsi_pt(ptvp, sg_fd, DEF_PT_TIMEOUT, verbose);
     ret = sg_cmds_process_resp(ptvp, "Read buffer(16)", res, mx_resp_len,
                                sense_b, noisy, verbose, &sense_cat);
     if (-1 == ret)
-        ;
+        ret = sg_convert_errno(get_scsi_pt_os_err(ptvp));
     else if (-2 == ret) {
         switch (sense_cat) {
         case SG_LIB_CAT_RECOVERED:
@@ -252,9 +256,9 @@ ll_read_buffer_16(int sg_fd, int rb_mode, int rb_mode_sp, int rb_id,
         }
     } else {
         if ((verbose > 2) && (ret > 0)) {
-            fprintf(stderr, "    Read buffer(16): response%s\n",
+            pr2serr("    Read buffer(16): response%s\n",
                     (ret > 256 ? ", first 256 bytes" : ""));
-            dStrHexErr((const char *)resp, (ret > 256 ? 256 : ret), -1);
+            hex2stderr((const uint8_t *)resp, (ret > 256 ? 256 : ret), -1);
         }
         ret = 0;
     }
@@ -265,35 +269,35 @@ ll_read_buffer_16(int sg_fd, int rb_mode, int rb_mode_sp, int rb_id,
 }
 
 static void
-dStrRaw(const char* str, int len)
+dStrRaw(const uint8_t * str, int len)
 {
     int k;
 
-    for (k = 0 ; k < len; ++k)
+    for (k = 0; k < len; ++k)
         printf("%c", str[k]);
 }
 
 int
 main(int argc, char * argv[])
 {
+    bool do_long = false;
+    bool o_readonly = false;
+    bool do_raw = false;
     int res, c, len, k;
     int sg_fd = -1;
     int do_help = 0;
     int do_hex = 0;
-    int do_long = 0;
-    int o_readonly = 0;
     int rb_id = 0;
     int rb_len = 4;
     int rb_mode = 0;
     int rb_mode_sp = 0;
-    int64_t ll;
-    uint64_t rb_offset = 0;
-    int do_raw = 0;
     int resid = 0;
     int verbose = 0;
     int ret = 0;
+    int64_t ll;
+    uint64_t rb_offset = 0;
     const char * device_name = NULL;
-    unsigned char * resp;
+    uint8_t * resp;
     const struct mode_s * mp;
 
     while (1) {
@@ -315,32 +319,31 @@ main(int argc, char * argv[])
         case 'i':
             rb_id = sg_get_num(optarg);
             if ((rb_id < 0) || (rb_id > 255)) {
-                fprintf(stderr, "argument to '--id' should be in the range "
-                        "0 to 255\n");
+                pr2serr("argument to '--id' should be in the range 0 to "
+                        "255\n");
                 return SG_LIB_SYNTAX_ERROR;
             }
             break;
         case 'l':
             rb_len = sg_get_num(optarg);
             if (rb_len < 0) {
-                fprintf(stderr, "bad argument to '--length'\n");
+                pr2serr("bad argument to '--length'\n");
                 return SG_LIB_SYNTAX_ERROR;
              }
              if (rb_len > 0xffffff) {
-                fprintf(stderr, "argument to '--length' must be <= "
-                        "0xffffff\n");
+                pr2serr("argument to '--length' must be <= 0xffffff\n");
                 return SG_LIB_SYNTAX_ERROR;
              }
              break;
         case 'L':
-            ++do_long;
+            do_long = true;
             break;
         case 'm':
             if (isdigit(*optarg)) {
                 rb_mode = sg_get_num(optarg);
                 if ((rb_mode < 0) || (rb_mode > 31)) {
-                    fprintf(stderr, "argument to '--mode' should be in the "
-                            "range 0 to 31\n");
+                    pr2serr("argument to '--mode' should be in the range 0 "
+                            "to 31\n");
                     return SG_LIB_SYNTAX_ERROR;
                 }
             } else {
@@ -351,7 +354,7 @@ main(int argc, char * argv[])
                         break;
                     }
                 }
-                if (NULL == mp) {
+                if (NULL == mp->mode_string) {
                     print_modes();
                     return SG_LIB_SYNTAX_ERROR;
                 }
@@ -360,22 +363,21 @@ main(int argc, char * argv[])
         case 'o':
            ll = sg_get_llnum(optarg);
            if (ll < 0) {
-                fprintf(stderr, "bad argument to '--offset'\n");
+                pr2serr("bad argument to '--offset'\n");
                 return SG_LIB_SYNTAX_ERROR;
             }
             rb_offset = ll;
             break;
         case 'r':
-            ++do_raw;
+            do_raw = true;
             break;
         case 'R':
-            ++o_readonly;
+            o_readonly = true;
             break;
         case 'S':
            rb_mode_sp = sg_get_num(optarg);
            if ((rb_mode_sp < 0) || (rb_mode_sp > 7)) {
-                fprintf(stderr, "expected argument to '--specific' to be 0 "
-                        "to 7\n");
+                pr2serr("expected argument to '--specific' to be 0 to 7\n");
                 return SG_LIB_SYNTAX_ERROR;
             }
             break;
@@ -383,10 +385,10 @@ main(int argc, char * argv[])
             ++verbose;
             break;
         case 'V':
-            fprintf(stderr, "version: %s\n", version_str);
+            pr2serr("version: %s\n", version_str);
             return 0;
         default:
-            fprintf(stderr, "unrecognised option code 0x%x ??\n", c);
+            pr2serr("unrecognised option code 0x%x ??\n", c);
             usage();
             return SG_LIB_SYNTAX_ERROR;
         }
@@ -394,7 +396,7 @@ main(int argc, char * argv[])
     if (do_help) {
         if (do_help > 1) {
             usage();
-            fprintf(stderr, "\n");
+            pr2serr("\n");
             print_modes();
         } else
             usage();
@@ -407,29 +409,25 @@ main(int argc, char * argv[])
         }
         if (optind < argc) {
             for (; optind < argc; ++optind)
-                fprintf(stderr, "Unexpected extra argument: %s\n",
-                        argv[optind]);
+                pr2serr("Unexpected extra argument: %s\n", argv[optind]);
             usage();
             return SG_LIB_SYNTAX_ERROR;
         }
     }
 
     if (NULL == device_name) {
-        fprintf(stderr, "missing device name!\n");
+        pr2serr("missing device name!\n");
         usage();
         return SG_LIB_SYNTAX_ERROR;
     }
 
-    if (rb_len > 0) {
-        resp = (unsigned char *)malloc(rb_len);
-        if (NULL == resp) {
-            fprintf(stderr, "unable to allocate %d bytes on the heap\n",
-                    rb_len);
-            return SG_LIB_CAT_OTHER;
-        }
-        memset(resp, 0, rb_len);
-    } else
-        resp = NULL;
+    len = rb_len ? rb_len : 8;
+    resp = (uint8_t *)malloc(len);
+    if (NULL == resp) {
+        pr2serr("unable to allocate %d bytes on the heap\n", len);
+        return SG_LIB_CAT_OTHER;
+    }
+    memset(resp, 0, len);
 
     if (do_raw) {
         if (sg_set_binary_mode(STDOUT_FILENO) < 0) {
@@ -442,7 +440,7 @@ main(int argc, char * argv[])
 #ifdef SG_LIB_WIN32
 #ifdef SG_LIB_WIN32_DIRECT
     if (verbose > 4)
-        fprintf(stderr, "Initial win32 SPT interface state: %s\n",
+        pr2serr("Initial win32 SPT interface state: %s\n",
                 scsi_pt_win32_spt_state() ? "direct" : "indirect");
     scsi_pt_win32_direct(SG_LIB_WIN32_DIRECT /* SPT pt interface */);
 #endif
@@ -450,32 +448,31 @@ main(int argc, char * argv[])
 
     sg_fd = sg_cmds_open_device(device_name, o_readonly, verbose);
     if (sg_fd < 0) {
-        fprintf(stderr, "open error: %s: %s\n", device_name,
-                safe_strerror(-sg_fd));
+        pr2serr("open error: %s: %s\n", device_name, safe_strerror(-sg_fd));
         ret = SG_LIB_FILE_ERROR;
         goto fini;
     }
 
     if (do_long)
-        res = ll_read_buffer_16(sg_fd, rb_mode, rb_mode_sp, rb_id, rb_offset,
-                                resp, rb_len, &resid, 1, verbose);
+        res = sg_ll_read_buffer_16(sg_fd, rb_mode, rb_mode_sp, rb_id,
+                                   rb_offset, resp, rb_len, &resid, true,
+                                   verbose);
     else if (rb_offset > 0xffffff) {
-        fprintf(stderr, "--offset value is too large for READ BUFFER(10), "
-                "try --16\n");
+        pr2serr("--offset value is too large for READ BUFFER(10), try "
+                "--16\n");
         ret = SG_LIB_SYNTAX_ERROR;
         goto fini;
     } else
-        res = ll_read_buffer_10(sg_fd, rb_mode, rb_mode_sp, rb_id,
-                                (uint32_t)rb_offset, resp, rb_len, &resid, 1,
-                                verbose);
+        res = sg_ll_read_buffer_10(sg_fd, rb_mode, rb_mode_sp, rb_id,
+                                   (uint32_t)rb_offset, resp, rb_len, &resid,
+                                   true, verbose);
     if (0 != res) {
         char b[80];
 
         ret = res;
         if (res > 0) {
             sg_get_category_sense_str(res, sizeof(b), b, verbose);
-            fprintf(stderr, "Read buffer(%d) failed: %s\n",
-                    (do_long ? 16 : 10), b);
+            pr2serr("Read buffer(%d) failed: %s\n", (do_long ? 16 : 10), b);
         }
         goto fini;
     }
@@ -483,25 +480,25 @@ main(int argc, char * argv[])
         rb_len -= resid;        /* got back less than requested */
     if (rb_len > 0) {
         if (do_raw)
-            dStrRaw((const char *)resp, rb_len);
+            dStrRaw(resp, rb_len);
         else if (do_hex || (rb_len < 4))
-            dStrHex((const char *)resp, rb_len, ((do_hex > 1) ? 0 : 1));
+            hex2stdout((const uint8_t *)resp, rb_len, ((do_hex > 1) ? 0 : 1));
         else {
             switch (rb_mode) {
             case MODE_DESCRIPTOR:
-                k = (resp[1] << 16) | (resp[2] << 8) | resp[3];
+                k = sg_get_unaligned_be24(resp + 1);
                 printf("OFFSET BOUNDARY: %d, Buffer offset alignment: "
                        "%d-byte\n", resp[0], (1 << resp[0]));
                 printf("BUFFER CAPACITY: %d (0x%x)\n", k, k);
                 break;
             case MODE_ECHO_BDESC:
-                k = ((resp[2] & 0x1F) << 8) | resp[3];
-
+                k = sg_get_unaligned_be16(resp + 2) & 0x1fff;
                 printf("EBOS:%d\n", resp[0] & 1 ? 1 : 0);
                 printf("Echo buffer capacity: %d (0x%x)\n", k, k);
                 break;
             default:
-                dStrHex((const char *)resp, rb_len, (verbose > 1 ? 0 : 1));
+                hex2stdout((const uint8_t *)resp, rb_len,
+                           (verbose > 1 ? 0 : 1));
                 break;
             }
         }
@@ -513,7 +510,7 @@ fini:
     if (sg_fd >= 0) {
         res = sg_cmds_close_device(sg_fd);
         if (res < 0) {
-            fprintf(stderr, "close error: %s\n", safe_strerror(-res));
+            pr2serr("close error: %s\n", safe_strerror(-res));
             if (0 == ret)
                 return SG_LIB_FILE_ERROR;
         }
