@@ -1,23 +1,25 @@
 /* This code is does a SCSI READ CAPACITY command on the given device
-   and outputs the result.
-
-*  Copyright (C) 1999 - 2014 D. Gilbert
-*  This program is free software; you can redistribute it and/or modify
-*  it under the terms of the GNU General Public License as published by
-*  the Free Software Foundation; either version 2, or (at your option)
-*  any later version.
-
-   This program was originally written with Linux 2.4 kernel series.
-   It now builds for the Linux 2.6 and 3 kernel series and various other
-   operating systems.
-
-*/
+ * and outputs the result.
+ *
+ * Copyright (C) 1999 - 2018 D. Gilbert
+ * This program is free software; you can redistribute it and/or modify
+ * it under the terms of the GNU General Public License as published by
+ * the Free Software Foundation; either version 2, or (at your option)
+ * any later version.
+ *
+ * This program was originally written with Linux 2.4 kernel series.
+ * It now builds for the Linux 2.6, 3 and 4 kernel series and various other
+ * operating systems.
+ */
 
 #include <stdio.h>
 #include <stdlib.h>
+#include <stdarg.h>
+#include <stdbool.h>
 #include <unistd.h>
 #include <string.h>
 #include <fcntl.h>
+#include <errno.h>
 #include <getopt.h>
 #define __STDC_FORMAT_MACROS 1
 #include <inttypes.h>
@@ -25,12 +27,14 @@
 #ifdef HAVE_CONFIG_H
 #include "config.h"
 #endif
+
 #include "sg_lib.h"
 #include "sg_cmds_basic.h"
 #include "sg_unaligned.h"
+#include "sg_pr2serr.h"
 
 
-static const char * version_str = "3.93 20141220";
+static const char * version_str = "4.00 20180302";
 
 #define ME "sg_readcap: "
 
@@ -38,46 +42,45 @@ static const char * version_str = "3.93 20141220";
 #define RCAP16_REPLY_LEN 32
 
 static struct option long_options[] = {
-    {"brief", 0, 0, 'b'},
-    {"help", 0, 0, 'h'},
-    {"hex", 0, 0, 'H'},
-    {"lba", 1, 0, 'L'},
-    {"long", 0, 0, 'l'},
-    {"16", 0, 0, 'l'},
-    {"new", 0, 0, 'N'},
-    {"old", 0, 0, 'O'},
-    {"pmi", 0, 0, 'p'},
-    {"raw", 0, 0, 'r'},
-    {"readonly", 0, 0, 'R'},
-    {"verbose", 0, 0, 'v'},
-    {"version", 0, 0, 'V'},
-    {"zbc", 0, 0, 'z'},
+    {"brief", no_argument, 0, 'b'},
+    {"help", no_argument, 0, 'h'},
+    {"hex", no_argument, 0, 'H'},
+    {"lba", required_argument, 0, 'L'},
+    {"long", no_argument, 0, 'l'},
+    {"16", no_argument, 0, 'l'},
+    {"new", no_argument, 0, 'N'},
+    {"old", no_argument, 0, 'O'},
+    {"pmi", no_argument, 0, 'p'},
+    {"raw", no_argument, 0, 'r'},
+    {"readonly", no_argument, 0, 'R'},
+    {"verbose", no_argument, 0, 'v'},
+    {"version", no_argument, 0, 'V'},
+    {"zbc", no_argument, 0, 'z'},
     {0, 0, 0, 0},
 };
 
 struct opts_t {
-    int do_brief;
+    bool do_brief;
+    bool do_long;
+    bool do_pmi;
+    bool do_raw;
+    bool o_readonly;
+    bool do_version;
+    bool do_zbc;
+    bool opt_new;
     int do_help;
     int do_hex;
     int do_lba;
-    int do_long;
-    int do_pmi;
-    int do_raw;
-    int o_readonly;
     int do_verbose;
-    int do_version;
-    int do_zbc;
     uint64_t llba;
     const char * device_name;
-    int opt_new;
 };
 
 
 static void
 usage()
 {
-    fprintf(stderr,
-            "Usage: sg_readcap [--brief] [--help] [--hex] [--lba=LBA] "
+    pr2serr("Usage: sg_readcap [--brief] [--help] [--hex] [--lba=LBA] "
             "[--long] [--16]\n"
             "                  [--pmi] [--raw] [--readonly] [--verbose] "
             "[--version]\n"
@@ -104,6 +107,7 @@ usage()
             "read-write)\n"
             "    --verbose|-v    increase verbosity\n"
             "    --version|-V    print version string and exit\n"
+            "    --old|-O        use old interface (use as first option)\n"
             "    --zbc|-z        show rc_basis ZBC field (implies --16)\n\n"
             "Perform a SCSI READ CAPACITY (10 or 16) command\n");
 }
@@ -111,7 +115,7 @@ usage()
 static void
 usage_old()
 {
-    fprintf(stderr, "Usage:  sg_readcap [-16] [-b] [-h] [-H] [-lba=LBA] "
+    pr2serr("Usage:  sg_readcap [-16] [-b] [-h] [-H] [-lba=LBA] "
             "[-pmi] [-r] [-R]\n"
             "                   [-v] [-V] [-z] DEVICE\n"
             "  where:\n"
@@ -132,6 +136,7 @@ usage_old()
             "    -R     open DEVICE read-only (def: RCAP(16) read-write)\n"
             "    -v     increase verbosity\n"
             "    -V     print version string and exit\n"
+            "    -N|--new   use new interface\n"
             "    -z     show rc_basis ZBC field (implies -16)\n\n"
             "Perform a SCSI READ CAPACITY (10 or 16) command\n");
 }
@@ -146,7 +151,7 @@ usage_for(const struct opts_t * op)
 }
 
 static int
-process_cl_new(struct opts_t * op, int argc, char * argv[])
+new_parse_cmd_line(struct opts_t * op, int argc, char * argv[])
 {
     int c;
     int a_one = 0;
@@ -166,10 +171,10 @@ process_cl_new(struct opts_t * op, int argc, char * argv[])
             break;
         case '6':
             if (a_one)
-                ++op->do_long;
+                op->do_long = true;
             break;
         case 'b':
-            ++op->do_brief;
+            op->do_brief = true;
             break;
         case 'h':
         case '?':
@@ -179,46 +184,46 @@ process_cl_new(struct opts_t * op, int argc, char * argv[])
             ++op->do_hex;
             break;
         case 'l':
-            ++op->do_long;
+            op->do_long = true;
             break;
         case 'L':
             nn = sg_get_llnum(optarg);
             if (-1 == nn) {
-                fprintf(stderr, "bad argument to '--lba='\n");
+                pr2serr("bad argument to '--lba='\n");
                 usage();
                 return SG_LIB_SYNTAX_ERROR;
             }
             op->llba = nn;
             /* force READ_CAPACITY16 for large lbas */
             if (op->llba > 0xfffffffeULL)
-                ++op->do_long;
+                op->do_long = true;
             ++op->do_lba;
             break;
         case 'N':
             break;      /* ignore */
         case 'O':
-            op->opt_new = 0;
+            op->opt_new = false;
             return 0;
         case 'p':
-            ++op->do_pmi;
+            op->do_pmi = true;
             break;
         case 'r':
-            ++op->do_raw;
+            op->do_raw = true;
             break;
         case 'R':
-            ++op->o_readonly;
+            op->o_readonly = true;
             break;
         case 'v':
             ++op->do_verbose;
             break;
         case 'V':
-            ++op->do_version;
+            op->do_version = true;
             break;
         case 'z':
-            ++op->do_zbc;
+            op->do_zbc = true;
             break;
         default:
-            fprintf(stderr, "unrecognised option code %c [0x%x]\n", c, c);
+            pr2serr("unrecognised option code %c [0x%x]\n", c, c);
             if (op->do_help)
                 break;
             usage();
@@ -232,8 +237,7 @@ process_cl_new(struct opts_t * op, int argc, char * argv[])
         }
         if (optind < argc) {
             for (; optind < argc; ++optind)
-                fprintf(stderr, "Unexpected extra argument: %s\n",
-                        argv[optind]);
+                pr2serr("Unexpected extra argument: %s\n", argv[optind]);
             usage();
             return SG_LIB_SYNTAX_ERROR;
         }
@@ -242,9 +246,10 @@ process_cl_new(struct opts_t * op, int argc, char * argv[])
 }
 
 static int
-process_cl_old(struct opts_t * op, int argc, char * argv[])
+old_parse_cmd_line(struct opts_t * op, int argc, char * argv[])
 {
-    int k, jmp_out, plen, num;
+    bool jmp_out;
+    int k, plen, num;
     const char * cp;
     uint64_t uu;
 
@@ -254,18 +259,18 @@ process_cl_old(struct opts_t * op, int argc, char * argv[])
         if (plen <= 0)
             continue;
         if ('-' == *cp) {
-            for (--plen, ++cp, jmp_out = 0; plen > 0; --plen, ++cp) {
+            for (--plen, ++cp, jmp_out = false; plen > 0; --plen, ++cp) {
                 switch (*cp) {
                 case '1':
                     if ('6' == *(cp + 1)) {
-                        ++op->do_long;
+                        op->do_long = true;
                         ++cp;
                         --plen;
                     } else
-                        jmp_out = 1;
+                        jmp_out = true;
                     break;
                 case 'b':
-                    ++op->do_brief;
+                    op->do_brief = true;
                     break;
                 case 'h':
                 case '?':
@@ -275,35 +280,35 @@ process_cl_old(struct opts_t * op, int argc, char * argv[])
                     ++op->do_hex;
                     break;
                 case 'N':
-                    op->opt_new = 1;
+                    op->opt_new = true;
                     return 0;
                 case 'O':
                     break;
                 case 'p':
                     if (0 == strncmp("pmi", cp, 3)) {
-                        ++op->do_pmi;
+                        op->do_pmi = true;
                         cp += 2;
                         plen -= 2;
                     } else
-                        jmp_out = 1;
+                        jmp_out = true;
                     break;
                 case 'r':
-                    ++op->do_raw;
+                    op->do_raw = true;
                     break;
                 case 'R':
-                    ++op->o_readonly;
+                    op->o_readonly = true;
                     break;
                 case 'v':
                     ++op->do_verbose;
                     break;
                 case 'V':
-                    ++op->do_version;
+                    op->do_version = true;
                     break;
                 case 'z':
-                    ++op->do_zbc;
+                    op->do_zbc = true;
                     break;
                 default:
-                    jmp_out = 1;
+                    jmp_out = true;
                     break;
                 }
                 if (jmp_out)
@@ -320,21 +325,21 @@ process_cl_old(struct opts_t * op, int argc, char * argv[])
                 }
                 /* force READ_CAPACITY16 for large lbas */
                 if (uu > 0xfffffffeULL)
-                    ++op->do_long;
+                    op->do_long = true;
                 op->llba = uu;
                 ++op->do_lba;
             } else if (0 == strncmp("-old", cp, 4))
                 ;
             else if (jmp_out) {
-                fprintf(stderr, "Unrecognized option: %s\n", cp);
+                pr2serr("Unrecognized option: %s\n", cp);
                 usage();
                 return SG_LIB_SYNTAX_ERROR;
             }
         } else if (0 == op->device_name)
             op->device_name = cp;
         else {
-            fprintf(stderr, "too many arguments, got: %s, not expecting: "
-                    "%s\n", op->device_name, cp);
+            pr2serr("too many arguments, got: %s, not expecting: %s\n",
+                    op->device_name, cp);
             usage();
             return SG_LIB_SYNTAX_ERROR;
         }
@@ -343,32 +348,32 @@ process_cl_old(struct opts_t * op, int argc, char * argv[])
 }
 
 static int
-process_cl(struct opts_t * op, int argc, char * argv[])
+parse_cmd_line(struct opts_t * op, int argc, char * argv[])
 {
     int res;
     char * cp;
 
     cp = getenv("SG3_UTILS_OLD_OPTS");
     if (cp) {
-        op->opt_new = 0;
-        res = process_cl_old(op, argc, argv);
+        op->opt_new = false;
+        res = old_parse_cmd_line(op, argc, argv);
         if ((0 == res) && op->opt_new)
-            res = process_cl_new(op, argc, argv);
+            res = new_parse_cmd_line(op, argc, argv);
     } else {
-        op->opt_new = 1;
-        res = process_cl_new(op, argc, argv);
-        if ((0 == res) && (0 == op->opt_new))
-            res = process_cl_old(op, argc, argv);
+        op->opt_new = true;
+        res = new_parse_cmd_line(op, argc, argv);
+        if ((0 == res) && (! op->opt_new))
+            res = old_parse_cmd_line(op, argc, argv);
     }
     return res;
 }
 
 static void
-dStrRaw(const char* str, int len)
+dStrRaw(const uint8_t * str, int len)
 {
     int k;
 
-    for (k = 0 ; k < len; ++k)
+    for (k = 0; k < len; ++k)
         printf("%c", str[k]);
 }
 
@@ -380,7 +385,7 @@ rc_basis_str(int rc_basis, char * b, int blen)
         snprintf(b, blen, "last contiguous that's not seq write required");
         break;
     case 1:
-        snprintf(b, blen, "last LBA on device");
+        snprintf(b, blen, "last LBA on logical unit");
         break;
     default:
         snprintf(b, blen, "reserved (0x%x)", rc_basis);
@@ -393,18 +398,22 @@ rc_basis_str(int rc_basis, char * b, int blen)
 int
 main(int argc, char * argv[])
 {
-    int sg_fd, res, prot_en, p_type, lbppbe, rw_0_flag;
-    uint64_t llast_blk_addr;
+    bool rw_0_flag;
+    int res, prot_en, p_type, lbppbe;
+    int sg_fd = -1;
     int ret = 0;
     uint32_t last_blk_addr, block_size;
-    unsigned char resp_buff[RCAP16_REPLY_LEN];
+    uint64_t llast_blk_addr;
+    uint8_t * resp_buff;
+    uint8_t * free_resp_buff;
+    const int resp_buff_sz = RCAP16_REPLY_LEN;
     char b[80];
     struct opts_t opts;
     struct opts_t * op;
 
     op = &opts;
     memset(op, 0, sizeof(opts));
-    res = process_cl(op, argc, argv);
+    res = parse_cmd_line(op, argc, argv);
     if (res)
         return SG_LIB_SYNTAX_ERROR;
     if (op->do_help) {
@@ -412,12 +421,12 @@ main(int argc, char * argv[])
         return 0;
     }
     if (op->do_version) {
-        fprintf(stderr, "Version string: %s\n", version_str);
+        pr2serr("Version string: %s\n", version_str);
         return 0;
     }
 
     if (NULL == op->device_name) {
-        fprintf(stderr, "No DEVICE argument given\n");
+        pr2serr("No DEVICE argument given\n");
         usage_for(op);
         return SG_LIB_SYNTAX_ERROR;
     }
@@ -429,40 +438,46 @@ main(int argc, char * argv[])
     }
     if (op->do_zbc) {
         if (! op->do_long)
-            ++op->do_long;
+            op->do_long = true;
     }
 
-    memset(resp_buff, 0, sizeof(resp_buff));
-
-    if ((0 == op->do_pmi) && (op->llba > 0)) {
-        fprintf(stderr, ME "lba can only be non-zero when '--pmi' is set\n");
+    resp_buff = sg_memalign(resp_buff_sz, 0, &free_resp_buff, false);
+    if (NULL == resp_buff) {
+        pr2serr("Unable to allocate %d bytes on heap\n", resp_buff_sz);
+        return sg_convert_errno(ENOMEM);
+    }
+    if ((! op->do_pmi) && (op->llba > 0)) {
+        pr2serr(ME "lba can only be non-zero when '--pmi' is set\n");
         usage_for(op);
-        return SG_LIB_SYNTAX_ERROR;
+        ret = SG_LIB_SYNTAX_ERROR;
+        goto fini;
     }
     if (op->do_long)
         rw_0_flag = op->o_readonly;
     else
-        rw_0_flag = 1;  /* RCAP(10) has opened RO in past, so leave */
+        rw_0_flag = true;  /* RCAP(10) has opened RO in past, so leave */
     if ((sg_fd = sg_cmds_open_device(op->device_name, rw_0_flag,
                                      op->do_verbose)) < 0) {
-        fprintf(stderr, ME "error opening file: %s: %s\n", op->device_name,
+        pr2serr(ME "error opening file: %s: %s\n", op->device_name,
                 safe_strerror(-sg_fd));
-        return SG_LIB_FILE_ERROR;
+        ret = sg_convert_errno(-sg_fd);
+        goto fini;
     }
 
     if (! op->do_long) {
         res = sg_ll_readcap_10(sg_fd, op->do_pmi, (unsigned int)op->llba,
-                               resp_buff, RCAP_REPLY_LEN, 1, op->do_verbose);
+                               resp_buff, RCAP_REPLY_LEN, true,
+                               op->do_verbose);
         ret = res;
         if (0 == res) {
             if (op->do_hex || op->do_raw) {
                 if (op->do_raw)
-                    dStrRaw((const char *)resp_buff, RCAP_REPLY_LEN);
+                    dStrRaw(resp_buff, RCAP_REPLY_LEN);
                 else if (op->do_hex > 2)
-                    dStrHex((const char *)resp_buff, RCAP_REPLY_LEN, -1);
+                    hex2stdout(resp_buff, RCAP_REPLY_LEN, -1);
                 else
-                    dStrHex((const char *)resp_buff, RCAP_REPLY_LEN, 1);
-                goto good;
+                    hex2stdout(resp_buff, RCAP_REPLY_LEN, 1);
+                goto fini;
             }
             last_blk_addr = sg_get_unaligned_be32(resp_buff + 0);
             if (0xffffffff != last_blk_addr) {
@@ -470,7 +485,7 @@ main(int argc, char * argv[])
                 if (op->do_brief) {
                     printf("0x%" PRIx32 " 0x%" PRIx32 "\n",
                            last_blk_addr + 1, block_size);
-                    goto good;
+                    goto fini;
                 }
                 printf("Read Capacity results:\n");
                 if (op->do_pmi)
@@ -500,49 +515,50 @@ main(int argc, char * argv[])
                            "%.2f GB\n", total_sz, sz_mb, sz_gb);
 #endif
                 }
-                goto good;
+                goto fini;
             } else {
                 printf("READ CAPACITY (10) indicates device capacity too "
                        "large\n  now trying 16 byte cdb variant\n");
-                op->do_long = 1;
+                op->do_long = true;
             }
         } else if (SG_LIB_CAT_INVALID_OP == res) {
-            op->do_long = 1;
+            op->do_long = true;
             sg_cmds_close_device(sg_fd);
             if ((sg_fd = sg_cmds_open_device(op->device_name, op->o_readonly,
                                              op->do_verbose)) < 0) {
-                fprintf(stderr, ME "error re-opening file: %s (rw): %s\n",
+                pr2serr(ME "error re-opening file: %s (rw): %s\n",
                         op->device_name, safe_strerror(-sg_fd));
-                return SG_LIB_FILE_ERROR;
+                ret = sg_convert_errno(-sg_fd);
+                goto fini;
             }
             if (op->do_verbose)
-                fprintf(stderr, "READ CAPACITY (10) not supported, trying "
-                        "READ CAPACITY (16)\n");
+                pr2serr("READ CAPACITY (10) not supported, trying READ "
+                        "CAPACITY (16)\n");
         } else if (res) {
             sg_get_category_sense_str(res, sizeof(b), b, op->do_verbose);
-            fprintf(stderr, "READ CAPACITY (10) failed: %s\n", b);
+            pr2serr("READ CAPACITY (10) failed: %s\n", b);
         }
     }
     if (op->do_long) {
         res = sg_ll_readcap_16(sg_fd, op->do_pmi, op->llba, resp_buff,
-                               RCAP16_REPLY_LEN, 1, op->do_verbose);
+                               RCAP16_REPLY_LEN, true, op->do_verbose);
         ret = res;
         if (0 == res) {
             if (op->do_hex || op->do_raw) {
                 if (op->do_raw)
-                    dStrRaw((const char *)resp_buff, RCAP16_REPLY_LEN);
+                    dStrRaw(resp_buff, RCAP16_REPLY_LEN);
                 else if (op->do_hex > 2)
-                    dStrHex((const char *)resp_buff, RCAP16_REPLY_LEN, -1);
+                    hex2stdout(resp_buff, RCAP16_REPLY_LEN, -1);
                 else
-                    dStrHex((const char *)resp_buff, RCAP16_REPLY_LEN, 1);
-                goto good;
+                    hex2stdout(resp_buff, RCAP16_REPLY_LEN, 1);
+                goto fini;
             }
             llast_blk_addr = sg_get_unaligned_be64(resp_buff + 0);
             block_size = sg_get_unaligned_be32(resp_buff + 8);
             if (op->do_brief) {
                 printf("0x%" PRIx64 " 0x%" PRIx32 "\n", llast_blk_addr + 1,
                        block_size);
-                goto good;
+                goto fini;
             }
             prot_en = !!(resp_buff[12] & 0x1);
             p_type = ((resp_buff[12] >> 1) & 0x7);
@@ -598,24 +614,27 @@ main(int argc, char * argv[])
                        "GB\n", total_sz, sz_mb, sz_gb);
 #endif
             }
-            goto good;
+            goto fini;
         } else if (SG_LIB_CAT_ILLEGAL_REQ == res)
-            fprintf(stderr, "bad field in READ CAPACITY (16) cdb "
-                    "including unsupported service action\n");
+            pr2serr("bad field in READ CAPACITY (16) cdb including "
+                    "unsupported service action\n");
         else if (res) {
             sg_get_category_sense_str(res, sizeof(b), b, op->do_verbose);
-            fprintf(stderr, "READ CAPACITY (16) failed: %s\n", b);
+            pr2serr("READ CAPACITY (16) failed: %s\n", b);
         }
     }
     if (op->do_brief)
         printf("0x0 0x0\n");
-
-good:
-    res = sg_cmds_close_device(sg_fd);
-    if (res < 0) {
-        fprintf(stderr, "close error: %s\n", safe_strerror(-res));
-        if (0 == ret)
-            return SG_LIB_FILE_ERROR;
+fini:
+    if (free_resp_buff)
+        free(free_resp_buff);
+    if (sg_fd >= 0) {
+        res = sg_cmds_close_device(sg_fd);
+        if (res < 0) {
+            pr2serr("close error: %s\n", safe_strerror(-res));
+            if (0 == ret)
+                return SG_LIB_FILE_ERROR;
+        }
     }
     return (ret >= 0) ? ret : SG_LIB_CAT_OTHER;
 }
