@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2004-2013 Hannes Reinecke and Douglas Gilbert.
+ * Copyright (c) 2004-2018 Hannes Reinecke and Douglas Gilbert.
  * All rights reserved.
  * Use of this source code is governed by a BSD-style
  * license that can be found in the BSD_LICENSE file.
@@ -9,6 +9,9 @@
 #include <fcntl.h>
 #include <stdio.h>
 #include <stdlib.h>
+#include <stdarg.h>
+#include <stdbool.h>
+#include <stdint.h>
 #include <string.h>
 #include <ctype.h>
 #include <getopt.h>
@@ -16,9 +19,12 @@
 #ifdef HAVE_CONFIG_H
 #include "config.h"
 #endif
+
 #include "sg_lib.h"
 #include "sg_cmds_basic.h"
 #include "sg_cmds_extra.h"
+#include "sg_unaligned.h"
+#include "sg_pr2serr.h"
 
 /* A utility program for the Linux OS SCSI subsystem.
  *
@@ -26,7 +32,7 @@
  *  to the 'SCSI Accessed Fault-Tolerant Enclosures' (SAF-TE) spec.
  */
 
-static const char * version_str = "0.24 20130507";
+static const char * version_str = "0.31 20180219";
 
 
 #define SENSE_BUFF_LEN 64       /* Arbitrary, could be larger */
@@ -57,32 +63,32 @@ struct safte_cfg_t safte_cfg;
 static unsigned int buf_capacity = 64;
 
 static void
-dStrRaw(const char* str, int len)
+dStrRaw(const uint8_t * str, int len)
 {
     int k;
 
-    for (k = 0 ; k < len; ++k)
+    for (k = 0; k < len; ++k)
         printf("%c", str[k]);
 }
 
 /* Buffer ID 0x0: Read Enclosure Configuration (mandatory) */
 static int
-read_safte_configuration(int sg_fd, unsigned char *rb_buff,
+read_safte_configuration(int sg_fd, uint8_t *rb_buff,
                          unsigned int rb_len, int verbose)
 {
     int res;
 
     if (rb_len < buf_capacity) {
-        fprintf(stderr,"SCSI BUFFER size too small (%d/%d bytes)\n",
-                rb_len, buf_capacity);
+        pr2serr("SCSI BUFFER size too small (%d/%d bytes)\n", rb_len,
+                buf_capacity);
         return SG_LIB_CAT_ILLEGAL_REQ;
     }
 
     if (verbose > 1)
-        fprintf(stderr, "Use READ BUFFER,mode=vendor_specific,buff_id=0 "
-                "to fetch configuration\n");
+        pr2serr("Use READ BUFFER,mode=vendor_specific,buff_id=0 to fetch "
+                "configuration\n");
     res = sg_ll_read_buffer(sg_fd, RWB_MODE_VENDOR, 0, 0,
-                            rb_buff, rb_len, 1, verbose);
+                            rb_buff, rb_len, true, verbose);
     if (res && res != SG_LIB_CAT_RECOVERED)
         return res;
 
@@ -123,27 +129,27 @@ do_safte_encl_status(int sg_fd, int do_hex, int do_raw, int verbose)
 {
     int res, i, offset;
     unsigned int rb_len;
-    unsigned char *rb_buff;
+    uint8_t *rb_buff;
 
     rb_len = safte_cfg.fans + safte_cfg.psupplies + safte_cfg.slots +
         safte_cfg.temps + 5 + safte_cfg.vendor_specific;
-    rb_buff = (unsigned char *)malloc(rb_len);
+    rb_buff = (uint8_t *)malloc(rb_len);
 
 
     if (verbose > 1)
-        fprintf(stderr, "Use READ BUFFER,mode=vendor_specific,buff_id=1 "
-                "to read enclosure status\n");
+        pr2serr("Use READ BUFFER,mode=vendor_specific,buff_id=1 to read "
+                "enclosure status\n");
     res = sg_ll_read_buffer(sg_fd, RWB_MODE_VENDOR, 1, 0,
-                            rb_buff, rb_len, 0, verbose);
+                            rb_buff, rb_len, false, verbose);
     if (res && res != SG_LIB_CAT_RECOVERED)
         return res;
 
     if (do_raw > 1) {
-        dStrRaw((const char *)rb_buff, buf_capacity);
+        dStrRaw(rb_buff, buf_capacity);
         return 0;
     }
     if (do_hex > 1) {
-        dStrHex((const char *)rb_buff, buf_capacity, 1);
+        hex2stdout(rb_buff, buf_capacity, 1);
         return 0;
     }
     printf("Enclosure Status:\n");
@@ -238,13 +244,14 @@ do_safte_encl_status(int sg_fd, int do_hex, int do_raw, int verbose)
 
     offset++;
     for (i = 0; i < safte_cfg.temps; i++) {
-        int temp = 0;
+        int temp = rb_buff[i + offset];
+        int is_celsius = !!(safte_cfg.flags & SAFTE_CFG_FLAG_CELSIUS);
 
-        if (!(safte_cfg.flags & SAFTE_CFG_FLAG_CELSIUS))
+        if (! is_celsius)
             temp -= 10;
 
-        printf("\tTemperature sensor %d: %d deg %c\n", i, rb_buff[i + offset],
-               safte_cfg.flags & SAFTE_CFG_FLAG_CELSIUS?'C':'F');
+        printf("\tTemperature sensor %d: %d deg %c\n", i, temp,
+               is_celsius ? 'C' : 'F');
     }
 
     offset += safte_cfg.temps;
@@ -264,17 +271,17 @@ do_safte_usage_statistics(int sg_fd, int do_hex, int do_raw, int verbose)
 {
     int res;
     unsigned int rb_len;
-    unsigned char *rb_buff;
+    uint8_t *rb_buff;
     unsigned int minutes;
 
     rb_len = 16 + safte_cfg.vendor_specific;
-    rb_buff = (unsigned char *)malloc(rb_len);
+    rb_buff = (uint8_t *)malloc(rb_len);
 
     if (verbose > 1)
-        fprintf(stderr, "Use READ BUFFER,mode=vendor_specific,buff_id=2 "
-                "to read usage statistics\n");
+        pr2serr("Use READ BUFFER,mode=vendor_specific,buff_id=2 to read "
+                "usage statistics\n");
     res = sg_ll_read_buffer(sg_fd, RWB_MODE_VENDOR, 2, 0,
-                            rb_buff, rb_len, 0, verbose);
+                            rb_buff, rb_len, false, verbose);
     if (res) {
         if (res == SG_LIB_CAT_ILLEGAL_REQ) {
             printf("Usage Statistics:\n\tNot implemented\n");
@@ -287,19 +294,17 @@ do_safte_usage_statistics(int sg_fd, int do_hex, int do_raw, int verbose)
     }
 
     if (do_raw > 1) {
-        dStrRaw((const char *)rb_buff, buf_capacity);
+        dStrRaw(rb_buff, buf_capacity);
         return 0;
     }
     if (do_hex > 1) {
-        dStrHex((const char *)rb_buff, buf_capacity, 1);
+        hex2stdout(rb_buff, buf_capacity, 1);
         return 0;
     }
     printf("Usage Statistics:\n");
-    minutes = (rb_buff[0] << 24) + (rb_buff[1] << 16) +
-        (rb_buff[2] <<  8) + rb_buff[3];
+    minutes = sg_get_unaligned_be32(rb_buff + 0);
     printf("\tPower on Minutes: %u\n", minutes);
-    minutes = (rb_buff[4] << 24) + (rb_buff[5] << 16) +
-        (rb_buff[6] <<  8) + rb_buff[7];
+    minutes = sg_get_unaligned_be32(rb_buff + 4);
     printf("\tPower on Cycles: %u\n", minutes);
 
     free(rb_buff);
@@ -312,16 +317,16 @@ do_safte_slot_insertions(int sg_fd, int do_hex, int do_raw, int verbose)
 {
     int res, i;
     unsigned int rb_len;
-    unsigned char *rb_buff, slot_status;
+    uint8_t *rb_buff, slot_status;
 
     rb_len = safte_cfg.slots * 2;
-    rb_buff = (unsigned char *)malloc(rb_len);
+    rb_buff = (uint8_t *)malloc(rb_len);
 
     if (verbose > 1)
-        fprintf(stderr, "Use READ BUFFER,mode=vendor_specific,buff_id=3 "
-                "to read device insertions\n");
+        pr2serr("Use READ BUFFER,mode=vendor_specific,buff_id=3 to read "
+                "device insertions\n");
     res = sg_ll_read_buffer(sg_fd, RWB_MODE_VENDOR, 3, 0,
-                            rb_buff, rb_len, 0, verbose);
+                            rb_buff, rb_len, false, verbose);
     if (res ) {
         if (res == SG_LIB_CAT_ILLEGAL_REQ) {
                 printf("Slot insertions:\n\tNot implemented\n");
@@ -334,16 +339,16 @@ do_safte_slot_insertions(int sg_fd, int do_hex, int do_raw, int verbose)
     }
 
     if (do_raw > 1) {
-        dStrRaw((const char *)rb_buff, buf_capacity);
+        dStrRaw(rb_buff, buf_capacity);
         return 0;
     }
     if (do_hex > 1) {
-        dStrHex((const char *)rb_buff, buf_capacity, 1);
+        hex2stdout(rb_buff, buf_capacity, 1);
         return 0;
     }
     printf("Slot insertions:\n");
     for (i = 0; i < safte_cfg.slots; i++) {
-        slot_status = (rb_buff[i * 2] << 8) + rb_buff[i * 2];
+        slot_status = sg_get_unaligned_be16(rb_buff + (i * 2));
         printf("\tSlot %d: %d insertions", i, slot_status);
     }
     free(rb_buff);
@@ -356,27 +361,27 @@ do_safte_slot_status(int sg_fd, int do_hex, int do_raw, int verbose)
 {
     int res, i;
     unsigned int rb_len;
-    unsigned char *rb_buff, slot_status;
+    uint8_t *rb_buff, slot_status;
 
     rb_len = safte_cfg.slots * 4;
-    rb_buff = (unsigned char *)malloc(rb_len);
+    rb_buff = (uint8_t *)malloc(rb_len);
 
     if (verbose > 1)
-        fprintf(stderr, "Use READ BUFFER,mode=vendor_specific,buff_id=4 "
-                "to read device slot status\n");
+        pr2serr("Use READ BUFFER,mode=vendor_specific,buff_id=4 to read "
+                "device slot status\n");
     res = sg_ll_read_buffer(sg_fd, RWB_MODE_VENDOR, 4, 0,
-                            rb_buff, rb_len, 0, verbose);
+                            rb_buff, rb_len, false, verbose);
     if (res && res != SG_LIB_CAT_RECOVERED) {
         free(rb_buff);
         return res;
     }
 
     if (do_raw > 1) {
-        dStrRaw((const char *)rb_buff, buf_capacity);
+        dStrRaw(rb_buff, buf_capacity);
         return 0;
     }
     if (do_hex > 1) {
-        dStrHex((const char *)rb_buff, buf_capacity, 1);
+        hex2stdout(rb_buff, buf_capacity, 1);
         return 0;
     }
     printf("Slot status:\n");
@@ -405,16 +410,16 @@ do_safte_global_flags(int sg_fd, int do_hex, int do_raw, int verbose)
 {
     int res;
     unsigned int rb_len;
-    unsigned char *rb_buff;
+    uint8_t *rb_buff;
 
     rb_len = 16;
-    rb_buff = (unsigned char *)malloc(rb_len);
+    rb_buff = (uint8_t *)malloc(rb_len);
 
     if (verbose > 1)
-        fprintf(stderr, "Use READ BUFFER,mode=vendor_specific,buff_id=5 "
-                "to read global flags\n");
+        pr2serr("Use READ BUFFER,mode=vendor_specific,buff_id=5 to read "
+                "global flags\n");
     res = sg_ll_read_buffer(sg_fd, RWB_MODE_VENDOR, 5, 0,
-                            rb_buff, rb_len, 0, verbose);
+                            rb_buff, rb_len, false, verbose);
     if (res ) {
         if (res == SG_LIB_CAT_ILLEGAL_REQ) {
                 printf("Global Flags:\n\tNot implemented\n");
@@ -427,11 +432,11 @@ do_safte_global_flags(int sg_fd, int do_hex, int do_raw, int verbose)
     }
 
     if (do_raw > 1) {
-        dStrRaw((const char *)rb_buff, buf_capacity);
+        dStrRaw(rb_buff, buf_capacity);
         return 0;
     }
     if (do_hex > 1) {
-        dStrHex((const char *)rb_buff, buf_capacity, 1);
+        hex2stdout(rb_buff, buf_capacity, 1);
         return 0;
     }
     printf("Global Flags:\n");
@@ -467,8 +472,7 @@ do_safte_global_flags(int sg_fd, int do_hex, int do_raw, int verbose)
 static
 void usage()
 {
-    fprintf(stderr,
-            "Usage:  sg_safte [--config] [--devstatus] [--encstatus] "
+    pr2serr("Usage:  sg_safte [--config] [--devstatus] [--encstatus] "
             "[--flags] [--help]\n"
             "                 [--hex] [--insertions] [--raw] [--usage] "
             "[--verbose]\n"
@@ -508,22 +512,24 @@ static struct option long_options[] = {
 int
 main(int argc, char * argv[])
 {
-    int sg_fd, c, ret, peri_type, no_hex_raw;
+    bool do_insertions = false;
+    bool no_hex_raw;
+    int sg_fd, c, ret, peri_type;
     int res = SG_LIB_CAT_OTHER;
     const char * device_name = NULL;
     char ebuff[EBUFF_SZ];
-    unsigned char *rb_buff;
-    int do_config = 0;
-    int do_status = 0;
-    int do_slots = 0;
-    int do_flags = 0;
-    int do_usage = 0;
+    uint8_t *rb_buff;
+    bool do_config = false;
+    bool do_status = false;
+    bool do_slots = false;
+    bool do_flags = false;
+    bool do_usage = false;
     int do_hex = 0;
     int do_raw = 0;
     int verbose = 0;
-    int do_insertions = 0;
     const char * cp;
     char buff[48];
+    char b[80];
     struct sg_simple_inquiry_resp inq_resp;
     const char op_name[] = "READ BUFFER";
 
@@ -538,13 +544,13 @@ main(int argc, char * argv[])
 
         switch (c) {
             case 'c':
-                do_config = 1;
+                do_config = true;
                 break;
             case 'd':
-                do_slots = 1;
+                do_slots = true;
                 break;
             case 'f':
-                do_flags = 1;
+                do_flags = true;
                 break;
             case 'h':
             case '?':
@@ -554,25 +560,25 @@ main(int argc, char * argv[])
                 ++do_hex;
                 break;
             case 'i':
-                do_insertions = 1;
+                do_insertions = true;
                 break;
             case 'r':
                 ++do_raw;
                 break;
             case 's':
-                do_status = 1;
+                do_status = true;
                 break;
             case 'u':
-                do_usage = 1;
+                do_usage = true;
                 break;
             case 'v':
                 ++verbose;
                 break;
             case 'V':
-                fprintf(stderr, "Version string: %s\n", version_str);
+                pr2serr("Version string: %s\n", version_str);
                 exit(0);
             default:
-                fprintf(stderr, "unrecognised option code 0x%x ??\n", c);
+                pr2serr("unrecognised option code 0x%x ??\n", c);
                 usage();
                 return SG_LIB_SYNTAX_ERROR;
         }
@@ -584,15 +590,14 @@ main(int argc, char * argv[])
         }
         if (optind < argc) {
             for (; optind < argc; ++optind)
-                fprintf(stderr, "Unexpected extra argument: %s\n",
-                        argv[optind]);
+                pr2serr("Unexpected extra argument: %s\n", argv[optind]);
             usage();
             return SG_LIB_SYNTAX_ERROR;
         }
     }
 
     if (NULL == device_name) {
-        fprintf(stderr, "missing device name!\n");
+        pr2serr("missing device name!\n");
         usage();
         return SG_LIB_SYNTAX_ERROR;
     }
@@ -603,7 +608,8 @@ main(int argc, char * argv[])
         }
     }
 
-    if ((sg_fd = sg_cmds_open_device(device_name, 0 /* rw */, verbose)) < 0) {
+    if ((sg_fd = sg_cmds_open_device(device_name, false /* rw */, verbose))
+        < 0) {
         snprintf(ebuff, EBUFF_SZ, "sg_safte: error opening file: %s (rw)",
                  device_name);
         perror(ebuff);
@@ -622,13 +628,13 @@ main(int argc, char * argv[])
             else
                 printf("  Peripheral device type: 0x%x\n", peri_type);
         } else {
-            fprintf(stderr, "sg_safte: %s doesn't respond to a SCSI "
-                    "INQUIRY\n", device_name);
+            pr2serr("sg_safte: %s doesn't respond to a SCSI INQUIRY\n",
+                    device_name);
             return SG_LIB_CAT_OTHER;
         }
     }
 
-    rb_buff = (unsigned char *)malloc(buf_capacity);
+    rb_buff = (uint8_t *)malloc(buf_capacity);
     if (!rb_buff)
         goto err_out;
 
@@ -643,13 +649,11 @@ main(int argc, char * argv[])
         goto err_out;
     }
     if (1 == do_raw) {
-        dStrRaw((const char *)rb_buff, buf_capacity);
-        res = 0;
+        dStrRaw(rb_buff, buf_capacity);
         goto finish;
     }
     if (1 == do_hex) {
-        dStrHex((const char *)rb_buff, buf_capacity, 1);
-        res = 0;
+        hex2stdout(rb_buff, buf_capacity, 1);
         goto finish;
     }
 
@@ -718,29 +722,15 @@ err_out:
     case 0:
     case SG_LIB_CAT_RECOVERED:
         break;
-    case SG_LIB_CAT_ABORTED_COMMAND:
-        fprintf(stderr, "%s: aborted command\n", op_name);
-        break;
-    case SG_LIB_CAT_NOT_READY:
-        fprintf(stderr, "%s: device not ready\n", op_name);
-        break;
-    case SG_LIB_CAT_UNIT_ATTENTION:
-        fprintf(stderr, "%s: unit attention\n", op_name);
-        break;
-    case SG_LIB_CAT_INVALID_OP:
-        fprintf(stderr, "%s: operation not supported\n", op_name);
-        break;
-    case SG_LIB_CAT_ILLEGAL_REQ:
-        fprintf(stderr, "%s: bad field in cdb\n", op_name);
-        break;
     default:
-        fprintf(stderr, "%s failed\n", op_name);
+        sg_get_category_sense_str(res, sizeof(b), b, verbose);
+        pr2serr("%s failed: %s\n", op_name, b);
         break;
     }
     ret = res;
     res = sg_cmds_close_device(sg_fd);
     if (res < 0) {
-        fprintf(stderr, "close error: %s\n", safe_strerror(-res));
+        pr2serr("close error: %s\n", safe_strerror(-res));
         if (0 == ret)
             return SG_LIB_FILE_ERROR;
     }
