@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2011-2015 Douglas Gilbert.
+ * Copyright (c) 2011-2018 Douglas Gilbert.
  * All rights reserved.
  * Use of this source code is governed by a BSD-style
  * license that can be found in the BSD_LICENSE file.
@@ -9,6 +9,8 @@
 #include <fcntl.h>
 #include <stdio.h>
 #include <stdlib.h>
+#include <stdarg.h>
+#include <stdbool.h>
 #include <string.h>
 #include <errno.h>
 #include <limits.h>
@@ -21,12 +23,15 @@
 #ifdef HAVE_CONFIG_H
 #include "config.h"
 #endif
+
 #include "sg_lib.h"
 #include "sg_pt.h"
 #include "sg_cmds_basic.h"
 #include "sg_cmds_extra.h"
+#include "sg_unaligned.h"
+#include "sg_pr2serr.h"
 
-static const char * version_str = "0.97 20150511";
+static const char * version_str = "1.07 20180219";
 
 /* Not all environments support the Unix sleep() */
 #if defined(MSC_VER) || defined(__MINGW32__)
@@ -74,6 +79,7 @@ static struct option long_options[] = {
     {"pattern", required_argument, 0, 'p'},
     {"quick", no_argument, 0, 'Q'},
     {"test", required_argument, 0, 'T'},
+    {"timeout", required_argument, 0, 't'},
     {"verbose", no_argument, 0, 'v'},
     {"version", no_argument, 0, 'V'},
     {"wait", no_argument, 0, 'w'},
@@ -82,36 +88,38 @@ static struct option long_options[] = {
 };
 
 struct opts_t {
-    int ause;
-    int block;
+    bool ause;
+    bool block;
+    bool crypto;
+    bool desc;
+    bool early;
+    bool fail;
+    bool invert;
+    bool overwrite;
+    bool quick;
+    bool wait;
+    bool znr;
     int count;
-    int crypto;
-    int desc;
-    int early;
-    int fail;
-    int invert;
     int ipl;    /* initialization pattern length */
-    int overwrite;
     int test;
-    int quick;
+    int timeout;        /* in seconds */
     int verbose;
-    int wait;
     int zero;
     const char * pattern_fn;
 };
 
 
-
 static void
 usage()
 {
-  fprintf(stderr, "Usage: "
-          "sg_sanitize [--ause] [--block] [--count=OC] [--crypto] [--early]\n"
+  pr2serr("Usage: sg_sanitize [--ause] [--block] [--count=OC] [--crypto] "
+          "[--early]\n"
           "                   [--fail] [--help] [--invert] [--ipl=LEN] "
           "[--overwrite]\n"
           "                   [--pattern=PF] [--quick] [--test=TE] "
-          "[--verbose]\n"
-          "                   [--version] [--wait] DEVICE\n"
+          "[--timeout=SEC]\n"
+          "                   [--verbose] [--version] [--wait] [--zero] "
+          "[--znr] DEVICE\n"
           "  where:\n"
           "    --ause|-A            set AUSE bit in cdb\n"
           "    --block|-B           do BLOCK ERASE sanitize\n"
@@ -141,12 +149,14 @@ usage()
           "    --test=TE|-T TE      TE is placed in TEST field of "
           "OVERWRITE\n"
           "                         parameter list (def: 0)\n"
+          "    --timeout=SEC|-t SEC    SANITIZE command timeout in seconds\n"
           "    --verbose|-v         increase verbosity\n"
           "    --version|-V         print version string then exit\n"
           "    --wait|-w            wait for command to finish (could "
           "take hours)\n"
           "    --zero|-z            use pattern of zeros for "
-          "OVERWRITE\n\n"
+          "OVERWRITE\n"
+          "    --znr|-Z             set ZNR (zone no reset) bit in cdb\n\n"
           "Performs a SCSI SANITIZE command.\n    <<<WARNING>>>: all data "
           "on DEVICE will be lost.\nDefault action is to give user time to "
           "reconsider; then execute SANITIZE\ncommand with IMMED bit set; "
@@ -161,58 +171,66 @@ static int
 do_sanitize(int sg_fd, const struct opts_t * op, const void * param_lstp,
             int param_lst_len)
 {
-    int k, ret, res, sense_cat, immed;
-    unsigned char sanCmdBlk[SANITIZE_OP_LEN];
-    unsigned char sense_b[SENSE_BUFF_LEN];
+    bool immed;
+    int k, ret, res, sense_cat, timeout;
+    uint8_t san_cdb[SANITIZE_OP_LEN];
+    uint8_t sense_b[SENSE_BUFF_LEN];
     struct sg_pt_base * ptvp;
 
     if (op->early || op->wait)
-        immed = op->early ? 1 : 0;
+        immed = op->early;
     else
-        immed = 1;
-    memset(sanCmdBlk, 0, sizeof(sanCmdBlk));
-    sanCmdBlk[0] = SANITIZE_OP;
+        immed = true;
+    timeout = (immed ? SHORT_TIMEOUT : LONG_TIMEOUT);
+    /* only use command line timeout if it exceeds previous defaults */
+    if (op->timeout > timeout)
+        timeout = op->timeout;
+    memset(san_cdb, 0, sizeof(san_cdb));
+    san_cdb[0] = SANITIZE_OP;
     if (op->overwrite)
-        sanCmdBlk[1] = SANITIZE_SA_OVERWRITE;
+        san_cdb[1] = SANITIZE_SA_OVERWRITE;
     else if (op->block)
-        sanCmdBlk[1] = SANITIZE_SA_BLOCK_ERASE;
+        san_cdb[1] = SANITIZE_SA_BLOCK_ERASE;
     else if (op->crypto)
-        sanCmdBlk[1] = SANITIZE_SA_CRYPTO_ERASE;
+        san_cdb[1] = SANITIZE_SA_CRYPTO_ERASE;
     else if (op->fail)
-        sanCmdBlk[1] = SANITIZE_SA_EXIT_FAIL_MODE;
+        san_cdb[1] = SANITIZE_SA_EXIT_FAIL_MODE;
     else
         return SG_LIB_SYNTAX_ERROR;
     if (immed)
-        sanCmdBlk[1] |= 0x80;
+        san_cdb[1] |= 0x80;
+    if (op->znr)        /* added sbc4r07 */
+        san_cdb[1] |= 0x40;
     if (op->ause)
-        sanCmdBlk[1] |= 0x20;
-    sanCmdBlk[7] = ((param_lst_len >> 8) & 0xff);
-    sanCmdBlk[8] = (param_lst_len & 0xff);
+        san_cdb[1] |= 0x20;
+    sg_put_unaligned_be16((uint16_t)param_lst_len, san_cdb + 7);
 
     if (op->verbose > 1) {
-        fprintf(stderr, "    Sanitize cmd: ");
+        pr2serr("    Sanitize cdb: ");
         for (k = 0; k < SANITIZE_OP_LEN; ++k)
-            fprintf(stderr, "%02x ", sanCmdBlk[k]);
-        fprintf(stderr, "\n");
-    }
-    if ((op->verbose > 2) && (param_lst_len > 0)) {
-        fprintf(stderr, "    Parameter list contents:\n");
-        dStrHexErr((const char *)param_lstp, param_lst_len, 1);
+            pr2serr("%02x ", san_cdb[k]);
+        pr2serr("\n");
+        if (op->verbose > 2) {
+            if (param_lst_len > 0) {
+                pr2serr("    Parameter list contents:\n");
+                hex2stderr((const uint8_t *)param_lstp, param_lst_len, 1);
+            }
+            pr2serr("    Sanitize command timeout: %d seconds\n", timeout);
+        }
     }
     ptvp = construct_scsi_pt_obj();
     if (NULL == ptvp) {
-        fprintf(stderr, "Sanitize: out of memory\n");
+        pr2serr("Sanitize: out of memory\n");
         return -1;
     }
-    set_scsi_pt_cdb(ptvp, sanCmdBlk, sizeof(sanCmdBlk));
+    set_scsi_pt_cdb(ptvp, san_cdb, sizeof(san_cdb));
     set_scsi_pt_sense(ptvp, sense_b, sizeof(sense_b));
-    set_scsi_pt_data_out(ptvp, (unsigned char *)param_lstp, param_lst_len);
-    res = do_scsi_pt(ptvp, sg_fd, (immed ? SHORT_TIMEOUT : LONG_TIMEOUT),
-                     op->verbose);
-    ret = sg_cmds_process_resp(ptvp, "Sanitize", res, 0, sense_b,
-                               1 /*noisy */, op->verbose, &sense_cat);
+    set_scsi_pt_data_out(ptvp, (uint8_t *)param_lstp, param_lst_len);
+    res = do_scsi_pt(ptvp, sg_fd, timeout, op->verbose);
+    ret = sg_cmds_process_resp(ptvp, "Sanitize", res, SG_NO_DATA_IN, sense_b,
+                               true /*noisy */, op->verbose, &sense_cat);
     if (-1 == ret)
-        ;
+        ret = sg_convert_errno(get_scsi_pt_os_err(ptvp));
     else if (-2 == ret) {
         switch (sense_cat) {
         case SG_LIB_CAT_RECOVERED:
@@ -221,13 +239,14 @@ do_sanitize(int sg_fd, const struct opts_t * op, const void * param_lstp,
             break;
         case SG_LIB_CAT_MEDIUM_HARD:
             {
-                int valid, slen;
+                bool valid;
+                int slen;
                 uint64_t ull = 0;
 
                 slen = get_scsi_pt_sense_len(ptvp);
                 valid = sg_get_sense_info_fld(sense_b, slen, &ull);
                 if (valid)
-                    fprintf(stderr, "Medium or hardware error starting at "
+                    pr2serr("Medium or hardware error starting at "
                             "lba=%" PRIu64 " [0x%" PRIx64 "]\n", ull, ull);
             }
             ret = sense_cat;
@@ -249,51 +268,51 @@ do_sanitize(int sg_fd, const struct opts_t * op, const void * param_lstp,
 #define TPROTO_ISCSI 5
 
 static char *
-get_lu_name(const unsigned char * ucp, int u_len, char * b, int b_len)
+get_lu_name(const uint8_t * bp, int u_len, char * b, int b_len)
 {
     int len, off, sns_dlen, dlen, k;
-    unsigned char u_sns[512];
+    uint8_t u_sns[512];
     char * cp;
 
     len = u_len - 4;
-    ucp += 4;
+    bp += 4;
     off = -1;
-    if (0 == sg_vpd_dev_id_iter(ucp, len, &off, VPD_ASSOC_LU,
+    if (0 == sg_vpd_dev_id_iter(bp, len, &off, VPD_ASSOC_LU,
                                 8 /* SCSI name string (sns) */,
                                 3 /* UTF-8 */)) {
-        sns_dlen = ucp[off + 3];
-        memcpy(u_sns, ucp + off + 4, sns_dlen);
+        sns_dlen = bp[off + 3];
+        memcpy(u_sns, bp + off + 4, sns_dlen);
         /* now want to check if this is iSCSI */
         off = -1;
-        if (0 == sg_vpd_dev_id_iter(ucp, len, &off, VPD_ASSOC_TPORT,
+        if (0 == sg_vpd_dev_id_iter(bp, len, &off, VPD_ASSOC_TPORT,
                                     8 /* SCSI name string (sns) */,
                                     3 /* UTF-8 */)) {
-            if ((0x80 & ucp[1]) && (TPROTO_ISCSI == (ucp[0] >> 4))) {
+            if ((0x80 & bp[1]) && (TPROTO_ISCSI == (bp[0] >> 4))) {
                 snprintf(b, b_len, "%.*s", sns_dlen, u_sns);
                 return b;
             }
         }
     } else
         sns_dlen = 0;
-    if (0 == sg_vpd_dev_id_iter(ucp, len, &off, VPD_ASSOC_LU,
+    if (0 == sg_vpd_dev_id_iter(bp, len, &off, VPD_ASSOC_LU,
                                 3 /* NAA */, 1 /* binary */)) {
-        dlen = ucp[off + 3];
+        dlen = bp[off + 3];
         if (! ((8 == dlen) || (16 ==dlen)))
             return b;
         cp = b;
         for (k = 0; ((k < dlen) && (b_len > 1)); ++k) {
-            snprintf(cp, b_len, "%02x", ucp[off + 4 + k]);
+            snprintf(cp, b_len, "%02x", bp[off + 4 + k]);
             cp += 2;
             b_len -= 2;
         }
-    } else if (0 == sg_vpd_dev_id_iter(ucp, len, &off, VPD_ASSOC_LU,
+    } else if (0 == sg_vpd_dev_id_iter(bp, len, &off, VPD_ASSOC_LU,
                                        2 /* EUI */, 1 /* binary */)) {
-        dlen = ucp[off + 3];
+        dlen = bp[off + 3];
         if (! ((8 == dlen) || (12 == dlen) || (16 ==dlen)))
             return b;
         cp = b;
         for (k = 0; ((k < dlen) && (b_len > 1)); ++k) {
-            snprintf(cp, b_len, "%02x", ucp[off + 4 + k]);
+            snprintf(cp, b_len, "%02x", bp[off + 4 + k]);
             cp += 2;
             b_len -= 2;
         }
@@ -308,16 +327,16 @@ get_lu_name(const unsigned char * ucp, int u_len, char * b, int b_len)
 #define VPD_DEVICE_ID 0x83
 
 static int
-print_dev_id(int fd, unsigned char * sinq_resp, int max_rlen, int verbose)
+print_dev_id(int fd, uint8_t * sinq_resp, int max_rlen, int verbose)
 {
     int res, k, n, verb, pdt, has_sn, has_di;
-    unsigned char b[256];
+    uint8_t b[256];
     char a[256];
     char pdt_name[64];
 
     verb = (verbose > 1) ? verbose - 1 : 0;
     memset(sinq_resp, 0, max_rlen);
-    res = sg_ll_inquiry(fd, 0, 0 /* evpd */, 0 /* pg_op */, b,
+    res = sg_ll_inquiry(fd, false, false /* evpd */, 0 /* pg_op */, b,
                         SAFE_STD_INQ_RESP_LEN, 1, verb);
     if (res)
         return res;
@@ -336,30 +355,29 @@ print_dev_id(int fd, unsigned char * sinq_resp, int max_rlen, int verbose)
         if (b[5] & 1)
             printf("      << supports protection information>>\n");
     } else {
-        fprintf(stderr, "Short INQUIRY response: %d bytes, expect at least "
-                "36\n", n);
+        pr2serr("Short INQUIRY response: %d bytes, expect at least 36\n", n);
         return SG_LIB_CAT_OTHER;
     }
-    res = sg_ll_inquiry(fd, 0, 1 /* evpd */, VPD_SUPPORTED_VPDS, b,
+    res = sg_ll_inquiry(fd, false, true /* evpd */, VPD_SUPPORTED_VPDS, b,
                         SAFE_STD_INQ_RESP_LEN, 1, verb);
     if (res) {
         if (verbose)
-            fprintf(stderr, "VPD_SUPPORTED_VPDS gave res=%d\n", res);
+            pr2serr("VPD_SUPPORTED_VPDS gave res=%d\n", res);
         return 0;
     }
     if (VPD_SUPPORTED_VPDS != b[1]) {
         if (verbose)
-            fprintf(stderr, "VPD_SUPPORTED_VPDS corrupted\n");
+            pr2serr("VPD_SUPPORTED_VPDS corrupted\n");
         return 0;
     }
-    n = (b[2] << 8) + b[3];
+    n = sg_get_unaligned_be16(b + 2);
     if (n > (SAFE_STD_INQ_RESP_LEN - 4))
         n = (SAFE_STD_INQ_RESP_LEN - 4);
     for (k = 0, has_sn = 0, has_di = 0; k < n; ++k) {
         if (VPD_UNIT_SERIAL_NUM == b[4 + k]) {
             if (has_di) {
                 if (verbose)
-                    fprintf(stderr, "VPD_SUPPORTED_VPDS dis-ordered\n");
+                    pr2serr("VPD_SUPPORTED_VPDS dis-ordered\n");
                 return 0;
             }
             ++has_sn;
@@ -369,37 +387,37 @@ print_dev_id(int fd, unsigned char * sinq_resp, int max_rlen, int verbose)
         }
     }
     if (has_sn) {
-        res = sg_ll_inquiry(fd, 0, 1 /* evpd */, VPD_UNIT_SERIAL_NUM, b,
-                            sizeof(b), 1, verb);
+        res = sg_ll_inquiry(fd, false, true /* evpd */, VPD_UNIT_SERIAL_NUM,
+                            b, sizeof(b), 1, verb);
         if (res) {
             if (verbose)
-                fprintf(stderr, "VPD_UNIT_SERIAL_NUM gave res=%d\n", res);
+                pr2serr("VPD_UNIT_SERIAL_NUM gave res=%d\n", res);
             return 0;
         }
         if (VPD_UNIT_SERIAL_NUM != b[1]) {
             if (verbose)
-                fprintf(stderr, "VPD_UNIT_SERIAL_NUM corrupted\n");
+                pr2serr("VPD_UNIT_SERIAL_NUM corrupted\n");
             return 0;
         }
-        n = (b[2] << 8) + b[3];
+        n = sg_get_unaligned_be16(b + 2);
         if (n > (int)(sizeof(b) - 4))
             n = (sizeof(b) - 4);
         printf("      Unit serial number: %.*s\n", n, (const char *)(b + 4));
     }
     if (has_di) {
-        res = sg_ll_inquiry(fd, 0, 1 /* evpd */, VPD_DEVICE_ID, b,
+        res = sg_ll_inquiry(fd, false, true /* evpd */, VPD_DEVICE_ID, b,
                             sizeof(b), 1, verb);
         if (res) {
             if (verbose)
-                fprintf(stderr, "VPD_DEVICE_ID gave res=%d\n", res);
+                pr2serr("VPD_DEVICE_ID gave res=%d\n", res);
             return 0;
         }
         if (VPD_DEVICE_ID != b[1]) {
             if (verbose)
-                fprintf(stderr, "VPD_DEVICE_ID corrupted\n");
+                pr2serr("VPD_DEVICE_ID corrupted\n");
             return 0;
         }
-        n = (b[2] << 8) + b[3];
+        n = sg_get_unaligned_be16(b + 2);
         if (n > (int)(sizeof(b) - 4))
             n = (sizeof(b) - 4);
         n = strlen(get_lu_name(b, n + 4, a, sizeof(a)));
@@ -413,19 +431,19 @@ print_dev_id(int fd, unsigned char * sinq_resp, int max_rlen, int verbose)
 int
 main(int argc, char * argv[])
 {
+    bool got_stdin = false;
     int sg_fd, k, res, c, infd, progress, vb, n, resp_len;
-    int got_stdin = 0;
     int param_lst_len = 0;
+    int ret = -1;
     const char * device_name = NULL;
     char ebuff[EBUFF_SZ];
     char b[80];
-    unsigned char requestSenseBuff[DEF_REQS_RESP_LEN];
-    unsigned char * wBuff = NULL;
-    int ret = -1;
+    uint8_t rsBuff[DEF_REQS_RESP_LEN];
+    uint8_t * wBuff = NULL;
     struct opts_t opts;
     struct opts_t * op;
     struct stat a_stat;
-    unsigned char inq_resp[SAFE_STD_INQ_RESP_LEN];
+    uint8_t inq_resp[SAFE_STD_INQ_RESP_LEN];
 
     op = &opts;
     memset(op, 0, sizeof(opts));
@@ -433,37 +451,36 @@ main(int argc, char * argv[])
     while (1) {
         int option_index = 0;
 
-        c = getopt_long(argc, argv, "ABc:CdeFhi:IOp:QT:vVwz", long_options,
+        c = getopt_long(argc, argv, "ABc:CdeFhi:IOp:Qt:T:vVwzZ", long_options,
                         &option_index);
         if (c == -1)
             break;
 
         switch (c) {
         case 'A':
-            ++op->ause;
+            op->ause = true;
             break;
         case 'B':
-            ++op->block;
+            op->block = true;
             break;
         case 'c':
             op->count = sg_get_num(optarg);
             if ((op->count < 1) || (op->count > 31))  {
-                fprintf(stderr, "bad argument to '--count', expect 1 to "
-                        "31\n");
+                pr2serr("bad argument to '--count', expect 1 to 31\n");
                 return SG_LIB_SYNTAX_ERROR;
             }
             break;
         case 'C':
-            ++op->crypto;
+            op->crypto = true;
             break;
         case 'd':
-            ++op->desc;
+            op->desc = true;
             break;
         case 'e':
-            ++op->early;
+            op->early = true;
             break;
         case 'F':
-            ++op->fail;
+            op->fail = true;
             break;
         case 'h':
         case '?':
@@ -472,27 +489,33 @@ main(int argc, char * argv[])
         case 'i':
             op->ipl = sg_get_num(optarg);
             if ((op->ipl < 1) || (op->ipl > 65535))  {
-                fprintf(stderr, "bad argument to '--ipl', expect 1 to "
-                        "65535\n");
+                pr2serr("bad argument to '--ipl', expect 1 to 65535\n");
                 return SG_LIB_SYNTAX_ERROR;
             }
             break;
         case 'I':
-            ++op->invert;
+            op->invert = true;
             break;
         case 'O':
-            ++op->overwrite;
+            op->overwrite = true;
             break;
         case 'p':
             op->pattern_fn = optarg;
             break;
         case 'Q':
-            ++op->quick;
+            op->quick = true;
+            break;
+        case 't':
+            op->timeout = sg_get_num(optarg);
+            if (op->timeout < 0) {
+                pr2serr("bad argument to '--timeout=SEC', want 0 or more\n");
+                return SG_LIB_SYNTAX_ERROR;
+            }
             break;
         case 'T':
             op->test = sg_get_num(optarg);
             if ((op->test < 0) || (op->test > 3))  {
-                fprintf(stderr, "bad argument to '--test', expect 0 to 3\n");
+                pr2serr("bad argument to '--test', expect 0 to 3\n");
                 return SG_LIB_SYNTAX_ERROR;
             }
             break;
@@ -500,16 +523,19 @@ main(int argc, char * argv[])
             ++op->verbose;
             break;
         case 'V':
-            fprintf(stderr, ME "version: %s\n", version_str);
+            pr2serr(ME "version: %s\n", version_str);
             return 0;
         case 'w':
-            ++op->wait;
+            op->wait = true;
             break;
         case 'z':
             ++op->zero;
             break;
+        case 'Z':
+            op->znr = true;
+            break;
         default:
-            fprintf(stderr, "unrecognised option code 0x%x ??\n", c);
+            pr2serr("unrecognised option code 0x%x ??\n", c);
             usage();
             return SG_LIB_SYNTAX_ERROR;
         }
@@ -521,66 +547,65 @@ main(int argc, char * argv[])
         }
         if (optind < argc) {
             for (; optind < argc; ++optind)
-                fprintf(stderr, "Unexpected extra argument: %s\n",
-                        argv[optind]);
+                pr2serr("Unexpected extra argument: %s\n", argv[optind]);
             usage();
             return SG_LIB_SYNTAX_ERROR;
         }
     }
     if (NULL == device_name) {
-        fprintf(stderr, "missing device name!\n");
+        pr2serr("missing device name!\n");
         usage();
         return SG_LIB_SYNTAX_ERROR;
     }
     vb = op->verbose;
-    n = !!op->block + !!op->crypto + !!op->fail + !!op->overwrite;
+    n = (int)op->block + (int)op->crypto + (int)op->fail + (int)op->overwrite;
     if (1 != n) {
-        fprintf(stderr, "one and only one of '--block', '--crypto', "
-                "'--fail' or '--overwrite' please\n");
+        pr2serr("one and only one of '--block', '--crypto', '--fail' or "
+                "'--overwrite' please\n");
         return SG_LIB_SYNTAX_ERROR;
     }
     if (op->overwrite) {
         if (op->zero) {
             if (op->pattern_fn) {
-                fprintf(stderr, "confused: both '--pattern=PF' and '--zero' "
+                pr2serr("confused: both '--pattern=PF' and '--zero' "
                         "options\n");
                 return SG_LIB_SYNTAX_ERROR;
             }
             op->ipl = 4;
         } else {
             if (NULL == op->pattern_fn) {
-                fprintf(stderr, "'--overwrite' requires '--pattern=PF' "
-                        "or '--zero' option\n");
+                pr2serr("'--overwrite' requires '--pattern=PF' or '--zero' "
+                        "option\n");
                 return SG_LIB_SYNTAX_ERROR;
             }
-            got_stdin = (0 == strcmp(op->pattern_fn, "-")) ? 1 : 0;
+            got_stdin = (0 == strcmp(op->pattern_fn, "-"));
             if (! got_stdin) {
                 memset(&a_stat, 0, sizeof(a_stat));
                 if (stat(op->pattern_fn, &a_stat) < 0) {
-                    fprintf(stderr, "pattern file: unable to stat(%s): %s\n",
+                    pr2serr("pattern file: unable to stat(%s): %s\n",
                             op->pattern_fn, safe_strerror(errno));
                     return SG_LIB_FILE_ERROR;
                 }
                 if (op->ipl <= 0) {
                     op->ipl = (int)a_stat.st_size;
                     if (op->ipl > MAX_XFER_LEN) {
-                        fprintf(stderr, "pattern file length exceeds 65535 "
-                                "bytes, need '--ipl=LEN' option\n");
+                        pr2serr("pattern file length exceeds 65535 bytes, "
+                                "need '--ipl=LEN' option\n");
                          return SG_LIB_FILE_ERROR;
                     }
                 }
             }
             if (op->ipl < 1) {
-                fprintf(stderr, "'--overwrite' requires '--ipl=LEN' "
-                        "option if can't get PF length\n");
+                pr2serr("'--overwrite' requires '--ipl=LEN' option if can't "
+                        "get PF length\n");
                 return SG_LIB_SYNTAX_ERROR;
             }
         }
     }
 
-    sg_fd = sg_cmds_open_device(device_name, 0 /* rw */, vb);
+    sg_fd = sg_cmds_open_device(device_name, false /* rw */, vb);
     if (sg_fd < 0) {
-        fprintf(stderr, ME "open error: %s: %s\n", device_name,
+        pr2serr(ME "open error: %s: %s\n", device_name,
                 safe_strerror(-sg_fd));
         return SG_LIB_FILE_ERROR;
     }
@@ -591,10 +616,10 @@ main(int argc, char * argv[])
 
     if (op->overwrite) {
         param_lst_len = op->ipl + 4;
-        wBuff = (unsigned char*)calloc(op->ipl + 4, 1);
+        wBuff = (uint8_t*)calloc(op->ipl + 4, 1);
         if (NULL == wBuff) {
-            fprintf(stderr, "unable to allocate %d bytes of memory with "
-                    "calloc()\n", op->ipl + 4);
+            pr2serr("unable to allocate %d bytes of memory with calloc()\n",
+                    op->ipl + 4);
             ret = SG_LIB_SYNTAX_ERROR;
             goto err_out;
         }
@@ -629,9 +654,9 @@ main(int argc, char * argv[])
                 goto err_out;
             }
             if (res < op->ipl) {
-                fprintf(stderr, "tried to read %d bytes from %s, got %d "
-                        "bytes\n", op->ipl, op->pattern_fn, res);
-                fprintf(stderr, "  so pad with 0x0 bytes and continue\n");
+                pr2serr("tried to read %d bytes from %s, got %d bytes\n",
+                         op->ipl, op->pattern_fn, res);
+                pr2serr("  so pad with 0x0 bytes and continue\n");
             }
             if (! got_stdin)
                 close(infd);
@@ -641,11 +666,10 @@ main(int argc, char * argv[])
             wBuff[0] |= ((op->test & 0x3) << 5);
         if (op->invert)
             wBuff[0] |= 0x80;
-        wBuff[2] = ((op->ipl >> 8) & 0xff);
-        wBuff[3] = (op->ipl & 0xff);
+        sg_put_unaligned_be16((uint16_t)op->ipl, wBuff + 2);
     }
 
-    if ((0 == op->quick) && (! op->fail)) {
+    if ((! op->quick) && (! op->fail)) {
         printf("\nA SANITIZE will commence in 15 seconds\n");
         printf("    ALL data on %s will be DESTROYED\n", device_name);
         printf("        Press control-C to abort\n");
@@ -663,50 +687,49 @@ main(int argc, char * argv[])
     ret = do_sanitize(sg_fd, op, wBuff, param_lst_len);
     if (ret) {
         sg_get_category_sense_str(ret, sizeof(b), b, vb);
-        fprintf(stderr, "Sanitize failed: %s\n", b);
+        pr2serr("Sanitize failed: %s\n", b);
     }
 
-    if ((0 == ret) && (0 == op->early) && (0 == op->wait)) {
-        for (k = 0 ;; ++k) {
+    if ((0 == ret) && (! op->early) && (! op->wait)) {
+        for (k = 0; ;++k) {     /* unbounded, exits via break */
             sleep_for(POLL_DURATION_SECS);
-            memset(requestSenseBuff, 0x0, sizeof(requestSenseBuff));
-            res = sg_ll_request_sense(sg_fd, op->desc, requestSenseBuff,
-                                      sizeof(requestSenseBuff), 1, vb);
+            memset(rsBuff, 0x0, sizeof(rsBuff));
+            res = sg_ll_request_sense(sg_fd, op->desc, rsBuff, sizeof(rsBuff),
+                                      1, vb);
             if (res) {
                 ret = res;
                 if (SG_LIB_CAT_INVALID_OP == res)
-                    fprintf(stderr, "Request Sense command not supported\n");
+                    pr2serr("Request Sense command not supported\n");
                 else if (SG_LIB_CAT_ILLEGAL_REQ == res) {
-                    fprintf(stderr, "bad field in Request Sense cdb\n");
-                    if (1 == op->desc) {
-                        fprintf(stderr, "Descriptor type sense may not be "
-                                "supported, try again with fixed type\n");
-                        op->desc = 0;
+                    pr2serr("bad field in Request Sense cdb\n");
+                    if (op->desc) {
+                        pr2serr("Descriptor type sense may not be supported, "
+                                "try again with fixed type\n");
+                        op->desc = false;
                         continue;
                     }
                 } else {
                     sg_get_category_sense_str(res, sizeof(b), b, vb);
-                    fprintf(stderr, "Request Sense: %s\n", b);
+                    pr2serr("Request Sense: %s\n", b);
                     if (0 == vb)
-                        fprintf(stderr, "    try the '-v' option for "
-                                "more information\n");
+                        pr2serr("    try the '-v' option for more "
+                                "information\n");
                 }
                 break;
             }
             /* "Additional sense length" same in descriptor and fixed */
-            resp_len = requestSenseBuff[7] + 8;
+            resp_len = rsBuff[7] + 8;
             if (vb > 2) {
-                fprintf(stderr, "Parameter data in hex\n");
-                dStrHexErr((const char *)requestSenseBuff, resp_len, 1);
+                pr2serr("Parameter data in hex\n");
+                hex2stderr(rsBuff, resp_len, 1);
             }
             progress = -1;
-            sg_get_sense_progress_fld(requestSenseBuff, resp_len,
-                                      &progress);
+            sg_get_sense_progress_fld(rsBuff, resp_len, &progress);
             if (progress < 0) {
                 ret = res;
                 if (vb > 1)
-                     fprintf(stderr, "No progress indication found, "
-                             "iteration %d\n", k + 1);
+                     pr2serr("No progress indication found, iteration %d\n",
+                             k + 1);
                 /* N.B. exits first time there isn't a progress indication */
                 break;
             } else
@@ -720,7 +743,7 @@ err_out:
         free(wBuff);
     res = sg_cmds_close_device(sg_fd);
     if (res < 0) {
-        fprintf(stderr, "close error: %s\n", safe_strerror(-res));
+        pr2serr("close error: %s\n", safe_strerror(-res));
         if (0 == ret)
             return SG_LIB_FILE_ERROR;
     }
