@@ -1,5 +1,5 @@
 /* A utility program originally written for the Linux OS SCSI subsystem
-*  Copyright (C) 2003-2015 D. Gilbert
+*  Copyright (C) 2003-2018 D. Gilbert
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
@@ -20,14 +20,18 @@
 #ifdef HAVE_CONFIG_H
 #include "config.h"
 #endif
+
 #include "sg_lib.h"
 #include "sg_cmds_basic.h"
 #include "sg_cmds_extra.h"
+#if SG_LIB_WIN32
 #include "sg_pt.h"      /* needed for scsi_pt_win32_direct() */
+#endif
 #include "sg_unaligned.h"
+#include "sg_pr2serr.h"
 
 
-static const char * version_str = "0.46 20150427";
+static const char * version_str = "0.59 20180302";
 
 #define ME "sg_senddiag: "
 
@@ -47,6 +51,7 @@ static struct option long_options[] = {
         {"raw", required_argument, 0, 'r'},
         {"selftest", required_argument, 0, 's'},
         {"test", no_argument, 0, 't'},
+        {"timeout", required_argument, 0, 'T'},
         {"uoff", no_argument, 0, 'u'},
         {"verbose", no_argument, 0, 'v'},
         {"version", no_argument, 0, 'V'},
@@ -54,23 +59,24 @@ static struct option long_options[] = {
 };
 
 struct opts_t {
-    int do_doff;
-    int do_extdur;
+    bool do_deftest;
+    bool do_doff;
+    bool do_extdur;
+    bool do_list;
+    bool do_pf;
+    bool do_raw;
+    bool do_uoff;
+    bool do_version;
+    bool opt_new;
     int do_help;
     int do_hex;
-    int do_list;
     int maxlen;
     int page_code;
-    int do_pf;
-    int do_raw;
     int do_selftest;
-    int do_deftest;
-    int do_uoff;
+    int timeout;
     int do_verbose;
-    int do_version;
     const char * device_name;
     const char * raw_arg;
-    int opt_new;
 };
 
 
@@ -81,9 +87,9 @@ usage()
            "[--list]\n"
            "                   [--maxlen=LEN] [--page=PG] [--pf] "
            "[--raw=H,H...]\n"
-           "                   [--selftest=ST] [--test] [--uoff] "
-           "[--verbose] [--version]\n"
-           "                   [DEVICE]\n"
+           "                   [--selftest=ST] [--test] [--timeout=SEC] "
+           "[--uoff]\n"
+           "                   [--verbose] [--version] [DEVICE]\n"
            "  where:\n"
            "    --doff|-d       device online (def: 0, only with '--test')\n"
            "    --extdur|-e     duration of an extended self-test (from mode "
@@ -111,8 +117,11 @@ usage()
            "                           5->foreground short, 6->foreground "
            "extended\n"
            "    --test|-t       default self-test\n"
+           "    --timeout=SEC|-T SEC    timeout for foreground self tests\n"
+           "                            unit: second (def: 7200 seconds)\n"
            "    --uoff|-u       unit offline (def: 0, only with '--test')\n"
            "    --verbose|-v    increase verbosity\n"
+           "    --old|-O        use old interface (use as first option)\n"
            "    --version|-V    output version string then exit\n\n"
            "Performs a SCSI SEND DIAGNOSTIC (and/or a RECEIVE DIAGNOSTIC "
            "RESULTS) command\n"
@@ -124,7 +133,7 @@ usage_old()
 {
     printf("Usage: sg_senddiag [-doff] [-e] [-h] [-H] [-l] [-pf]"
            " [-raw=H,H...]\n"
-           "                   [-s=SF] [-t] [-uoff] [-v] [-V] "
+           "                   [-s=SF] [-t] [-T=SEC] [-uoff] [-v] [-V] "
            "[DEVICE]\n"
            "  where:\n"
            "    -doff   device online (def: 0, only with '-t')\n"
@@ -142,9 +151,11 @@ usage_old()
            " 4->abort test\n"
            "            5->foreground short, 6->foreground extended\n"
            "    -t      default self-test\n"
+           "    -T SEC    timeout for foreground self tests\n"
            "    -uoff   unit offline (def: 0, only with '-t')\n"
            "    -v      increase verbosity (print issued SCSI cmds)\n"
            "    -V      output version string\n"
+           "    -N|--new   use new interface\n"
            "    -?      output this usage message\n\n"
            "Performs a SCSI SEND DIAGNOSTIC (and/or a RECEIVE DIAGNOSTIC "
            "RESULTS) command\n"
@@ -152,24 +163,24 @@ usage_old()
 }
 
 static int
-process_cl_new(struct opts_t * op, int argc, char * argv[])
+new_parse_cmd_line(struct opts_t * op, int argc, char * argv[])
 {
     int c, n;
 
     while (1) {
         int option_index = 0;
 
-        c = getopt_long(argc, argv, "dehHlm:NOpP:r:s:tuvV", long_options,
+        c = getopt_long(argc, argv, "dehHlm:NOpP:r:s:tT:uvV", long_options,
                         &option_index);
         if (c == -1)
             break;
 
         switch (c) {
         case 'd':
-            op->do_doff = 1;
+            op->do_doff = true;
             break;
         case 'e':
-            op->do_extdur = 1;
+            op->do_extdur = true;
             break;
         case 'h':
         case '?':
@@ -179,13 +190,13 @@ process_cl_new(struct opts_t * op, int argc, char * argv[])
             ++op->do_hex;
             break;
         case 'l':
-            ++op->do_list;
+            op->do_list = true;
             break;
         case 'm':
             n = sg_get_num(optarg);
             if ((n < 0) || (n > 0xffff)) {
-                fprintf(stderr, "bad argument to '--maxlen=' or greater "
-                        "than 65535 [0xffff]\n");
+                pr2serr("bad argument to '--maxlen=' or greater than 65535 "
+                        "[0xffff]\n");
                 return SG_LIB_SYNTAX_ERROR;
             }
             op->maxlen = n;
@@ -193,47 +204,55 @@ process_cl_new(struct opts_t * op, int argc, char * argv[])
         case 'N':
             break;      /* ignore */
         case 'O':
-            op->opt_new = 0;
+            op->opt_new = false;
             return 0;
         case 'p':
-            op->do_pf = 1;
+            op->do_pf = true;
             break;
         case 'P':
             n = sg_get_num(optarg);
             if ((n < 0) || (n > 0xff)) {
-                fprintf(stderr, "bad argument to '--page=' or greater "
-                        "than 255 [0xff]\n");
+                pr2serr("bad argument to '--page=' or greater than 255 "
+                        "[0xff]\n");
                 return SG_LIB_SYNTAX_ERROR;
             }
             op->page_code = n;
             break;
         case 'r':
             op->raw_arg = optarg;
-            op->do_raw = 1;
+            op->do_raw = true;
             break;
         case 's':
             n = sg_get_num(optarg);
             if ((n < 0) || (n > 7)) {
-                fprintf(stderr, "bad argument to '--selftest='\n");
+                pr2serr("bad argument to '--selftest='\n");
                 usage();
                 return SG_LIB_SYNTAX_ERROR;
             }
             op->do_selftest = n;
             break;
         case 't':
-            op->do_deftest = 1;
+            op->do_deftest = true;
+            break;
+        case 'T':
+            n = sg_get_num(optarg);
+            if (n < 0) {
+                pr2serr("bad argument to '--timeout=SEC'\n");
+                return SG_LIB_SYNTAX_ERROR;
+            }
+            op->timeout = n;
             break;
         case 'u':
-            op->do_uoff = 1;
+            op->do_uoff = true;
             break;
         case 'v':
             ++op->do_verbose;
             break;
         case 'V':
-            ++op->do_version;
+            op->do_version = true;
             break;
         default:
-            fprintf(stderr, "unrecognised option code %c [0x%x]\n", c, c);
+            pr2serr("unrecognised option code %c [0x%x]\n", c, c);
             if (op->do_help)
                 break;
             usage();
@@ -247,8 +266,7 @@ process_cl_new(struct opts_t * op, int argc, char * argv[])
         }
         if (optind < argc) {
             for (; optind < argc; ++optind)
-                fprintf(stderr, "Unexpected extra argument: %s\n",
-                        argv[optind]);
+                pr2serr("Unexpected extra argument: %s\n", argv[optind]);
             usage();
             return SG_LIB_SYNTAX_ERROR;
         }
@@ -257,9 +275,10 @@ process_cl_new(struct opts_t * op, int argc, char * argv[])
 }
 
 static int
-process_cl_old(struct opts_t * op, int argc, char * argv[])
+old_parse_cmd_line(struct opts_t * op, int argc, char * argv[])
 {
-    int k, jmp_out, plen, num;
+    bool jmp_out;
+    int k, plen, num, n;
     unsigned int u;
     const char * cp;
 
@@ -269,61 +288,61 @@ process_cl_old(struct opts_t * op, int argc, char * argv[])
         if (plen <= 0)
             continue;
         if ('-' == *cp) {
-            for (--plen, ++cp, jmp_out = 0; plen > 0; --plen, ++cp) {
+            for (--plen, ++cp, jmp_out = false; plen > 0; --plen, ++cp) {
                 switch (*cp) {
                 case 'd':
                     if (0 == strncmp("doff", cp, 4)) {
-                        op->do_doff = 1;
+                        op->do_doff = true;
                         cp += 3;
                         plen -= 3;
                     } else
-                        jmp_out = 1;
+                        jmp_out = true;
                     break;
                 case 'e':
-                    op->do_extdur = 1;
+                    op->do_extdur = true;
                     break;
                 case 'h':
                 case 'H':
                     ++op->do_hex;
                     break;
                 case 'l':
-                    ++op->do_list;
+                    op->do_list = true;
                     break;
                 case 'N':
-                    op->opt_new = 1;
+                    op->opt_new = true;
                     return 0;
                 case 'O':
                     break;
                 case 'p':
                     if (0 == strncmp("pf", cp, 2)) {
-                        op->do_pf = 1;
+                        op->do_pf = true;
                         ++cp;
                         --plen;
                     } else
-                        jmp_out = 1;
+                        jmp_out = true;
                     break;
                 case 't':
-                    op->do_deftest = 1;
+                    op->do_deftest = true;
                     break;
                 case 'u':
                     if (0 == strncmp("uoff", cp, 4)) {
-                        op->do_uoff = 1;
+                        op->do_uoff = true;
                         cp += 3;
                         plen -= 3;
                     } else
-                        jmp_out = 1;
+                        jmp_out = true;
                     break;
                 case 'v':
                     ++op->do_verbose;
                     break;
                 case 'V':
-                    ++op->do_version;
+                    op->do_version = true;
                     break;
                 case '?':
                     ++op->do_help;
                     break;
                 default:
-                    jmp_out = 1;
+                    jmp_out = true;
                     break;
                 }
                 if (jmp_out)
@@ -333,7 +352,7 @@ process_cl_old(struct opts_t * op, int argc, char * argv[])
                 continue;
             if (0 == strncmp("raw=", cp, 4)) {
                 op->raw_arg = cp + 4;
-                op->do_raw = 1;
+                op->do_raw = true;
             } else if (0 == strncmp("s=", cp, 2)) {
                 num = sscanf(cp + 2, "%x", &u);
                 if ((1 != num) || (u > 7)) {
@@ -342,18 +361,26 @@ process_cl_old(struct opts_t * op, int argc, char * argv[])
                     return SG_LIB_SYNTAX_ERROR;
                 }
                 op->do_selftest = u;
+            } else if (0 == strncmp("T=", cp, 2)) {
+                num = sscanf(cp + 2, "%d", &n);
+                if ((1 != num) || (n < 0)) {
+                    printf("Bad page code after '-T=SEC' option\n");
+                    usage_old();
+                    return SG_LIB_SYNTAX_ERROR;
+                }
+                op->timeout = n;
             } else if (0 == strncmp("-old", cp, 5))
                 ;
             else if (jmp_out) {
-                fprintf(stderr, "Unrecognized option: %s\n", cp);
+                pr2serr("Unrecognized option: %s\n", cp);
                 usage_old();
                 return SG_LIB_SYNTAX_ERROR;
             }
         } else if (0 == op->device_name)
             op->device_name = cp;
         else {
-            fprintf(stderr, "too many arguments, got: %s, not expecting: "
-                    "%s\n", op->device_name, cp);
+            pr2serr("too many arguments, got: %s, not expecting: %s\n",
+                    op->device_name, cp);
             usage_old();
             return SG_LIB_SYNTAX_ERROR;
         }
@@ -362,36 +389,41 @@ process_cl_old(struct opts_t * op, int argc, char * argv[])
 }
 
 static int
-process_cl(struct opts_t * op, int argc, char * argv[])
+parse_cmd_line(struct opts_t * op, int argc, char * argv[])
 {
     int res;
     char * cp;
 
     cp = getenv("SG3_UTILS_OLD_OPTS");
     if (cp) {
-        op->opt_new = 0;
-        res = process_cl_old(op, argc, argv);
+        op->opt_new = false;
+        res = old_parse_cmd_line(op, argc, argv);
         if ((0 == res) && op->opt_new)
-            res = process_cl_new(op, argc, argv);
+            res = new_parse_cmd_line(op, argc, argv);
     } else {
-        op->opt_new = 1;
-        res = process_cl_new(op, argc, argv);
-        if ((0 == res) && (0 == op->opt_new))
-            res = process_cl_old(op, argc, argv);
+        op->opt_new = true;
+        res = new_parse_cmd_line(op, argc, argv);
+        if ((0 == res) && (! op->opt_new))
+            res = old_parse_cmd_line(op, argc, argv);
     }
     return res;
 }
 
 /* Return of 0 -> success, otherwise see sg_ll_send_diag() */
 static int
-do_senddiag(int sg_fd, int sf_code, int pf_bit, int sf_bit, int devofl_bit,
-            int unitofl_bit, void * outgoing_pg, int outgoing_len,
-            int noisy, int verbose)
+do_senddiag(int sg_fd, int sf_code, bool pf_bit, bool sf_bit,
+            bool devofl_bit, bool unitofl_bit, void * outgoing_pg,
+            int outgoing_len, int tmout, bool noisy, int verbose)
 {
     int long_duration = 0;
 
-    if ((0 == sf_bit) && ((5 == sf_code) || (6 == sf_code)))
-        long_duration = 1;      /* foreground self-tests */
+    if ((0 == sf_bit) && ((5 == sf_code) || (6 == sf_code))) {
+        /* foreground self-tests */
+        if (tmout <= 0)
+            long_duration = 1;
+        else
+            long_duration = tmout;
+    }
     return sg_ll_send_diag(sg_fd, sf_code, pf_bit, sf_bit, devofl_bit,
                            unitofl_bit, long_duration, outgoing_pg,
                            outgoing_len, noisy, verbose);
@@ -399,22 +431,32 @@ do_senddiag(int sg_fd, int sf_code, int pf_bit, int sf_bit, int devofl_bit,
 
 /* Get expected extended self-test time from mode page 0xa (for '-e') */
 static int
-do_modes_0a(int sg_fd, void * resp, int mx_resp_len, int noisy, int mode6,
+do_modes_0a(int sg_fd, void * resp, int mx_resp_len, bool mode6, bool noisy,
             int verbose)
 {
     int res;
+    int resid = 0;
 
     if (mode6)
-        res = sg_ll_mode_sense6(sg_fd, 1 /* dbd */, 0 /* pc */, 0xa /* page */,
-                                0, resp, mx_resp_len, noisy, verbose);
+        res = sg_ll_mode_sense6(sg_fd, true /* dbd */, false /* pc */,
+                                0xa /* page */, false, resp, mx_resp_len,
+                                noisy, verbose);
     else
-        res = sg_ll_mode_sense10(sg_fd, 0 /* llbaa */, 1 /* dbd */, 0, 0xa, 0,
-                                 resp, mx_resp_len, noisy, verbose);
+        res = sg_ll_mode_sense10_v2(sg_fd, false /* llbaa */, true /* dbd */,
+                                    false, 0xa, false, resp, mx_resp_len,
+                                    0, &resid, noisy, verbose);
     if (res) {
         char b[80];
 
         sg_get_category_sense_str(res, sizeof(b), b, verbose);
-        fprintf(stderr, "Mode sense (%s): %s\n", (mode6 ? "6" : "10"), b);
+        pr2serr("Mode sense (%s): %s\n", (mode6 ? "6" : "10"), b);
+    } else {
+        mx_resp_len -= resid;
+        if (mx_resp_len < 4) {
+            pr2serr("%s: response length (%d) too small (resid=%d)\n",
+                    __func__, mx_resp_len, resid);
+            res = SG_LIB_WILD_RESID;
+        }
     }
     return res;
 }
@@ -423,7 +465,7 @@ do_modes_0a(int sg_fd, void * resp, int mx_resp_len, int noisy, int mode6,
 /* stdin (one per line, comma separated list or space separated list). */
 /* Returns 0 if ok, or 1 if error. */
 static int
-build_diag_page(const char * inp, unsigned char * mp_arr, int * mp_arr_len,
+build_diag_page(const char * inp, uint8_t * mp_arr, int * mp_arr_len,
                 int max_arr_len)
 {
     int in_len, k, j, m;
@@ -440,10 +482,10 @@ build_diag_page(const char * inp, unsigned char * mp_arr, int * mp_arr_len,
     if (0 == in_len)
         *mp_arr_len = 0;
     if ('-' == inp[0]) {        /* read from stdin */
+        bool split_line;
+        int off = 0;
         char line[512];
         char carry_over[4];
-        int off = 0;
-        int split_line;
 
         carry_over[0] = 0;
         for (j = 0; j < 512; ++j) {
@@ -454,9 +496,9 @@ build_diag_page(const char * inp, unsigned char * mp_arr, int * mp_arr_len,
                 if ('\n' == line[in_len - 1]) {
                     --in_len;
                     line[in_len] = '\0';
-                    split_line = 0;
+                    split_line = false;
                 } else
-                    split_line = 1;
+                    split_line = true;
             }
             if (in_len < 1) {
                 carry_over[0] = 0;
@@ -469,8 +511,8 @@ build_diag_page(const char * inp, unsigned char * mp_arr, int * mp_arr_len,
                     if (1 == sscanf(carry_over, "%x", &h))
                         mp_arr[off - 1] = h;       /* back up and overwrite */
                     else {
-                        fprintf(stderr, "build_diag_page: carry_over error "
-                                "['%s'] around line %d\n", carry_over, j + 1);
+                        pr2serr("build_diag_page: carry_over error ['%s'] "
+                                "around line %d\n", carry_over, j + 1);
                         return 1;
                     }
                     lcp = line + 1;
@@ -489,16 +531,16 @@ build_diag_page(const char * inp, unsigned char * mp_arr, int * mp_arr_len,
                 continue;
             k = strspn(lcp, "0123456789aAbBcCdDeEfF ,\t");
             if ((k < in_len) && ('#' != lcp[k])) {
-                fprintf(stderr, "build_diag_page: syntax error at "
-                        "line %d, pos %d\n", j + 1, m + k + 1);
+                pr2serr("build_diag_page: syntax error at line %d, pos %d\n",
+                        j + 1, m + k + 1);
                 return 1;
             }
             for (k = 0; k < 1024; ++k) {
                 if (1 == sscanf(lcp, "%x", &h)) {
                     if (h > 0xff) {
-                        fprintf(stderr, "build_diag_page: hex number "
-                                "larger than 0xff in line %d, pos %d\n",
-                                j + 1, (int)(lcp - line + 1));
+                        pr2serr("build_diag_page: hex number larger than "
+                                "0xff in line %d, pos %d\n", j + 1,
+                                (int)(lcp - line + 1));
                         return 1;
                     }
                     if (split_line && (1 == strlen(lcp))) {
@@ -506,8 +548,7 @@ build_diag_page(const char * inp, unsigned char * mp_arr, int * mp_arr_len,
                         carry_over[0] = *lcp;
                     }
                     if ((off + k) >= max_arr_len) {
-                        fprintf(stderr, "build_diag_page: array length "
-                                "exceeded\n");
+                        pr2serr("build_diag_page: array length exceeded\n");
                         return 1;
                     }
                     mp_arr[off + k] = h;
@@ -522,9 +563,8 @@ build_diag_page(const char * inp, unsigned char * mp_arr, int * mp_arr_len,
                         --k;
                         break;
                     }
-                    fprintf(stderr, "build_diag_page: error in "
-                            "line %d, at pos %d\n", j + 1,
-                            (int)(lcp - line + 1));
+                    pr2serr("build_diag_page: error in line %d, at pos %d\n",
+                            j + 1, (int)(lcp - line + 1));
                     return 1;
                 }
             }
@@ -534,14 +574,14 @@ build_diag_page(const char * inp, unsigned char * mp_arr, int * mp_arr_len,
     } else {        /* hex string on command line */
         k = strspn(inp, "0123456789aAbBcCdDeEfF, ");
         if (in_len != k) {
-            fprintf(stderr, "build_diag_page: error at pos %d\n", k + 1);
+            pr2serr("build_diag_page: error at pos %d\n", k + 1);
             return 1;
         }
         for (k = 0; k < max_arr_len; ++k) {
             if (1 == sscanf(lcp, "%x", &h)) {
                 if (h > 0xff) {
-                    fprintf(stderr, "build_diag_page: hex number larger "
-                            "than 0xff at pos %d\n", (int)(lcp - inp + 1));
+                    pr2serr("build_diag_page: hex number larger than 0xff at "
+                            "pos %d\n", (int)(lcp - inp + 1));
                     return 1;
                 }
                 mp_arr[k] = h;
@@ -555,14 +595,14 @@ build_diag_page(const char * inp, unsigned char * mp_arr, int * mp_arr_len,
                     cp = c2p;
                 lcp = cp + 1;
             } else {
-                fprintf(stderr, "build_diag_page: error at pos %d\n",
+                pr2serr("build_diag_page: error at pos %d\n",
                         (int)(lcp - inp + 1));
                 return 1;
             }
         }
         *mp_arr_len = k + 1;
         if (k == max_arr_len) {
-            fprintf(stderr, "build_diag_page: array length exceeded\n");
+            pr2serr("build_diag_page: array length exceeded\n");
             return 1;
         }
     }
@@ -630,20 +670,22 @@ list_page_codes()
 int
 main(int argc, char * argv[])
 {
-    int sg_fd, k, num, rsp_len, res, rsp_buff_size, pg;
+    int sg_fd, k, num, rsp_len, res, rsp_buff_size, pg, bd_len, resid, vb;
     int read_in_len = 0;
     int ret = 0;
     struct opts_t opts;
     struct opts_t * op;
-    unsigned char * rsp_buff = NULL;
+    uint8_t * rsp_buff = NULL;
+    uint8_t * free_rsp_buff = NULL;
     const char * cp;
-    unsigned char * read_in = NULL;
+    uint8_t * read_in = NULL;
+    uint8_t * free_read_in = NULL;
 
     op = &opts;
     memset(op, 0, sizeof(opts));
     op->maxlen = DEF_ALLOC_LEN;
     op->page_code = -1;
-    res = process_cl(op, argc, argv);
+    res = parse_cmd_line(op, argc, argv);
     if (res)
         return SG_LIB_SYNTAX_ERROR;
     if (op->do_help) {
@@ -654,7 +696,7 @@ main(int argc, char * argv[])
         return 0;
     }
     if (op->do_version) {
-        fprintf(stderr, "Version string: %s\n", version_str);
+        pr2serr("Version string: %s\n", version_str);
         return 0;
     }
     rsp_buff_size = op->maxlen;
@@ -664,17 +706,18 @@ main(int argc, char * argv[])
             list_page_codes();
             return 0;
         }
-        fprintf(stderr, "No DEVICE argument given\n");
+        pr2serr("No DEVICE argument given\n");
         if (op->opt_new)
             usage();
         else
             usage_old();
         return SG_LIB_SYNTAX_ERROR;
     }
+    vb = op->do_verbose;
     if (op->do_raw) {
-        read_in = (unsigned char *)calloc(op->maxlen, 1);
+        read_in = sg_memalign(op->maxlen, 0, &free_read_in, vb > 3);
         if (NULL == read_in) {
-            fprintf(stderr, "unable to allocate %d bytes\n", op->maxlen);
+            pr2serr("unable to allocate %d bytes\n", op->maxlen);
             return SG_LIB_CAT_OTHER;
         }
         if (build_diag_page(op->raw_arg, read_in, &read_in_len, op->maxlen)) {
@@ -685,7 +728,8 @@ main(int argc, char * argv[])
                 printf("Bad sequence after '-raw=' option\n");
                 usage_old();
             }
-            return SG_LIB_SYNTAX_ERROR;
+            ret = SG_LIB_SYNTAX_ERROR;
+            goto fini;
         }
     }
 
@@ -697,7 +741,8 @@ main(int argc, char * argv[])
             printf("setting -doff or -uoff only useful when -t is set\n");
             usage_old();
         }
-        return SG_LIB_SYNTAX_ERROR;
+        ret = SG_LIB_SYNTAX_ERROR;
+        goto fini;
     }
     if ((op->do_selftest > 0) && op->do_deftest) {
         if (op->opt_new) {
@@ -707,7 +752,8 @@ main(int argc, char * argv[])
             printf("either set -s=SF or -t (not both)\n");
             usage_old();
         }
-        return SG_LIB_SYNTAX_ERROR;
+        ret = SG_LIB_SYNTAX_ERROR;
+        goto fini;
     }
     if (op->do_raw) {
         if ((op->do_selftest > 0) || op->do_deftest || op->do_extdur ||
@@ -721,7 +767,8 @@ main(int argc, char * argv[])
                        "'-l'\n");
                 usage_old();
             }
-            return SG_LIB_SYNTAX_ERROR;
+            ret = SG_LIB_SYNTAX_ERROR;
+            goto fini;
         }
         if (! op->do_pf) {
             if (op->opt_new)
@@ -734,34 +781,39 @@ main(int argc, char * argv[])
     }
 #ifdef SG_LIB_WIN32
 #ifdef SG_LIB_WIN32_DIRECT
-    if (op->do_verbose > 4)
-        fprintf(stderr, "Initial win32 SPT interface state: %s\n",
+    if (vb > 4)
+        pr2serr("Initial win32 SPT interface state: %s\n",
                 scsi_pt_win32_spt_state() ? "direct" : "indirect");
     if (op->maxlen >= 16384)
         scsi_pt_win32_direct(SG_LIB_WIN32_DIRECT /* SPT pt interface */);
 #endif
 #endif
 
-    if ((sg_fd = sg_cmds_open_device(op->device_name, 0 /* rw */,
-                                     op->do_verbose)) < 0) {
-        fprintf(stderr, ME "error opening file: %s: %s\n", op->device_name,
+    if ((sg_fd = sg_cmds_open_device(op->device_name, false /* rw */, vb)) <
+         0) {
+        pr2serr(ME "error opening file: %s: %s\n", op->device_name,
                 safe_strerror(-sg_fd));
-        return SG_LIB_FILE_ERROR;
+        ret = sg_convert_errno(-sg_fd);
+        goto fini;
     }
-    rsp_buff = (unsigned char *)calloc(op->maxlen, 1);
+    rsp_buff = sg_memalign(op->maxlen, 0, &free_rsp_buff, vb > 3);
     if (NULL == rsp_buff) {
-        fprintf(stderr, "unable to allocate %d bytes (2)\n", op->maxlen);
-        return SG_LIB_CAT_OTHER;
+        pr2serr("unable to allocate %d bytes (2)\n", op->maxlen);
+        ret = SG_LIB_CAT_OTHER;
+        goto close_fini;
     }
-    if (op->do_extdur) {
-        res = do_modes_0a(sg_fd, rsp_buff, 32, 1, 0, op->do_verbose);
+    if (op->do_extdur) {  /* fetch Extended self-test time from Control
+                           * mode page with Mode Sense(10) command*/
+        res = do_modes_0a(sg_fd, rsp_buff, 32, false /* mode6 */,
+                          true /* noisy */, vb);
         if (0 == res) {
-            /* Assume mode sense(10) response without block descriptors */
-            num = sg_get_unaligned_be16(rsp_buff) - 6;
+            /* Mode sense(10) response, step over any block descriptors */
+            num = sg_msense_calc_length(rsp_buff, 32, false, &bd_len);
+            num -= (8 /* MS(10) header length */ + bd_len);
             if (num >= 0xc) {
                 int secs;
 
-                secs = sg_get_unaligned_be16(rsp_buff + 18);
+                secs = sg_get_unaligned_be16(rsp_buff + 8 + bd_len + 10);
 #ifdef SG_LIB_MINGW
                 printf("Expected extended self-test duration=%d seconds "
                        "(%g minutes)\n", secs, secs / 60.0);
@@ -776,25 +828,34 @@ main(int argc, char * argv[])
             printf("Extended self-test duration (mode page 0xa) failed\n");
             goto err_out9;
         }
-    } else if ((op->do_list) || (op->page_code >= 0x0)) {
+    } else if (op->do_list || (op->page_code >= 0x0)) {
         pg = op->page_code;
         if (pg < 0)
-            res = do_senddiag(sg_fd, 0, 1 /* pf */, 0, 0, 0, rsp_buff, 4, 1,
-                              op->do_verbose);
+            res = do_senddiag(sg_fd, 0, true /* pf */, false, false, false,
+                              rsp_buff, 4, op->timeout, 1, vb);
         else
             res = 0;
         if (0 == res) {
-            if (0 == sg_ll_receive_diag(sg_fd, (pg >= 0x0),
-                                        ((pg >= 0x0) ? pg : 0), rsp_buff,
-                                        rsp_buff_size, 1, op->do_verbose)) {
+            resid = 0;
+            if (0 == sg_ll_receive_diag_v2(sg_fd, (pg >= 0x0),
+                                           ((pg >= 0x0) ? pg : 0), rsp_buff,
+                                           rsp_buff_size, 0, &resid,
+                                           true, vb)) {
+                rsp_buff_size -= resid;
+                if (rsp_buff_size < 4) {
+                    pr2serr("RD resid (%d) indicates response too small "
+                            "(lem=%d)\n", resid, rsp_buff_size);
+                    goto err_out;
+                }
                 rsp_len = sg_get_unaligned_be16(rsp_buff + 2) + 4;
+                rsp_len= (rsp_len < rsp_buff_size) ? rsp_len : rsp_buff_size;
                 if (op->do_hex > 1)
-                    dStrHex((const char *)rsp_buff, rsp_len,
+                    hex2stdout(rsp_buff, rsp_len,
                             (2 == op->do_hex) ? 0 : -1);
                 else if (pg < 0x1) {
                     printf("Supported diagnostic pages response:\n");
                     if (op->do_hex)
-                        dStrHex((const char *)rsp_buff, rsp_len, 1);
+                        hex2stdout(rsp_buff, rsp_len, 1);
                     else {
                         for (k = 0; k < (rsp_len - 4); ++k) {
                             cp = find_page_code_desc(rsp_buff[k + 4]);
@@ -809,12 +870,11 @@ main(int argc, char * argv[])
                                "hex:\n", cp, pg);
                     else
                         printf("diagnostic page 0x%x response in hex:\n", pg);
-                    dStrHex((const char *)rsp_buff, rsp_len, 1);
+                    hex2stdout(rsp_buff, rsp_len, 1);
                 }
             } else {
                 ret = res;
-                fprintf(stderr, "RECEIVE DIAGNOSTIC RESULTS command "
-                        "failed\n");
+                pr2serr("RECEIVE DIAGNOSTIC RESULTS command failed\n");
                 goto err_out9;
             }
         } else {
@@ -822,16 +882,16 @@ main(int argc, char * argv[])
             goto err_out;
         }
     } else if (op->do_raw) {
-        res = do_senddiag(sg_fd, 0, op->do_pf, 0, 0, 0, read_in,
-                          read_in_len, 1, op->do_verbose);
+        res = do_senddiag(sg_fd, 0, op->do_pf, false, false, false, read_in,
+                          read_in_len, op->timeout, 1, vb);
         if (res) {
             ret = res;
             goto err_out;
         }
     } else {
         res = do_senddiag(sg_fd, op->do_selftest, op->do_pf, op->do_deftest,
-                          op->do_doff, op->do_uoff, NULL, 0, 1,
-                          op->do_verbose);
+                          op->do_doff, op->do_uoff, NULL, 0, op->timeout, 1,
+                          vb);
         if (0 == res) {
             if ((5 == op->do_selftest) || (6 == op->do_selftest))
                 printf("Foreground self-test returned GOOD status\n");
@@ -842,30 +902,28 @@ main(int argc, char * argv[])
             goto err_out;
         }
     }
-    res = sg_cmds_close_device(sg_fd);
-    if ((res < 0) && (0 == ret))
-        return SG_LIB_SYNTAX_ERROR;
-    return (ret >= 0) ? ret : SG_LIB_CAT_OTHER;
+    goto close_fini;
 
 err_out:
     if (SG_LIB_CAT_UNIT_ATTENTION == res)
-        fprintf(stderr, "SEND DIAGNOSTIC, unit attention\n");
+        pr2serr("SEND DIAGNOSTIC, unit attention\n");
     else if (SG_LIB_CAT_ABORTED_COMMAND == res)
-        fprintf(stderr, "SEND DIAGNOSTIC, aborted command\n");
+        pr2serr("SEND DIAGNOSTIC, aborted command\n");
     else if (SG_LIB_CAT_NOT_READY == res)
-        fprintf(stderr, "SEND DIAGNOSTIC, device not "
-                "ready\n");
+        pr2serr("SEND DIAGNOSTIC, device not ready\n");
     else
-        fprintf(stderr, "SEND DIAGNOSTIC command, failed\n");
+        pr2serr("SEND DIAGNOSTIC command, failed\n");
 err_out9:
-    if (op->do_verbose < 2)
-        fprintf(stderr, "  try again with '-vv' for more information\n");
+    if (vb < 2)
+        pr2serr("  try again with '-vv' for more information\n");
+close_fini:
     res = sg_cmds_close_device(sg_fd);
     if ((res < 0) && (0 == ret))
-        return SG_LIB_FILE_ERROR;
-    if (read_in)
-        free(read_in);
-    if (rsp_buff)
-        free(rsp_buff);
+        ret = SG_LIB_FILE_ERROR;
+fini:
+    if (free_read_in)
+        free(free_read_in);
+    if (free_rsp_buff)
+        free(free_rsp_buff);
     return (ret >= 0) ? ret : SG_LIB_CAT_OTHER;
 }
