@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 1999-2013 D. Gilbert
+ *  Copyright (C) 1999-2018 D. Gilbert
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
  *  the Free Software Foundation; either version 2, or (at your option)
@@ -19,6 +19,8 @@
 
 #include <stdio.h>
 #include <stdlib.h>
+#include <stdarg.h>
+#include <stdbool.h>
 #include <unistd.h>
 #include <string.h>
 #include <fcntl.h>
@@ -29,53 +31,54 @@
 #endif
 #include "sg_lib.h"
 #include "sg_cmds_basic.h"
+#include "sg_pr2serr.h"
 
 
-static const char * version_str = "0.59 20130507";  /* sbc3r14; mmc6r01a */
+static const char * version_str = "0.64 20180302";  /* sbc3r14; mmc6r01a */
 
 static struct option long_options[] = {
-        {"eject", 0, 0, 'e'},
-        {"fl", 1, 0, 'f'},
-        {"help", 0, 0, 'h'},
-        {"immed", 0, 0, 'i'},
-        {"load", 0, 0, 'l'},
-        {"loej", 0, 0, 'L'},
-        {"mod", 1, 0, 'm'},
-        {"noflush", 0, 0, 'n'},
-        {"new", 0, 0, 'N'},
-        {"old", 0, 0, 'O'},
-        {"pc", 1, 0, 'p'},
-        {"readonly", 0, 0, 'r'},
-        {"start", 0, 0, 's'},
-        {"stop", 0, 0, 'S'},
-        {"verbose", 0, 0, 'v'},
-        {"version", 0, 0, 'V'},
+        {"eject", no_argument, 0, 'e'},
+        {"fl", required_argument, 0, 'f'},
+        {"help", no_argument, 0, 'h'},
+        {"immed", no_argument, 0, 'i'},
+        {"load", no_argument, 0, 'l'},
+        {"loej", no_argument, 0, 'L'},
+        {"mod", required_argument, 0, 'm'},
+        {"noflush", no_argument, 0, 'n'},
+        {"new", no_argument, 0, 'N'},
+        {"old", no_argument, 0, 'O'},
+        {"pc", required_argument, 0, 'p'},
+        {"readonly", no_argument, 0, 'r'},
+        {"start", no_argument, 0, 's'},
+        {"stop", no_argument, 0, 'S'},
+        {"verbose", no_argument, 0, 'v'},
+        {"version", no_argument, 0, 'V'},
         {0, 0, 0, 0},
 };
 
 struct opts_t {
-    int do_eject;
+    bool do_eject;
+    bool do_immed;
+    bool do_load;
+    bool do_loej;
+    bool do_noflush;
+    bool do_readonly;
+    bool do_start;
+    bool do_stop;
+    bool do_version;
+    bool opt_new;
     int do_fl;
     int do_help;
-    int do_immed;
-    int do_load;
-    int do_loej;
     int do_mod;
-    int do_noflush;
-    int do_readonly;
     int do_pc;
-    int do_start;
-    int do_stop;
     int do_verbose;
-    int do_version;
     const char * device_name;
-    int opt_new;
 };
 
 static void
 usage()
 {
-    fprintf(stderr, "Usage: sg_start [--eject] [--fl=FL] [--help] "
+    pr2serr("Usage: sg_start [--eject] [--fl=FL] [--help] "
             "[--immed] [--load] [--loej]\n"
             "                [--mod=PC_MOD] [--noflush] [--pc=PC] "
             "[--readonly]\n"
@@ -110,6 +113,7 @@ usage()
             "given\n"
             "    --stop|-S       stop unit (e.g. spin down disk)\n"
             "    --verbose|-v    increase verbosity\n"
+            "    --old|-O        use old interface (use as first option)\n"
             "    --version|-V    print version string then exit\n\n"
             "    Example: 'sg_start --stop /dev/sdb'    stops unit\n"
             "             'sg_start --eject /dev/scd0'  stops unit and "
@@ -121,7 +125,7 @@ usage()
 static void
 usage_old()
 {
-    fprintf(stderr, "Usage:  sg_start [0] [1] [--eject] [--fl=FL] "
+    pr2serr("Usage:  sg_start [0] [1] [--eject] [--fl=FL] "
             "[-i] [--imm=0|1]\n"
             "                 [--load] [--loej] [--mod=PC_MOD] "
             "[--noflush] [--pc=PC]\n"
@@ -155,6 +159,7 @@ usage_old()
             "action\n"
             "    --stop     stop unit (same as '0')\n"
             "    -v         verbose (print out SCSI commands)\n"
+            "    -N|--new   use new interface\n"
             "    -V         print version string then exit\n\n"
             "    Example: 'sg_start --stop /dev/sdb'    stops unit\n"
             "             'sg_start --eject /dev/scd0'  stops unit and "
@@ -164,7 +169,7 @@ usage_old()
 }
 
 static int
-process_cl_new(struct opts_t * optsp, int argc, char * argv[])
+news_parse_cmd_line(struct opts_t * op, int argc, char * argv[])
 {
     int c, n, err;
 
@@ -178,78 +183,78 @@ process_cl_new(struct opts_t * optsp, int argc, char * argv[])
 
         switch (c) {
         case 'e':
-            ++optsp->do_eject;
-            ++optsp->do_loej;
+            op->do_eject = true;
+            op->do_loej = true;
             break;
         case 'f':
             n = sg_get_num(optarg);
             if ((n < 0) || (n > 3)) {
-                fprintf(stderr, "bad argument to '--fl='\n");
+                pr2serr("bad argument to '--fl='\n");
                 usage();
                 return SG_LIB_SYNTAX_ERROR;
             }
-            ++optsp->do_loej;
-            ++optsp->do_start;
-            optsp->do_fl = n;
+            op->do_loej = true;
+            op->do_start = true;
+            op->do_fl = n;
             break;
         case 'h':
         case '?':
-            ++optsp->do_help;
+            ++op->do_help;
             break;
         case 'i':
-            ++optsp->do_immed;
+            op->do_immed = true;
             break;
         case 'l':
-            ++optsp->do_load;
-            ++optsp->do_loej;
+            op->do_load = true;
+            op->do_loej = true;
             break;
         case 'L':
-            ++optsp->do_loej;
+            op->do_loej = true;
             break;
         case 'm':
             n = sg_get_num(optarg);
             if ((n < 0) || (n > 15)) {
-                fprintf(stderr, "bad argument to '--mod='\n");
+                pr2serr("bad argument to '--mod='\n");
                 usage();
                 return SG_LIB_SYNTAX_ERROR;
             }
-            optsp->do_mod = n;
+            op->do_mod = n;
             break;
         case 'n':
-            ++optsp->do_noflush;
+            op->do_noflush = true;
             break;
         case 'N':
             break;      /* ignore */
         case 'O':
-            optsp->opt_new = 0;
+            op->opt_new = false;
             return 0;
         case 'p':
             n = sg_get_num(optarg);
             if ((n < 0) || (n > 15)) {
-                fprintf(stderr, "bad argument to '--pc='\n");
+                pr2serr("bad argument to '--pc='\n");
                 usage();
                 return SG_LIB_SYNTAX_ERROR;
             }
-            optsp->do_pc = n;
+            op->do_pc = n;
             break;
         case 'r':
-            ++optsp->do_readonly;
+            op->do_readonly = true;
             break;
         case 's':
-            ++optsp->do_start;
+            op->do_start = true;
             break;
         case 'S':
-            ++optsp->do_stop;
+            op->do_stop = true;
             break;
         case 'v':
-            ++optsp->do_verbose;
+            ++op->do_verbose;
             break;
         case 'V':
-            ++optsp->do_version;
+            op->do_version = true;
             break;
         default:
-            fprintf(stderr, "unrecognised option code %c [0x%x]\n", c, c);
-            if (optsp->do_help)
+            pr2serr("unrecognised option code %c [0x%x]\n", c, c);
+            if (op->do_help)
                 break;
             usage();
             return SG_LIB_SYNTAX_ERROR;
@@ -259,17 +264,17 @@ process_cl_new(struct opts_t * optsp, int argc, char * argv[])
     for (; optind < argc; ++optind) {
         if (1 == strlen(argv[optind])) {
             if (0 == strcmp("0", argv[optind])) {
-                ++optsp->do_stop;
+                op->do_stop = true;
                 continue;
             } else if (0 == strcmp("1", argv[optind])) {
-                ++optsp->do_start;
+                op->do_start = true;
                 continue;
             }
         }
-        if (NULL == optsp->device_name)
-            optsp->device_name = argv[optind];
+        if (NULL == op->device_name)
+            op->device_name = argv[optind];
         else {
-            fprintf(stderr, "Unexpected extra argument: %s\n", argv[optind]);
+            pr2serr("Unexpected extra argument: %s\n", argv[optind]);
             ++err;
         }
     }
@@ -281,11 +286,13 @@ process_cl_new(struct opts_t * optsp, int argc, char * argv[])
 }
 
 static int
-process_cl_old(struct opts_t * optsp, int argc, char * argv[])
+old_parse_cmd_line(struct opts_t * op, int argc, char * argv[])
 {
-    int k, jmp_out, plen, num;
-    int ambigu = 0;
-    int startstop = -1;
+    bool ambigu = false;
+    bool jmp_out;
+    bool startstop = false;
+    bool startstop_set = false;
+    int k, plen, num;
     unsigned int u;
     const char * cp;
 
@@ -295,40 +302,40 @@ process_cl_old(struct opts_t * optsp, int argc, char * argv[])
         if (plen <= 0)
             continue;
         if ('-' == *cp) {
-            for (--plen, ++cp, jmp_out = 0; plen > 0;
+            for (--plen, ++cp, jmp_out = false; plen > 0;
                  --plen, ++cp) {
                 switch (*cp) {
                 case 'i':
                     if ('\0' == *(cp + 1))
-                        optsp->do_immed = 1;
+                        op->do_immed = true;
                     else
-                        jmp_out = 1;
+                        jmp_out = true;
                     break;
                 case 'r':
-                    ++optsp->do_readonly;
+                    op->do_readonly = true;
                     break;
                 case 'v':
-                    ++optsp->do_verbose;
+                    ++op->do_verbose;
                     break;
                 case 'V':
-                    ++optsp->do_version;
+                    op->do_version = true;
                     break;
                 case 'h':
                 case '?':
-                    ++optsp->do_help;
+                    ++op->do_help;
                     break;
                 case 'N':
-                    optsp->opt_new = 1;
+                    op->opt_new = true;
                     return 0;
                 case 'O':
                     break;
                 case '-':
                     ++cp;
                     --plen;
-                    jmp_out = 1;
+                    jmp_out = true;
                     break;
                 default:
-                    jmp_out = 1;
+                    jmp_out = true;
                     break;
                 }
                 if (jmp_out)
@@ -338,120 +345,135 @@ process_cl_old(struct opts_t * optsp, int argc, char * argv[])
                 continue;
 
             if (0 == strncmp(cp, "eject", 5)) {
-                optsp->do_loej = 1;
-                if (startstop == 1)
-                    ambigu = 1;
-                else
-                    startstop = 0;
+                op->do_loej = true;
+                if (startstop_set && startstop)
+                    ambigu = true;
+                else {
+                    startstop = false;
+                    startstop_set = true;
+                }
             } else if (0 == strncmp("fl=", cp, 3)) {
                 num = sscanf(cp + 3, "%x", &u);
                 if (1 != num) {
-                    fprintf(stderr, "Bad value after 'fl=' option\n");
+                    pr2serr("Bad value after 'fl=' option\n");
                     usage_old();
                     return SG_LIB_SYNTAX_ERROR;
                 }
-                startstop = 1;
-                optsp->do_loej = 1;
-                optsp->do_fl = u;
+                startstop = true;
+                startstop_set = true;
+                op->do_loej = true;
+                op->do_fl = u;
             } else if (0 == strncmp("imm=", cp, 4)) {
                 num = sscanf(cp + 4, "%x", &u);
                 if ((1 != num) || (u > 1)) {
-                    fprintf(stderr, "Bad value after 'imm=' option\n");
+                    pr2serr("Bad value after 'imm=' option\n");
                     usage_old();
                     return SG_LIB_SYNTAX_ERROR;
                 }
-                optsp->do_immed = u;
+                op->do_immed = !! u;
             } else if (0 == strncmp(cp, "load", 4)) {
-                optsp->do_loej = 1;
-                if (startstop == 0)
-                    ambigu = 1;
-                else
-                    startstop = 1;
+                op->do_loej = true;
+                if (startstop_set && (! startstop))
+                    ambigu = true;
+                else {
+                    startstop = true;
+                    startstop_set = true;
+                }
             } else if (0 == strncmp(cp, "loej", 4))
-                optsp->do_loej = 1;
+                op->do_loej = true;
             else if (0 == strncmp("pc=", cp, 3)) {
                 num = sscanf(cp + 3, "%x", &u);
                 if ((1 != num) || (u > 15)) {
-                    fprintf(stderr, "Bad value after after 'pc=' option\n");
+                    pr2serr("Bad value after after 'pc=' option\n");
                     usage_old();
                     return SG_LIB_SYNTAX_ERROR;
                 }
-                optsp->do_pc = u;
+                op->do_pc = u;
             } else if (0 == strncmp("mod=", cp, 4)) {
                 num = sscanf(cp + 3, "%x", &u);
                 if (1 != num) {
-                    fprintf(stderr, "Bad value after 'mod=' option\n");
+                    pr2serr("Bad value after 'mod=' option\n");
                     usage_old();
                     return SG_LIB_SYNTAX_ERROR;
                 }
-                optsp->do_mod = u;
+                op->do_mod = u;
             } else if (0 == strncmp(cp, "noflush", 7)) {
-                optsp->do_noflush = 1;
+                op->do_noflush = true;
             } else if (0 == strncmp(cp, "start", 5)) {
-                if (startstop == 0)
-                    ambigu = 1;
-                else
-                    startstop = 1;
+                if (startstop_set && (! startstop))
+                    ambigu = true;
+                else {
+                    startstop = true;
+                    startstop_set = true;
+                }
             } else if (0 == strncmp(cp, "stop", 4)) {
-                if (startstop == 1)
-                    ambigu = 1;
-                else
-                    startstop = 0;
+                if (startstop_set && startstop)
+                    ambigu = true;
+                else {
+                    startstop = false;
+                    startstop_set = true;
+                }
             } else if (0 == strncmp(cp, "old", 3))
                 ;
             else if (jmp_out) {
-                fprintf(stderr, "Unrecognized option: %s\n", cp);
+                pr2serr("Unrecognized option: %s\n", cp);
                 usage_old();
                 return SG_LIB_SYNTAX_ERROR;
             }
         } else if (0 == strcmp("0", cp)) {
-            if (1 == startstop)
-                ambigu = 1;
-            else
-                startstop = 0;
+            if (startstop_set && startstop)
+                ambigu = true;
+            else {
+                startstop = false;
+                startstop_set = true;
+            }
         } else if (0 == strcmp("1", cp)) {
-            if (0 == startstop)
-                ambigu = 1;
-            else
-                startstop = 1;
-        } else if (0 == optsp->device_name)
-                optsp->device_name = cp;
+            if (startstop_set && (! startstop))
+                ambigu = true;
+            else {
+                startstop = true;
+                startstop_set = true;
+            }
+        } else if (0 == op->device_name)
+                op->device_name = cp;
         else {
-            fprintf(stderr, "too many arguments, got: %s, not "
-                    "expecting: %s\n", optsp->device_name, cp);
+            pr2serr("too many arguments, got: %s, not "
+                    "expecting: %s\n", op->device_name, cp);
             usage_old();
             return SG_LIB_SYNTAX_ERROR;
         }
         if (ambigu) {
-            fprintf(stderr, "please, only one of 0, 1, --eject, "
+            pr2serr("please, only one of 0, 1, --eject, "
                     "--load, --start or --stop\n");
             usage_old();
             return SG_LIB_SYNTAX_ERROR;
-        } else if (0 == startstop)
-            ++optsp->do_stop;
-        else if (1 == startstop)
-            ++optsp->do_start;
+        } else if (startstop_set) {
+            if (startstop)
+                op->do_start = true;
+            else
+                op->do_stop = true;
+        }
     }
     return 0;
 }
 
 static int
-process_cl(struct opts_t * optsp, int argc, char * argv[])
+parse_cmd_line(struct opts_t * op, int argc, char * argv[])
 {
     int res;
     char * cp;
 
     cp = getenv("SG3_UTILS_OLD_OPTS");
     if (cp) {
-        optsp->opt_new = 0;
-        res = process_cl_old(optsp, argc, argv);
-        if ((0 == res) && optsp->opt_new)
-            res = process_cl_new(optsp, argc, argv);
+        op->opt_new = false;
+        res = old_parse_cmd_line(op, argc, argv);
+        if ((0 == res) && op->opt_new)
+            res = news_parse_cmd_line(op, argc, argv);
     } else {
-        optsp->opt_new = 1;
-        res = process_cl_new(optsp, argc, argv);
-        if ((0 == res) && (0 == optsp->opt_new))
-            res = process_cl_old(optsp, argc, argv);
+        op->opt_new = true;
+        res = news_parse_cmd_line(op, argc, argv);
+        if ((0 == res) && (! op->opt_new))
+            res = old_parse_cmd_line(op, argc, argv);
     }
     return res;
 }
@@ -463,101 +485,96 @@ main(int argc, char * argv[])
     int fd, res;
     int ret = 0;
     struct opts_t opts;
+    struct opts_t * op;
 
-    memset(&opts, 0, sizeof(opts));
-    opts.do_fl = -1;    /* only when >= 0 set FL bit */
-    res = process_cl(&opts, argc, argv);
+    op = &opts;
+    memset(op, 0, sizeof(opts));
+    op->do_fl = -1;    /* only when >= 0 set FL bit */
+    res = parse_cmd_line(op, argc, argv);
     if (res)
         return SG_LIB_SYNTAX_ERROR;
-    if (opts.do_help) {
-        if (opts.opt_new)
+    if (op->do_help) {
+        if (op->opt_new)
             usage();
         else
             usage_old();
         return 0;
     }
-    if (opts.do_version) {
-        fprintf(stderr, "Version string: %s\n", version_str);
+    if (op->do_version) {
+        pr2serr("Version string: %s\n", version_str);
         return 0;
     }
 
-    if (opts.do_start && opts.do_stop) {
-        fprintf(stderr, "Ambiguous to give both '--start' and '--stop'\n");
+    if (op->do_start && op->do_stop) {
+        pr2serr("Ambiguous to give both '--start' and '--stop'\n");
         return SG_LIB_SYNTAX_ERROR;
     }
-    if (opts.do_load && opts.do_eject) {
-        fprintf(stderr, "Ambiguous to give both '--load' and '--eject'\n");
+    if (op->do_load && op->do_eject) {
+        pr2serr("Ambiguous to give both '--load' and '--eject'\n");
         return SG_LIB_SYNTAX_ERROR;
     }
-    if (opts.do_load)
-       opts.do_start = 1;
-    else if ((opts.do_eject) || (opts.do_stop))
-       opts.do_start = 0;
-    else if (opts.opt_new && opts.do_loej && (0 == opts.do_start))
-        opts.do_start = 1;      /* --loej alone in new interface is load */
-    else if ((0 == opts.do_loej) && (-1 == opts.do_fl) && (0 == opts.do_pc))
-       opts.do_start = 1;
+    if (op->do_load)
+       op->do_start = true;
+    else if ((op->do_eject) || op->do_stop)
+       op->do_start = false;
+    else if (op->opt_new && op->do_loej && (! op->do_start))
+        op->do_start = true;      /* --loej alone in new interface is load */
+    else if ((! op->do_loej) && (-1 == op->do_fl) && (0 == op->do_pc))
+       op->do_start = true;
     /* default action is to start when no other active options */
 
-    if (0 == opts.device_name) {
-        fprintf(stderr, "No DEVICE argument given\n");
-        if (opts.opt_new)
+    if (0 == op->device_name) {
+        pr2serr("No DEVICE argument given\n");
+        if (op->opt_new)
             usage();
         else
             usage_old();
         return SG_LIB_SYNTAX_ERROR;
     }
 
-    if (opts.do_fl >= 0) {
-        if (opts.do_start == 0) {
-            fprintf(stderr, "Giving '--fl=FL' with '--stop' (or "
-                    "'--eject') is invalid\n");
+    if (op->do_fl >= 0) {
+        if (! op->do_start) {
+            pr2serr("Giving '--fl=FL' with '--stop' (or '--eject') is "
+                    "invalid\n");
             return SG_LIB_SYNTAX_ERROR;
         }
-        if (opts.do_pc > 0) {
-            fprintf(stderr, "Giving '--fl=FL' with '--pc=PC' "
-                    "when PC is non-zero is invalid\n");
+        if (op->do_pc > 0) {
+            pr2serr("Giving '--fl=FL' with '--pc=PC' when PC is non-zero "
+                    "is invalid\n");
             return SG_LIB_SYNTAX_ERROR;
         }
     }
 
-    fd = sg_cmds_open_device(opts.device_name,  opts.do_readonly /* rw */,
-                             opts.do_verbose);
+    fd = sg_cmds_open_device(op->device_name, op->do_readonly,
+                             op->do_verbose);
     if (fd < 0) {
-        fprintf(stderr, "Error trying to open %s: %s\n",
-                opts.device_name, safe_strerror(-fd));
+        pr2serr("Error trying to open %s: %s\n", op->device_name,
+                safe_strerror(-fd));
         return SG_LIB_FILE_ERROR;
     }
 
-    res = 0;
-    if (opts.do_fl >= 0)
-        res = sg_ll_start_stop_unit(fd, opts.do_immed, opts.do_fl, 0 /* pc */,
-                                    1 /* fl */, 1 /* loej */,
-                                    1 /*start */, 1 /* noisy */,
-                                    opts.do_verbose);
-    else if (opts.do_pc > 0)
-        res = sg_ll_start_stop_unit(fd, opts.do_immed, opts.do_mod,
-                                    opts.do_pc, opts.do_noflush, 0, 0, 1,
-                                    opts.do_verbose);
+    if (op->do_fl >= 0)
+        res = sg_ll_start_stop_unit(fd, op->do_immed, op->do_fl, 0 /* pc */,
+                                    true /* fl */, true /* loej */,
+                                    true /*start */, true /* noisy */,
+                                    op->do_verbose);
+    else if (op->do_pc > 0)
+        res = sg_ll_start_stop_unit(fd, op->do_immed, op->do_mod,
+                                    op->do_pc, op->do_noflush, false, false,
+                                    true, op->do_verbose);
     else
-        res = sg_ll_start_stop_unit(fd, opts.do_immed, 0, 0, opts.do_noflush,
-                                    opts.do_loej, opts.do_start, 1,
-                                    opts.do_verbose);
+        res = sg_ll_start_stop_unit(fd, op->do_immed, 0, false,
+                                    op->do_noflush, op->do_loej,
+                                    op->do_start, true, op->do_verbose);
     ret = res;
     if (res) {
-        if (opts.do_verbose < 2) {
-            if (SG_LIB_CAT_INVALID_OP == res)
-                fprintf(stderr, "command not supported\n");
-            else if (SG_LIB_CAT_NOT_READY == res)
-                fprintf(stderr, "device not ready\n");
-            else if (SG_LIB_CAT_UNIT_ATTENTION == res)
-                fprintf(stderr, "unit attention\n");
-            else if (SG_LIB_CAT_ABORTED_COMMAND == res)
-                fprintf(stderr, "aborted command\n");
-            else if (SG_LIB_CAT_ILLEGAL_REQ == res)
-                fprintf(stderr, "invalid field in cdb\n");
+        if (op->do_verbose < 2) {
+            char b[80];
+
+            sg_get_category_sense_str(res, sizeof(b), b, op->do_verbose);
+            pr2serr("%s\n", b);
         }
-        fprintf(stderr, "START STOP UNIT command failed\n");
+        pr2serr("START STOP UNIT command failed\n");
     }
     res = sg_cmds_close_device(fd);
     if ((res < 0) && (0 == ret))
