@@ -1,15 +1,15 @@
-/* sg_test_rwbuf.c */
 /*
- * Program to test the SCSI host adapter by issueing
+ * (c) 2000 Kurt Garloff <garloff at suse dot de>
+ * heavily based on Douglas Gilbert's sg_rbuf program.
+ * (c) 1999-2018 Douglas Gilbert
+ *
+ * Program to test the SCSI host adapter by issuing
  * write and read operations on a device's buffer
  * and calculating checksums.
  * NOTE: If you can not reserve the buffer of the device
  * for this purpose (SG_GET_RESERVED_SIZE), you risk
  * serious data corruption, if the device is accessed by
  * somebody else in the meantime.
- * (c) 2000 Kurt Garloff <garloff at suse dot de>
- * heavily based on Douglas Gilbert's sg_rbuf program.
- * (c) 1999-2008 Douglas Gilbert
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -26,6 +26,8 @@
 #include <fcntl.h>
 #include <stdio.h>
 #include <stdlib.h>
+#include <stdarg.h>
+#include <stdbool.h>
 #include <string.h>
 #include <getopt.h>
 #include <sys/ioctl.h>
@@ -35,11 +37,14 @@
 #ifdef HAVE_CONFIG_H
 #include "config.h"
 #endif
+
 #include "sg_lib.h"
 #include "sg_io_linux.h"
+#include "sg_unaligned.h"
+#include "sg_pr2serr.h"
 
 
-static char * version_str = "1.07 20100423";
+static const char * version_str = "1.15 20180219";
 
 #define BPI (signed)(sizeof(int))
 
@@ -58,54 +63,55 @@ static char * version_str = "1.07 20100423";
 static int base = 0x12345678;
 static int buf_capacity = 0;
 static int buf_granul = 255;
-static unsigned char *cmpbuf = NULL;
+static uint8_t *cmpbuf = NULL;
 
 
 /* Options */
 static int size = -1;
-static char do_quick = 0;
+static bool do_quick = false;
 static int addwrite  = 0;
 static int addread   = 0;
 static int verbose   = 0;
 
 static struct option long_options[] = {
-        {"help", 0, 0, 'h'},
-        {"quick", 0, 0, 'q'},
-        {"addrd", 1, 0, 'r'},
-        {"size", 1, 0, 's'},
-        {"times", 1, 0, 't'},
-        {"verbose", 0, 0, 'v'},
-        {"version", 0, 0, 'V'},
-        {"addwr", 1, 0, 'w'},
+        {"help", no_argument, 0, 'h'},
+        {"quick", no_argument, 0, 'q'},
+        {"addrd", required_argument, 0, 'r'},
+        {"size", required_argument, 0, 's'},
+        {"times", required_argument, 0, 't'},
+        {"verbose", no_argument, 0, 'v'},
+        {"version", no_argument, 0, 'V'},
+        {"addwr", required_argument, 0, 'w'},
         {0, 0, 0, 0},
 };
 
-int find_out_about_buffer (int sg_fd)
+static int
+find_out_about_buffer(int sg_fd)
 {
-        unsigned char rbCmdBlk[] = {READ_BUFFER, 0, 0, 0, 0, 0, 0, 0, 0, 0};
-        unsigned char rbBuff[RB_DESC_LEN];
-        unsigned char sense_buffer[32];
+        uint8_t rb_cdb[] = {READ_BUFFER, 0, 0, 0, 0, 0, 0, 0, 0, 0};
+        uint8_t rbBuff[RB_DESC_LEN];
+        uint8_t sense_buffer[32];
         struct sg_io_hdr io_hdr;
         int k, res;
 
-        rbCmdBlk[1] = RB_MODE_DESC;
-        rbCmdBlk[8] = RB_DESC_LEN;
+        rb_cdb[1] = RB_MODE_DESC;
+        rb_cdb[8] = RB_DESC_LEN;
         memset(&io_hdr, 0, sizeof(struct sg_io_hdr));
         io_hdr.interface_id = 'S';
-        io_hdr.cmd_len = sizeof(rbCmdBlk);
+        io_hdr.cmd_len = sizeof(rb_cdb);
         io_hdr.mx_sb_len = sizeof(sense_buffer);
         io_hdr.dxfer_direction = SG_DXFER_FROM_DEV;
         io_hdr.dxfer_len = RB_DESC_LEN;
         io_hdr.dxferp = rbBuff;
-        io_hdr.cmdp = rbCmdBlk;
+        io_hdr.cmdp = rb_cdb;
         io_hdr.sbp = sense_buffer;
         io_hdr.timeout = 60000;     /* 60000 millisecs == 60 seconds */
 
         if (verbose) {
-                fprintf(stderr, "    read buffer [mode desc] cdb: ");
-                for (k = 0; k < (int)sizeof(rbCmdBlk); ++k)
-                        fprintf(stderr, "%02x ", rbCmdBlk[k]);
-                fprintf(stderr, "\n");
+                pr2serr("    read buffer [mode desc] cdb: ");
+                for (k = 0; k < (int)sizeof(rb_cdb); ++k)
+                        pr2serr("%02x ", rb_cdb[k]);
+                pr2serr("\n");
         }
         if (ioctl(sg_fd, SG_IO, &io_hdr) < 0) {
                 perror(ME "SG_IO READ BUFFER descriptor error");
@@ -116,17 +122,23 @@ int find_out_about_buffer (int sg_fd)
         switch (res) {
         case SG_LIB_CAT_RECOVERED:
                 sg_chk_n_print3("READ BUFFER descriptor, continuing",
-                                &io_hdr, 1);
-                /* fall through */
+                                &io_hdr, true);
+#if defined(__GNUC__)
+#if (__GNUC__ >= 7)
+                __attribute__((fallthrough));
+                /* FALL THROUGH */
+#endif
+#endif
         case SG_LIB_CAT_CLEAN:
                 break;
         default: /* won't bother decoding other categories */
-                sg_chk_n_print3("READ BUFFER descriptor error", &io_hdr, 1);
+                sg_chk_n_print3("READ BUFFER descriptor error", &io_hdr,
+                                true);
                 return res;
         }
 
-        buf_capacity = ((rbBuff[1] << 16) | (rbBuff[2] << 8) | rbBuff[3]);
-        buf_granul = (unsigned char)rbBuff[0];
+        buf_capacity = sg_get_unaligned_be24(rbBuff + 1);
+        buf_granul = (uint8_t)rbBuff[0];
 #if 0
         printf("READ BUFFER reports: %02x %02x %02x %02x\n",
                rbBuff[0], rbBuff[1], rbBuff[2], rbBuff[3]);
@@ -137,7 +149,8 @@ int find_out_about_buffer (int sg_fd)
         return 0;
 }
 
-int mymemcmp (unsigned char *bf1, unsigned char *bf2, int len)
+static int
+mymemcmp (uint8_t *bf1, uint8_t *bf2, int len)
 {
         int df;
         for (df = 0; df < len; df++)
@@ -146,7 +159,8 @@ int mymemcmp (unsigned char *bf1, unsigned char *bf2, int len)
 }
 
 /* return 0 if good, else 2222 */
-int do_checksum (int *buf, int len, int quiet)
+static int
+do_checksum(int *buf, int len, bool quiet)
 {
         int sum = base;
         int i; int rln = len;
@@ -157,7 +171,7 @@ int do_checksum (int *buf, int len, int quiet)
                 if (!quiet) printf ("sg_test_rwbuf: Checksum error (sz=%i):"
                                     " %08x\n", len, sum);
                 if (cmpbuf && !quiet) {
-                        int diff = mymemcmp (cmpbuf, (unsigned char*)buf,
+                        int diff = mymemcmp (cmpbuf, (uint8_t*)buf,
                                              len);
                         printf ("Differ at pos %i/%i:\n", diff, len);
                         for (i = 0; i < 24 && i+diff < len; i++)
@@ -165,7 +179,7 @@ int do_checksum (int *buf, int len, int quiet)
                         printf ("\n");
                         for (i = 0; i < 24 && i+diff < len; i++)
                                 printf (" %02x",
-                                        ((unsigned char*)buf)[i+diff]);
+                                        ((uint8_t*)buf)[i+diff]);
                         printf ("\n");
                 }
                 return 2222;
@@ -201,7 +215,7 @@ void do_fill_buffer (int *buf, int len)
         }
         if (len >= BPI) buf[len/BPI - 1] = 0x12345678 - sum;
         else ((char*)buf)[0] = 0x12345678 + ((char*)buf)[0] - sum;
-        if (do_checksum (buf, len, 1)) {
+        if (do_checksum(buf, len, true)) {
                 if (len < BPI) goto retry;
                 printf ("sg_test_rwbuf: Memory corruption?\n");
                 exit (1);
@@ -210,37 +224,35 @@ void do_fill_buffer (int *buf, int len)
 }
 
 
-int read_buffer (int sg_fd, unsigned size)
+int read_buffer (int sg_fd, unsigned ssize)
 {
         int res, k;
-        unsigned char rbCmdBlk[] = {READ_BUFFER, 0, 0, 0, 0, 0, 0, 0, 0, 0};
-        int bufSize = size + addread;
-        unsigned char * rbBuff = (unsigned char *)malloc(bufSize);
-        unsigned char sense_buffer[32];
+        uint8_t rb_cdb[] = {READ_BUFFER, 0, 0, 0, 0, 0, 0, 0, 0, 0};
+        int bufSize = ssize + addread;
+        uint8_t * rbBuff = (uint8_t *)malloc(bufSize);
+        uint8_t sense_buffer[32];
         struct sg_io_hdr io_hdr;
 
         if (NULL == rbBuff)
                 return -1;
-        rbCmdBlk[1] = RWB_MODE_DATA;
-        rbCmdBlk[6] = 0xff & (bufSize >> 16);
-        rbCmdBlk[7] = 0xff & (bufSize >> 8);
-        rbCmdBlk[8] = 0xff & bufSize;
+        rb_cdb[1] = RWB_MODE_DATA;
+        sg_put_unaligned_be24((uint32_t)bufSize, rb_cdb + 6);
         memset(&io_hdr, 0, sizeof(struct sg_io_hdr));
         io_hdr.interface_id = 'S';
-        io_hdr.cmd_len = sizeof(rbCmdBlk);
+        io_hdr.cmd_len = sizeof(rb_cdb);
         io_hdr.mx_sb_len = sizeof(sense_buffer);
         io_hdr.dxfer_direction = SG_DXFER_FROM_DEV;
         io_hdr.dxfer_len = bufSize;
         io_hdr.dxferp = rbBuff;
-        io_hdr.cmdp = rbCmdBlk;
+        io_hdr.cmdp = rb_cdb;
         io_hdr.sbp = sense_buffer;
         io_hdr.pack_id = 2;
         io_hdr.timeout = 60000;     /* 60000 millisecs == 60 seconds */
         if (verbose) {
-                fprintf(stderr, "    read buffer [mode data] cdb: ");
-                for (k = 0; k < (int)sizeof(rbCmdBlk); ++k)
-                        fprintf(stderr, "%02x ", rbCmdBlk[k]);
-                fprintf(stderr, "\n");
+                pr2serr("    read buffer [mode data] cdb: ");
+                for (k = 0; k < (int)sizeof(rb_cdb); ++k)
+                        pr2serr("%02x ", rb_cdb[k]);
+                pr2serr("\n");
         }
 
         if (ioctl(sg_fd, SG_IO, &io_hdr) < 0) {
@@ -252,54 +264,57 @@ int read_buffer (int sg_fd, unsigned size)
         res = sg_err_category3(&io_hdr);
         switch (res) {
         case SG_LIB_CAT_RECOVERED:
-            sg_chk_n_print3("READ BUFFER data, continuing", &io_hdr, 1);
-            /* fall through */
+            sg_chk_n_print3("READ BUFFER data, continuing", &io_hdr, true);
+#if defined(__GNUC__)
+#if (__GNUC__ >= 7)
+            __attribute__((fallthrough));
+            /* FALL THROUGH */
+#endif
+#endif
         case SG_LIB_CAT_CLEAN:
                 break;
         default: /* won't bother decoding other categories */
-                sg_chk_n_print3("READ BUFFER data error", &io_hdr, 1);
+                sg_chk_n_print3("READ BUFFER data error", &io_hdr, true);
                 free(rbBuff);
                 return res;
         }
 
-        res = do_checksum ((int*)rbBuff, size, 0);
+        res = do_checksum((int*)rbBuff, ssize, false);
         free(rbBuff);
         return res;
 }
 
-int write_buffer (int sg_fd, unsigned size)
+int write_buffer (int sg_fd, unsigned ssize)
 {
-        unsigned char wbCmdBlk[] = {WRITE_BUFFER, 0, 0, 0, 0, 0, 0, 0, 0, 0};
-        int bufSize = size + addwrite;
-        unsigned char * wbBuff = (unsigned char *)malloc(bufSize);
-        unsigned char sense_buffer[32];
+        uint8_t wb_cdb[] = {WRITE_BUFFER, 0, 0, 0, 0, 0, 0, 0, 0, 0};
+        int bufSize = ssize + addwrite;
+        uint8_t * wbBuff = (uint8_t *)malloc(bufSize);
+        uint8_t sense_buffer[32];
         struct sg_io_hdr io_hdr;
         int k, res;
 
         if (NULL == wbBuff)
                 return -1;
         memset(wbBuff, 0, bufSize);
-        do_fill_buffer ((int*)wbBuff, size);
-        wbCmdBlk[1] = RWB_MODE_DATA;
-        wbCmdBlk[6] = 0xff & (bufSize >> 16);
-        wbCmdBlk[7] = 0xff & (bufSize >> 8);
-        wbCmdBlk[8] = 0xff & bufSize;
+        do_fill_buffer ((int*)wbBuff, ssize);
+        wb_cdb[1] = RWB_MODE_DATA;
+        sg_put_unaligned_be24((uint32_t)bufSize, wb_cdb + 6);
         memset(&io_hdr, 0, sizeof(struct sg_io_hdr));
         io_hdr.interface_id = 'S';
-        io_hdr.cmd_len = sizeof(wbCmdBlk);
+        io_hdr.cmd_len = sizeof(wb_cdb);
         io_hdr.mx_sb_len = sizeof(sense_buffer);
         io_hdr.dxfer_direction = SG_DXFER_TO_DEV;
         io_hdr.dxfer_len = bufSize;
         io_hdr.dxferp = wbBuff;
-        io_hdr.cmdp = wbCmdBlk;
+        io_hdr.cmdp = wb_cdb;
         io_hdr.sbp = sense_buffer;
         io_hdr.pack_id = 1;
         io_hdr.timeout = 60000;     /* 60000 millisecs == 60 seconds */
         if (verbose) {
-                fprintf(stderr, "    write buffer [mode data] cdb: ");
-                for (k = 0; k < (int)sizeof(wbCmdBlk); ++k)
-                        fprintf(stderr, "%02x ", wbCmdBlk[k]);
-                fprintf(stderr, "\n");
+                pr2serr("    write buffer [mode data] cdb: ");
+                for (k = 0; k < (int)sizeof(wb_cdb); ++k)
+                        pr2serr("%02x ", wb_cdb[k]);
+                pr2serr("\n");
         }
 
         if (ioctl(sg_fd, SG_IO, &io_hdr) < 0) {
@@ -311,12 +326,17 @@ int write_buffer (int sg_fd, unsigned size)
         res = sg_err_category3(&io_hdr);
         switch (res) {
         case SG_LIB_CAT_RECOVERED:
-            sg_chk_n_print3("WRITE BUFFER data, continuing", &io_hdr, 1);
-            /* fall through */
+            sg_chk_n_print3("WRITE BUFFER data, continuing", &io_hdr, true);
+#if defined(__GNUC__)
+#if (__GNUC__ >= 7)
+            __attribute__((fallthrough));
+            /* FALL THROUGH */
+#endif
+#endif
         case SG_LIB_CAT_CLEAN:
                 break;
         default: /* won't bother decoding other categories */
-                sg_chk_n_print3("WRITE BUFFER data error", &io_hdr, 1);
+                sg_chk_n_print3("WRITE BUFFER data error", &io_hdr, true);
                 free(wbBuff);
                 return res;
         }
@@ -334,17 +354,17 @@ void usage ()
                 " or\n"
                 "       sg_test_rwbuf DEVICE SZ [AW] [AR]\n");
         printf ("  where:\n"
-                "    --addrd=AR       extra bytes to fetch during READ "
+                "    --addrd=AR|-r    extra bytes to fetch during READ "
                 "BUFFER\n"
-                "    --addwr=AW       extra bytes to send to WRITE BUFFER\n"
-                "    --help           output this usage message then exit\n"
-                "    --quick          output read buffer size then exit\n"
-                "    --size=SZ        size of buffer (in bytes) to write "
+                "    --addwr=AW|-w    extra bytes to send to WRITE BUFFER\n"
+                "    --help|-l        output this usage message then exit\n"
+                "    --quick|-q       output read buffer size then exit\n"
+                "    --size=SZ|-s     size of buffer (in bytes) to write "
                 "then read back\n"
-                "    --times=NUM      number of times to run test "
+                "    --times=NUM|-t   number of times to run test "
                 "(default 1)\n"
-                "    --verbose        increase verbosity of output\n"
-                "    --version        output version then exit\n");
+                "    --verbose|-v     increase verbosity of output\n"
+                "    --version|-V     output version then exit\n");
         printf ("\nWARNING: If you access the device at the same time, e.g. "
                 "because it's a\n");
         printf (" mounted hard disk, the device's buffer may be used by the "
@@ -378,26 +398,26 @@ int main (int argc, char * argv[])
                         usage();
                         return 0;
                 case 'q':
-                        do_quick = 1;
+                        do_quick = true;
                         break;
                 case 'r':
                         addread = sg_get_num(optarg);
                         if (-1 == addread) {
-                                fprintf(stderr, "bad argument to '--addrd'\n");
+                                pr2serr("bad argument to '--addrd'\n");
                                 return SG_LIB_SYNTAX_ERROR;
                         }
                         break;
                 case 's':
                         size = sg_get_num(optarg);
                         if (-1 == size) {
-                                fprintf(stderr, "bad argument to '--size'\n");
+                                pr2serr("bad argument to '--size'\n");
                                 return SG_LIB_SYNTAX_ERROR;
                         }
                         break;
                 case 't':
                         times = sg_get_num(optarg);
                         if (-1 == times) {
-                                fprintf(stderr, "bad argument to '--times'\n");
+                                pr2serr("bad argument to '--times'\n");
                                 return SG_LIB_SYNTAX_ERROR;
                         }
                         break;
@@ -405,13 +425,12 @@ int main (int argc, char * argv[])
                         verbose++;
                         break;
                 case 'V':
-                        fprintf(stderr, ME "version: %s\n",
-                                version_str);
+                        pr2serr(ME "version: %s\n", version_str);
                         return 0;
                 case 'w':
                         addwrite = sg_get_num(optarg);
                         if (-1 == addwrite) {
-                                fprintf(stderr, "bad argument to '--addwr'\n");
+                                pr2serr("bad argument to '--addwr'\n");
                                 return SG_LIB_SYNTAX_ERROR;
                         }
                         break;
@@ -430,22 +449,21 @@ int main (int argc, char * argv[])
                 if (-1 == size) {
                         size = sg_get_num(argv[optind]);
                         if (-1 == size) {
-                                fprintf(stderr, "bad <sz>\n");
+                                pr2serr("bad <sz>\n");
                                 usage();
                                 return SG_LIB_SYNTAX_ERROR;
                         }
                         if (++optind < argc) {
                                 addwrite = sg_get_num(argv[optind]);
                                 if (-1 == addwrite) {
-                                        fprintf(stderr, "bad [addwr]\n");
+                                        pr2serr("bad [addwr]\n");
                                         usage();
                                         return SG_LIB_SYNTAX_ERROR;
                                 }
                                 if (++optind < argc) {
                                         addread = sg_get_num(argv[optind]);
                                         if (-1 == addread) {
-                                                fprintf(stderr,
-                                                        "bad [addrd]\n");
+                                                pr2serr("bad [addrd]\n");
                                                 usage();
                                                 return SG_LIB_SYNTAX_ERROR;
                                         }
@@ -455,20 +473,20 @@ int main (int argc, char * argv[])
                 }
                 if (optind < argc) {
                         for (; optind < argc; ++optind)
-                                fprintf(stderr, "Unexpected extra argument"
-                                        ": %s\n", argv[optind]);
+                                pr2serr("Unexpected extra argument" ": %s\n",
+                                        argv[optind]);
                         usage();
                         return SG_LIB_SYNTAX_ERROR;
                 }
         }
         if (NULL == device_name) {
-                fprintf(stderr, "no device name given\n");
+                pr2serr("no device name given\n");
                 usage();
                 return SG_LIB_SYNTAX_ERROR;
         }
         if ((size <= 0) && (! do_quick)) {
-                fprintf(stderr, "must give '--size' or '--quick' options "
-                        "or <sz> argument\n");
+                pr2serr("must give '--size' or '--quick' options or <sz> "
+                        "argument\n");
                 usage();
                 return SG_LIB_SYNTAX_ERROR;
         }
@@ -488,13 +506,12 @@ int main (int argc, char * argv[])
                 goto err_out;
         }
         if (size > buf_capacity) {
-                fprintf (stderr, ME "sz=%i > buf_capacity=%i\n",
-                        size, buf_capacity);
+                pr2serr (ME "sz=%i > buf_capacity=%i\n", size, buf_capacity);
                 ret = SG_LIB_CAT_OTHER;
                 goto err_out;
         }
 
-        cmpbuf = (unsigned char *)malloc(size);
+        cmpbuf = (uint8_t *)malloc(size);
         for (k = 0; k < times; ++k) {
                 ret = write_buffer (sg_fd, size);
                 if (ret) {
