@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2006-2015 Douglas Gilbert.
+ * Copyright (c) 2006-2018 Douglas Gilbert.
  * All rights reserved.
  * Use of this source code is governed by a BSD-style
  * license that can be found in the BSD_LICENSE file.
@@ -9,7 +9,6 @@
 #include <fcntl.h>
 #include <stdio.h>
 #include <stdlib.h>
-#include <stdarg.h>
 #include <string.h>
 #define __STDC_FORMAT_MACROS 1
 #include <inttypes.h>
@@ -23,6 +22,8 @@
 #endif
 #include "sg_lib.h"
 #include "sg_cmds_basic.h"
+#include "sg_unaligned.h"
+#include "sg_pr2serr.h"
 
 /* This is a companion file to sg_vpd.c . It contains logic to output and
    decode vendor specific VPD pages
@@ -48,8 +49,8 @@
 #define VPD_VP_EMC 2
 #define VPD_VP_DDS 3
 #define VPD_VP_HP3PAR 4
-#define VPD_VP_LTO5 5
-#define VPD_VP_LTO6 6
+#define VPD_VP_IBM_LTO 5
+#define VPD_VP_HP_LTO 6
 
 
 /* vendor VPD pages */
@@ -58,24 +59,24 @@
 #define VPD_V_UPR_EMC  0xc0
 #define VPD_V_HVER_RDAC  0xc0
 #define VPD_V_FVER_DDS 0xc0
-#define VPD_V_FVER_LTO6 0xc0
-#define VPD_V_DCRL_LTO5 0xc0
+#define VPD_V_FVER_LTO 0xc0
+#define VPD_V_DCRL_LTO 0xc0
 #define VPD_V_DATC_SEA  0xc1
 #define VPD_V_FVER_RDAC  0xc1
-#define VPD_V_HVER_LTO6 0xc1
-#define VPD_V_DSN_LTO5 0xc1
+#define VPD_V_HVER_LTO 0xc1
+#define VPD_V_DSN_LTO 0xc1
 #define VPD_V_JUMP_SEA 0xc2
 #define VPD_V_SVER_RDAC 0xc2
-#define VPD_V_PCA_LTO6 0xc2
+#define VPD_V_PCA_LTO 0xc2
 #define VPD_V_DEV_BEH_SEA 0xc3
 #define VPD_V_FEAT_RDAC 0xc3
-#define VPD_V_MECH_LTO6 0xc3
+#define VPD_V_MECH_LTO 0xc3
 #define VPD_V_SUBS_RDAC 0xc4
-#define VPD_V_HEAD_LTO6 0xc4
-#define VPD_V_ACI_LTO6 0xc5
-#define VPD_V_DUCD_LTO5 0xc7
+#define VPD_V_HEAD_LTO 0xc4
+#define VPD_V_ACI_LTO 0xc5
+#define VPD_V_DUCD_LTO 0xc7
 #define VPD_V_EDID_RDAC 0xc8
-#define VPD_V_MPDS_LTO5 0xc8
+#define VPD_V_MPDS_LTO 0xc8
 #define VPD_V_VAC_RDAC 0xc9
 #define VPD_V_RVSI_RDAC 0xca
 #define VPD_V_SAID_RDAC 0xd0
@@ -86,15 +87,17 @@
 
 /* These structures are duplicates of those of the same name in
  * sg_vpd.c . Take care that both are the same. */
+
 struct opts_t {
-    int do_all;
-    int do_enum;
+    bool do_all;
+    bool do_enum;
+    bool do_force;
+    bool do_long;
+    bool do_quiet;
     int do_hex;
-    int num_vpd;
+    int vpd_pn;
     int do_ident;
-    int do_long;
     int maxlen;
-    int do_quiet;
     int do_raw;
     int vend_prod_num;
     int verbose;
@@ -113,11 +116,11 @@ struct svpd_values_name_t {
     const char * name;
 };
 
-int vpd_fetch_page_from_dev(int sg_fd, unsigned char * rp, int page,
-                            int mxlen, int vb, int * rlenp);
+int vpd_fetch_page(int sg_fd, uint8_t * rp, int page, int mxlen,
+                   bool qt, int vb, int * rlenp);
 
 /* sharing large global buffer, defined in sg_vpd.c */
-extern unsigned char rsp_buff[];
+extern uint8_t * rsp_buff;
 
 /* end of section copied from sg_vpd.c . Maybe sg_vpd.h is needed */
 
@@ -134,8 +137,8 @@ static struct svpd_vp_name_t vp_arr[] = {
     {VPD_VP_DDS, "dds", "DDS tape family from IBM"},
     {VPD_VP_EMC, "emc", "EMC (company)"},
     {VPD_VP_HP3PAR, "hp3par", "3PAR array (HP was Left Hand)"},
-    {VPD_VP_LTO5, "lto5", "LTO-5 tape/systems (IBM)"},
-    {VPD_VP_LTO6, "lto6", "LTO-6 tape/systems (IBM)"},
+    {VPD_VP_IBM_LTO, "ibm_lto", "IBM LTO tape/systems"},
+    {VPD_VP_HP_LTO, "hp_lto", "HP LTO tape/systems"},
     {VPD_VP_RDAC, "rdac", "RDAC array (NetApp E-Series)"},
     {VPD_VP_SEAGATE, "sea", "Seagate disk"},
     {0, NULL, NULL},
@@ -145,37 +148,38 @@ static struct svpd_vp_name_t vp_arr[] = {
 /* 'subvalue' holds vendor/product number to disambiguate */
 /* Arrange in alphabetical order by acronym */
 static struct svpd_values_name_t vendor_vpd_pg[] = {
-    {VPD_V_ACI_LTO6, VPD_VP_LTO6, 1, "aci", "ACI revision level (LTO-6)"},
+    {VPD_V_ACI_LTO, VPD_VP_HP_LTO, 1, "aci", "ACI revision level (HP LTO)"},
     {VPD_V_DATC_SEA, VPD_VP_SEAGATE, 0, "datc", "Date code (Seagate)"},
-    {VPD_V_DCRL_LTO5, VPD_VP_LTO5, 1, "dcrl" , "Drive component revision "
-     "levels (LTO-5)"},
+    {VPD_V_DCRL_LTO, VPD_VP_IBM_LTO, 1, "dcrl", "Drive component revision "
+     "levels (IBM LTO)"},
     {VPD_V_FVER_DDS, VPD_VP_DDS, 1, "ddsver", "Firmware revision (DDS)"},
     {VPD_V_DEV_BEH_SEA, VPD_VP_SEAGATE, 0, "devb", "Device behavior "
      "(Seagate)"},
-    {VPD_V_DSN_LTO5, VPD_VP_LTO5, 1, "dsn" , "Drive serial numbers (LTO-5)"},
-    {VPD_V_DUCD_LTO5, VPD_VP_LTO5, 1, "ducd" , "Device unique "
-     "configuration data (LTO-5)"},
+    {VPD_V_DSN_LTO, VPD_VP_IBM_LTO, 1, "dsn", "Drive serial numbers (IBM "
+     "LTO)"},
+    {VPD_V_DUCD_LTO, VPD_VP_IBM_LTO, 1, "ducd", "Device unique "
+     "configuration data (IBM LTO)"},
     {VPD_V_EDID_RDAC, VPD_VP_RDAC, 0, "edid", "Extended device "
      "identification (RDAC)"},
     {VPD_V_FEAT_RDAC, VPD_VP_RDAC, 0, "prm4", "Feature Parameters (RDAC)"},
     {VPD_V_FIRM_SEA, VPD_VP_SEAGATE, 0, "firm", "Firmware numbers "
      "(Seagate)"},
-    {VPD_V_FVER_LTO6, VPD_VP_LTO6, 0, "frl" , "Firmware revision level "
-     "(LTO-6)"},
+    {VPD_V_FVER_LTO, VPD_VP_HP_LTO, 0, "frl", "Firmware revision level "
+     "(HP LTO)"},
     {VPD_V_FVER_RDAC, VPD_VP_RDAC, 0, "fwr4", "Firmware version (RDAC)"},
-    {VPD_V_HEAD_LTO6, VPD_VP_LTO6, 1, "head", "Head Assy revision level "
-     "(LTO-6)"},
+    {VPD_V_HEAD_LTO, VPD_VP_HP_LTO, 1, "head", "Head Assy revision level "
+     "(HP LTO)"},
     {VPD_V_HP3PAR, VPD_VP_HP3PAR, 0, "hp3par", "Volume information "
      "(HP/3PAR)"},
-    {VPD_V_HVER_LTO6, VPD_VP_LTO6, 1, "hrl", "Hardware revision level "
-     "(LTO-6)"},
+    {VPD_V_HVER_LTO, VPD_VP_HP_LTO, 1, "hrl", "Hardware revision level "
+     "(HP LTO)"},
     {VPD_V_HVER_RDAC, VPD_VP_RDAC, 0, "hwr4", "Hardware version (RDAC)"},
     {VPD_V_JUMP_SEA, VPD_VP_SEAGATE, 0, "jump", "Jump setting (Seagate)"},
-    {VPD_V_MECH_LTO6, VPD_VP_LTO6, 1, "mech", "Mechanism revision level "
-     "(LTO-6)"},
-    {VPD_V_MPDS_LTO5, VPD_VP_LTO5, 1, "mpds" , "Mode parameter default "
-     "settings (LTO-5)"},
-    {VPD_V_PCA_LTO6, VPD_VP_LTO6, 1, "pca", "PCA revision level (LTO-6)"},
+    {VPD_V_MECH_LTO, VPD_VP_HP_LTO, 1, "mech", "Mechanism revision level "
+     "(HP LTO)"},
+    {VPD_V_MPDS_LTO, VPD_VP_IBM_LTO, 1, "mpds", "Mode parameter default "
+     "settings (IBM LTO)"},
+    {VPD_V_PCA_LTO, VPD_VP_HP_LTO, 1, "pca", "PCA revision level (HP LTO)"},
     {VPD_V_RVSI_RDAC, VPD_VP_RDAC, 0, "rvsi", "Replicated volume source "
      "identifier (RDAC)"},
     {VPD_V_SAID_RDAC, VPD_VP_RDAC, 0, "said", "Storage array world wide "
@@ -188,24 +192,19 @@ static struct svpd_values_name_t vendor_vpd_pg[] = {
 };
 
 
-#ifdef __GNUC__
-static int pr2serr(const char * fmt, ...)
-        __attribute__ ((format (printf, 1, 2)));
-#else
-static int pr2serr(const char * fmt, ...);
-#endif
-
-
-static int
-pr2serr(const char * fmt, ...)
+void
+dup_sanity_chk(int sz_opts_t, int sz_values_name_t)
 {
-    va_list args;
-    int n;
+    const size_t my_sz_opts_t = sizeof(struct opts_t);
+    const size_t my_sz_values_name_t = sizeof(struct svpd_values_name_t);
 
-    va_start(args, fmt);
-    n = vfprintf(stderr, fmt, args);
-    va_end(args);
-    return n;
+    if (sz_opts_t != (int)my_sz_opts_t)
+        pr2serr(">>> struct opts_t differs in size from sg_vpd.c [%d != "
+                "%d]\n", (int)my_sz_opts_t, sz_opts_t);
+    if (sz_values_name_t != (int)my_sz_values_name_t)
+        pr2serr(">>> struct svpd_values_name_t differs in size from "
+                "sg_vpd.c [%d != %d]\n", (int)my_sz_values_name_t,
+                sz_values_name_t);
 }
 
 static int
@@ -306,16 +305,16 @@ svpd_find_vendor_by_acron(const char * ap)
 void
 svpd_enumerate_vendor(int vend_prod_num)
 {
+    bool seen;
     const struct svpd_vp_name_t * vpp;
     const struct svpd_values_name_t * vnp;
-    int seen;
 
     if (vend_prod_num < 0) {
-        for (seen = 0, vpp = vp_arr; vpp->acron; ++vpp) {
+        for (seen = false, vpp = vp_arr; vpp->acron; ++vpp) {
             if (vpp->name) {
                 if (! seen) {
                     printf("\nVendor/product identifiers:\n");
-                    seen = 1;
+                    seen = true;
                 }
                 printf("  %-10s %d      %s\n", vpp->acron,
                        vpp->vend_prod_num, vpp->name);
@@ -324,13 +323,13 @@ svpd_enumerate_vendor(int vend_prod_num)
     }
     if (-1 == vend_prod_num)
         return;
-    for (seen = 0, vnp = vendor_vpd_pg; vnp->acron; ++vnp) {
+    for (seen = false, vnp = vendor_vpd_pg; vnp->acron; ++vnp) {
         if ((vend_prod_num >= 0) && (vend_prod_num != vnp->subvalue))
             continue;
         if (vnp->name) {
             if (! seen) {
                 printf("\nVendor specific VPD pages:\n");
-                seen = 1;
+                seen = true;
             }
             printf("  %-10s 0x%02x,%d      %s\n", vnp->acron,
                    vnp->value, vnp->subvalue, vnp->name);
@@ -339,13 +338,13 @@ svpd_enumerate_vendor(int vend_prod_num)
 }
 
 int
-svpd_count_vendor_vpds(int num_vpd, int vend_prod_num)
+svpd_count_vendor_vpds(int vpd_pn, int vend_prod_num)
 {
     const struct svpd_values_name_t * vnp;
     int matches;
 
     for (vnp = vendor_vpd_pg, matches = 0; vnp->acron; ++vnp) {
-        if ((num_vpd == vnp->value) && vnp->name) {
+        if ((vpd_pn == vnp->value) && vnp->name) {
             if ((vend_prod_num < 0) || (vend_prod_num == vnp->subvalue)) {
                 if (0 == matches)
                     printf("Matching vendor specific VPD pages:\n");
@@ -359,16 +358,16 @@ svpd_count_vendor_vpds(int num_vpd, int vend_prod_num)
 }
 
 static void
-dStrRaw(const char* str, int len)
+dStrRaw(const uint8_t * str, int len)
 {
     int k;
 
-    for (k = 0 ; k < len; ++k)
+    for (k = 0; k < len; ++k)
         printf("%c", str[k]);
 }
 
 static void
-decode_vpd_c0_hp3par(unsigned char * buff, int len)
+decode_vpd_c0_hp3par(uint8_t * buff, int len)
 {
     int rev;
     long offset;
@@ -389,31 +388,22 @@ decode_vpd_c0_hp3par(unsigned char * buff, int len)
     printf("  XCopy supported: %s\n", (buff[5] & 0x20) ? "yes" : "no");
 
     if (rev > 3) {
-        printf("  VV ID: %" PRIu64 "\n", ((uint64_t) buff[28] << 56) +
-                ((uint64_t) buff[29] << 48) + ((uint64_t) buff[30] << 40) +
-                ((uint64_t) buff[31] << 32) + ((uint64_t) buff[32] << 24) +
-                (buff[33] << 16) + (buff[34] << 8) + buff[35]);
-
+        printf("  VV ID: %" PRIu64 "\n", sg_get_unaligned_be64(buff + 28));
         offset = 44;
         printf("  Volume name: %s\n", &buff[offset]);
 
-        printf("  Domain ID: %d\n", (buff[36] << 24) +  (buff[37] << 16) +
-                (buff[38] << 8) + buff[39]);
+        printf("  Domain ID: %d\n", sg_get_unaligned_be32(buff + 36));
 
-        offset += (buff[offset - 4] << 24) + (buff[offset - 3] << 16) +
-                (buff[offset - 2] << 8) + buff[offset - 1] + 4;
+        offset += sg_get_unaligned_be32(buff + offset - 4) + 4;
         printf("  Domain Name: %s\n", &buff[offset]);
 
-        offset += (buff[offset - 4] << 24) + (buff[offset - 3] << 16) +
-                (buff[offset - 2] << 8) + buff[offset - 1] + 4;
+        offset += sg_get_unaligned_be32(buff + offset - 4) + 4;
         printf("  User CPG: %s\n", &buff[offset]);
 
-        offset += (buff[offset - 4] << 24) + (buff[offset - 3] << 16) +
-                (buff[offset - 2] << 8) + buff[offset - 1] + 4;
+        offset += sg_get_unaligned_be32(buff + offset - 4) + 4;
         printf("  Snap CPG: %s\n", &buff[offset]);
 
-        offset += (buff[offset - 4] << 24) + (buff[offset - 3] << 16) +
-                (buff[offset - 2] << 8) + buff[offset - 1];
+        offset += sg_get_unaligned_be32(buff + offset - 4);
 
         printf("  VV policies: %s,%s,%s,%s\n",
                 (buff[offset + 3] & 0x01) ? "stale_ss" : "no_stale_ss",
@@ -424,27 +414,20 @@ decode_vpd_c0_hp3par(unsigned char * buff, int len)
     }
 
     if (buff[5] & 0x04) {
-        printf("  Allocation unit: %d\n", (buff[8] << 24) +  (buff[9] << 16) +
-                (buff[10] << 8) + buff[11]);
+        printf("  Allocation unit: %d\n", sg_get_unaligned_be32(buff + 8));
 
         printf("  Data pool size: %" PRIu64 "\n",
-               (((uint64_t)buff[12]) << 56) + (((uint64_t)buff[13]) << 48) +
-               (((uint64_t)buff[14]) << 40) + (((uint64_t)buff[15]) << 32) +
-               (((uint64_t)buff[16]) << 24) + (buff[17] << 16) +
-               (buff[18] << 8) + buff[19]);
+               sg_get_unaligned_be64(buff + 12));
 
         printf("  Space allocated: %" PRIu64 "\n",
-               (((uint64_t)buff[20]) << 56) + (((uint64_t)buff[21]) << 48) +
-               (((uint64_t)buff[22]) << 40) + (((uint64_t)buff[23]) << 32) +
-               (((uint64_t)buff[24]) << 24) + (buff[25] << 16) +
-               (buff[26] << 8) + buff[27]);
+               sg_get_unaligned_be64(buff + 20));
     }
     return;
 }
 
 
 static void
-decode_firm_vpd_c0_sea(unsigned char * buff, int len)
+decode_firm_vpd_c0_sea(uint8_t * buff, int len)
 {
     if (len < 28) {
         pr2serr("Seagate firmware numbers VPD page length too short=%d\n",
@@ -478,7 +461,7 @@ decode_firm_vpd_c0_sea(unsigned char * buff, int len)
 }
 
 static void
-decode_date_code_vpd_c1_sea(unsigned char * buff, int len)
+decode_date_code_vpd_c1_sea(uint8_t * buff, int len)
 {
     if (len < 20) {
         pr2serr("Seagate Data code VPD page length too short=%d\n",
@@ -490,7 +473,7 @@ decode_date_code_vpd_c1_sea(unsigned char * buff, int len)
 }
 
 static void
-decode_dev_beh_vpd_c3_sea(unsigned char * buff, int len)
+decode_dev_beh_vpd_c3_sea(uint8_t * buff, int len)
 {
     if (len < 25) {
         pr2serr("Seagate Device behaviour VPD page length too short=%d\n",
@@ -553,7 +536,7 @@ static const char * failover_mode_arr[] =
 };
 
 static void
-decode_upr_vpd_c0_emc(unsigned char * buff, int len)
+decode_upr_vpd_c0_emc(uint8_t * buff, int len)
 {
     int k, ip_mgmt, vpp80, lun_z;
 
@@ -571,7 +554,7 @@ decode_upr_vpd_c0_emc(unsigned char * buff, int len)
         printf("%02x", buff[10 + k]);
     printf("\n");
     printf("  Array Serial Number: ");
-    dStrRaw((const char *)&buff[50], buff[49]);
+    dStrRaw(&buff[50], buff[49]);
     printf("\n");
 
     printf("  LUN State: ");
@@ -627,7 +610,7 @@ decode_upr_vpd_c0_emc(unsigned char * buff, int len)
 }
 
 static void
-decode_rdac_vpd_c0(unsigned char * buff, int len)
+decode_rdac_vpd_c0(uint8_t * buff, int len)
 {
     int memsize;
     char name[65];
@@ -642,7 +625,7 @@ decode_rdac_vpd_c0(unsigned char * buff, int len)
         return;
     }
     printf("  Number of channels: %x\n", buff[8]);
-    memsize = buff[10] << 8 | buff[11];
+    memsize = sg_get_unaligned_be16(buff + 10);
     printf("  Processor Memory Size: %d\n", memsize);
     memset(name, 0, 65);
     memcpy(name, buff + 16, 64);
@@ -673,7 +656,7 @@ decode_rdac_vpd_c0(unsigned char * buff, int len)
 }
 
 static void
-decode_rdac_vpd_c1(unsigned char * buff, int len)
+decode_rdac_vpd_c1(uint8_t * buff, int len)
 {
     int i, n, v, r, m, p, d, y, num_part;
     char part[5];
@@ -715,7 +698,7 @@ decode_rdac_vpd_c1(unsigned char * buff, int len)
 }
 
 static void
-decode_rdac_vpd_c2(unsigned char * buff, int len)
+decode_rdac_vpd_c2(uint8_t * buff, int len)
 {
     int i, n, v, r, m, p, d, y, num_part;
     char part[5];
@@ -770,7 +753,7 @@ decode_rdac_vpd_c2(unsigned char * buff, int len)
 }
 
 static void
-decode_rdac_vpd_c3(unsigned char * buff, int len)
+decode_rdac_vpd_c3(uint8_t * buff, int len)
 {
     if (len < 0x2c) {
         pr2serr("Feature parameters VPD page length too short=%d\n", len);
@@ -792,7 +775,7 @@ decode_rdac_vpd_c3(unsigned char * buff, int len)
 }
 
 static void
-decode_rdac_vpd_c4(unsigned char * buff, int len)
+decode_rdac_vpd_c4(uint8_t * buff, int len)
 {
     char subsystem_id[17];
     char subsystem_rev[5];
@@ -862,7 +845,7 @@ decode_rdac_vpd_c4(unsigned char * buff, int len)
 }
 
 static void
-convert_binary_to_ascii(unsigned char * src, unsigned char * dst,  int len)
+convert_binary_to_ascii(uint8_t * src, uint8_t * dst,  int len)
 {
     int i;
 
@@ -872,7 +855,7 @@ convert_binary_to_ascii(unsigned char * src, unsigned char * dst,  int len)
 }
 
 static void
-decode_rdac_vpd_c8(unsigned char * buff, int len)
+decode_rdac_vpd_c8(uint8_t * buff, int len)
 {
     int i;
 #ifndef SG_LIB_MINGW
@@ -883,7 +866,7 @@ decode_rdac_vpd_c8(unsigned char * buff, int len)
     int label_len;
     char uuid[33];
     int uuid_len;
-    unsigned char port_id[128];
+    uint8_t port_id[128];
     int n;
 
     if (len < 0xab) {
@@ -906,13 +889,13 @@ decode_rdac_vpd_c8(unsigned char * buff, int len)
 
     printf("  Volume Unique Identifier: %s\n", uuid);
 #ifndef SG_LIB_MINGW
-    tstamp = (buff[24] << 24) + (buff[25] << 16) + (buff[26] << 8) + buff[27];
+    tstamp = sg_get_unaligned_be32(buff + 24);
     printf("    Creation Number: %d, Timestamp: %s",
-           (buff[22] << 8) + buff[23], ctime(&tstamp));
+           sg_get_unaligned_be16(buff + 22), ctime(&tstamp));
 #else
     printf("    Creation Number: %d, Timestamp value: %u",
-           (buff[22] << 8) + buff[23],
-           (buff[24] << 24) + (buff[25] << 16) + (buff[26] << 8) + buff[27]);
+           sg_get_unaligned_be16(buff + 22),
+           sg_get_unaligned_be32(buff + 24));
 #endif
     memset(label, 0, 61);
     label_len = buff[28];
@@ -958,7 +941,7 @@ decode_rdac_vpd_c8(unsigned char * buff, int len)
             break;
         case TPROTO_ISCSI: /* iSCSI */
             printf("iSCSI\n");
-            n = (buff[177] << 8) + buff[178];
+            n = sg_get_unaligned_be32(buff + 177);
             memcpy(port_id, &buff[179], n);
             n = 179 + n;
             break;
@@ -983,7 +966,7 @@ decode_rdac_vpd_c8(unsigned char * buff, int len)
 }
 
 static void
-decode_rdac_vpd_c9_rtpg_data(unsigned char aas, unsigned char vendor)
+decode_rdac_vpd_c9_rtpg_data(uint8_t aas, uint8_t vendor)
 {
     printf("  Asymmetric Access State:");
     switch(aas & 0x0F) {
@@ -1045,7 +1028,7 @@ decode_rdac_vpd_c9_rtpg_data(unsigned char aas, unsigned char vendor)
 }
 
 static void
-decode_rdac_vpd_c9(unsigned char * buff, int len)
+decode_rdac_vpd_c9(uint8_t * buff, int len)
 {
     if (len < 3) {
         pr2serr("Volume Access Control VPD page length too short=%d\n", len);
@@ -1149,7 +1132,7 @@ decode_rdac_vpd_c9(unsigned char * buff, int len)
 }
 
 static void
-decode_rdac_vpd_ca(unsigned char * buff, int len)
+decode_rdac_vpd_ca(uint8_t * buff, int len)
 {
     int i;
 
@@ -1182,7 +1165,7 @@ decode_rdac_vpd_ca(unsigned char * buff, int len)
 }
 
 static void
-decode_rdac_vpd_d0(unsigned char * buff, int len)
+decode_rdac_vpd_d0(uint8_t * buff, int len)
 {
     int i;
 
@@ -1201,7 +1184,7 @@ decode_rdac_vpd_d0(unsigned char * buff, int len)
 
 
 static void
-decode_dds_vpd_c0(unsigned char * buff, int len)
+decode_dds_vpd_c0(uint8_t * buff, int len)
 {
     char firmware_rev[25];
     char build_date[43];
@@ -1234,7 +1217,7 @@ decode_dds_vpd_c0(unsigned char * buff, int len)
 }
 
 static void
-decode_lto6_vpd_cx(unsigned char * buff, int len, int page)
+decode_hp_lto_vpd_cx(uint8_t * buff, int len, int page)
 {
     char str[32];
     const char *comp = NULL;
@@ -1289,10 +1272,10 @@ decode_lto6_vpd_cx(unsigned char * buff, int len, int page)
 }
 
 static void
-decode_lto5_dcrl(unsigned char * buff, int len)
+decode_ibm_lto_dcrl(uint8_t * buff, int len)
 {
     if (len < 0x2b) {
-        pr2serr("Driver Component Revision Levels page (LTO-5) invalid "
+        pr2serr("Driver Component Revision Levels page (IBM LTO) invalid "
                 "length=%d\n", len);
         return;
     }
@@ -1303,10 +1286,10 @@ decode_lto5_dcrl(unsigned char * buff, int len)
 }
 
 static void
-decode_lto5_dsn(unsigned char * buff, int len)
+decode_ibm_lto_dsn(uint8_t * buff, int len)
 {
     if (len < 0x1c) {
-        pr2serr("Driver Serial Numbers page (LTO-5) invalid "
+        pr2serr("Driver Serial Numbers page (IBM LTO) invalid "
                 "length=%d\n", len);
         return;
     }
@@ -1314,7 +1297,7 @@ decode_lto5_dsn(unsigned char * buff, int len)
     printf("  Reported serial number: %.12s\n", buff + 16);
 }
 
-/* Returns 0 if successful, see sg_ll_inquiry() plus SG_LIB_SYNTAX_ERROR for
+/* Returns 0 if successful, see sg_ll_inquiry() plus SG_LIB_CAT_OTHER for
    unsupported page */
 int
 svpd_decode_vendor(int sg_fd, struct opts_t * op, int off)
@@ -1323,31 +1306,45 @@ svpd_decode_vendor(int sg_fd, struct opts_t * op, int off)
     char name[64];
     const struct svpd_values_name_t * vnp;
     int alloc_len = op->maxlen;
-    unsigned char * rp;
+    uint8_t * rp;
 
+    switch (op->vpd_pn) {
+    case 0xc0:
+    case 0xc1:
+    case 0xc2:
+    case 0xc3:
+    case 0xc4:
+    case 0xc5:
+    case 0xc8:
+    case 0xc9:
+    case 0xca:
+    case 0xd0:
+        break;
+    default:    /* not known so return prior to fetching page */
+        return SG_LIB_CAT_OTHER;
+    }
     rp = rsp_buff + off;
     if (sg_fd >= 0) {
         if (0 == alloc_len)
             alloc_len = DEF_ALLOC_LEN;
     }
-    res = vpd_fetch_page_from_dev(sg_fd, rp, op->num_vpd, alloc_len,
+    res = vpd_fetch_page(sg_fd, rp, op->vpd_pn, alloc_len, op->do_quiet,
                                   op->verbose, &len);
     if (0 == res) {
-        vnp = svpd_get_v_detail(op->num_vpd, op->vend_prod_num, 0xf & rp[0]);
+        vnp = svpd_get_v_detail(op->vpd_pn, op->vend_prod_num, 0xf & rp[0]);
         if (vnp && vnp->name)
             strcpy(name, vnp->name);
         else
             snprintf(name, sizeof(name) - 1, "Vendor VPD page=0x%x",
-                     op->num_vpd);
+                     op->vpd_pn);
         if ((! op->do_raw) && (! op->do_quiet) && (op->do_hex < 2))
             printf("%s VPD Page:\n", name);
         if (op->do_raw)
-            dStrRaw((const char *)rp, len);
+            dStrRaw(rp, len);
         else if (op->do_hex)
-            dStrHex((const char *)rp, len,
-                    ((1 == op->do_hex) ? 0 : -1));
+            hex2stdout(rp, len, ((1 == op->do_hex) ? 0 : -1));
         else {
-            switch(op->num_vpd) {
+            switch(op->vpd_pn) {
                 case 0xc0:
                     if (VPD_VP_SEAGATE == op->vend_prod_num)
                         decode_firm_vpd_c0_sea(rp, len);
@@ -1359,87 +1356,89 @@ svpd_decode_vendor(int sg_fd, struct opts_t * op, int off)
                         decode_rdac_vpd_c0(rp, len);
                     else if (VPD_VP_DDS == op->vend_prod_num)
                         decode_dds_vpd_c0(rp, len);
-                    else if (VPD_VP_LTO5 == op->vend_prod_num)
-                        decode_lto5_dcrl(rp, len);
-                    else if (VPD_VP_LTO6 == op->vend_prod_num)
-                        decode_lto6_vpd_cx(rp, len, op->num_vpd);
+                    else if (VPD_VP_IBM_LTO == op->vend_prod_num)
+                        decode_ibm_lto_dcrl(rp, len);
+                    else if (VPD_VP_HP_LTO == op->vend_prod_num)
+                        decode_hp_lto_vpd_cx(rp, len, op->vpd_pn);
                     else
-                        dStrHex((const char *)rp, len, 0);
+                        hex2stdout(rp, len, 0);
                     break;
                 case 0xc1:
                     if (VPD_VP_SEAGATE == op->vend_prod_num)
                         decode_date_code_vpd_c1_sea(rp, len);
                     else if (VPD_VP_RDAC == op->vend_prod_num)
                         decode_rdac_vpd_c1(rp, len);
-                    else if (VPD_VP_LTO5 == op->vend_prod_num)
-                        decode_lto5_dsn(rp, len);
-                    else if (VPD_VP_LTO6 == op->vend_prod_num)
-                        decode_lto6_vpd_cx(rp, len, op->num_vpd);
+                    else if (VPD_VP_IBM_LTO == op->vend_prod_num)
+                        decode_ibm_lto_dsn(rp, len);
+                    else if (VPD_VP_HP_LTO == op->vend_prod_num)
+                        decode_hp_lto_vpd_cx(rp, len, op->vpd_pn);
                     else
-                        dStrHex((const char *)rp, len, 0);
+                        hex2stdout(rp, len, 0);
                     break;
                 case 0xc2:
                     if (VPD_VP_RDAC == op->vend_prod_num)
                         decode_rdac_vpd_c2(rp, len);
-                    else if (VPD_VP_LTO6 == op->vend_prod_num)
-                        decode_lto6_vpd_cx(rp, len, op->num_vpd);
+                    else if (VPD_VP_HP_LTO == op->vend_prod_num)
+                        decode_hp_lto_vpd_cx(rp, len, op->vpd_pn);
                     else
-                        dStrHex((const char *)rp, len, 0);
+                        hex2stdout(rp, len, 0);
                     break;
                 case 0xc3:
                     if (VPD_VP_SEAGATE == op->vend_prod_num)
                         decode_dev_beh_vpd_c3_sea(rp, len);
                     else if (VPD_VP_RDAC == op->vend_prod_num)
                         decode_rdac_vpd_c3(rp, len);
-                    else if (VPD_VP_LTO6 == op->vend_prod_num)
-                        decode_lto6_vpd_cx(rp, len, op->num_vpd);
+                    else if (VPD_VP_HP_LTO == op->vend_prod_num)
+                        decode_hp_lto_vpd_cx(rp, len, op->vpd_pn);
                     else
-                        dStrHex((const char *)rp, len, 0);
+                        hex2stdout(rp, len, 0);
                     break;
                 case 0xc4:
                     if (VPD_VP_RDAC == op->vend_prod_num)
                         decode_rdac_vpd_c4(rp, len);
-                    else if (VPD_VP_LTO6 == op->vend_prod_num)
-                        decode_lto6_vpd_cx(rp, len, op->num_vpd);
+                    else if (VPD_VP_HP_LTO == op->vend_prod_num)
+                        decode_hp_lto_vpd_cx(rp, len, op->vpd_pn);
                     else
-                        dStrHex((const char *)rp, len, 0);
+                        hex2stdout(rp, len, 0);
                     break;
                 case 0xc5:
-                    if (VPD_VP_LTO6 == op->vend_prod_num)
-                        decode_lto6_vpd_cx(rp, len, op->num_vpd);
+                    if (VPD_VP_HP_LTO == op->vend_prod_num)
+                        decode_hp_lto_vpd_cx(rp, len, op->vpd_pn);
                     else
-                        dStrHex((const char *)rp, len, 0);
+                        hex2stdout(rp, len, 0);
                     break;
                 case 0xc8:
                     if (VPD_VP_RDAC == op->vend_prod_num)
                         decode_rdac_vpd_c8(rp, len);
                     else
-                        dStrHex((const char *)rp, len, 0);
+                        hex2stdout(rp, len, 0);
                     break;
                 case 0xc9:
                     if (VPD_VP_RDAC == op->vend_prod_num)
                         decode_rdac_vpd_c9(rp, len);
                     else
-                        dStrHex((const char *)rp, len, 0);
+                        hex2stdout(rp, len, 0);
                     break;
                 case 0xca:
                     if (VPD_VP_RDAC == op->vend_prod_num)
                         decode_rdac_vpd_ca(rp, len);
                     else
-                        dStrHex((const char *)rp, len, 0);
+                        hex2stdout(rp, len, 0);
                     break;
                 case 0xd0:
                     if (VPD_VP_RDAC == op->vend_prod_num)
                         decode_rdac_vpd_d0(rp, len);
                     else
-                        dStrHex((const char *)rp, len, 0);
+                        hex2stdout(rp, len, 0);
                     break;
                 default:
-                    return SG_LIB_SYNTAX_ERROR;
+                    pr2serr("%s: logic error, should know can't decode "
+                            "pn=0x%x\n", __func__, op->vpd_pn);
+                    return SG_LIB_CAT_OTHER;
             }
             return 0;
         }
     } else
-        pr2serr("Vendor VPD page=0x%x  failed to fetch", op->num_vpd);
+        pr2serr("Vendor VPD page=0x%x  failed to fetch", op->vpd_pn);
     return res;
 }
