@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2004-2016 Douglas Gilbert.
+ * Copyright (c) 2004-2018 Douglas Gilbert.
  * All rights reserved.
  * Use of this source code is governed by a BSD-style
  * license that can be found in the BSD_LICENSE file.
@@ -9,6 +9,8 @@
 #include <fcntl.h>
 #include <stdio.h>
 #include <stdlib.h>
+#include <stdarg.h>
+#include <stdbool.h>
 #include <string.h>
 #include <getopt.h>
 #include <ctype.h>
@@ -16,6 +18,7 @@
 #ifdef HAVE_CONFIG_H
 #include "config.h"
 #endif
+
 #include "sg_lib.h"
 #include "sg_cmds_basic.h"
 #include "sg_unaligned.h"
@@ -27,7 +30,7 @@
  * mode page on the given device.
  */
 
-static const char * version_str = "1.16 20160208";
+static const char * version_str = "1.22 20180219";
 
 #define ME "sg_wr_mode: "
 
@@ -38,16 +41,16 @@ static const char * version_str = "1.16 20160208";
 
 
 static struct option long_options[] = {
-        {"contents", 1, 0, 'c'},
-        {"dbd", 0, 0, 'd'},
-        {"force", 0, 0, 'f'},
-        {"help", 0, 0, 'h'},
-        {"len", 1, 0, 'l'},
-        {"mask", 1, 0, 'm'},
-        {"page", 1, 0, 'p'},
-        {"save", 0, 0, 's'},
-        {"verbose", 0, 0, 'v'},
-        {"version", 0, 0, 'V'},
+        {"contents", required_argument, 0, 'c'},
+        {"dbd", no_argument, 0, 'd'},
+        {"force", no_argument, 0, 'f'},
+        {"help", no_argument, 0, 'h'},
+        {"len", required_argument, 0, 'l'},
+        {"mask", required_argument, 0, 'm'},
+        {"page", required_argument, 0, 'p'},
+        {"save", no_argument, 0, 's'},
+        {"verbose", no_argument, 0, 'v'},
+        {"version", no_argument, 0, 'V'},
         {0, 0, 0, 0},
 };
 
@@ -95,7 +98,7 @@ static void usage()
  * quoted. For stdin (indicated by *inp=='-') there should be either
  * one entry per line, a comma separated list or space separated list.
  * Returns 0 if ok, or 1 if error. */
-static int build_mode_page(const char * inp, unsigned char * mp_arr,
+static int build_mode_page(const char * inp, uint8_t * mp_arr,
                            int * mp_arr_len, int max_arr_len)
 {
     int in_len, k, j, m;
@@ -112,10 +115,10 @@ static int build_mode_page(const char * inp, unsigned char * mp_arr,
     if (0 == in_len)
         *mp_arr_len = 0;
     if ('-' == inp[0]) {        /* read from stdin */
-        char line[512];
-        char carry_over[4];
+        bool split_line;
         int off = 0;
-        int split_line;
+        char carry_over[4];
+        char line[512];
 
         carry_over[0] = 0;
         for (j = 0; j < 512; ++j) {
@@ -126,9 +129,9 @@ static int build_mode_page(const char * inp, unsigned char * mp_arr,
                 if ('\n' == line[in_len - 1]) {
                     --in_len;
                     line[in_len] = '\0';
-                    split_line = 0;
+                    split_line = false;
                 } else
-                    split_line = 1;
+                    split_line = true;
             }
             if (in_len < 1) {
                 carry_over[0] = 0;
@@ -242,7 +245,7 @@ static int build_mode_page(const char * inp, unsigned char * mp_arr,
 /* Read hex numbers from command line (comma separated list).
  * Can also be (single) space separated list but needs to be quoted on the
  * command line. Returns 0 if ok, or 1 if error. */
-static int build_mask(const char * inp, unsigned char * mask_arr,
+static int build_mask(const char * inp, uint8_t * mask_arr,
                       int * mask_arr_len, int max_arr_len)
 {
     int in_len, k;
@@ -302,24 +305,25 @@ static int build_mask(const char * inp, unsigned char * mask_arr,
 
 int main(int argc, char * argv[])
 {
+    bool dbd = false;
+    bool force = false;
+    bool got_contents = false;
+    bool got_mask = false;
+    bool mode_6 = false;
+    bool save = false;
     int sg_fd, res, c, num, alloc_len, off, pdt;
     int k, md_len, hdr_len, bd_len, mask_in_len;
-    unsigned u, uu;
-    int dbd = 0;
-    int got_contents = 0;
-    int force = 0;
-    int got_mask = 0;
-    int mode_6 = 0;
     int pg_code = -1;
     int sub_pg_code = 0;
-    int save = 0;
     int verbose = 0;
     int read_in_len = 0;
+    unsigned u, uu;
     const char * device_name = NULL;
-    unsigned char read_in[MX_ALLOC_LEN];
-    unsigned char mask_in[MX_ALLOC_LEN];
-    unsigned char ref_md[MX_ALLOC_LEN];
+    uint8_t read_in[MX_ALLOC_LEN];
+    uint8_t mask_in[MX_ALLOC_LEN];
+    uint8_t ref_md[MX_ALLOC_LEN];
     char ebuff[EBUFF_SZ];
+    char errStr[128];
     char b[80];
     struct sg_simple_inquiry_resp inq_data;
     int ret = 0;
@@ -340,13 +344,13 @@ int main(int argc, char * argv[])
                 pr2serr("bad argument to '--contents'\n");
                 return SG_LIB_SYNTAX_ERROR;
             }
-            got_contents = 1;
+            got_contents = true;
             break;
         case 'd':
-            dbd = 1;
+            dbd = true;
             break;
         case 'f':
-            force = 1;
+            force = true;
             break;
         case 'h':
         case '?':
@@ -355,7 +359,7 @@ int main(int argc, char * argv[])
         case 'l':
             num = sscanf(optarg, "%d", &res);
             if ((1 == num) && ((6 == res) || (10 == res)))
-                mode_6 = (6 == res) ? 1 : 0;
+                mode_6 = (6 == res);
             else {
                 pr2serr("length (of cdb) must be 6 or 10\n");
                 return SG_LIB_SYNTAX_ERROR;
@@ -368,7 +372,7 @@ int main(int argc, char * argv[])
                 pr2serr("bad argument to '--mask'\n");
                 return SG_LIB_SYNTAX_ERROR;
             }
-            got_mask = 1;
+            got_mask = true;
             break;
         case 'p':
            if (NULL == strchr(optarg, ',')) {
@@ -394,7 +398,7 @@ int main(int argc, char * argv[])
             }
             break;
         case 's':
-            save = 1;
+            save = true;
             break;
         case 'v':
             ++verbose;
@@ -436,7 +440,7 @@ int main(int argc, char * argv[])
         return SG_LIB_SYNTAX_ERROR;
     }
 
-    sg_fd = sg_cmds_open_device(device_name, 0 /* rw */, verbose);
+    sg_fd = sg_cmds_open_device(device_name, false /* rw */, verbose);
     if (sg_fd < 0) {
         pr2serr(ME "open error: %s: %s\n", device_name, safe_strerror(-sg_fd));
         return SG_LIB_FILE_ERROR;
@@ -448,39 +452,38 @@ int main(int argc, char * argv[])
 
     /* do MODE SENSE to fetch current values */
     memset(ref_md, 0, MX_ALLOC_LEN);
+    snprintf(errStr, sizeof(errStr), "MODE SENSE (%d): ", mode_6 ? 6 : 10);
     alloc_len = mode_6 ? SHORT_ALLOC_LEN : MX_ALLOC_LEN;
     if (mode_6)
         res = sg_ll_mode_sense6(sg_fd, dbd, 0 /*current */, pg_code,
-                                sub_pg_code, ref_md, alloc_len, 1, verbose);
+                                sub_pg_code, ref_md, alloc_len, true,
+                                verbose);
      else
-        res = sg_ll_mode_sense10(sg_fd, 0 /* llbaa */, dbd, 0 /* current */,
-                                 pg_code, sub_pg_code, ref_md, alloc_len, 1,
-                                 verbose);
+        res = sg_ll_mode_sense10(sg_fd, false /* llbaa */, dbd,
+                                 0 /* current */, pg_code, sub_pg_code,
+                                 ref_md, alloc_len, true, verbose);
     ret = res;
     if (res) {
         if (SG_LIB_CAT_INVALID_OP == res)
-            pr2serr("MODE SENSE (%d) not supported, try '--len=%d'\n",
-                    (mode_6 ? 6 : 10), (mode_6 ? 10 : 6));
+            pr2serr("%snot supported, try '--len=%d'\n", errStr,
+                    (mode_6 ? 10 : 6));
         else {
             sg_get_category_sense_str(res, sizeof(b), b, verbose);
-            pr2serr("MODE SENSE (%d): %s\n", (mode_6 ? 6 : 10), b);
+            pr2serr("%s%s\n", errStr, b);
         }
         goto err_out;
     }
     off = sg_mode_page_offset(ref_md, alloc_len, mode_6, ebuff, EBUFF_SZ);
     if (off < 0) {
-        pr2serr("MODE SENSE (%d): %s\n", (mode_6 ? 6 : 10), ebuff);
+        pr2serr("%s%s\n", errStr, ebuff);
         goto err_out;
     }
-    if (mode_6) {
-        hdr_len = 4;
-        md_len = ref_md[0] + 1;
-        bd_len = ref_md[3];
-    } else {
-        hdr_len = 8;
-        md_len = sg_get_unaligned_be16(ref_md + 0) + 2;
-        bd_len = sg_get_unaligned_be16(ref_md + 6);
+    md_len = sg_msense_calc_length(ref_md, alloc_len, mode_6, &bd_len);
+    if (md_len < 0) {
+        pr2serr("%ssg_msense_calc_length() failed\n", errStr);
+        goto err_out;
     }
+    hdr_len = mode_6 ? 4 : 8;
     if (got_contents) {
         if (read_in_len < 2) {
             pr2serr("contents length=%d too short\n", read_in_len);
@@ -539,11 +542,11 @@ int main(int argc, char * argv[])
 
         memcpy(ref_md + off, read_in, read_in_len);
         if (mode_6)
-            res = sg_ll_mode_select6(sg_fd, 1 /* PF */, save, ref_md, md_len,
-                                     1, verbose);
+            res = sg_ll_mode_select6(sg_fd, true /* PF */, save, ref_md,
+                                     md_len, true, verbose);
         else
-            res = sg_ll_mode_select10(sg_fd, 1 /* PF */, save, ref_md,
-                                      md_len, 1, verbose);
+            res = sg_ll_mode_select10(sg_fd, true /* PF */, save, ref_md,
+                                      md_len, true, verbose);
         ret = res;
         if (res) {
             sg_get_category_sense_str(res, sizeof(b), b, verbose);
@@ -553,14 +556,14 @@ int main(int argc, char * argv[])
     } else {
         printf(">>> No contents given, so show current mode page data:\n");
         printf("  header:\n");
-        dStrHex((const char *)ref_md, hdr_len, -1);
+        hex2stdout(ref_md, hdr_len, -1);
         if (bd_len) {
             printf("  block descriptor(s):\n");
-            dStrHex((const char *)(ref_md + hdr_len), bd_len, -1);
+            hex2stdout(ref_md + hdr_len, bd_len, -1);
         } else
             printf("  << no block descriptors >>\n");
         printf("  mode page:\n");
-        dStrHex((const char *)(ref_md + off), md_len - off, -1);
+        hex2stdout(ref_md + off, md_len - off, -1);
     }
 err_out:
     res = sg_cmds_close_device(sg_fd);
