@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2006-2015 Luben Tuikov and Douglas Gilbert.
+ * Copyright (c) 2006-2018 Luben Tuikov and Douglas Gilbert.
  * All rights reserved.
  * Use of this source code is governed by a BSD-style
  * license that can be found in the BSD_LICENSE file.
@@ -10,6 +10,8 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <stdarg.h>
+#include <stdint.h>
+#include <stdbool.h>
 #include <ctype.h>
 #include <string.h>
 #include <getopt.h>
@@ -22,14 +24,20 @@
 #include "sg_lib.h"
 #include "sg_cmds_basic.h"
 #include "sg_cmds_extra.h"
-#include "sg_pt.h"      /* needed for scsi_pt_win32_direct() */
 #include "sg_unaligned.h"
+#include "sg_pr2serr.h"
+
+#ifdef SG_LIB_WIN32
+#ifdef SG_LIB_WIN32_DIRECT
+#include "sg_pt.h"      /* needed for scsi_pt_win32_direct() */
+#endif
+#endif
 
 /*
  * This utility issues the SCSI WRITE BUFFER command to the given device.
  */
 
-static const char * version_str = "1.20 20150217";    /* spc5r02 */
+static const char * version_str = "1.25 20180217";    /* spc5r19 */
 
 #define ME "sg_write_buffer: "
 #define DEF_XFER_LEN (8 * 1024 * 1024)
@@ -42,12 +50,16 @@ static const char * version_str = "1.20 20150217";    /* spc5r02 */
 
 static struct option long_options[] = {
         {"bpw", required_argument, 0, 'b'},
+        {"dry-run", no_argument, 0, 'd'},
+        {"dry_run", no_argument, 0, 'd'},
         {"help", no_argument, 0, 'h'},
         {"id", required_argument, 0, 'i'},
         {"in", required_argument, 0, 'I'},
         {"length", required_argument, 0, 'l'},
         {"mode", required_argument, 0, 'm'},
         {"offset", required_argument, 0, 'o'},
+        {"read-stdin", no_argument, 0, 'r'},
+        {"read_stdin", no_argument, 0, 'r'},
         {"raw", no_argument, 0, 'r'},
         {"skip", required_argument, 0, 's'},
         {"specific", required_argument, 0, 'S'},
@@ -57,42 +69,26 @@ static struct option long_options[] = {
         {0, 0, 0, 0},
 };
 
-#ifdef __GNUC__
-static int pr2serr(const char * fmt, ...)
-        __attribute__ ((format (printf, 1, 2)));
-#else
-static int pr2serr(const char * fmt, ...);
-#endif
-
-
-static int
-pr2serr(const char * fmt, ...)
-{
-    va_list args;
-    int n;
-
-    va_start(args, fmt);
-    n = vfprintf(stderr, fmt, args);
-    va_end(args);
-    return n;
-}
-
 
 static void
 usage()
 {
     pr2serr("Usage: "
-            "sg_write_buffer [--bpw=CS] [--help] [--id=ID] [--in=FILE]\n"
+            "sg_write_buffer [--bpw=CS] [--dry-run] [--help] [--id=ID] "
+            "[--in=FILE]\n"
             "                       [--length=LEN] [--mode=MO] "
-            "[--offset=OFF] [--raw]\n"
-            "                       [--skip=SKIP] [--specific=MS] "
-            "[--timeout=TO]\n"
-            "                       [--verbose] [--version] DEVICE\n"
+            "[--offset=OFF]\n"
+            "                       [--read-stdin] [--skip=SKIP] "
+            "[--specific=MS]\n"
+            "                       [--timeout=TO] [--verbose] [--version] "
+            "DEVICE\n"
             "  where:\n"
             "    --bpw=CS|-b CS         CS is chunk size: bytes per write "
             "buffer\n"
             "                           command (def: 0 -> as many as "
             "possible)\n"
+            "    --dry-run|-d           skip WRITE BUFFER commands, do "
+            "everything else\n"
             "    --help|-h              print out usage message then exit\n"
             "    --id=ID|-i ID          buffer identifier (0 (default) to "
             "255)\n"
@@ -106,7 +102,7 @@ usage()
             "                           (def: 0 -> 'combined header and "
             "data' (obs))\n"
             "    --offset=OFF|-o OFF    buffer offset (unit: bytes, def: 0)\n"
-            "    --raw|-r               read from stdin (same as '-I -')\n"
+            "    --read-stdin|-r        read from stdin (same as '-I -')\n"
             "    --skip=SKIP|-s SKIP    bytes in file FILE to skip before "
             "reading\n"
             "    --specific=MS|-S MS    mode specific value; 3 bit field "
@@ -189,108 +185,38 @@ print_modes(void)
             "dmc_offs_ev_defer mode downloads.\n");
 }
 
-/* <<<< This function may be moved to the library in the future >>> */
-/* Invokes a SCSI WRITE BUFFER command (SPC). Return of 0 ->
- * success, SG_LIB_CAT_INVALID_OP -> invalid opcode,
- * SG_LIB_CAT_ILLEGAL_REQ -> bad field in cdb, SG_LIB_CAT_UNIT_ATTENTION,
- * SG_LIB_CAT_NOT_READY -> device not ready, SG_LIB_CAT_ABORTED_COMMAND,
- * -1 -> other failure */
-static int
-sg_ll_write_buffer_v2(int sg_fd, int mode, int m_specific, int buffer_id,
-                      uint32_t buffer_offset, void * paramp,
-                      uint32_t param_len, int to_secs, int noisy, int verbose)
-{
-    int k, res, ret, sense_cat;
-    uint8_t wbufCmdBlk[WRITE_BUFFER_CMDLEN] =
-        {WRITE_BUFFER_CMD, 0, 0, 0, 0, 0, 0, 0, 0, 0};
-    uint8_t sense_b[SENSE_BUFF_LEN];
-    struct sg_pt_base * ptvp;
-
-    if (buffer_offset > 0xffffff) {
-        pr2serr("%s: buffer_offset value too large for 24 bits\n", __func__);
-        return -1;
-    }
-    if (param_len > 0xffffff) {
-        pr2serr("%s: param_len value too large for 24 bits\n", __func__);
-        return -1;
-    }
-    wbufCmdBlk[1] = (uint8_t)(mode & 0x1f);
-    wbufCmdBlk[1] |= (uint8_t)((m_specific & 0x7) << 5);
-    wbufCmdBlk[2] = (uint8_t)(buffer_id & 0xff);
-    sg_put_unaligned_be24(buffer_offset, wbufCmdBlk + 3);
-    sg_put_unaligned_be24(param_len, wbufCmdBlk + 6);
-    if (verbose) {
-        pr2serr("    Write buffer cmd: ");
-        for (k = 0; k < WRITE_BUFFER_CMDLEN; ++k)
-            pr2serr("%02x ", wbufCmdBlk[k]);
-        pr2serr("\n");
-        if ((verbose > 1) && paramp && param_len) {
-            pr2serr("    Write buffer parameter list%s:\n",
-                    ((param_len > 256) ? " (first 256 bytes)" : ""));
-            dStrHexErr((const char *)paramp,
-                       ((param_len > 256) ? 256 : param_len), -1);
-        }
-    }
-
-    ptvp = construct_scsi_pt_obj();
-    if (NULL == ptvp) {
-        pr2serr("%s: out of memory\n", __func__);
-        return -1;
-    }
-    set_scsi_pt_cdb(ptvp, wbufCmdBlk, sizeof(wbufCmdBlk));
-    set_scsi_pt_sense(ptvp, sense_b, sizeof(sense_b));
-    set_scsi_pt_data_out(ptvp, (uint8_t *)paramp, param_len);
-    res = do_scsi_pt(ptvp, sg_fd, to_secs, verbose);
-    ret = sg_cmds_process_resp(ptvp, "Write buffer", res, 0, sense_b,
-                               noisy, verbose, &sense_cat);
-    if (-1 == ret)
-        ;
-    else if (-2 == ret) {
-        switch (sense_cat) {
-        case SG_LIB_CAT_RECOVERED:
-        case SG_LIB_CAT_NO_SENSE:
-            ret = 0;
-            break;
-        default:
-            ret = sense_cat;
-            break;
-        }
-    } else
-        ret = 0;
-
-    destruct_scsi_pt_obj(ptvp);
-    return ret;
-}
-
 
 int
 main(int argc, char * argv[])
 {
-    int sg_fd, infd, res, c, len, k, n, got_stdin;
+    bool bpw_then_activate = false;
+    bool dry_run = false;
+    bool got_stdin = false;
+    bool wb_len_given = false;
+    int sg_fd, infd, res, c, len, k, n;
     int bpw = 0;
-    int bpw_then_activate = 0;
     int do_help = 0;
+    int ret = 0;
+    int verbose = 0;
     int wb_id = 0;
     int wb_len = 0;
-    int wb_len_given = 0;
     int wb_mode = 0;
     int wb_offset = 0;
     int wb_skip = 0;
     int wb_timeout = DEF_PT_TIMEOUT;
     int wb_mspec = 0;
-    int verbose = 0;
     const char * device_name = NULL;
     const char * file_name = NULL;
-    unsigned char * dop = NULL;
+    uint8_t * dop = NULL;
+    uint8_t * free_dop = NULL;
     char * cp;
     const struct mode_s * mp;
     char ebuff[EBUFF_SZ];
-    int ret = 0;
 
     while (1) {
         int option_index = 0;
 
-        c = getopt_long(argc, argv, "b:hi:I:l:m:o:rs:S:t:vV", long_options,
+        c = getopt_long(argc, argv, "b:dhi:I:l:m:o:rs:S:t:vV", long_options,
                         &option_index);
         if (c == -1)
             break;
@@ -305,9 +231,12 @@ main(int argc, char * argv[])
             }
             if ((cp = strchr(optarg, ','))) {
                 if (0 == strncmp("act", cp + 1, 3))
-                    ++bpw_then_activate;
+                    bpw_then_activate = true;
             }
             break;
+        case 'd':
+            dry_run = true;
+            break;
         case 'h':
         case '?':
             ++do_help;
@@ -329,7 +258,7 @@ main(int argc, char * argv[])
                 pr2serr("bad argument to '--length'\n");
                 return SG_LIB_SYNTAX_ERROR;
              }
-             wb_len_given = 1;
+             wb_len_given = true;
              break;
         case 'm':
             if (isdigit(*optarg)) {
@@ -360,7 +289,7 @@ main(int argc, char * argv[])
                 return SG_LIB_SYNTAX_ERROR;
             }
             break;
-        case 'r':
+        case 'r':       /* --read-stdin and --raw (previous name) */
             file_name = "-";
             break;
         case 's':
@@ -438,7 +367,7 @@ main(int argc, char * argv[])
 #endif
 #endif
 
-    sg_fd = sg_cmds_open_device(device_name, 0 /* rw */, verbose);
+    sg_fd = sg_cmds_open_device(device_name, false /* rw */, verbose);
     if (sg_fd < 0) {
         pr2serr(ME "open error: %s: %s\n", device_name,
                 safe_strerror(-sg_fd));
@@ -447,14 +376,15 @@ main(int argc, char * argv[])
     if (file_name || (wb_len > 0)) {
         if (0 == wb_len)
             wb_len = DEF_XFER_LEN;
-        if (NULL == (dop = (unsigned char *)malloc(wb_len))) {
+        dop = sg_memalign(wb_len, 0, &free_dop, false);
+        if (NULL == dop) {
             pr2serr(ME "out of memory\n");
             ret = SG_LIB_SYNTAX_ERROR;
             goto err_out;
         }
         memset(dop, 0xff, wb_len);
         if (file_name) {
-            got_stdin = (0 == strcmp(file_name, "-")) ? 1 : 0;
+            got_stdin = (0 == strcmp(file_name, "-"));
             if (got_stdin) {
                 if (wb_skip > 0) {
                     pr2serr("Can't skip on stdin\n");
@@ -526,26 +456,47 @@ main(int argc, char * argv[])
                 pr2serr("sending write buffer, mode=0x%x, mspec=%d, id=%d, "
                         " offset=%d, len=%d\n", wb_mode, wb_mspec, wb_id,
                         wb_offset + k, n);
-            res = sg_ll_write_buffer_v2(sg_fd, wb_mode, wb_mspec, wb_id,
-                                        wb_offset + k, dop + k, n,
-                                        wb_timeout, 1, verbose);
+            if (dry_run) {
+                if (verbose)
+                    pr2serr("skipping WRITE BUFFER command due to "
+                            "--dry-run\n");
+                res = 0;
+            } else
+                res = sg_ll_write_buffer_v2(sg_fd, wb_mode, wb_mspec, wb_id,
+                                            wb_offset + k, dop + k, n,
+                                            wb_timeout, true, verbose);
             if (res)
                 break;
         }
         if (bpw_then_activate) {
             if (verbose)
                 pr2serr("sending Activate deferred microcode [0xf]\n");
-            res = sg_ll_write_buffer_v2(sg_fd, MODE_ACTIVATE_MC, 0, 0, 0,
-                                        NULL, 0, wb_timeout, 1, verbose);
+            if (dry_run) {
+                if (verbose)
+                    pr2serr("skipping WRITE BUFFER(ACTIVATE) command due to "
+                            "--dry-run\n");
+                res = 0;
+            } else
+                res = sg_ll_write_buffer_v2(sg_fd, MODE_ACTIVATE_MC,
+                                            0 /* buffer_id */,
+                                            0 /* buffer_offset */, 0,
+                                            NULL, 0, wb_timeout, true,
+                                            verbose);
         }
     } else {
         if (verbose)
             pr2serr("sending single write buffer, mode=0x%x, mpsec=%d, "
                     "id=%d, offset=%d, len=%d\n", wb_mode, wb_mspec, wb_id,
                     wb_offset, wb_len);
-        res = sg_ll_write_buffer_v2(sg_fd, wb_mode, wb_mspec, wb_id,
-                                    wb_offset, dop, wb_len, wb_timeout, 1,
-                                    verbose);
+        if (dry_run) {
+            if (verbose)
+                pr2serr("skipping WRITE BUFFER(all in one) command due to "
+                        "--dry-run\n");
+            res = 0;
+        } else
+            res = sg_ll_write_buffer_v2(sg_fd, wb_mode, wb_mspec, wb_id,
+                                        wb_offset, dop, wb_len, wb_timeout,
+                                        true, verbose);
     }
     if (0 != res) {
         char b[80];
@@ -556,8 +507,8 @@ main(int argc, char * argv[])
     }
 
 err_out:
-    if (dop)
-        free(dop);
+    if (free_dop)
+        free(free_dop);
     res = sg_cmds_close_device(sg_fd);
     if (res < 0) {
         pr2serr("close error: %s\n", safe_strerror(-res));
