@@ -1,5 +1,5 @@
 /* A utility program for the Linux OS SCSI subsystem.
- *  Copyright (C) 2004-2016 D. Gilbert
+ *  Copyright (C) 2004-2018 D. Gilbert
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
  *  the Free Software Foundation; either version 2, or (at your option)
@@ -8,7 +8,7 @@
  * This program issues the SCSI command WRITE LONG to a given SCSI device.
  * It sends the command with the logical block address passed as the lba
  * argument, and the transfer length set to the xfer_len argument. the
- * buffer to be writen to the device filled with 0xff, this buffer includes
+ * buffer to be written to the device filled with 0xff, this buffer includes
  * the sector data and the ECC bytes.
  *
  * This code was contributed by Saeed Bishara
@@ -18,6 +18,8 @@
 #include <fcntl.h>
 #include <stdio.h>
 #include <stdlib.h>
+#include <stdarg.h>
+#include <stdbool.h>
 #include <string.h>
 #include <errno.h>
 #include <getopt.h>
@@ -27,12 +29,13 @@
 #ifdef HAVE_CONFIG_H
 #include "config.h"
 #endif
+
 #include "sg_lib.h"
 #include "sg_cmds_basic.h"
 #include "sg_cmds_extra.h"
 #include "sg_pr2serr.h"
 
-static const char * version_str = "1.12 20160121";
+static const char * version_str = "1.16 20180219";
 
 
 #define MAX_XFER_LEN 10000
@@ -44,16 +47,16 @@ static const char * version_str = "1.12 20160121";
 #define EBUFF_SZ 256
 
 static struct option long_options[] = {
-        {"16", 0, 0, 'S'},
-        {"cor_dis", 0, 0, 'c'},
-        {"help", 0, 0, 'h'},
-        {"in", 1, 0, 'i'},
-        {"lba", 1, 0, 'l'},
-        {"pblock", 0, 0, 'p'},
-        {"verbose", 0, 0, 'v'},
-        {"version", 0, 0, 'V'},
-        {"wr_uncor", 0, 0, 'w'},
-        {"xfer_len", 1, 0, 'x'},
+        {"16", no_argument, 0, 'S'},
+        {"cor_dis", no_argument, 0, 'c'},
+        {"help", no_argument, 0, 'h'},
+        {"in", required_argument, 0, 'i'},
+        {"lba", required_argument, 0, 'l'},
+        {"pblock", no_argument, 0, 'p'},
+        {"verbose", no_argument, 0, 'v'},
+        {"version", no_argument, 0, 'V'},
+        {"wr_uncor", no_argument, 0, 'w'},
+        {"xfer_len", required_argument, 0, 'x'},
         {0, 0, 0, 0},
 };
 
@@ -94,24 +97,24 @@ usage()
 int
 main(int argc, char * argv[])
 {
+    bool do_16 = false;
+    bool cor_dis = false;
+    bool got_stdin;
+    bool pblock = false;
+    bool wr_uncor = false;
     int sg_fd, res, c, infd, offset;
-    unsigned char * writeLongBuff = NULL;
-    void * rawp = NULL;
     int xfer_len = 520;
-    int cor_dis = 0;
-    int pblock = 0;
-    int wr_uncor = 0;
-    int do_16 = 0;
-    uint64_t llba = 0;
+    int ret = 1;
     int verbose = 0;
     int64_t ll;
-    int got_stdin;
+    uint64_t llba = 0;
     const char * device_name = NULL;
+    uint8_t * writeLongBuff = NULL;
+    void * rawp = NULL;
+    const char * ten_or;
     char file_name[256];
     char b[80];
     char ebuff[EBUFF_SZ];
-    const char * ten_or;
-    int ret = 1;
 
     memset(file_name, 0, sizeof file_name);
     while (1) {
@@ -124,7 +127,7 @@ main(int argc, char * argv[])
 
         switch (c) {
         case 'c':
-            cor_dis = 1;
+            cor_dis = true;
             break;
         case 'h':
         case '?':
@@ -142,10 +145,10 @@ main(int argc, char * argv[])
             llba = (uint64_t)ll;
             break;
         case 'p':
-            pblock = 1;
+            pblock = true;
             break;
         case 'S':
-            do_16 = 1;
+            do_16 = true;
             break;
         case 'v':
             ++verbose;
@@ -154,7 +157,7 @@ main(int argc, char * argv[])
             pr2serr(ME "version: %s\n", version_str);
             return 0;
         case 'w':
-            wr_uncor = 1;
+            wr_uncor = true;
             break;
         case 'x':
             xfer_len = sg_get_num(optarg);
@@ -195,7 +198,7 @@ main(int argc, char * argv[])
         usage();
         return SG_LIB_SYNTAX_ERROR;
     }
-    sg_fd = sg_cmds_open_device(device_name, 0 /* rw */, verbose);
+    sg_fd = sg_cmds_open_device(device_name, false /* rw */, verbose);
     if (sg_fd < 0) {
         pr2serr(ME "open error: %s: %s\n", device_name, safe_strerror(-sg_fd));
         return SG_LIB_FILE_ERROR;
@@ -211,10 +214,10 @@ main(int argc, char * argv[])
             ret = SG_LIB_FILE_ERROR;
             goto err_out;
         }
-        writeLongBuff = (unsigned char *)rawp;
+        writeLongBuff = (uint8_t *)rawp;
         memset(rawp, 0xff, MAX_XFER_LEN);
         if (file_name[0]) {
-            got_stdin = (0 == strcmp(file_name, "-")) ? 1 : 0;
+            got_stdin = (0 == strcmp(file_name, "-"));
             if (got_stdin) {
                 infd = STDIN_FILENO;
                 if (sg_set_binary_mode(STDIN_FILENO) < 0)
@@ -250,16 +253,17 @@ main(int argc, char * argv[])
         pr2serr(ME "issue write long to device %s\n\t\txfer_len= %d (0x%x), "
                 "lba=%" PRIu64 " (0x%" PRIx64 ")\n    cor_dis=%d, "
                 "wr_uncor=%d, pblock=%d\n", device_name, xfer_len, xfer_len,
-                llba, llba, cor_dis, wr_uncor, pblock);
+                llba, llba, (int)cor_dis, (int)wr_uncor, (int)pblock);
 
     ten_or = do_16 ? "16" : "10";
     if (do_16)
         res = sg_ll_write_long16(sg_fd, cor_dis, wr_uncor, pblock, llba,
-                                 writeLongBuff, xfer_len, &offset, 1, verbose);
+                                 writeLongBuff, xfer_len, &offset, true,
+                                 verbose);
     else
         res = sg_ll_write_long10(sg_fd, cor_dis, wr_uncor, pblock,
                                  (unsigned int)llba, writeLongBuff, xfer_len,
-                                 &offset, 1, verbose);
+                                 &offset, true, verbose);
     ret = res;
     switch (res) {
     case 0:
