@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2014 Douglas Gilbert
+ * Copyright (c) 2014-2018 Douglas Gilbert
  * All rights reserved.
  * Use of this source code is governed by a BSD-style
  * license that can be found in the BSD_LICENSE file.
@@ -17,6 +17,8 @@
 #include <fcntl.h>
 #include <stdio.h>
 #include <stdlib.h>
+#include <stdarg.h>
+#include <stdbool.h>
 #include <string.h>
 #include <errno.h>
 #include <limits.h>
@@ -29,11 +31,14 @@
 #ifdef HAVE_CONFIG_H
 #include "config.h"
 #endif
+
 #include "sg_lib.h"
 #include "sg_pt.h"
 #include "sg_cmds_basic.h"
+#include "sg_unaligned.h"
+#include "sg_pr2serr.h"
 
-static const char * version_str = "1.06 20141217";
+static const char * version_str = "1.13 20180219";
 
 
 #define ME "sg_write_verify: "
@@ -73,8 +78,7 @@ static struct option long_options[] = {
 static void
 usage()
 {
-    fprintf(stderr, "Usage: "
-            "sg_write_verify [--16] [--bytchk=BC] [--dpo] [--group=GN] "
+    pr2serr("Usage: sg_write_verify [--16] [--bytchk=BC] [--dpo] [--group=GN] "
             "[--help]\n"
             "                       [--ilen=IL] [--in=IF] --lba=LBA "
             "[--num=NUM]\n"
@@ -117,39 +121,39 @@ usage()
 /* Invokes a SCSI WRITE AND VERIFY according with CDB. Returns 0 -> success,
  * various SG_LIB_CAT_* positive values or -1 -> other errors */
 static int
-run_scsi_transaction(int sg_fd, const unsigned char *cdbp, int cdb_len,
-                     unsigned char *dop, int do_len, int timeout, int verbose)
+run_scsi_transaction(int sg_fd, const uint8_t *cdbp, int cdb_len,
+                     uint8_t *dop, int do_len, int timeout,
+                     bool noisy, int verbose)
 {
     int res, k, sense_cat, ret;
-    unsigned char sense_b[SENSE_BUFF_LEN];
-    int noisy = 1;
     struct sg_pt_base * ptvp;
+    uint8_t sense_b[SENSE_BUFF_LEN];
     char b[32];
 
     snprintf(b, sizeof(b), "Write and verify(%d)", cdb_len);
     if (verbose) {
-       fprintf(stderr, "    %s cmd: ", b);
+       pr2serr("    %s cdb: ", b);
        for (k = 0; k < cdb_len; ++k)
-           fprintf(stderr, "%02x ", cdbp[k]);
-       fprintf(stderr, "\n");
+           pr2serr("%02x ", cdbp[k]);
+       pr2serr("\n");
        if ((verbose > 2) && dop && do_len) {
-            fprintf(stderr, "    Data out buffer [%d bytes]:\n", do_len);
-            dStrHexErr((const char *)dop, do_len, -1);
+            pr2serr("    Data out buffer [%d bytes]:\n", do_len);
+            hex2stderr(dop, do_len, -1);
         }
     }
     ptvp = construct_scsi_pt_obj();
     if (NULL == ptvp) {
-        fprintf(stderr, "%s: out of memory\n", b);
+        pr2serr("%s: out of memory\n", b);
         return -1;
     }
     set_scsi_pt_cdb(ptvp, cdbp, cdb_len);
     set_scsi_pt_sense(ptvp, sense_b, sizeof(sense_b));
     set_scsi_pt_data_out(ptvp, dop, do_len);
     res = do_scsi_pt(ptvp, sg_fd, timeout, verbose);
-    ret = sg_cmds_process_resp(ptvp, b, res, 0, sense_b, noisy, verbose,
-                               &sense_cat);
+    ret = sg_cmds_process_resp(ptvp, b, res, SG_NO_DATA_IN, sense_b, noisy,
+                               verbose, &sense_cat);
     if (-1 == ret)
-        ;
+        ret = sg_convert_errno(get_scsi_pt_os_err(ptvp));
     else if (-2 == ret) {
         switch (sense_cat) {
         case SG_LIB_CAT_RECOVERED:
@@ -158,14 +162,15 @@ run_scsi_transaction(int sg_fd, const unsigned char *cdbp, int cdb_len,
             break;
         case SG_LIB_CAT_MEDIUM_HARD:    /* write or verify failed */
             {
-                int valid, slen;
+                bool valid;
+                int slen;
                 uint64_t ull = 0;
 
                 slen = get_scsi_pt_sense_len(ptvp);
                 valid = sg_get_sense_info_fld(sense_b, slen, &ull);
                 if (valid)
-                    fprintf(stderr, "Medium or hardware error starting at "
-                            "lba=%" PRIu64 " [0x%" PRIx64 "]\n", ull, ull);
+                    pr2serr("Medium or hardware error starting at lba=%"
+                            PRIu64 " [0x%" PRIx64 "]\n", ull, ull);
             }
             ret = sense_cat;
             break;
@@ -185,12 +190,13 @@ run_scsi_transaction(int sg_fd, const unsigned char *cdbp, int cdb_len,
 /* Invokes a SCSI WRITE AND VERIFY (10) command (SBC). Returns 0 -> success,
 * various SG_LIB_CAT_* positive values or -1 -> other errors */
 static int
-sg_ll_write_verify10(int sg_fd, int wrprotect, int dpo, int bytchk,
+sg_ll_write_verify10(int sg_fd, int wrprotect, bool dpo, int bytchk,
                      unsigned int lba, int num_lb, int group,
-                     unsigned char *dop, int do_len, int timeout, int verbose)
+                     uint8_t *dop, int do_len, int timeout,
+                     bool noisy, int verbose)
 {
     int ret;
-    unsigned char wv_cdb[WRITE_VERIFY10_CMDLEN];
+    uint8_t wv_cdb[WRITE_VERIFY10_CMDLEN];
 
     memset(wv_cdb, 0, WRITE_VERIFY10_CMDLEN);
     wv_cdb[0] = WRITE_VERIFY10_CMD;
@@ -200,27 +206,23 @@ sg_ll_write_verify10(int sg_fd, int wrprotect, int dpo, int bytchk,
     if (bytchk)
        wv_cdb[1] |= ((bytchk & 0x3) << 1);
 
-    wv_cdb[2] = (lba >> 24) & 0xff;
-    wv_cdb[3] = (lba >> 16) & 0xff;
-    wv_cdb[4] = (lba >> 8) & 0xff;
-    wv_cdb[5] = lba & 0xff;
+    sg_put_unaligned_be32((uint32_t)lba, wv_cdb + 2);
     wv_cdb[6] = group & 0x1f;
-    wv_cdb[7] = (num_lb >> 8) & 0xff;
-    wv_cdb[8] = num_lb & 0xff;
+    sg_put_unaligned_be16((uint16_t)num_lb, wv_cdb + 7);
     ret = run_scsi_transaction(sg_fd, wv_cdb, sizeof(wv_cdb), dop, do_len,
-                               timeout, verbose);
+                               timeout, noisy, verbose);
     return ret;
 }
 
 /* Invokes a SCSI WRITE AND VERIFY (16) command (SBC). Returns 0 -> success,
 * various SG_LIB_CAT_* positive values or -1 -> other errors */
 static int
-sg_ll_write_verify16(int sg_fd, int wrprotect, int dpo, int bytchk,
-                     uint64_t llba, int num_lb, int group, unsigned char *dop,
-                     int do_len, int timeout, int verbose)
+sg_ll_write_verify16(int sg_fd, int wrprotect, bool dpo, int bytchk,
+                     uint64_t llba, int num_lb, int group, uint8_t *dop,
+                     int do_len, int timeout, bool noisy, int verbose)
 {
     int ret;
-    unsigned char wv_cdb[WRITE_VERIFY16_CMDLEN];
+    uint8_t wv_cdb[WRITE_VERIFY16_CMDLEN];
 
 
     memset(wv_cdb, 0, sizeof(wv_cdb));
@@ -231,21 +233,11 @@ sg_ll_write_verify16(int sg_fd, int wrprotect, int dpo, int bytchk,
     if (bytchk)
         wv_cdb[1] |= ((bytchk & 0x3) << 1);
 
-    wv_cdb[2] = (llba >> 56) & 0xff;
-    wv_cdb[3] = (llba >> 48) & 0xff;
-    wv_cdb[4] = (llba >> 40) & 0xff;
-    wv_cdb[5] = (llba >> 32) & 0xff;
-    wv_cdb[6] = (llba >> 24) & 0xff;
-    wv_cdb[7] = (llba >> 16) & 0xff;
-    wv_cdb[8] = (llba >> 8) & 0xff;
-    wv_cdb[9] = llba & 0xff;
-    wv_cdb[10] = (num_lb >> 24) & 0xff;
-    wv_cdb[11] = (num_lb >> 16) & 0xff;
-    wv_cdb[12] = (num_lb >> 8) & 0xff;
-    wv_cdb[13] = num_lb & 0xff;
+    sg_put_unaligned_be64(llba, wv_cdb + 2);
+    sg_put_unaligned_be32((uint32_t)num_lb, wv_cdb + 10);
     wv_cdb[14] = group & 0x1f;
     ret = run_scsi_transaction(sg_fd, wv_cdb, sizeof(wv_cdb), dop, do_len,
-                               timeout, verbose);
+                               timeout, noisy, verbose);
     return ret;
 }
 
@@ -259,8 +251,7 @@ open_if(const char * fn, int got_stdin)
     else {
         fd = open(fn, O_RDONLY);
         if (fd < 0) {
-            fprintf(stderr, ME "open error: %s: %s\n", fn,
-                    safe_strerror(errno));
+            pr2serr(ME "open error: %s: %s\n", fn, safe_strerror(errno));
             return -SG_LIB_FILE_ERROR;
         }
     }
@@ -274,31 +265,32 @@ open_if(const char * fn, int got_stdin)
 int
 main(int argc, char * argv[])
 {
-    int sg_fd, res, c, n, first_time;
-    unsigned char * wvb = NULL;
-    void * wrkBuff = NULL;
-    int dpo = 0;
+    bool do_16 = false;
+    bool dpo = false;
+    bool first_time;
+    bool given_do_16 = false;
+    bool has_filename = false;
+    bool lba_given = false;
+    bool repeat = false;
+    int sg_fd, res, c, n;
     int bytchk = 0;
     int group = 0;
-    int do_16 = 0;
-    int given_do_16 = 0;
-    uint64_t llba = 0;
-    int lba_given = 0;
-    uint32_t num_lb = 1;
-    uint32_t snum_lb = 1;
-    int repeat = 0;
-    int timeout = DEF_TIMEOUT_SECS;
-    int verbose = 0;
-    int64_t ll;
-    int wrprotect = 0;
-    const char * device_name = NULL;
-    const char * ifnp;
-    int has_filename = 0;
     int ilen = -1;
     int ifd = -1;
-    int ret = 1;
     int b_p_lb = 512;
+    int ret = 1;
+    int timeout = DEF_TIMEOUT_SECS;
     int tnum_lb_wr = 0;
+    int verbose = 0;
+    int wrprotect = 0;
+    uint32_t num_lb = 1;
+    uint32_t snum_lb = 1;
+    uint64_t llba = 0;
+    int64_t ll;
+    uint8_t * wvb = NULL;
+    void * wrkBuff = NULL;
+    const char * device_name = NULL;
+    const char * ifnp;
     char cmd_name[32];
 
     ifnp = "";          /* keep MinGW quiet */
@@ -316,19 +308,17 @@ main(int argc, char * argv[])
              * sbc4r02 (not =2 nor =3) but that may change in the future. */
             bytchk = sg_get_num(optarg);
             if ((bytchk < 0) || (bytchk > 3))  {
-                fprintf(stderr, "argument to '--bytchk' expected to be 0 "
-                        "to 3\n");
+                pr2serr("argument to '--bytchk' expected to be 0 to 3\n");
                 return SG_LIB_SYNTAX_ERROR;
             }
             break;
         case 'd':
-            dpo = 1;
+            dpo = true;
             break;
         case 'g':
             group = sg_get_num(optarg);
-            if ((group < 0) || (group > 31))  {
-                fprintf(stderr, "argument to '--group' expected to be 0 "
-                        "to 31\n");
+            if ((group < 0) || (group > 63))  {
+                pr2serr("argument to '--group' expected to be 0 to 63\n");
                 return SG_LIB_SYNTAX_ERROR;
             }
             break;
@@ -338,47 +328,47 @@ main(int argc, char * argv[])
             return 0;
         case 'i':
             ifnp = optarg;
-            has_filename = 1;
+            has_filename = true;
             break;
         case 'I':
             ilen = sg_get_num(optarg);
             if (-1 == ilen) {
-                fprintf(stderr, "bad argument to '--ilen'\n");
+                pr2serr("bad argument to '--ilen'\n");
                 return SG_LIB_SYNTAX_ERROR;
             }
             break;
         case 'l':
             if (lba_given) {
-                fprintf(stderr, "must have one and only one '--lba'\n");
+                pr2serr("must have one and only one '--lba'\n");
                 return SG_LIB_SYNTAX_ERROR;
             }
             ll = sg_get_llnum(optarg);
             if (ll < 0) {
-                fprintf(stderr, "bad argument to '--lba'\n");
+                pr2serr("bad argument to '--lba'\n");
                 return SG_LIB_SYNTAX_ERROR;
             }
             llba = (uint64_t)ll;
-            ++lba_given;
+            lba_given = true;
             break;
         case 'n':
             n = sg_get_num(optarg);
             if (-1 == n) {
-                fprintf(stderr, "bad argument to '--num'\n");
+                pr2serr("bad argument to '--num'\n");
                 return SG_LIB_SYNTAX_ERROR;
             }
             num_lb = (uint32_t)n;
             break;
         case 'R':
-            ++repeat;
+            repeat = true;
             break;
         case 'S':
-            do_16 = 1;
-            given_do_16 = 1;
+            do_16 = true;
+            given_do_16 = true;
             break;
         case 't':
             timeout = sg_get_num(optarg);
             if (timeout < 1) {
-                fprintf(stderr, "bad argument to '--timeout'\n");
+                pr2serr("bad argument to '--timeout'\n");
                 return SG_LIB_SYNTAX_ERROR;
             }
             break;
@@ -386,19 +376,19 @@ main(int argc, char * argv[])
             ++verbose;
             break;
         case 'V':
-            fprintf(stderr, ME "version: %s\n", version_str);
+            pr2serr(ME "version: %s\n", version_str);
             return 0;
         case 'w':
             wrprotect = sg_get_num(optarg);
             if ((wrprotect < 0) || (wrprotect > 7))  {
-                fprintf(stderr, "wrprotect (%d) is out of range ( < %d)\n",
-                        wrprotect, 7);
+                pr2serr("wrprotect (%d) is out of range ( < %d)\n", wrprotect,
+                        7);
                 return SG_LIB_SYNTAX_ERROR;
             }
 
             break;
         default:
-            fprintf(stderr, "unrecognised option code 0x%x ??\n", c);
+            pr2serr("unrecognised option code 0x%x ??\n", c);
             usage();
             return SG_LIB_SYNTAX_ERROR;
         }
@@ -410,71 +400,68 @@ main(int argc, char * argv[])
         }
         if (optind < argc) {
             for (; optind < argc; ++optind)
-                fprintf(stderr, "Unexpected extra argument: %s\n",
-                        argv[optind]);
+                pr2serr("Unexpected extra argument: %s\n", argv[optind]);
             usage();
             return SG_LIB_SYNTAX_ERROR;
        }
     }
 
     if (NULL == device_name) {
-        fprintf(stderr, "missing device name!\n");
+        pr2serr("missing device name!\n");
         usage();
         return SG_LIB_SYNTAX_ERROR;
     }
     if (! lba_given) {
-        fprintf(stderr, "need a --lba=LBA option\n");
+        pr2serr("need a --lba=LBA option\n");
         usage();
         return SG_LIB_SYNTAX_ERROR;
     }
     if (repeat) {
         if (! has_filename) {
-            fprintf(stderr, "with '--repeat' need '--in=IF' option\n");
+            pr2serr("with '--repeat' need '--in=IF' option\n");
             usage();
             return SG_LIB_SYNTAX_ERROR;
         }
         if (ilen < 1) {
-            fprintf(stderr, "with '--repeat' need '--ilen=ILEN' option\n");
+            pr2serr("with '--repeat' need '--ilen=ILEN' option\n");
             usage();
             return SG_LIB_SYNTAX_ERROR;
         } else {
             b_p_lb = ilen / num_lb;
             if (b_p_lb < 64) {
-                fprintf(stderr, "calculated %d bytes per logical block, "
-                        "too small\n", b_p_lb);
+                pr2serr("calculated %d bytes per logical block, too small\n",
+                        b_p_lb);
                 usage();
                 return SG_LIB_SYNTAX_ERROR;
             }
         }
     }
 
-    sg_fd = sg_cmds_open_device(device_name, 0 /* rw */, verbose);
+    sg_fd = sg_cmds_open_device(device_name, false /* rw */, verbose);
     if (sg_fd < 0) {
-        fprintf(stderr, ME "open error: %s: %s\n", device_name,
-                safe_strerror(-sg_fd));
+        pr2serr(ME "open error: %s: %s\n", device_name, safe_strerror(-sg_fd));
         return SG_LIB_FILE_ERROR;
     }
 
-    if ((0 == do_16) && (llba > UINT_MAX))
-        do_16 = 1;
-    if ((0 == do_16) && (num_lb > 0xffff))
-        do_16 = 1;
+    if ((! do_16) && (llba > UINT_MAX))
+        do_16 = true;
+    if ((! do_16) && (num_lb > 0xffff))
+        do_16 = true;
     snprintf(cmd_name, sizeof(cmd_name), "Write and verify(%d)",
              (do_16 ? 16 : 10));
-    if (verbose && (0 == given_do_16) && do_16)
-        fprintf(stderr, "Switching to %s because LBA or NUM too large\n",
-                cmd_name);
+    if (verbose && (! given_do_16) && do_16)
+        pr2serr("Switching to %s because LBA or NUM too large\n", cmd_name);
     if (verbose) {
-        fprintf(stderr, "Issue %s to device %s\n\tilen=%d", cmd_name,
-                device_name, ilen);
+        pr2serr("Issue %s to device %s\n\tilen=%d", cmd_name, device_name,
+                ilen);
         if (ilen > 0)
-            fprintf(stderr, " [0x%x]", ilen);
-        fprintf(stderr, ", lba=%" PRIu64 " [0x%" PRIx64 "]\n\twrprotect=%d, "
-                "dpo=%d, bytchk=%d, group=%d, repeat=%d\n", llba, llba,
-                wrprotect, dpo, bytchk, group, repeat);
+            pr2serr(" [0x%x]", ilen);
+        pr2serr(", lba=%" PRIu64 " [0x%" PRIx64 "]\n\twrprotect=%d, dpo=%d, "
+                "bytchk=%d, group=%d, repeat=%d\n", llba, llba, wrprotect,
+                (int)dpo, bytchk, group, (int)repeat);
     }
 
-    first_time = 1;
+    first_time = true;
     do {
         if (first_time) {
             //If a file with data to write has been provided
@@ -485,7 +472,7 @@ main(int argc, char * argv[])
                     ifd = STDIN_FILENO;
                     ifnp = "<stdin>";
                     if (verbose > 1)
-                        fprintf(stderr, "Reading input data from stdin\n");
+                        pr2serr("Reading input data from stdin\n");
                 } else {
                     ifd = open_if(ifnp, 0);
                     if (ifd < 0) {
@@ -495,76 +482,75 @@ main(int argc, char * argv[])
                 }
                 if (ilen < 1) {
                     if (fstat(ifd, &a_stat) < 0) {
-                        fprintf(stderr, "Could not fstat(%s)\n", ifnp);
+                        pr2serr("Could not fstat(%s)\n", ifnp);
                         goto err_out;
                     }
                     if (! S_ISREG(a_stat.st_mode)) {
-                        fprintf(stderr, "Cannot determine IF size, please "
-                                "give '--ilen='\n");
+                        pr2serr("Cannot determine IF size, please give "
+                                "'--ilen='\n");
                         goto err_out;
                     }
                     ilen = (int)a_stat.st_size;
                     if (ilen < 1) {
-                        fprintf(stderr, "%s file size too small\n", ifnp);
+                        pr2serr("%s file size too small\n", ifnp);
                         goto err_out;
                     } else if (verbose)
-                        fprintf(stderr, "Using file size of %d bytes\n", ilen);
+                        pr2serr("Using file size of %d bytes\n", ilen);
                 }
                 if (NULL == (wrkBuff = malloc(ilen))) {
-                    fprintf(stderr, ME "out of memory\n");
+                    pr2serr(ME "out of memory\n");
                     ret = SG_LIB_CAT_OTHER;
                     goto err_out;
                 }
-                wvb = (unsigned char *)wrkBuff;
+                wvb = (uint8_t *)wrkBuff;
                 res = read(ifd, wvb, ilen);
                 if (res < 0) {
-                    fprintf(stderr, "Could not read from %s", ifnp);
+                    pr2serr("Could not read from %s", ifnp);
                     goto err_out;
                 }
                 if (res < ilen) {
-                    fprintf(stderr, "Read only %d bytes (expected %d) from "
-                            "%s\n", res, ilen, ifnp);
+                    pr2serr("Read only %d bytes (expected %d) from %s\n", res,
+                            ilen, ifnp);
                     if (repeat)
-                        fprintf(stderr, "Will scale subsequent pieces when "
-                                "repeat=1, but this is first\n");
+                        pr2serr("Will scale subsequent pieces when "
+                                "repeat=true, but this is first\n");
                     goto err_out;
                 }
             } else {
                 if (ilen < 1) {
                     if (verbose)
-                        fprintf(stderr, "Default write length to %d*%d=%d "
-                                "bytes\n", num_lb, 512, 512 * num_lb);
+                        pr2serr("Default write length to %d*%d=%d bytes\n",
+                                num_lb, 512, 512 * num_lb);
                     ilen = 512 * num_lb;
                 }
                 if (NULL == (wrkBuff = malloc(ilen))) {
-                    fprintf(stderr, ME "out of memory\n");
+                    pr2serr(ME "out of memory\n");
                     ret = SG_LIB_CAT_OTHER;
                     goto err_out;
                 }
-                wvb = (unsigned char *)wrkBuff;
+                wvb = (uint8_t *)wrkBuff;
                 /* Not sure about this: default contents to 0xff bytes */
                 memset(wrkBuff, 0xff, ilen);
             }
-            first_time = 0;
+            first_time = false;
             snum_lb = num_lb;
-        } else {        /* repeat=1, first_time=0, must be reading file */
+        } else {    /* repeat=true, first_time=false, must be reading file */
             llba += snum_lb;
             res = read(ifd, wvb, ilen);
             if (res < 0) {
-                fprintf(stderr, "Could not read from %s", ifnp);
+                pr2serr("Could not read from %s", ifnp);
                 goto err_out;
             } else {
                 if (verbose > 1)
-                fprintf(stderr, "Subsequent read from %s got %d bytes\n",
-                        ifnp, res);
+                pr2serr("Subsequent read from %s got %d bytes\n", ifnp, res);
                 if (0 == res)
                     break;
                 if (res < ilen) {
                     snum_lb = (uint32_t)(res / b_p_lb);
                     n = res % b_p_lb;
                     if (0 != n)
-                        fprintf(stderr, ">>> warning: ignoring last %d "
-                                "bytes of %s\n", n, ifnp);
+                        pr2serr(">>> warning: ignoring last %d bytes of %s\n",
+                                n, ifnp);
                     if (snum_lb < 1)
                         break;
                 }
@@ -573,11 +559,12 @@ main(int argc, char * argv[])
         if (do_16)
             res = sg_ll_write_verify16(sg_fd, wrprotect, dpo, bytchk, llba,
                                        snum_lb, group, wvb, ilen, timeout,
-                                       verbose);
+                                       verbose > 0, verbose);
         else
             res = sg_ll_write_verify10(sg_fd, wrprotect, dpo, bytchk,
                                        (unsigned int)llba, snum_lb, group,
-                                       wvb, ilen, timeout, verbose);
+                                       wvb, ilen, timeout, verbose > 0,
+                                       verbose);
         ret = res;
         if (repeat && (0 == ret))
             tnum_lb_wr += snum_lb;
@@ -587,25 +574,25 @@ main(int argc, char * argv[])
 
 err_out:
     if (repeat)
-        fprintf(stderr, "%d [0x%x] logical blocks written, in total\n",
-                tnum_lb_wr, tnum_lb_wr);
+        pr2serr("%d [0x%x] logical blocks written, in total\n", tnum_lb_wr,
+                tnum_lb_wr);
     if (wrkBuff)
         free(wrkBuff);
     if ((ifd >= 0) && (STDIN_FILENO != ifd))
         close(ifd);
     res = sg_cmds_close_device(sg_fd);
     if (res < 0) {
-        fprintf(stderr, "close error: %s\n", safe_strerror(-res));
+        pr2serr("close error: %s\n", safe_strerror(-res));
         if (0 == ret)
             return SG_LIB_FILE_ERROR;
     }
     if (ret && (0 == verbose)) {
         if (SG_LIB_CAT_INVALID_OP == ret)
-            fprintf(stderr, "%s command not supported\n", cmd_name);
+            pr2serr("%s command not supported\n", cmd_name);
         else if (ret > 0)
-            fprintf(stderr, "%s, exit status %d\n", cmd_name, ret);
+            pr2serr("%s, exit status %d\n", cmd_name, ret);
         else if (ret < 0)
-            fprintf(stderr, "Some error occurred\n");
+            pr2serr("Some error occurred\n");
     }
     return (ret >= 0) ? ret : SG_LIB_CAT_OTHER;
 }
