@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2014-2015 Douglas Gilbert.
+ * Copyright (c) 2014-2018 Douglas Gilbert.
  * All rights reserved.
  * Use of this source code is governed by a BSD-style
  * license that can be found in the BSD_LICENSE file.
@@ -9,6 +9,8 @@
 #include <fcntl.h>
 #include <stdio.h>
 #include <stdlib.h>
+#include <stdarg.h>
+#include <stdbool.h>
 #include <string.h>
 #include <ctype.h>
 #include <getopt.h>
@@ -18,6 +20,7 @@
 #ifdef HAVE_CONFIG_H
 #include "config.h"
 #endif
+
 #include "sg_lib.h"
 #include "sg_lib_data.h"
 #include "sg_pt.h"
@@ -32,7 +35,7 @@
  * to the given SCSI device. Based on zbc-r04c.pdf .
  */
 
-static const char * version_str = "1.02 20151219";
+static const char * version_str = "1.07 20180219";
 
 #define SG_ZONING_OUT_CMDLEN 16
 #define CLOSE_ZONE_SA 0x1
@@ -80,7 +83,8 @@ usage()
             "    --open|-o          issue OPEN ZONE command\n"
             "    --verbose|-v       increase verbosity\n"
             "    --version|-V       print version string and exit\n"
-            "    --zone=ID|-z ID    ID is the starting LBA of the zone\n\n"
+            "    --zone=ID|-z ID    ID is the starting LBA of the zone "
+            "(def: 0)\n\n"
             "Performs a SCSI OPEN ZONE, CLOSE ZONE or FINISH ZONE command. "
             "ID is\ndecimal by default, for hex use a leading '0x' or a "
             "trailing 'h'.\nEither --close, --finish, or --open option "
@@ -90,38 +94,41 @@ usage()
 /* Invokes the zone out command indicated by 'sa' (ZBC).  Return of 0
  * -> success, various SG_LIB_CAT_* positive values or -1 -> other errors */
 static int
-sg_ll_zone_out(int sg_fd, int sa, uint64_t zid, int all, int noisy,
+sg_ll_zone_out(int sg_fd, int sa, uint64_t zid, bool all, bool noisy,
                int verbose)
 {
     int k, ret, res, sense_cat;
-    unsigned char zoCmdBlk[SG_ZONING_OUT_CMDLEN] =
-          {SG_ZONING_OUT, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0};
-    unsigned char sense_b[SENSE_BUFF_LEN];
     struct sg_pt_base * ptvp;
+    uint8_t zo_cdb[SG_ZONING_OUT_CMDLEN] =
+          {SG_ZONING_OUT, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0};
+    uint8_t sense_b[SENSE_BUFF_LEN];
+    char b[64];
 
-    zoCmdBlk[1] = 0x1f & sa;
-    sg_put_unaligned_be64(zid, zoCmdBlk + 2);
+    zo_cdb[1] = 0x1f & sa;
+    sg_put_unaligned_be64(zid, zo_cdb + 2);
     if (all)
-        zoCmdBlk[14] = 0x1;
+        zo_cdb[14] = 0x1;
+    sg_get_opcode_sa_name(zo_cdb[0], sa, -1, sizeof(b), b);
     if (verbose) {
-        pr2serr("    Reset write pointer cdb: ");
+        pr2serr("    %s cdb: ", b);
         for (k = 0; k < SG_ZONING_OUT_CMDLEN; ++k)
-            pr2serr("%02x ", zoCmdBlk[k]);
+            pr2serr("%02x ", zo_cdb[k]);
         pr2serr("\n");
     }
 
     ptvp = construct_scsi_pt_obj();
     if (NULL == ptvp) {
-        pr2serr("Reset write pointer: out of memory\n");
+        pr2serr("%s: out of memory\n", b);
         return -1;
     }
-    set_scsi_pt_cdb(ptvp, zoCmdBlk, sizeof(zoCmdBlk));
+    set_scsi_pt_cdb(ptvp, zo_cdb, sizeof(zo_cdb));
     set_scsi_pt_sense(ptvp, sense_b, sizeof(sense_b));
     res = do_scsi_pt(ptvp, sg_fd, DEF_PT_TIMEOUT, verbose);
-    ret = sg_cmds_process_resp(ptvp, "reset write pointer", res, 0, sense_b,
-                               noisy, verbose, &sense_cat);
+    ret = sg_cmds_process_resp(ptvp, "reset write pointer", res,
+                               SG_NO_DATA_IN, sense_b, noisy, verbose,
+                               &sense_cat);
     if (-1 == ret)
-        ;
+        ret = sg_convert_errno(get_scsi_pt_os_err(ptvp));
     else if (-2 == ret) {
         switch (sense_cat) {
         case SG_LIB_CAT_RECOVERED:
@@ -142,19 +149,18 @@ sg_ll_zone_out(int sg_fd, int sa, uint64_t zid, int all, int noisy,
 int
 main(int argc, char * argv[])
 {
+    bool all = false;
+    bool close = false;
+    bool finish = false;
+    bool open = false;
     int sg_fd, res, c;
-    int all = 0;
-    int close = 0;
-    int finish = 0;
-    int open = 0;
     int verbose = 0;
-    int zid_given = 0;
+    int ret = 0;
     int sa = 0;
     uint64_t zid = 0;
     int64_t ll;
     const char * device_name = NULL;
     const char * sa_name;
-    int ret = 0;
 
     while (1) {
         int option_index = 0;
@@ -167,14 +173,14 @@ main(int argc, char * argv[])
         switch (c) {
         case 'a':
         case 'R':
-            ++all;
+            all = true;
             break;
         case 'c':
-            ++close;
+            close = true;
             sa = CLOSE_ZONE_SA;
             break;
         case 'f':
-            ++finish;
+            finish = true;
             sa = FINISH_ZONE_SA;
             break;
         case 'h':
@@ -182,7 +188,7 @@ main(int argc, char * argv[])
             usage();
             return 0;
         case 'o':
-            ++open;
+            open = true;
             sa = OPEN_ZONE_SA;
             break;
         case 'v':
@@ -198,7 +204,6 @@ main(int argc, char * argv[])
                 return SG_LIB_SYNTAX_ERROR;
             }
             zid = (uint64_t)ll;
-            ++zid_given;
             break;
         default:
             pr2serr("unrecognised option code 0x%x ??\n", c);
@@ -220,7 +225,7 @@ main(int argc, char * argv[])
         }
     }
 
-    if (1 != ((!! close) + (!! finish) + (!! open))) {
+    if (1 != ((int)close + (int)finish + (int)open)) {
         pr2serr("one from the --close, --finish and --open options must be "
                 "given\n");
         usage();
@@ -234,14 +239,14 @@ main(int argc, char * argv[])
         return SG_LIB_SYNTAX_ERROR;
     }
 
-    sg_fd = sg_cmds_open_device(device_name, 0, verbose);
+    sg_fd = sg_cmds_open_device(device_name, false /* rw */, verbose);
     if (sg_fd < 0) {
         pr2serr("open error: %s: %s\n", device_name,
                 safe_strerror(-sg_fd));
         return SG_LIB_FILE_ERROR;
     }
 
-    res = sg_ll_zone_out(sg_fd, sa, zid, all, 1, verbose);
+    res = sg_ll_zone_out(sg_fd, sa, zid, all, true, verbose);
     ret = res;
     if (res) {
         if (SG_LIB_CAT_INVALID_OP == res)
