@@ -1,7 +1,7 @@
 /* A utility program for copying files. Specialised for "files" that
  * represent devices that understand the SCSI command set.
  *
- * Copyright (C) 1999 - 2016 D. Gilbert and P. Allworth
+ * Copyright (C) 1999 - 2018 D. Gilbert and P. Allworth
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation; either version 2, or (at your option)
@@ -20,10 +20,10 @@
  * in this case) are transferred to or from the sg device in a single SCSI
  * command.
  *
- * This version is designed for the linux kernel 2.4, 2.6 and 3 series.
+ * This version is designed for the linux kernel 2.4, 2.6, 3 and 4 series.
  */
 
-#define _XOPEN_SOURCE 500
+#define _XOPEN_SOURCE 600
 #ifndef _GNU_SOURCE
 #define _GNU_SOURCE 1
 #endif
@@ -32,6 +32,8 @@
 #include <fcntl.h>
 #include <stdio.h>
 #include <stdlib.h>
+#include <stdarg.h>
+#include <stdbool.h>
 #include <string.h>
 #include <ctype.h>
 #include <errno.h>
@@ -41,9 +43,9 @@
 #define __STDC_FORMAT_MACROS 1
 #include <inttypes.h>
 #include <sys/ioctl.h>
-#include <sys/types.h>
 #include <sys/stat.h>
 #include <sys/sysmacros.h>
+#include <sys/types.h>  /* needed for lseek64() */
 #include <sys/time.h>
 #include <linux/major.h>
 #include <linux/fs.h>   /* <sys/mount.h> */
@@ -58,7 +60,7 @@
 #include "sg_pr2serr.h"
 
 
-static const char * version_str = "5.52 20160121";
+static const char * version_str = "5.62 20180219";
 
 #define DEF_BLOCK_SIZE 512
 #define DEF_BLOCKS_PER_TRANSFER 128
@@ -74,13 +76,6 @@ static const char * version_str = "5.52 20160121";
 #define READ_CAP_REPLY_LEN 8
 #define RCAP16_REPLY_LEN 32
 
-#ifndef SERVICE_ACTION_IN
-#define SERVICE_ACTION_IN     0x9e
-#endif
-#ifndef SAI_READ_CAPACITY_16
-#define SAI_READ_CAPACITY_16  0x10
-#endif
-
 #define DEF_TIMEOUT 60000       /* 60,000 millisecs == 60 seconds */
 
 #define SGP_READ10 0x28
@@ -105,14 +100,14 @@ static const char * version_str = "5.52 20160121";
 #define EBUFF_SZ 512
 
 struct flags_t {
-    int append;
-    int coe;
-    int dio;
-    int direct;
-    int dpo;
-    int dsync;
-    int excl;
-    int fua;
+    bool append;
+    bool coe;
+    bool dio;
+    bool direct;
+    bool dpo;
+    bool dsync;
+    bool excl;
+    bool fua;
 };
 
 typedef struct request_collection
@@ -126,7 +121,7 @@ typedef struct request_collection
     int64_t in_count;               /*  | blocks remaining for next read */
     int64_t in_rem_count;           /*  | count of remaining in blocks */
     int in_partial;                   /*  | */
-    int in_stop;                      /*  | */
+    bool in_stop;                     /*  | */
     pthread_mutex_t in_mutex;         /* -/ */
     int outfd;
     int64_t seek;
@@ -137,12 +132,12 @@ typedef struct request_collection
     int64_t out_count;              /*  | blocks remaining for next write */
     int64_t out_rem_count;          /*  | count of remaining out blocks */
     int out_partial;                  /*  | */
-    int out_stop;                     /*  | */
+    bool out_stop;                    /*  | */
     pthread_mutex_t out_mutex;        /*  | */
     pthread_cond_t out_sync_cv;       /* -/ hold writes until "in order" */
     int bs;
     int bpt;
-    int dio_incomplete;         /* -\ */
+    int dio_incomplete_count;   /* -\ */
     int sum_of_resids;          /*  | */
     pthread_mutex_t aux_mutex;  /* -/ (also serializes some printf()s */
     int debug;
@@ -150,18 +145,18 @@ typedef struct request_collection
 
 typedef struct request_element
 {       /* one instance per worker thread */
+    bool wr;
     int infd;
     int outfd;
-    int wr;
     int64_t blk;
     int num_blks;
-    unsigned char * buffp;
-    unsigned char * alloc_bp;
+    uint8_t * buffp;
+    uint8_t * alloc_bp;
     struct sg_io_hdr io_hdr;
-    unsigned char cmd[MAX_SCSI_CDBSZ];
-    unsigned char sb[SENSE_BUFF_LEN];
+    uint8_t cmd[MAX_SCSI_CDBSZ];
+    uint8_t sb[SENSE_BUFF_LEN];
     int bs;
-    int dio_incomplete;
+    int dio_incomplete_count;
     int resid;
     int cdbsz_in;
     int cdbsz_out;
@@ -177,21 +172,22 @@ static const char * proc_allow_dio = "/proc/scsi/sg/allow_dio";
 
 static void sg_in_operation(Rq_coll * clp, Rq_elem * rep);
 static void sg_out_operation(Rq_coll * clp, Rq_elem * rep);
-static int normal_in_operation(Rq_coll * clp, Rq_elem * rep, int blocks);
+static bool normal_in_operation(Rq_coll * clp, Rq_elem * rep, int blocks);
 static void normal_out_operation(Rq_coll * clp, Rq_elem * rep, int blocks);
 static int sg_start_io(Rq_elem * rep);
-static int sg_finish_io(int wr, Rq_elem * rep, pthread_mutex_t * a_mutp);
+static int sg_finish_io(bool wr, Rq_elem * rep, pthread_mutex_t * a_mutp);
 
 #define STRERR_BUFF_LEN 128
 
 static pthread_mutex_t strerr_mut = PTHREAD_MUTEX_INITIALIZER;
 
-static int do_time = 0;
+static bool shutting_down = false;
+static bool do_sync = false;
+static bool do_time = false;
 static Rq_coll rcoll;
 static struct timeval start_tm;
 static int64_t dd_count = -1;
 static int num_threads = DEF_NUM_THREADS;
-static int do_sync = 0;
 static int exit_status = 0;
 
 
@@ -277,6 +273,14 @@ install_handler(int sig_num, void (*sig_handler) (int sig))
     }
 }
 
+#ifdef SG_LIB_ANDROID
+static void
+thread_exit_handler(int sig)
+{
+    pthread_exit(0);
+}
+#endif
+
 /* Make safe_strerror() thread safe */
 static char *
 tsafe_strerror(int code, char * ebp)
@@ -383,7 +387,7 @@ static void
 guarded_stop_in(Rq_coll * clp)
 {
     pthread_mutex_lock(&clp->in_mutex);
-    clp->in_stop = 1;
+    clp->in_stop = true;
     pthread_mutex_unlock(&clp->in_mutex);
 }
 
@@ -391,7 +395,7 @@ static void
 guarded_stop_out(Rq_coll * clp)
 {
     pthread_mutex_lock(&clp->out_mutex);
-    clp->out_stop = 1;
+    clp->out_stop = true;
     pthread_mutex_unlock(&clp->out_mutex);
 }
 
@@ -407,16 +411,17 @@ static int
 scsi_read_capacity(int sg_fd, int64_t * num_sect, int * sect_sz)
 {
     int res;
-    unsigned char rcBuff[RCAP16_REPLY_LEN];
+    uint8_t rcBuff[RCAP16_REPLY_LEN];
 
-    res = sg_ll_readcap_10(sg_fd, 0, 0, rcBuff, READ_CAP_REPLY_LEN, 0, 0);
+    res = sg_ll_readcap_10(sg_fd, 0, 0, rcBuff, READ_CAP_REPLY_LEN, false, 0);
     if (0 != res)
         return res;
 
     if ((0xff == rcBuff[0]) && (0xff == rcBuff[1]) && (0xff == rcBuff[2]) &&
         (0xff == rcBuff[3])) {
 
-        res = sg_ll_readcap_16(sg_fd, 0, 0, rcBuff, RCAP16_REPLY_LEN, 0, 0);
+        res = sg_ll_readcap_16(sg_fd, 0, 0, rcBuff, RCAP16_REPLY_LEN, false,
+                               0);
         if (0 != res)
             return res;
         *num_sect = sg_get_unaligned_be64(rcBuff + 0) + 1;
@@ -474,6 +479,8 @@ sig_listen_thread(void * v_clp)
 
     while (1) {
         sigwait(&signal_set, &sig_number);
+        if (shutting_down)
+            break;
         if (SIGINT == sig_number) {
             pr2serr(ME "interrupted by SIGINT\n");
             guarded_stop_both(clp);
@@ -489,7 +496,7 @@ cleanup_in(void * v_clp)
     Rq_coll * clp = (Rq_coll *)v_clp;
 
     pr2serr("thread cancelled while in mutex held\n");
-    clp->in_stop = 1;
+    clp->in_stop = true;
     pthread_mutex_unlock(&clp->in_mutex);
     guarded_stop_out(clp);
     pthread_cond_broadcast(&clp->out_sync_cv);
@@ -501,7 +508,7 @@ cleanup_out(void * v_clp)
     Rq_coll * clp = (Rq_coll *)v_clp;
 
     pr2serr("thread cancelled while out mutex held\n");
-    clp->out_stop = 1;
+    clp->out_stop = true;
     pthread_mutex_unlock(&clp->out_mutex);
     guarded_stop_in(clp);
     pthread_cond_broadcast(&clp->out_sync_cv);
@@ -513,9 +520,8 @@ read_write_thread(void * v_clp)
     Rq_coll * clp;
     Rq_elem rel;
     Rq_elem * rep = &rel;
-    size_t psz = 0;
     int sz;
-    volatile int stop_after_write = 0;
+    volatile bool stop_after_write = false;
     int64_t seek_skip;
     int blocks, status;
 
@@ -523,16 +529,11 @@ read_write_thread(void * v_clp)
     sz = clp->bpt * clp->bs;
     seek_skip =  clp->seek - clp->skip;
     memset(rep, 0, sizeof(Rq_elem));
-#if defined(HAVE_SYSCONF) && defined(_SC_PAGESIZE)
-    psz = sysconf(_SC_PAGESIZE); /* POSIX.1 (was getpagesize()) */
-#else
-    psz = 4096;     /* give up, pick likely figure */
-#endif
-    if (NULL == (rep->alloc_bp = (unsigned char *)malloc(sz + psz)))
+    rep->buffp = sg_memalign(sz, 0 /* page align */, &rep->alloc_bp, false);
+    if (NULL == rep->buffp)
         err_exit(ENOMEM, "out of memory creating user buffers\n");
-    rep->buffp = (unsigned char *)(((uintptr_t)rep->alloc_bp + psz - 1) &
-                                   (~(psz - 1)));
-    /* Follow clp members are constant during lifetime of thread */
+
+    /* Following clp members are constant during lifetime of thread */
     rep->bs = clp->bs;
     rep->infd = clp->infd;
     rep->outfd = clp->outfd;
@@ -552,7 +553,7 @@ read_write_thread(void * v_clp)
             break;
         }
         blocks = (clp->in_count > clp->bpt) ? clp->bpt : clp->in_count;
-        rep->wr = 0;
+        rep->wr = false;
         rep->blk = clp->in_blk;
         rep->num_blks = blocks;
         clp->in_blk += blocks;
@@ -583,21 +584,21 @@ read_write_thread(void * v_clp)
 
         if (clp->out_stop || (clp->out_count <= 0)) {
             if (! clp->out_stop)
-                clp->out_stop = 1;
+                clp->out_stop = true;
             status = pthread_mutex_unlock(&clp->out_mutex);
             if (0 != status) err_exit(status, "unlock out_mutex");
             break;
         }
         if (stop_after_write)
-            clp->out_stop = 1;
-        rep->wr = 1;
+            clp->out_stop = true;
+        rep->wr = true;
         rep->blk = clp->out_blk;
         clp->out_blk += blocks;
         clp->out_count -= blocks;
 
         if (0 == rep->num_blks) {
-            clp->out_stop = 1;
-            stop_after_write = 1;
+            clp->out_stop = true;
+            stop_after_write = true;
             status = pthread_mutex_unlock(&clp->out_mutex);
             if (0 != status) err_exit(status, "unlock out_mutex");
             break;      /* read nothing so leave loop */
@@ -623,22 +624,23 @@ read_write_thread(void * v_clp)
             break;
         pthread_cond_broadcast(&clp->out_sync_cv);
     } /* end of while loop */
-    if (rep->alloc_bp) free(rep->alloc_bp);
+    if (rep->alloc_bp)
+        free(rep->alloc_bp);
     status = pthread_mutex_lock(&clp->in_mutex);
     if (0 != status) err_exit(status, "lock in_mutex");
     if (! clp->in_stop)
-        clp->in_stop = 1;  /* flag other workers to stop */
+        clp->in_stop = true;  /* flag other workers to stop */
     status = pthread_mutex_unlock(&clp->in_mutex);
     if (0 != status) err_exit(status, "unlock in_mutex");
     pthread_cond_broadcast(&clp->out_sync_cv);
     return stop_after_write ? NULL : clp;
 }
 
-static int
+static bool
 normal_in_operation(Rq_coll * clp, Rq_elem * rep, int blocks)
 {
+    bool stop_after_write = false;
     int res;
-    int stop_after_write = 0;
     char strerr_buff[STRERR_BUFF_LEN];
 
     /* enters holding in_mutex */
@@ -657,14 +659,14 @@ normal_in_operation(Rq_coll * clp, Rq_elem * rep, int blocks)
         else {
             pr2serr("error in normal read, %s\n",
                     tsafe_strerror(errno, strerr_buff));
-            clp->in_stop = 1;
+            clp->in_stop = true;
             guarded_stop_out(clp);
             return 1;
         }
     }
     if (res < blocks * clp->bs) {
         int o_blocks = blocks;
-        stop_after_write = 1;
+        stop_after_write = true;
         blocks = res / clp->bs;
         if ((res % clp->bs) > 0) {
             blocks++;
@@ -702,7 +704,7 @@ normal_out_operation(Rq_coll * clp, Rq_elem * rep, int blocks)
             pr2serr("error normal write, %s\n",
                     tsafe_strerror(errno, strerr_buff));
             guarded_stop_in(clp);
-            clp->out_stop = 1;
+            clp->out_stop = true;
             return;
         }
     }
@@ -718,8 +720,8 @@ normal_out_operation(Rq_coll * clp, Rq_elem * rep, int blocks)
 }
 
 static int
-sg_build_scsi_cdb(unsigned char * cdbp, int cdb_sz, unsigned int blocks,
-                  int64_t start_block, int write_true, int fua, int dpo)
+sg_build_scsi_cdb(uint8_t * cdbp, int cdb_sz, unsigned int blocks,
+                  int64_t start_block, bool write_true, bool fua, bool dpo)
 {
     int rd_opcode[] = {0x8, 0x28, 0xa8, 0x88};
     int wr_opcode[] = {0xa, 0x2a, 0xaa, 0x8a};
@@ -733,10 +735,10 @@ sg_build_scsi_cdb(unsigned char * cdbp, int cdb_sz, unsigned int blocks,
     switch (cdb_sz) {
     case 6:
         sz_ind = 0;
-        cdbp[0] = (unsigned char)(write_true ? wr_opcode[sz_ind] :
+        cdbp[0] = (uint8_t)(write_true ? wr_opcode[sz_ind] :
                                                rd_opcode[sz_ind]);
         sg_put_unaligned_be24(0x1fffff & start_block, cdbp + 1);
-        cdbp[4] = (256 == blocks) ? 0 : (unsigned char)blocks;
+        cdbp[4] = (256 == blocks) ? 0 : (uint8_t)blocks;
         if (blocks > 256) {
             pr2serr(ME "for 6 byte commands, maximum number of blocks is "
                     "256\n");
@@ -755,7 +757,7 @@ sg_build_scsi_cdb(unsigned char * cdbp, int cdb_sz, unsigned int blocks,
         break;
     case 10:
         sz_ind = 1;
-        cdbp[0] = (unsigned char)(write_true ? wr_opcode[sz_ind] :
+        cdbp[0] = (uint8_t)(write_true ? wr_opcode[sz_ind] :
                                                rd_opcode[sz_ind]);
         sg_put_unaligned_be32((uint32_t)start_block, cdbp + 2);
         sg_put_unaligned_be16((uint16_t)blocks, cdbp + 7);
@@ -767,14 +769,14 @@ sg_build_scsi_cdb(unsigned char * cdbp, int cdb_sz, unsigned int blocks,
         break;
     case 12:
         sz_ind = 2;
-        cdbp[0] = (unsigned char)(write_true ? wr_opcode[sz_ind] :
+        cdbp[0] = (uint8_t)(write_true ? wr_opcode[sz_ind] :
                                                rd_opcode[sz_ind]);
         sg_put_unaligned_be32((uint32_t)start_block, cdbp + 2);
         sg_put_unaligned_be32((uint32_t)blocks, cdbp + 6);
         break;
     case 16:
         sz_ind = 3;
-        cdbp[0] = (unsigned char)(write_true ? wr_opcode[sz_ind] :
+        cdbp[0] = (uint8_t)(write_true ? wr_opcode[sz_ind] :
                                                rd_opcode[sz_ind]);
         sg_put_unaligned_be64((uint64_t)start_block, cdbp + 2);
         sg_put_unaligned_be32((uint32_t)blocks, cdbp + 10);
@@ -831,12 +833,17 @@ sg_in_operation(Rq_coll * clp, Rq_elem * rep)
                 pr2serr(">> substituted zeros for in blk=%" PRId64 " for %d "
                         "bytes\n", rep->blk, rep->num_blks * rep->bs);
             }
-            /* fall through */
+#if defined(__GNUC__)
+#if (__GNUC__ >= 7)
+            __attribute__((fallthrough));
+            /* FALL THROUGH */
+#endif
+#endif
         case 0:
-            if (rep->dio_incomplete || rep->resid) {
+            if (rep->dio_incomplete_count || rep->resid) {
                 status = pthread_mutex_lock(&clp->aux_mutex);
                 if (0 != status) err_exit(status, "lock aux_mutex");
-                clp->dio_incomplete += rep->dio_incomplete;
+                clp->dio_incomplete_count += rep->dio_incomplete_count;
                 clp->sum_of_resids += rep->resid;
                 status = pthread_mutex_unlock(&clp->aux_mutex);
                 if (0 != status) err_exit(status, "unlock aux_mutex");
@@ -900,12 +907,17 @@ sg_out_operation(Rq_coll * clp, Rq_elem * rep)
             } else
                 pr2serr(">> ignored error for out blk=%" PRId64 " for %d "
                         "bytes\n", rep->blk, rep->num_blks * rep->bs);
-            /* fall through */
+#if defined(__GNUC__)
+#if (__GNUC__ >= 7)
+            __attribute__((fallthrough));
+            /* FALL THROUGH */
+#endif
+#endif
         case 0:
-            if (rep->dio_incomplete || rep->resid) {
+            if (rep->dio_incomplete_count || rep->resid) {
                 status = pthread_mutex_lock(&clp->aux_mutex);
                 if (0 != status) err_exit(status, "lock aux_mutex");
-                clp->dio_incomplete += rep->dio_incomplete;
+                clp->dio_incomplete_count += rep->dio_incomplete_count;
                 clp->sum_of_resids += rep->resid;
                 status = pthread_mutex_unlock(&clp->aux_mutex);
                 if (0 != status) err_exit(status, "unlock aux_mutex");
@@ -930,9 +942,9 @@ static int
 sg_start_io(Rq_elem * rep)
 {
     struct sg_io_hdr * hp = &rep->io_hdr;
-    int fua = rep->wr ? rep->out_flags.fua : rep->in_flags.fua;
-    int dpo = rep->wr ? rep->out_flags.dpo : rep->in_flags.dpo;
-    int dio = rep->wr ? rep->out_flags.dio : rep->in_flags.dio;
+    bool fua = rep->wr ? rep->out_flags.fua : rep->in_flags.fua;
+    bool dpo = rep->wr ? rep->out_flags.dpo : rep->in_flags.dpo;
+    bool dio = rep->wr ? rep->out_flags.dio : rep->in_flags.dio;
     int cdbsz = rep->wr ? rep->cdbsz_out : rep->cdbsz_in;
     int res;
 
@@ -979,7 +991,7 @@ sg_start_io(Rq_elem * rep)
    -> try again, SG_LIB_CAT_NOT_READY, SG_LIB_CAT_MEDIUM_HARD,
    -1 other errors */
 static int
-sg_finish_io(int wr, Rq_elem * rep, pthread_mutex_t * a_mutp)
+sg_finish_io(bool wr, Rq_elem * rep, pthread_mutex_t * a_mutp)
 {
     int res, status;
     struct sg_io_hdr io_hdr;
@@ -1013,12 +1025,12 @@ sg_finish_io(int wr, Rq_elem * rep, pthread_mutex_t * a_mutp)
             break;
         case SG_LIB_CAT_RECOVERED:
             sg_chk_n_print3((wr ? "writing continuing":
-                                       "reading continuing"), hp, 0);
+                                       "reading continuing"), hp, false);
             break;
         case SG_LIB_CAT_ABORTED_COMMAND:
         case SG_LIB_CAT_UNIT_ATTENTION:
             if (rep->debug > 8)
-                sg_chk_n_print3((wr ? "writing": "reading"), hp, 0);
+                sg_chk_n_print3((wr ? "writing": "reading"), hp, false);
             return res;
         case SG_LIB_CAT_NOT_READY:
         default:
@@ -1029,7 +1041,7 @@ sg_finish_io(int wr, Rq_elem * rep, pthread_mutex_t * a_mutp)
                          wr ? "writing": "reading", rep->blk);
                 status = pthread_mutex_lock(a_mutp);
                 if (0 != status) err_exit(status, "lock aux_mutex");
-                sg_chk_n_print3(ebuff, hp, 0);
+                sg_chk_n_print3(ebuff, hp, false);
                 status = pthread_mutex_unlock(a_mutp);
                 if (0 != status) err_exit(status, "unlock aux_mutex");
                 return res;
@@ -1040,9 +1052,9 @@ sg_finish_io(int wr, Rq_elem * rep, pthread_mutex_t * a_mutp)
 #endif
     if ((wr ? rep->out_flags.dio : rep->in_flags.dio) &&
         ((hp->info & SG_INFO_DIRECT_IO_MASK) != SG_INFO_DIRECT_IO))
-        rep->dio_incomplete = 1; /* count dios done as indirect IO */
+        rep->dio_incomplete_count = 1; /* count dios done as indirect IO */
     else
-        rep->dio_incomplete = 0;
+        rep->dio_incomplete_count = 0;
     rep->resid = hp->resid;
     if (rep->debug > 8)
         pr2serr("sg_finish_io: completed %s\n", wr ? "WRITE" : "READ");
@@ -1059,7 +1071,6 @@ sg_prepare(int fd, int bs, int bpt)
         pr2serr(ME "sg driver prior to 3.x.y\n");
         return 1;
     }
-    res = 0;
     t = bs * bpt;
     res = ioctl(fd, SG_SET_RESERVED_SIZE, &t);
     if (res < 0)
@@ -1090,21 +1101,21 @@ process_flags(const char * arg, struct flags_t * fp)
         if (np)
             *np++ = '\0';
         if (0 == strcmp(cp, "append"))
-            fp->append = 1;
+            fp->append = true;
         else if (0 == strcmp(cp, "coe"))
-            fp->coe = 1;
+            fp->coe = true;
         else if (0 == strcmp(cp, "dio"))
-            fp->dio = 1;
+            fp->dio = true;
         else if (0 == strcmp(cp, "direct"))
-            fp->direct = 1;
+            fp->direct = true;
         else if (0 == strcmp(cp, "dpo"))
-            fp->dpo = 1;
+            fp->dpo = true;
         else if (0 == strcmp(cp, "dsync"))
-            fp->dsync = 1;
+            fp->dsync = true;
         else if (0 == strcmp(cp, "excl"))
-            fp->excl = 1;
+            fp->excl = true;
         else if (0 == strcmp(cp, "fua"))
-            fp->fua = 1;
+            fp->fua = true;
         else if (0 == strcmp(cp, "null"))
             ;
         else {
@@ -1142,7 +1153,15 @@ main(int argc, char * argv[])
     int in_sect_sz, out_sect_sz, status, n, flags;
     void * vp;
     char ebuff[EBUFF_SZ];
+#if SG_LIB_ANDROID
+    struct sigaction actions;
 
+    memset(&actions, 0, sizeof(actions));
+    sigemptyset(&actions.sa_mask);
+    actions.sa_flags = 0;
+    actions.sa_handler = thread_exit_handler;
+    sigaction(SIGUSR1, &actions, NULL);
+#endif
     memset(&rcoll, 0, sizeof(Rq_coll));
     rcoll.bpt = DEF_BLOCKS_PER_TRANSFER;
     rcoll.in_type = FT_OTHER;
@@ -1181,7 +1200,7 @@ main(int argc, char * argv[])
             rcoll.cdbsz_out = rcoll.cdbsz_in;
             cdbsz_given = 1;
         } else if (0 == strcmp(key,"coe")) {
-            rcoll.in_flags.coe = sg_get_num(buf);
+            rcoll.in_flags.coe = !! sg_get_num(buf);
             rcoll.out_flags.coe = rcoll.in_flags.coe;
         } else if (0 == strcmp(key,"count")) {
             if (0 != strcmp("-1", buf)) {
@@ -1195,14 +1214,14 @@ main(int argc, char * argv[])
                    (0 == strncmp(key,"verb", 4)))
             rcoll.debug = sg_get_num(buf);
         else if (0 == strcmp(key,"dio")) {
-            rcoll.in_flags.dio = sg_get_num(buf);
+            rcoll.in_flags.dio = !! sg_get_num(buf);
             rcoll.out_flags.dio = rcoll.in_flags.dio;
         } else if (0 == strcmp(key,"fua")) {
             n = sg_get_num(buf);
             if (n & 1)
-                rcoll.out_flags.fua = 1;
+                rcoll.out_flags.fua = true;
             if (n & 2)
-                rcoll.in_flags.fua = 1;
+                rcoll.in_flags.fua = true;
         } else if (0 == strcmp(key,"ibs")) {
             ibs = sg_get_num(buf);
             if (-1 == ibs) {
@@ -1250,11 +1269,11 @@ main(int argc, char * argv[])
                 return SG_LIB_SYNTAX_ERROR;
             }
         } else if (0 == strcmp(key,"sync"))
-            do_sync = sg_get_num(buf);
+            do_sync = !! sg_get_num(buf);
         else if (0 == strcmp(key,"thr"))
             num_threads = sg_get_num(buf);
         else if (0 == strcmp(key,"time"))
-            do_time = sg_get_num(buf);
+            do_time = !! sg_get_num(buf);
         else if ((0 == strncmp(key, "--help", 7)) ||
                  (0 == strncmp(key, "-h", 2)) ||
                  (0 == strcmp(key, "-?"))) {
@@ -1284,7 +1303,7 @@ main(int argc, char * argv[])
         pr2serr("skip and seek cannot be negative\n");
         return SG_LIB_SYNTAX_ERROR;
     }
-    if ((rcoll.out_flags.append > 0) && (seek > 0)) {
+    if (rcoll.out_flags.append && (seek > 0)) {
         pr2serr("Can't use both append and seek switches\n");
         return SG_LIB_SYNTAX_ERROR;
     }
@@ -1607,24 +1626,37 @@ main(int argc, char * argv[])
         }
     }
 
-    if ((do_time) && (start_tm.tv_sec || start_tm.tv_usec))
+    if (do_time && (start_tm.tv_sec || start_tm.tv_usec))
         calc_duration_throughput(0);
 
     if (do_sync) {
         if (FT_SG == rcoll.out_type) {
             pr2serr(">> Synchronizing cache on %s\n", outf);
-            res = sg_ll_sync_cache_10(rcoll.outfd, 0, 0, 0, 0, 0, 0, 0);
+            res = sg_ll_sync_cache_10(rcoll.outfd, 0, 0, 0, 0, 0, false, 0);
             if (SG_LIB_CAT_UNIT_ATTENTION == res) {
                 pr2serr("Unit attention(out), continuing\n");
-                res = sg_ll_sync_cache_10(rcoll.outfd, 0, 0, 0, 0, 0, 0, 0);
+                res = sg_ll_sync_cache_10(rcoll.outfd, 0, 0, 0, 0, 0, false,
+                                          0);
             }
             if (0 != res)
                 pr2serr("Unable to synchronize cache\n");
         }
     }
 
+#if 0
+#if SG_LIB_ANDROID
+    /* Android doesn't have pthread_cancel() so use pthread_kill() instead.
+     * Also there is no need to link with -lpthread in Android */
+    status = pthread_kill(sig_listen_thread_id, SIGUSR1);
+#else
     status = pthread_cancel(sig_listen_thread_id);
+#endif
     if (0 != status) err_exit(status, "pthread_cancel");
+#endif          /* 0 */
+
+    shutting_down = true;
+    status = pthread_kill(sig_listen_thread_id, SIGINT);
+    if (0 != status) err_exit(status, "pthread_kill");
     if (STDIN_FILENO != rcoll.infd)
         close(rcoll.infd);
     if ((STDOUT_FILENO != rcoll.outfd) && (FT_DEV_NULL != rcoll.out_type))
@@ -1637,12 +1669,12 @@ main(int argc, char * argv[])
             res = SG_LIB_CAT_OTHER;
     }
     print_stats("");
-    if (rcoll.dio_incomplete) {
+    if (rcoll.dio_incomplete_count) {
         int fd;
         char c;
 
         pr2serr(">> Direct IO requested but incomplete %d times\n",
-                rcoll.dio_incomplete);
+                rcoll.dio_incomplete_count);
         if ((fd = open(proc_allow_dio, O_RDONLY)) >= 0) {
             if (1 == read(fd, &c, 1)) {
                 if ('0' == c)
